/* @(#)COMMONGABI/common/gabilib/abi_dbug.c	1.8 99/10/04 */
/*
* (C) Copyright 1998 UniSoft Limited.
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*/

/***********************************************************************

NAME:		abi_dbug.c

PROJECT:	Solaris ABI Test Suite

MODULE CONTAINS:
	xx_dbfail()
	xx_dbunresolved()
	xx_dbuntested()
	xx_dbunsupported()
	xx_dbnotinuse()
	xx_dbrfi()

SYNOPSIS:
	void xx_dbfail(char *fmt, ...)
	void xx_dbunresolved(char *fmt, ...)
	void xx_dbuntested(char *fmt, ...)
	void xx_dbunsupported(char *fmt, ...)
	void xx_dbnotinuse(char *fmt, ...)
	void xx_dbrfi(char *fmt, ...)

DESCRIPTION:
	Additions to the dbug package. The xx_db..() functions are
	accessed through the corresponding xx_.. macro in abi_dbug.h.
	Each one is essentially a shorthand for

		xx_rpt(result); in_rpt(fmt, args ...);

	The exception to this is when they are called in a startup or
	cleanup routine, in which case, xx_rpt() is not called.

AUTHOR:		Matthew Hails, UniSoft Ltd.
DATE CREATED:	October 1998
MODIFICATIONS:
		Matthew Hails, UniSoft Ltd., December 1998
			Modification to enable xx_vaprintf() to calculate
			length of format string when /dev/null is not
			accessible.

		Matthew Hails, UniSoft Ltd., March 1999
			Deal with case where xx_..() routines are called
			in a TET startup or cleanup routine.

		Matthew Hails, UniSoft Ltd., October 1999
			Added xx_dbrfi().

************************************************************************/

#include <std.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <ctype.h>
#include <unistd.h>
#include <dbug.h>
#include <report.h>
#include <abi_dbug.h>
#include <tet_api.h>

/* Maximum line length for strings printed using these routines. Lines longer
 * than this are split.
 */
#define LINELEN	256

/* Local prototypes */
static void xx_doresult(int result, char *fmt, va_list vlist);
static void xx_prline(char *line);
static char * xx_vaprintf(char *fmt, va_list vlist);

/* External declarations */
extern struct tet_testlist tet_testlist[];

/*
 * xx_dbfail
 *
 * Mark the current test with a FAILURE result code and log an error message to
 * the journal.
 *
 *	fmt	Format string, cf. printf().
 */
void
xx_dbfail(char *fmt, ...)
{
	int saverrno = errno;
	va_list vlist;

	va_start(vlist, fmt);
	xx_doresult(FAILURE, fmt, vlist);
	va_end(vlist);

	errno = saverrno;
}

/*
 * xx_dbunresolved
 *
 * Mark the current test with a DELETION result code and log an error message
 * to the journal.
 *
 *	fmt	Format string, cf. printf().
 */
void
xx_dbunresolved(char *fmt, ...)
{
	int saverrno = errno;
	va_list vlist;

	va_start(vlist, fmt);
	xx_doresult(DELETION, fmt, vlist);
	va_end(vlist);

	errno = saverrno;
}

/*
 * xx_dbuntested
 *
 * Mark the current test with an UNTESTED result code and log an error message
 * to the journal.
 *
 *	fmt	Format string, cf. printf().
 */
void
xx_dbuntested(char *fmt, ...)
{
	int saverrno = errno;
	va_list vlist;

	va_start(vlist, fmt);
	xx_doresult(UNTESTED, fmt, vlist);
	va_end(vlist);

	errno = saverrno;
}

/*
 * xx_dbunsupported
 *
 * Mark the current test with an UNSUPPORTED result code and log an error
 * message to the journal.
 *
 *	fmt	Format string, cf. printf().
 */
void
xx_dbunsupported(char *fmt, ...)
{
	int saverrno = errno;
	va_list vlist;

	va_start(vlist, fmt);
	xx_doresult(UNSUPPORTED, fmt, vlist);
	va_end(vlist);

	errno = saverrno;
}

/*
 * xx_dbnotinuse
 *
 * Mark the current test with an NOTINUSE result code and log an error
 * message to the journal.
 *
 *	fmt	Format string, cf. printf().
 */
void
xx_dbnotinuse(char *fmt, ...)
{
	int saverrno = errno;
	va_list vlist;

	va_start(vlist, fmt);
	xx_doresult(NOTINUSE, fmt, vlist);
	va_end(vlist);

	errno = saverrno;
}

/*
 * xx_dbrfi
 *
 * Mark the current test with an RFI result code and log an error
 * message to the journal.
 *
 *	fmt	Format string, cf. printf().
 */
void
xx_dbrfi(char *fmt, ...)
{
	int saverrno = errno;
	va_list vlist;

	va_start(vlist, fmt);
	xx_doresult(RFI, fmt, vlist);
	va_end(vlist);

	errno = saverrno;
}

/*
 * xx_doresult
 *
 * Log result code via TET and print an informational line also via TET.
 * These are also reporting through the dbug package if debugging is switched
 * on.
 *
 *	result		Result code.
 *	fmt		Format string, cf. printf().
 *			May be NULL, in which case no message is reported.
 *	vlist		Argument list, as defined in <stdarg.h>.
 */
static void
xx_doresult(int result, char *fmt, va_list vlist)
{
	char *buf;
	char *msg;
	char linebuf[LINELEN + 1];
	char *cp;
	char *start;
	char *end;
	char *space;

	/* fmt may be NULL, in which case no message is reported */
	if (fmt == NULL)
	{
		buf = NULL;
		msg = "";
	}
	else
	{
		/* Print format list into malloc'ed buffer */
		buf = xx_vaprintf(fmt, vlist);
		if (buf == NULL)
			msg = "(Internal error)";
		else
			msg = buf;
	}

	/* Log result code - unless we've been called in a startup or cleanup
	 * routine.
	 */
	if (tet_thistest != 0)
		xx_dbrpt(result);

	/* Parsing to be able to print multiple lines. The messages logged to
	 * TET could use tet_vprintf() if/when we are using TETware, but since
	 * we're doing the splitting for db_vprintf() anyway, it makes the code
	 * look neater to just use tet_infoline() [see xx_prline()].
 	 */
	for (cp = msg, start = msg, space = msg; *cp != '\0'; cp++)
	{
		if (isspace((unsigned char)*cp))
			space = cp;

		if (*cp == '\n' || cp - start >= sizeof(linebuf) - 1)
		{
			if (space > start)
				end = space;
			else
				end = cp;

			strncpy(linebuf, start, (size_t)(end - start));
			linebuf[end - start] = '\0';
			xx_prline(linebuf);

			start = end;
		}
	}

	if (cp > start)
		xx_prline(start);

	/* Free any memory returned by xx_vaprintf() */
	if (buf != NULL)
		free(buf);
}

/*
 * xx_prline
 *
 * Output a single report line.
 *
 *	line	Pointer to report string.
 */
static void
xx_prline(char *line)
{
	if (db_on)
		db_printf("Info", "%s", line);

	tet_infoline(line);
}

/*
 * xx_vaprintf
 *
 * Print a variable argument list format string, allocating the memory as
 * necessary using malloc().
 *
 *	fmt	Format string, cf. printf.
 *	vlist	Argument list, as defined in <stdarg.h>.
 *
 * Returns a pointer to the new string, allocated using malloc(). Freeing
 * this is the responsibility of the caller. Returns NULL on failure.
 */
static char *
xx_vaprintf(char *fmt, va_list vlist)
{
	static char devnull[] = "/dev/null";
	static char tmpfile[] = "./bitbucket.tmp";
	FILE *fp;
	int usingtmp = 0;
	int len;
	void *newstr;

	/* Open null device and print formatted output to it to determine
	 * length of resulting string. If we can't access null device, use a
	 * temporary file in the current directory and remove it later.
	 */
	fp = fopen(devnull, "w");
	if (fp == NULL)
	{
		if (access(tmpfile, F_OK) == 0 || errno != ENOENT)
			return NULL;

		fp = fopen(tmpfile, "w");
		if (fp == NULL)
			return NULL;

		usingtmp = 1;
	}

	len = vfprintf(fp, fmt, vlist);
	fclose(fp);

	if (usingtmp)
		unlink(tmpfile);

	/* Allocate memory for new string */
	newstr = malloc(len + 1);
	if (newstr == NULL)
		return NULL;

	/* Print formatted output to new string */
	if (vsprintf(newstr, fmt, vlist)
#if	defined(TEST_LIBUCB)
		== NULL)
#else	/* defined(TEST_LIBUCB) */
		!= len)
#endif	/* defined(TEST_LIBUCB) */
	{
		free(newstr);
		return NULL;
	}

	return newstr;
}
