/* dv @(#)COMMONGABI/common/gabilib/resmgr.c	1.10 99/11/12 */
/*
* (C) Copyright 1998 UniSoft Limited.
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*/

/***********************************************************************

NAME:		resmgr.c

PROJECT:	SCD

MODULE CONTAINS:
	res_register()
	res_unregister()
	res_unregtype()
	res_freeall()
	res_setenable()
	res_addtype()
	res_regnewtype()
	res_cmpptr()
	res_savestr()
	res_cmpstr()
	res_saveint()
	res_cmpint()
	res_remove()
	res_close()
	res_cmpfail()
	res_fclose()
	res_freenoop()
	rmalloc()
	rrealloc()
	rstrdup()
	ropen()
	rmkdir()

SYNOPSIS:
	void res_register(void *data, int type)
	void res_unregister(void *data, int type)
	void res_unregtype(int type)
	void res_freeall(void)
	int res_setenable(int type, int enable)
	int res_addtype(res_savefunc_t savefn, res_cmpfunc_t cmpfn,
		res_freefunc_t freefn)
	void res_regnewtype(int *type, res_savefunc_t savefn,
		res_cmpfunc_t cmpfn, res_freefunc_t freefn, void *data)
	res_freefunc_t freefn, void *data)
	void * res_saveptr(void *data)
	int res_cmpptr(void *data1, void *data2)
	void * res_savestr(void *data)
	int res_cmpstr(void *data1, void *data2)
	void * res_saveint(void *data)
	int res_cmpint(void *data1, void *data2)
	void res_remove(void *data)
	void res_close(void *data)
	int res_cmpfail(void *data1, void *data2)
	void res_fclose(void *data)
	void res_freenoop(void *data)
	void * rmalloc(size_t size)
	void * rrealloc(void *ptr, size_t size)
	char * rstrdup(char *str)
	int ropen(char *path, int oflag, ... [mode_t mode] )
	int rmkdir(char *path, mode_t mode)

DESCRIPTION:
	Resource manager: enables automatic freeing of resources at the
	end of a test.

	A resource is registered by a call to res_register(). A call to
	res_freeall() `free's all resources currently registered.
	res_unregister() and res_unregtype() can be used to unregister
	resources to avoid them being freed by res_freeall().

	Resource registration is enabled by default. It can be disabled
	and re-enabled using res_setenabled().

	The res_ functions may be called directly from a test, but are
	more usually used to build up allocation routines.

	The functions rmalloc(), rrealloc(), rstrdup(), ropen() and
	rmkdir() are all registered versions of their `r'-less equivalent.
	The only difference is in the memory allocation routines, which
	exit when no more memory is available.

	New resource types can be added at runtime using res_addtype(),
	or the convenience function res_regnewtype().
	The functions res_saveptr(), res_cmpptr(), res_savestr(),
	res_cmpstr(), res_saveint(), res_cmpint(), res_unlink(),
	res_close(), res_cmpfail(), res_freenoop() are provided to
	make this easier.

AUTHOR:		Matthew Hails, UniSoft Ltd.
DATE CREATED:	July 1998 (from MIPSBB3 branch)
MODIFICATIONS:

		Matthew Hails, UniSoft Ltd., 12 November 1999
			Added res_regnewtype().

************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdarg.h>
#include <limits.h>
#include <resmgr.h>

#if	!defined(NELEM)
# define NELEM(x)	((int)(sizeof(x)/sizeof((x)[0])))
#endif	/* !defined(NELEM) */

#if	!defined(FALSE)
# define FALSE	0
#endif	/* !defined(FALSE) */
#if	!defined(TRUE)
# define TRUE	1
#endif	/* !defined(TRUE) */

/* Number of resource structures to allocate at a time */
#define	ALLOCINCR	32

/* A resource class */
struct resclass
{
	int enabled;
	res_savefunc_t savefn;
	res_cmpfunc_t cmpfn;
	res_freefunc_t freefn;
};

/* A single registered resource */
struct resource
{
	void *data;
	int type;
};

/* Local prototypes */
static struct resclass * res_getclass(int type);
static void * res_emalloc(size_t size);
static void * res_erealloc(void *ptr, size_t size);

/*
 * Static data together comprising the current state of the resource manager:
 *
 *	stdrcs		Static array of standard resource classes.
 *
 *	xrcs		Array of extra resource classes; those added at
 *			runtime.
 *	nxrcs		Number of classes in xrcs array.
 *
 *	resbase		Stack of saved resources; as dynamic array.
 *	ressize		Allocated size of stack 
 *	rescount	Number of items used in stack
 */
static struct resclass stdrcs[] =
{
	/* RES_NONE */
	{ FALSE,	res_saveptr,	res_cmpfail,	res_freenoop },

	/* RES_MEMORY */
	{ TRUE,		res_saveptr,	res_cmpptr,	free },

	/* RES_FILE */
	{ TRUE,		res_savestr,	res_cmpstr,	res_remove },

	/* RES_FILEDES */
	{ TRUE,		res_saveint,	res_cmpint,	res_close },

	/* RES_DIR */
	{ TRUE,		res_savestr,	res_cmpstr,	res_remove },

	/* RES_STDIOSTREAM */
	{ TRUE,		res_saveptr,	res_cmpptr,	res_fclose }
};

/* Sanity check so array indices are OK */
#if RES_NONE != 0 || RES_MEMORY != 1 || RES_FILE != 2 || RES_FILEDES != 3 \
	|| RES_DIR != 4 || RES_STDIOSTREAM != 5
# error RES_ defines are out of step with __FILE__
#endif

static struct resclass *xrcs = NULL;
static int nxrcs = 0;

static struct resource *resbase = NULL;
static int ressize = 0;
static int rescount = 0;

/*
 * res_register
 *
 * Register a resource with the resource manager. This is usually called from
 * another routine, rather than directly from a test.
 *
 *	data	Generic pointer to the data to register.
 *	type	Type of data. Determines how the data is saved and freed.
 *		The intial types that can be used are:
 *
 *		RES_MEMORY		Memory returned from one of the
 *					malloc() family of functions.
 *		RES_FILE		A file.
 *		RES_FILEDES		A file descriptor.
 *		RES_DIR			A directory.
 *		RES_STDIOSTREAM		A stdio stream.
 *
 *		More types can be added using res_addtype().
 */
void
res_register(void *data, int type)
{
	struct resclass *rc;
	void *savedata;
	struct resource	*rp;

	rc = res_getclass(type);
	if (rc == NULL || !rc->enabled)
		return;

	/* N.B. This must happen before this resource is added to the list.
	 * Resources are freed in reverse order, so any extra resources used
	 * by the save function are freed after this resource.
	 */
	savedata = rc->savefn(data);

	if (ressize <= rescount)
	{
		ressize += ALLOCINCR;
		resbase = res_erealloc(resbase,
			ressize * sizeof(struct resource));
	}

	rp = &resbase[rescount++];
	rp->data = savedata;
	rp->type = type;
}

/*
 * res_unregister
 *
 * Remove a resource from the resource manager's registration record.
 * This would be called if it were neccessary to free the resource manually.
 * Would mainly be called from other routines rather than from a test. 
 *
 *	data	Generic pointer to the data to register.
 *	type	Type of data.
 */
void
res_unregister(void *data, int type)
{
	struct resource *rp;
	struct resclass *rc;

	rc = res_getclass(type);
	if (rc == NULL || resbase == NULL)
		return;

	for (rp = &resbase[rescount - 1]; rp >= resbase; rp--)
	{
		if (rp->type == type && rc->cmpfn(data, rp->data) == 0)
		{
			rp->type = RES_NONE;
			return;
		}
	}
}

/*
 * res_unregtype
 *
 * Unregisters all resources of a certain type without freeing them.
 *
 *	type	Resource type to unregister. RES_ALL unregisters all resources.
 */
void
res_unregtype(int type)
{
	struct resource *rp;

	if (resbase == NULL)
		return;

	for (rp = &resbase[rescount - 1]; rp >= resbase; rp--)
	{
		if (rp->type == type || type == RES_ALL)
			rp->type = RES_NONE;
	}
}

/*
 * res_freeall
 *
 * Free all resources currently registered with the resource manager.
 */
void
res_freeall(void)
{
	struct resource	*rp;

	/*
	 * Resources are freed in reverse order. See res_register().
	 * No need to check for enabled/disabled resource types - this is
	 * handled by res_setenable().
	 */
	while (rescount > 0)
	{
		rp = &resbase[--rescount];
		res_getclass(rp->type)->freefn(rp->data);
	}
}

/*
 * res_setenable
 *
 * Enable or disable resource registration for a certain type of resource.
 * Registration is enabled for all types by default.
 *
 *	type	Resource type for which to enable or disable registration.
 *		RES_ALL indicates all types.
 *	enable	Boolean value: true (non-zero) indicates that registration
 *		is to be enabled for the specified resource type; false
 *		(zero) indicates that registration is to be disabled.
 *
 * Returns the previous state of enablement. Note that for RES_ALL a non-zero
 * value is returned if any of the resource classes were previously enabled.
 */
int
res_setenable(int type, int enable)
{
	struct resclass *rc;
	int rv;

	rv = 0;

	switch (type)
	{
	case RES_NONE:
		break;

	case RES_ALL:
		/* Set enabled flag for standard resource classes, except for
		 * RES_NONE (index 0), which we never change.
		 */
		for (rc = &stdrcs[1]; rc < stdrcs + NELEM(stdrcs); rc++)
		{
			if (rc->enabled)
				rv = 1;

			rc->enabled = enable;
		}

		/* Set enabled flag for extra resource classes */
		for (rc = xrcs; rc < xrcs + nxrcs; rc++)
		{
			if (rc->enabled)
				rv = 1;

			rc->enabled = enable;
		}

		break;

	default:
		rc = res_getclass(type);
		if (rc != NULL)
		{
			rv = rc->enabled;
			rc->enabled = enable;
		}

		break;
	}

	return rv;
}

/*
 * res_addtype
 *
 * Add new resource type.
 *
 *	savefn		Pointer to function used to save resource for later
 *			deallocation.
 *	cmpfn		Pointer to function used to compare resources of this
 *			type to verify they are the same.
 *	freefn		Pointer to function used to free this type of resource.
 *
 * Returns the value of the new resource type.
 */
int
res_addtype(res_savefunc_t savefn, res_cmpfunc_t cmpfn, res_freefunc_t freefn)
{
	nxrcs++;
	xrcs = res_erealloc(xrcs, nxrcs * sizeof(*xrcs));
	xrcs[nxrcs - 1].enabled = TRUE;
	xrcs[nxrcs - 1].savefn = savefn;
	xrcs[nxrcs - 1].cmpfn = cmpfn;
	xrcs[nxrcs - 1].freefn = freefn;

	return NELEM(stdrcs) + nxrcs - 1;
}

/*
 * res_regnewtype
 *
 * Register data of a new resource type. This is a convenience routine,
 * combining res_addtype() and res_register() to add a new type and register
 * the data.
 *
 *	type		Pointer to the type of data. The address of a static
 *			variable should be passed here, initialized to
 *			RES_NONE. On the first call of this function, this
 *			value will be changed to the new resource type value.
 *	savefn		Pointer to function used to save resource for later
 *			deallocation.
 *	cmpfn		Pointer to function used to compare resources of this
 *			type to verify they are the same.
 *	freefn		Pointer to function used to free this type of resource.
 *	data		Generic pointer to the data to register.
 */
void
res_regnewtype(int *type, res_savefunc_t savefn, res_cmpfunc_t cmpfn,
	res_freefunc_t freefn, void *data)
{
	/* Add new type if it doesn't already exist */
	if (*type == RES_NONE)
		*type = res_addtype(savefn, cmpfn, freefn);

	/* Now register the data */
	res_register(data, *type);
}

/*
 * res_getclass
 *
 * Get resource class corresponding to a given type. Note that this succeeds
 * for RES_NONE (but not for RES_ALL).
 *
 *	type 	Type of resource.
 *
 * Returns a pointer to the resource class; or NULL if there is no resource
 * class corresponding to the given type value.
 */
static struct resclass *
res_getclass(int type)
{
	if (type >= 0)
	{
		if (type < NELEM(stdrcs))
			return &stdrcs[type];

		if (type < NELEM(stdrcs) + nxrcs)
			return &xrcs[type - NELEM(stdrcs)];
	}

	return NULL;
}

/* ========================================================================= */

void *
res_saveptr(void *data)
{
	return data;
}

int
res_cmpptr(void *data1, void *data2)
{
	return (data1 == data2) ? 0 : 1;
}

void *
res_savestr(void *data)
{
	return rstrdup(data);
}

int
res_cmpstr(void *data1, void *data2)
{
	return strcmp(data1, data2);
}

void *
res_saveint(void *data)
{
	int *savedata;

	savedata = rmalloc(sizeof(int));
	*savedata = *(int *)data;

	return savedata;
}

int
res_cmpint(void *data1, void *data2)
{
	return (*(int *)data1 == *(int *)data2) ? 0 : 1;
}

void
res_remove(void *data)
{
	remove(data);
}

void
res_close(void *data)
{
	close(*(int *)data);
}

int
res_cmpfail(void *data1, void *data2)
{
	return 1;
}

void
res_fclose(void *data)
{
	fclose(data);
}

void
res_freenoop(void *data)
{
}

/* ========================================================================= */

/*
 * rmalloc
 *
 * Registered malloc. Allocates memory using malloc and registers the memory
 * with the resource manager. If no memory is available, then exits.
 *
 *	size	Amount of memory to allocate, in bytes.
 *
 * Returns a pointer to the newly allocated memory.
 */
void *
rmalloc(size_t size)
{
	void *ptr;

	ptr = res_emalloc(size);
	res_register(ptr, RES_MEMORY);

	return ptr;
}

/*
 * rrealloc
 *
 * Registered realloc. Reallocates memory using realloc and registers the
 * memory with the resource manager. If no memory is available, then it exits.
 * The memory may have already been allocated with rmalloc; if so then the
 * original memory is unregistered to avoid illegal free calls later on.
 *
 *	ptr	Pointer to block of memory already allocated.
 *	size	Amount of memory to allocate, in bytes.
 *
 * Returns a pointer to the newly allocated memory.
 */
void *
rrealloc(void *ptr, size_t size)
{
	void *newmem;

	newmem = res_erealloc(ptr, size);

	/*
	 * Unregister original memory and register new pointer.
	 * If it was not originally allocated that's not a problem.
	 */
	if (ptr != NULL)
		res_unregister(ptr, RES_MEMORY);

	res_register(newmem, RES_MEMORY);

	return newmem;
}

/*
 * rstrdup
 *
 * Registered strdup(). Copies a string (cf. strdup()) and registers the memory
 * with the resource manager. If no memory is available, then it exits.
 *
 *	str	Null terminated string to be copied.
 *
 * Returns a pointer to the new string.
 */
char *
rstrdup(char *str)
{
	char *newstr;

	newstr = rmalloc(strlen(str) + 1);
	strcpy(newstr, str);

	return newstr;
}

/*
 * ropen
 *
 * Registered open(). Opens a file using open() and registers it with the
 * resource manager.
 *
 * Arguments and return value as for open().
 */
int
ropen(char *path, int oflag, ... /* [mode_t mode] */ )
{
	int fd;
	va_list vlist;
	mode_t mode;

	mode = 0;

	if (oflag & O_CREAT)
	{
		va_start(vlist, oflag);
		mode = va_arg(vlist, mode_t);
		va_end(vlist);
	}

	fd = open(path, oflag, mode);
	if (fd > 0)
		res_register(&fd, RES_FILEDES);

	/* errno should be set correctly, since if open() fails we don't
	 * do anything else.
	 */

	return fd;
}

/*
 * rmkdir
 *
 * Registered mkdir(). Creates a directory using mkdir() and registers it with
 * the resource manager.
 *
 * Arguments and return value as for mkdir().
 */
int
rmkdir(char *path, mode_t mode)
{
	int rv;

	rv = mkdir(path, mode);
	if (rv == 0)
		res_register(path, RES_DIR);

	/* errno should be set correctly, since if mkdir() fails we don't
	 * do anything else.
	 */
	return rv;
}

/* ========================================================================= */

static void *
res_emalloc(size_t size)
{
	void *ptr;

	ptr = malloc(size);
	if (ptr == NULL)
	{
		fprintf(stderr, "malloc failure\n");
		exit(EXIT_FAILURE);
	}

	return ptr;
}

static void *
res_erealloc(void *ptr, size_t size)
{
	void *newmem;

	newmem = realloc(ptr, size);
	if (newmem == NULL)
	{
		fprintf(stderr, "realloc failure\n");
		exit(EXIT_FAILURE);
	}

	return newmem;
}
