/* dv @(#)COMMONGABI/common/gabilib/resutils.c	1.1 98/10/14 */
/*
* (C) Copyright 1998 UniSoft Limited.
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*/

/***********************************************************************

NAME:		resutils.c

PROJECT:	Solaris ABI Test Suite

MODULE CONTAINS:
	res_mmap()
	res_getshmsegtype()
	res_getshmattype()

SYNOPSIS:
	void *res_mmap(void *addr, size_t len, int prot, int flags,
		int fildes, off_t off)
	int res_getshmsegtype(void)
	int res_getshmattype(void)

DESCRIPTION:
	Utility functions for various `non-standard' resources with the
	resource manager and which do not require linking with any
	additional system libraries.

	res_mmap() is a registered version of mmap(2), i.e. equivalent
	to mmap() except it also registers the resulting mapping with
	the resource manager.

	res_getshmsegtype() and res_getshmattype() obtain the
	registration types to use with the resource manager for shared
	memory segments and shared memory attachments respectively.

AUTHOR:		Matthew Hails, UniSoft Ltd.
DATE CREATED:	October 1998
MODIFICATIONS:

************************************************************************/

#include <std.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/shm.h>
#include <sys/ipc.h>
#include <dbug.h>
#include <report.h>
#include <resmgr.h>
#include <resutils.h>

/* Structure holding details of a memory mapping */
struct mapping
{
	void *addr;
	int size;
};

/* Local prototypes */
static void * savemapping(void *data);
static int cmpmapping(void *data1, void *data2);
static void rmmapping(void *data);
static void rmshm(void *data);
static void closeshm(void *data);

/*
 * Registered mmap().
 */
void *
res_mmap(void *addr, size_t len, int prot, int flags, int fildes, off_t off)
{
	static int type = RES_NONE;
	struct mapping map;

	DBUG_ENTER("res_mmap");

	map.addr = mmap(addr, len, prot, flags, fildes, off);
	map.size = len;

	if (map.addr != MAP_FAILED)
	{
		if (type == RES_NONE)
			type = res_addtype(savemapping, cmpmapping, rmmapping);

		res_register(&map, type);
	}

	DBUG_RETURN(map.addr);
}

static void *
savemapping(void *data)
{
	struct mapping *src;
	struct mapping *dst;

	src = data;
	dst = rmalloc(sizeof(*dst));
	dst->addr = src->addr;
	dst->size = src->size;

	return dst;
}

static int
cmpmapping(void *data1, void *data2)
{
	struct mapping *m1 = data1;
	struct mapping *m2 = data2;

	return (m1->addr == m2->addr && m1->size == m2->size) ? 0 : 1;
}

static void
rmmapping(void *data)
{
	struct mapping *map = data;

	munlock(map->addr, map->size);
	munmap(map->addr, map->size);
}

/*
 * res_getshmsegtype
 *
 * Get the resource type for a shared memory segment as returned from shmget().
 * This adds the type the first time it is called.
 */
int
res_getshmsegtype(void)
{
	static int type = RES_NONE;

	if (type == RES_NONE)
		type = res_addtype(res_saveint, res_cmpint, rmshm);

	return type;
}

static void
rmshm(void *data)
{
	shmctl(*(int *)data, IPC_RMID, 0);
}

/*
 * res_getshmattype
 *
 * Get the resource type for a shared memory attachment as returned from
 * shmat(). This adds the type the first time it is called.
 */
int
res_getshmattype(void)
{
	static int type = RES_NONE;

	if (type == RES_NONE)
		type = res_addtype(res_saveptr, res_cmpptr, closeshm);

	return type;
}

static void
closeshm(void *data)
{
	shmdt(data);
}
