/*
*      SCCS:   @(#)common/genlib(chktimes.c)	4.1 (01/21/97)  VSXgen release 1.4
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/


/***********************************************************************

NAME:		chktimes - Check updating of inode time stamps

PROJECT:	VSX (X/OPEN Validation Suite)

MODULE CONTAINS:
	savtimes()
	chktimes()
	fsavtimes()
	fchktimes()
SYNOPSIS:
	void savtimes(file)
		char *file;

	char *chktimes(file, flags)
		char *file;
		int  flags;

	void fsavtimes(fd)
		int	fd;

	char *fchktimes(fd, flags)
		int	fd;
		int  flags;

DESCRIPTION:
	savtimes() saves the current mtime, utime and ctime values
	for the specified file.

	chktimes() verifies that these time stamps have been updated
	since the call to savtimes(), as specified by flags.  Returns
	NULL if updates are correct, otherwise a message describing
	the failure.

	fsavtimes() has the same functionality as savtimes(), except
	that it uses a file descriptor instead of a path name to
	reference the file.

	fchktimes() has the same functionality as fchktimes(), except
	that it uses a file descriptor instead of a path name to
	reference the file.

SOURCE:		
AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	28 Sept 1988
MODIFICATIONS:	Joseph Nave 	3 Mar 1989
		included fsavtimes and fchktimes

		Anthony Lee, 11 July 1990
		Retrofitted resillience macros.

		Geoff Clare, 13 Oct 1992
		Change sleep(1) to sleep(2) for XNFS testing, to allow
		for client & server clocks "ticking over" into next
		second at different times.

************************************************************************/

#include <std.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <dbug.h>
#include <chktimes.h>
#include <sysdep.h>

extern char	*errname();

private void		checknew();
private void		checksame();
private char		*checkstat();

private struct stat	*savt,	*fsavt;
private int		errcnt;
private char		errbuf[100];

public void
savtimes(file)
char *file;
{
	static struct stat savtbuf;
	int	pathok = 0;

	DBUG_LENTER("savtimes", 2);
	DBUG_PRINTF("args", "savtimes(%s)", file);

	savt = &savtbuf;
	if (stat(file, savt) == SYSERROR)
	{
		vsxerror("stat(\"%s\") failed in savtimes() - errno %d(%s)",
			file, errno, errname(errno));
		savt = NULL;
	}
	else
	{
		/* allow for current time to change before doing the update */
		(void) sleep(2);
		PATH_TRACE;
	}

	/*globok = (pathok == 1) ? 1 : 0;*/
	PATH_FUNC_RPT(1); 

	DBUG_VOID_RETURN;
}


public void
fsavtimes(fd)
int	fd;
{
	static struct stat fsavtbuf;
	int	pathok = 0;

	DBUG_LENTER("fsavtimes", 2);
	DBUG_PRINTF("args", "fsavtimes(%d)", fd);

	fsavt = &fsavtbuf;
	if (fstat(fd, fsavt) == SYSERROR)
	{
		vsxerror("fstat(%d) failed in fsavtimes() - errno %d(%s)",
			fd, errno, errname(errno));
		fsavt = NULL;
	}
	else
	{
		/* allow for current time to change before doing the update */
		(void) sleep(2);
		PATH_TRACE;
	}

	/*globok = (pathok == 1) ? 1 : 0;*/
	PATH_FUNC_RPT(1); 

	DBUG_VOID_RETURN;
}

public char *
chktimes(file, flags)
char *file;
int flags;
{
	struct stat newtbuf, *newt = &newtbuf;
	char *err = NULL;
	int	pathok = 0;

	DBUG_LENTER("chktimes", 2);
	DBUG_PRINTF("args", "chktimes(%s, 0x%x)", file, flags);

	(void) strcpy(errbuf, "time stamps incorrect: ");

	if (savt == NULL)
		err = "stat() failed in savtimes()";
	else if (stat(file, newt) == SYSERROR)
	{
		vsxerror("stat(\"%s\") failed in chktimes() - errno %d(%s)",
			file, errno, errname(errno));
		err = "stat() failed in chktimes()";
	}
	else
	{
		PATH_TRACE;
		globok = 0;
		err = checkstat(savt , newt , flags);
		PATH_FUNC_TRACE; 
	}

	globok = 0;
	PATH_FUNC_RPT(2); 

	if (err != NULL)
	{
		DBUG_PRINTF("return", "chktimes = \"%s\"", err);
		DBUG_RETURN(err);
	}

	DBUG_PRINTF("return", "chktimes = NULL", errbuf);
	DBUG_RETURN((char *)NULL);
}


public char *
fchktimes(fd, flags)
int	fd;
int flags;
{
	struct stat newtbuf, *newt = &newtbuf;
	char *err = NULL;
	int	pathok = 0;

	DBUG_LENTER("fchktimes", 2);
	DBUG_PRINTF("args", "fchktimes(%d, 0x%x)", fd, flags);

	(void) strcpy(errbuf, "time stamps incorrect: ");

	if (fsavt == NULL)
		err = "fstat() failed in fsavtimes()";
	else if (fstat(fd, newt) == SYSERROR)
	{
		vsxerror("fstat(\"%d\") failed in fchktimes() - errno %d(%s)",
			fd, errno, errname(errno));
		err = "stat() failed in fchktimes()";
	}
	else
	{
		PATH_TRACE;
		globok = 0;
		err = checkstat(fsavt , newt , flags);
		PATH_FUNC_TRACE;
	}

	globok = 0;
	PATH_FUNC_RPT(2); 

	if (err != NULL)
	{
		DBUG_PRINTF("return", "fchktimes = \"%s\"", err);
		DBUG_RETURN(err);
	}

	DBUG_PRINTF("return", "fchktimes = NULL", errbuf);
	DBUG_RETURN((char *)NULL);
}

private char *
checkstat(oldt , newt , flags)
struct	stat	*oldt,	*newt;
int flags;
{
	char	*err = NULL;
	int	pathok = 0;

	DBUG_LENTER("checkstat", 2);
	DBUG_PRINTF("args", "checkstat(0x%x, 0x%x, %d)", oldt, newt, flags);

	errcnt = 0;
	if (flags & NEWATIME)
	{
		globok = 0;
		checknew(oldt->st_atime, newt->st_atime, "atime");
		PATH_FUNC_TRACE; 
	}

	if (flags & NEWMTIME)
	{
		globok = 0;
		checknew(oldt->st_mtime, newt->st_mtime, "mtime");
		PATH_FUNC_TRACE; 
	}

	if (flags & NEWCTIME)
	{
		globok = 0;
		checknew(oldt->st_ctime, newt->st_ctime, "ctime");
		PATH_FUNC_TRACE; 
	}

	if (errcnt > 0)
	{
		err = errbuf;
		(void) strcat(err, errcnt == 1 ? " was" : " were");
		(void) strcat(err, " not updated");
		errcnt = 100;
	}
	else
		PATH_TRACE;


	if (!(flags & NEWATIME))
	{
		globok = 0;
		checksame(oldt->st_atime, newt->st_atime, "atime");
		PATH_FUNC_TRACE; 
	}

	if (!(flags & NEWMTIME))
	{
		globok = 0;
		checksame(oldt->st_mtime, newt->st_mtime, "mtime");
		PATH_FUNC_TRACE; 
	}

	if (!(flags & NEWCTIME))
	{
		globok = 0;
		checksame(oldt->st_ctime, newt->st_ctime, "ctime");
		PATH_FUNC_TRACE; 
	}

	if (errcnt%100 > 0)
	{
		err = errbuf;
		(void) strcat(err, errcnt%100 == 1 ? " was" : " were");
		(void) strcat(err, " updated");
	}
	else
		PATH_TRACE;

	globok = 0;
	PATH_FUNC_RPT(5); 

	DBUG_RETURN(err);
}

private void
checknew(ot, nt, name)
time_t ot, nt;
char *name;
{
	/* check time stamp has been updated  */
	int pathok = 0; 	
	if (ot == nt)
	{
		if (errcnt++ != 0)
			(void) strcat(errbuf, ", ");
		(void) strcat(errbuf, name);
	}
	else
	{
		PATH_TRACE; 
		PATH_FUNC_RPT(1);
	}
}

private void
checksame(ot, nt, name)
time_t ot, nt;
char *name;
{
	/* check time stamp has NOT been updated  */

	int pathok = 0;
	
	DBUG_LENTER("checksame", 2);
	DBUG_PRINTF("args", "checksame(%ld, %ld, %s)", ot, nt, name);
	if (ot != nt)
	{
		if (errcnt++ != 0)
			(void) strcat(errbuf, ", ");
		(void) strcat(errbuf, name);
	}
	else
	{
		PATH_TRACE;
		PATH_FUNC_RPT(1);
	}

	DBUG_VOID_RETURN;
}
