/*
*      SCCS:   @(#)common/tet_startup/startup.c	4.1 (01/21/97)  VSXgen release 1.4
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/


/***********************************************************************

NAME:		startup.c - main testset tet_startup routine
PROJECT:	VSX (X/OPEN Validation Suite)
MODULE CONTAINS:
	Interface between TET tet_startup/tet_cleanup and VSX setfunc/clnfunc.
	The startup routine creates a work directory "d.<tset>" and
	chdir()s to it.  The cleanup routine removes the work directory.
	Startup also initialises vg_ident from tet_pname, vg_dbug from
	VSX_DBUG_FLAGS and VSX_DBUG_FILE, and checks the value of
	no_tests against tet_testlist[].
	If child1 or child2 are > 0, cleanup kills them.

RELATED DOCUMENTS:
AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	18 Oct 1990
MODIFICATIONS:
		Geoff Clare, 15 Nov 1991
			Add child1 and child2.
			Add set_pname() call.

		Geoff Clare, 17 July 1992
			Change child1 and child2 from int to pid_t.

		Geoff Clare, 10 Sept 1992
			Changes for testing XNFS.

		Geoff Clare, 23 June 1994
			Added #include <stdio.h> (for sprintf decl)

************************************************************************/

#include <std.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <dbug.h>
#include <tet_api.h>
#include <xnfs.h>
#include <sysdep.h>

#define MODEANY	(mode_t)(S_IRWXU|S_IRWXG|S_IRWXO)

extern	void	set_pname();
extern	char	*errname();
extern	char	*newstr();

extern	int	no_tests;
extern	void	(*setfunc)();
extern	void	(*clnfunc)();
extern	struct tet_testlist tet_testlist[];

private	void	startup();
private	void	cleanup();
private	int	creat_wdir();

private char	delreason[_POSIX_PATH_MAX+80];
private char	wdname[_POSIX_NAME_MAX+1];

#if TEST_XNFS
private int	snames[] = { 1, 0 };
public	char	startdir[_POSIX_PATH_MAX+1];
#endif

public	char	*vg_ident = NULL;
public	char	*vg_dbug = NULL;
public	void	(*tet_startup) () = startup;
public	void	(*tet_cleanup) () = cleanup;
public	pid_t	child1;
public	pid_t	child2;

private void
startup()
{
	char *dbflags, *dbfile;
	int i, len, fail = 0;
#if TEST_XNFS
	char *nfs_mount_dir;
#endif

	(void) umask((mode_t)0);

	vg_ident = tet_pname;
	set_pname(tet_pname);
	DBUG_PROCESS(tet_pname);

	dbflags = tet_getvar("VSX_DBUG_FLAGS");
	if (dbflags != NULL && *dbflags != '\0')
	{
		dbfile = NULL;

		/* If no "o,file" in flags, add one for VSX_DBUG_FILE */
		if (dbflags[0] != 'o')
		{
			char *cp;
			for (cp = dbflags; cp[1] != '\0'; cp++)
				if (cp[0] == ':' && cp[1] == 'o')
					break;
			if (cp[1] != 'o')
				dbfile = tet_getvar("VSX_DBUG_FILE");
		}
		if (dbfile == NULL)
			dbfile = "";

		len = strlen(dbflags) + strlen(dbfile) + 3; /* 3 for ":o," */
		vg_dbug = (char *) malloc((size_t)(len + 3)); /* 3 for "-D\0" */
		if (vg_dbug != NULL)
		{
			(void) strcpy(vg_dbug, "-D");
			(void) strcat(vg_dbug, dbflags);
			if (*dbfile != '\0')
			{
				(void) strcat(vg_dbug, ":o,");
				(void) strcat(vg_dbug, dbfile);
			}
			DBUG_PUSH(&vg_dbug[2]);
		}
		else
		{
			DBUG_PUSH(dbflags);
			DBUG_PRINTF("startup", "malloc() for vg_dbug failed");
		}
	}
	
	for (i = 0; tet_testlist[i].testfunc != NULL; i++)
		/* do nothing */ ;
	if (no_tests != i)
	{
		vsxerror("warning: no_tests incorrect - was %d, now %d",
			no_tests, i);
		no_tests = i;
	}

	/* If testing XNFS, save current directory in startdir[] and
	   chdir() to XNFS_MOUNT_DEFAULT */

#if TEST_XNFS
	nfs_mount_dir = tet_getvar("XNFS_MOUNT_DEFAULT");
	if (nfs_mount_dir == NULL || *nfs_mount_dir == '\0')
	{
		fail++;
		(void) strcpy(delreason,
			"parameter XNFS_MOUNT_DEFAULT is not set");
	}
	else if (getcwd(startdir, sizeof(startdir)) == NULL)
	{
		fail++;
		(void) sprintf(delreason,
			"getcwd() failed in startup with errno = %d (%s)",
				errno, errname(errno));
	}
	else if (chdir(nfs_mount_dir) != 0)
	{
		fail++;
		(void) sprintf(delreason,
			"could not change directory to XNFS_MOUNT_DEFAULT (%s)",
				nfs_mount_dir);
	}
	else
#endif
	     if (creat_wdir() != 0)
		fail++;
	else if (chdir(wdname) != 0)
	{
		fail++;
		(void) sprintf(delreason,
			"could not change directory to \"%s\"", wdname);
	}

	/* If testing XNFS, sync with slave before it chdirs to wdname */
#if TEST_XNFS
	(void) tet_sync(1, snames, SYNC_TIMEOUT);
	/* if the sync fails, the slave will report a chdir() failure */
#endif

	if (fail)
	{
		for (i = 1; i <= no_tests; i++)
			tet_delete(i, delreason);
	}
	else
	{
		if (setfunc != NULL)
			(*setfunc)();
	}
}

private void
cleanup()
{
	if (clnfunc != NULL)
		(*clnfunc)();

	if (child1 > 0)
		(void) killchild(child1, 5*SPEEDFACTOR);

	if (child2 > 0)
		(void) killchild(child2, 5*SPEEDFACTOR);

	/* If testing XNFS, sync with slave after it chdirs out of wdname */
#if TEST_XNFS
	/* The testset cleanup code uses sync point numbers starting
	   at (no_tests+1)<<8, so use (no_tests+2)<<8 here. */
	(void) tet_sync((no_tests+2)<<8, snames, SYNC_TIMEOUT);
#endif

	if (chdir("..") == 0)
		(void) rmdir(wdname);
}

private int
creat_wdir()
{
	/* create work directory "d.(tset)" */

	char *tsetname;
	struct stat statb;

	DBUG_LENTER("creat_wdir", 2);

	/* construct temp dir name from testset name */
	tsetname = strrchr(tet_pname, '/');
	if (tsetname == NULL)
		tsetname = tet_pname;
	else
		tsetname++;
	(void) strcpy(wdname, "d.");
	if (strncmp(tsetname, "T.", (size_t)2) == 0)
		tsetname += 2;
	(void) strncat(wdname, tsetname, (size_t)(_POSIX_NAME_MAX-2));

	/* check for existing directory */
	if (stat(wdname, &statb) != SYSERROR)
	{
		if(S_ISDIR(statb.st_mode))
		{
			/* check existing dir has suitable permissions */
			if ((statb.st_mode & MODEANY) == MODEANY)
			{
				DBUG_PRINTF("creat_wdir",
					"%s already exists and is usable",
					wdname);
				DBUG_RETURN(0);
			}
			else
				(void) rmdir(wdname);
		}
		else
			(void) unlink(wdname);
	}

	if (mkdir(wdname, MODEANY) != 0)
	{
		int err = errno;

		DBUG_PRINTF("creat_wdir",
			"mkdir(\"%s\", %#o) failed with errno = %d (%s)",
			wdname, (int)MODEANY, err, errname(err));
	
		(void) sprintf(delreason,
			"could not create work directory \"%s\"", wdname);
	
		DBUG_RETURN(-1);
	}

	DBUG_RETURN(0);
}
