/*
*      SCCS:   @(#)common/tsetlib(do_access.c)	4.1 (01/21/97)  VSXgen release 1.4
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/


/***********************************************************************

NAME:		do_access.c - set up file access tests

PROJECT:	VSX (X/OPEN Validation Suite)

MODULE CONTAINS:
	do_access()
	cln_access()
SYNOPSIS:
	int do_access(filetype, fileperm, pathperm, func, waittime)
		int		filetype;
		mode_t		fileperm;
		mode_t		pathperm;
		int		(*func)();
		int		waittime;
		int		spno;
	
	void cln_access()

DESCRIPTION:
	Do_access() sets up files with the required ownership and
	permissions for testing file access calls in a user-supplied
	function.  The calling process must have PRV_SETID and
	PRV_ACCESS privileges.

	Filetype specifies the type of target file, either 'd' for a
	directory or 'f' for a plain file.  Pathperm is the desired
	access to the target file's path prefix; fileperm specifies the
	access to the file itself.  Both are given as the bitwise OR of
	the components S_IRUSR, S_IWUSR and S_IXUSR.

	Func is a user supplied function which will be passed a pathname
	to the target file and a NULL-terminated list of strings describing
	which permissions are being used to access it.  It should return 0
	if the test succeeds, otherwise non-zero.

	Waittime (given in seconds) is used to time out child processes.
	This should be a typical child timeout, such as 5*SPEEDFACTOR.
	It is multiplied up as necessary for children that may take
	longer.  In some cases the multiplier is dependent on NGROUPS_MAX.
	If do_access() is itself executed from a child, the timeout on that
	child should be something like (waittime * (5 + NGROUPS_MAX/4)).

	Spno is a sync point number which is only used when testing XNFS.

	Do_access() returns 0 if all the (*func)() calls returned 0.  If
	a preparation error occurred (reported by do_access() as a DELETION),
	or any (*func)() calls returned non-zero, then do_access() returns -1.

	Cln_access() cleans up the files created by do_access().  The
	calling process should clean up any files created in target
	directories before exiting the user supplied function (*func)().

CONDITIONAL COMPILE PARAMS:
SIDE EFFECTS:
RELATED DOCUMENTS:
SOURCE:
AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	6 June 1989
MODIFICATIONS:
		Geoff Clare, 15 Nov 1989
			Add filetype argument

		David Sawyer
		Tue Jul 17 15:57:19 BST 1990
		Added resillience code.

		Peter Skan, 27 Sept 1991
			Call unsetprv() to remove ACCESS privilege

		Peter Skan, 9 Oct 1991
			Also remove SETID privilege

		John-Paul, 23 July 1992
			mode variables changed from 'int' to 'mode_t'.

		David Sawyer, 27 September 1992
		Added TEST_XNFS code

		Geoff Clare, 6 Jan 1993
			Create files on server when owner is mapped.
			Add chown() call to cr_target().

		Geoff Clare, 2 Feb 1993
			When testing XNFS, expect getgroups() to return
			min of NGROUPS_MAX values for server and client.

		Geoff Clare, 3 Nov 1993
			Implement interpretation XNFS/009/931001.

		Geoff Clare, 18 March 1994
			When testing XNFS, don't expect getgroups() to
			return more than 16 

		Geoff Clare, 23 June 1994
			Added #include <stdio.h> (for sprintf decl)

		Geoff Clare, 13 Oct 1995
			Make timeout on ch_normt() depend on ngroups.

************************************************************************/

#include	<std.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<sys/types.h>	
#include	<unistd.h>
#include	<errno.h>
#include	<sys/stat.h>	
#include	<dbug.h>
#include	<report.h>
#include	<setprv.h>
#include 	<tet_api.h>

#if	   TEST_XNFS

#include	<limits.h>
#include	<string.h>
#include	<xnfs.h>

#endif	/* TEST_XNFS */

#include	<sysdep.h>

#define	MODEANY		(mode_t)(S_IRWXU|S_IRWXG|S_IRWXO)
#define	NOK		2
#define	PATH_OK		3

extern	char	*errname();
extern	char	*newstr();

extern	int	vguid0, vggid0, vguid1, vggid1;

private int	normtests();
private void	ch_normt();
private int	privtests();
private void	gmodes();
private void	omodes();
private int	initgrps();
private void	cr_target();
private void	targdesc();

private	char	*nudir		= "do_acc-nu.d";
private	char	*nufile		= "do_acc-nu.d/do_acc-nu.t";
private	char	*ngdir		= "do_acc-ng.d";
private	char	*ngfile		= "do_acc-ng.d/do_acc-ng.t";
private	char	*nodir		= "do_acc-no.d";
private	char	*nofile		= "do_acc-no.d/do_acc-no.t";
private	char	**sgdir;
private	char	**sgfile;
private	char	*pudir		= "do_acc-pu.d";
private	char	*pufile		= "do_acc-pu.d/do_acc-pu.t";
private	char	*pgdir		= "do_acc-pg.d";
private	char	*pgfile		= "do_acc-pg.d/do_acc-pg.t";
private	char	*podir		= "do_acc-po.d";
private	char	*pofile		= "do_acc-po.d/do_acc-po.t";
#if	   TEST_XNFS
private	char	*audir		= "do_acc-au.d";
private	char	*aufile		= "do_acc-au.d/do_acc-au.t";
private	char	*aodir		= "do_acc-ao.d";
private	char	*aofile		= "do_acc-ao.d/do_acc-ao.t";
#endif	/* TEST_XNFS */

private	char	*tdir;
private	char	*tfile;
private uid_t	tuid;
private gid_t	tgid;
private mode_t	dirmode;
private mode_t	filemode;
private	char	descbuf[300];
private	char	*acc_desc[4] = { descbuf, descbuf+100, descbuf+200, NULL };
private int	targetisdir;
private	int	(*usrfunc)();
private gid_t	*suppgrps;
private int	ngroups = -1;
private int	ngroups_max = -1;
private int	userid;


#if	   TEST_XNFS

#define WAITTIME	(5 * SPEEDFACTOR)

private int	snames[] = { 1, 0 };

private char    srootdir[_POSIX_PATH_MAX+1];
private char    *nfs_mount_dir;
private mode_t	g_fileperm;
private int	g_waittime;

private void	nfs_privtests();
private void    cln_norm_uid0();
private void    cln_norm_uid1();
private void    cln_anon_root();
private void    cln_anon_uid1();
private void    cln_privs();

#endif	/* TEST_XNFS */


public int
do_access(filetype, fileperm, pathperm, func, waittime, spno)
int		filetype;
mode_t		fileperm;
mode_t		pathperm;
int		(*func)();
int		waittime;
int		spno;
{
	char 	*msg = NULL;
	int 	ret = 0;
	int	pathok = 0;
#if TEST_XNFS
	int	nvals;
	char	**strvals;
	extern char **xnfs_getsv_c();
#endif

	DBUG_LENTER("do_access", 2);
	DBUG_PRINTF("args", "do_access('%c', %#o, %#o, 0x%lx, %u, %d)",
		filetype, (int)fileperm, (int)pathperm, (long)func,
		waittime, spno);

	/* check for bad arguments */
	if (filetype != 'd' && filetype != 'f')
		msg = "bad filetype argument in call to do_access()";
	else if ((fileperm & ~S_IRWXU) != 0)
		msg = "bad fileperm argument in call to do_access()";
	else if ((pathperm & ~S_IRWXU) != 0)
		msg = "bad pathperm argument in call to do_access()";
	else if (setprv(PRV_SETID) != 0)
		msg = "setprv(PRV_SETID) failed in do_access()";
	else if (setprv(PRV_ACCESS) != 0)
		msg = "setprv(PRV_ACCESS) failed in do_access()";

	if (msg != NULL)
	{
		xx_rpt(DELETION);
		in_rpt("%s", msg);
		DBUG_PRINTF("return", "do_access = -1");
		DBUG_RETURN(-1);
	}
	else
		PATH_TRACE;

	targetisdir = (filetype == 'd');

	usrfunc = func;

	if(uids()!=0)
	{
		xx_rpt(DELETION);
		DBUG_PRINTF("return", "do_access = -1");
		DBUG_RETURN(-1);
	}
	else
		PATH_TRACE;

	if (ngroups < 0)
	{
		errno = 0;
		ngroups_max = sysconf(_SC_NGROUPS_MAX);
		if (ngroups_max < 0)
		{
			xx_rpt(DELETION);
			in_rpt("sysconf(_SC_NGROUPS_MAX) failed - errno %d (%s)",
				errno, errname(errno));
			DBUG_PRINTF("return", "do_access = -1");
			DBUG_RETURN(-1);
		}
		else 
			PATH_TRACE;

#if TEST_XNFS
		/* obtain server value for NGROUPS_MAX and set `ngroups' to
		   minimum of client and server values */

		nvals = 0;
		strvals = xnfs_getsv_c(&nvals, spno);
		if (strvals == NULL || nvals != 1)
		{
			xx_rpt(DELETION);
			in_rpt("xnfs_getsv_c() failed when getting NGROUPS_MAX from server");
			DBUG_PRINTF("return", "do_access = -1");
			DBUG_RETURN(-1);
		}
		else 
			PATH_TRACE;

		ngroups = atoi(*strvals);
		if (ngroups < 0)
		{
			xx_rpt(DELETION);
			in_rpt("could not obtain NGROUPS_MAX from server");
			DBUG_PRINTF("return", "do_access = -1");
			DBUG_RETURN(-1);
		}
		else 
			PATH_TRACE;

		if (ngroups > ngroups_max)
			ngroups = ngroups_max;
		
		/* maximum usable by NFS is 16 */
		if (ngroups > 16)
			ngroups = 16;
#else
		ngroups = ngroups_max;
		PATH_TRACE;
		PATH_TRACE;
#endif

		if (ngroups > 0)
		{
			globok = 0;
			if (initgrps() != 0)
				ret = -1;
			else
				PATH_FUNC_TRACE;
		}
		else
			PATH_TRACE;
	}
	else
	{
#if TEST_XNFS
		/* server always sends its NGROUPS_MAX, so this is
		   to satisfy the syncs */
		(void) xnfs_getsv_c(&nvals, spno);
#endif
		PATH_TRACE; PATH_TRACE;
		PATH_TRACE; PATH_TRACE;
	}

	if (ret == 0)
	{
		userid = vguid0;
		globok = 0;
		if (normtests(fileperm, pathperm, waittime, spno) != 0)
			ret = -1;
		else
			PATH_FUNC_TRACE;

#if	   TEST_XNFS

		userid = XNFS_MAPPED_UID;
		globok = 0;
		if (normtests(fileperm, pathperm, waittime, spno) != 0)
			ret = -1;
		else
			PATH_FUNC_TRACE;

		nfs_mount_dir = tet_getvar("XNFS_MOUNT_ROOT");
		if (nfs_mount_dir == NULL || *nfs_mount_dir == '\0')
		{
			xx_rpt(DELETION);
			in_rpt("parameter XNFS_MOUNT_ROOT is not set");
			ret = -1;
		}
		else
		{
        		char 	*tsetname;

			PATH_TRACE;

			(void) strcpy(srootdir, nfs_mount_dir);

			/* construct temp dir name from testset name */
			tsetname = strrchr(tet_pname, '/');

			if (tsetname == NULL)
				tsetname = tet_pname;
			else
				tsetname++;

			(void) strcat(srootdir, "/d.");
			if (strncmp(tsetname, "T.", (size_t)2) == 0)
			tsetname += 2;
			(void) strncat(srootdir, tsetname, (size_t)(_POSIX_NAME_MAX-2));

			DBUG_PRINTF("do_access", "srootdir = \"%s\"", srootdir);

			errno = 0;
			if (mkdir(srootdir, MODEANY) != 0 && errno != EEXIST)
			{
				xx_rpt(DELETION);
				in_rpt("mkdir failed for \"%s\"", srootdir);
				ret = -1;
			}
			else
				PATH_TRACE;
		}

		g_fileperm = fileperm;
		g_waittime = waittime;
		switch (tet_fork(nfs_privtests, NULLFN, 2*waittime, NOK|PATH_OK))
		{
		case PATH_OK:
			PATH_TRACE;
			break;

		case 0:
			break;

		default:
			ret = -1;
			break;
		}

#else	/* TEST_XNFS */

		PATH_TRACE; PATH_TRACE; PATH_TRACE;

		globok = 0;
		if (privtests(fileperm, waittime) != 0)
			ret = -1;
		else
			PATH_FUNC_TRACE;

#endif	/* TEST_XNFS */

	}

	PATH_FUNC_RPT(11);

	DBUG_PRINTF("return", "do_access = %d", ret);
	DBUG_RETURN(ret);
}


private int
normtests(fileperm, pathperm, waittime, spno)
mode_t	fileperm;
mode_t	pathperm;
int	waittime;
int	spno;
{
	int 	i;
	int	retval;
	int	pathok = 0;

	/* create test files */

	DBUG_LENTER("normtests", 2);

	/* owner */
	dirmode = pathperm;
	filemode = fileperm;
#if TEST_XNFS
	if(userid == XNFS_MAPPED_UID)
	{
		tdir = audir; 
		tfile = aufile; 

		/* wait for server to create file owned by anonymous uid */
		if (tet_sync(spno+SP_GETPUT_END+1, snames, SYNC_TIMEOUT) < 0)
		{
			xx_rpt(DELETION);
			in_rpt("tet_sync() failed in do_access()");
			DBUG_RETURN(-1);
		}
		else
			PATH_TRACE;
	}
	else
#endif /* TEST_XNFS */
	{
		tdir = nudir; 
		tfile = nufile; 
		tuid = userid; tgid = vggid0;
		retval = cppair(cr_target, NULLFN, waittime, NOK);
		if ( retval == PATH_OK)
		{
			PATH_TRACE;
		}
		else
		{
			if (retval != OK)
				DBUG_RETURN(-1);
		}
	}

	/* effective group */

#if TEST_XNFS
	/* not tested when user ID is mapped */
	if(userid != XNFS_MAPPED_UID)
#endif
	{
		gmodes(pathperm, fileperm, &dirmode, &filemode);

		tdir = ngdir; tfile = ngfile; 
		tuid = vguid1; tgid = vggid0;
		retval = cppair(cr_target, NULLFN, waittime, NOK);
		if ( retval == PATH_OK)
		{
			PATH_TRACE;
		}
		else
		{
			if (retval != OK)
				DBUG_RETURN(-1);
		}
	}
#if TEST_XNFS
	else
		PATH_TRACE;
#endif

	/* supplementary groups */

#if TEST_XNFS
	/* not tested when user ID is mapped */
	if(userid != XNFS_MAPPED_UID)
#endif
	{
		for (i = 0; i < ngroups; i++)
		{
			/* dirmode and filemode same as for effective group */
			tdir = sgdir[i]; 
			tfile = sgfile[i];
			tuid = vguid1; tgid = suppgrps[i];
			retval = cppair(cr_target, NULLFN, waittime, NOK);
			if ( retval == PATH_OK)
			{
				PATH_TRACE;
			}
			else
			{
				if (retval != OK)
					DBUG_RETURN(-1);
			}
		}
	}
#if TEST_XNFS
	else
	{
		for (i = 0; i < ngroups; i++)
			PATH_TRACE;
	}
#endif

	/* other */

	omodes(pathperm, fileperm, &dirmode, &filemode);
#if TEST_XNFS
	if(userid == XNFS_MAPPED_UID)
	{
		tdir = aodir; tfile = aofile; 
	}
	else
#endif	/* TEST_XNFS */
	{
		tdir = nodir; tfile = nofile; 
	}
	tuid = vguid1; tgid = vggid1;
	retval = cppair(cr_target, NULLFN, waittime, NOK);
	if ( retval == PATH_OK)
	{
		PATH_TRACE;
	}
	else
	{
		if (retval != OK)
			DBUG_RETURN(-1);
	}

	/* Execute the tests */

	dirmode = pathperm;
	filemode = fileperm;
	retval = cppair(ch_normt, NULLFN, waittime + (ngroups*waittime)/10, NOK);
	if ( retval == PATH_OK)
	{
		PATH_TRACE;
	}
	else
	{
		if (retval != OK)
			DBUG_RETURN(-1);
	}

	PATH_FUNC_RPT(2+ngroups+2);
	
	DBUG_RETURN(0);
}


private void
ch_normt()
{
	mode_t	pathperm;
	mode_t	fileperm;
	int	i;
	int	excode = OK;
	int	pathok = 0;
	char 	*filename;
	char 	*dirname;

	DBUG_LENTER("ch_normt" , 2);

	pathperm = dirmode;
	fileperm = filemode;

	(void) setgid((gid_t)vggid0);
	if (getegid() != vggid0)
	{
		xx_rpt(DELETION);
		in_rpt("setgid(%d) did not set effective GID to %d",
			vggid0, vggid0);
		DBUG_EXIT(NOK);
	}
	else
		PATH_TRACE;

	(void) setuid((uid_t) userid);
	if (geteuid() != userid)
	{
		xx_rpt(DELETION);
		in_rpt("setuid(%d) did not set effective UID to %d",
			userid, userid);
		DBUG_EXIT(NOK);
	}
	else
		PATH_TRACE;

	if (unsetprv(PRV_ACCESS) != 0)
	{
		xx_rpt(DELETION);
		in_rpt("unsetprv(PRV_ACCESS) failed in do_access()");
		DBUG_EXIT(NOK);
	}
	else
		PATH_TRACE;

	if (unsetprv(PRV_SETID) != 0)
	{
		xx_rpt(DELETION);
		in_rpt("unsetprv(PRV_SETID) failed in do_access()");
		DBUG_EXIT(NOK);
	}
	else
		PATH_TRACE;

	(void) sprintf(acc_desc[0],
		"non-privileged process with effective uid = %d, gid = %d",
		(int) geteuid(), (int) getegid());


	/* owner */
#if TEST_XNFS
	if(userid == XNFS_MAPPED_UID)
	{
		filename = aufile;
		dirname = audir;
	}
	else
#endif	/* TEST_XNFS */
	{
		filename = nufile;
		dirname = nudir;
	}

	targdesc(dirname,filename,userid,vggid0,(int) pathperm,(int) fileperm);

	globok = 0;
	if ((*usrfunc)(filename, acc_desc) != 0)
		excode = NOK;
	else
		PATH_FUNC_TRACE;


	/* effective group */

#if TEST_XNFS
	/* not tested when user ID is mapped */
	if(userid != XNFS_MAPPED_UID)
#endif
	{
		gmodes(pathperm, fileperm, &dirmode, &filemode);

		filename = ngfile;
		dirname = ngdir;

		targdesc(dirname,filename,vguid1,vggid0,(int) dirmode,(int) filemode);

		globok = 0;
		if ((*usrfunc)(filename, acc_desc) != 0)
			excode = NOK;
		else
			PATH_FUNC_TRACE;
	}
#if TEST_XNFS
	else
		PATH_TRACE;
#endif


	/* supplementary groups */

#if TEST_XNFS
	/* not tested when user ID is mapped */
	if(userid != XNFS_MAPPED_UID)
#endif
	{
		for (i = 0; i < ngroups; i++)
		{
			filename = sgfile[i];
			dirname = sgdir[i];

			targdesc(dirname, filename, vguid1, (int)suppgrps[i],
				 (int)dirmode, (int)filemode);

			globok = 0;
			if ((*usrfunc)(filename, acc_desc) != 0)
				excode = NOK;
			else
				PATH_FUNC_TRACE;
		}
	}
#if TEST_XNFS
	else
	{
		for (i = 0; i < ngroups; i++)
			PATH_TRACE;
	}
#endif

	/* other */

	omodes(pathperm, fileperm, &dirmode, &filemode);

#if TEST_XNFS
	if(userid == XNFS_MAPPED_UID)
	{
		filename = aofile;
		dirname = aodir;
	}
	else
#endif	/* TEST_XNFS */
	{
		filename = nofile;
		dirname = nodir;
	}

	targdesc(dirname,filename,vguid1,vggid1,(int) dirmode,(int) filemode);

	globok = 0;
	if ((*usrfunc)(filename, acc_desc) != 0)
		excode = NOK;
	else
		PATH_FUNC_TRACE;

	if (pathok == 6+ngroups+1)
		DBUG_EXIT(PATH_OK);

	DBUG_EXIT(excode);
}

#if	   TEST_XNFS

private void
nfs_privtests()
{
	int 	ret = 0;
	int	pathok = 0;

	DBUG_LENTER("nfs_privtests",2);

	if (chdir(srootdir) != 0)
	{
		xx_rpt(DELETION);
		in_rpt("chdir failed for \"%s\"", srootdir);
		DBUG_EXIT(NOK);
	}
	else
		PATH_TRACE;

	/*
	 *  Call the function that does the privileged tests
	 */
	globok = 0;
	ret = privtests(g_fileperm, g_waittime);
	if (ret == 0)
	{
		PATH_FUNC_TRACE;
		if (pathok == 2)
			DBUG_EXIT(PATH_OK);
		else
			DBUG_EXIT(0);
	}
	
	DBUG_EXIT(NOK);
}

#endif	/* TEST_XNFS */

private int
privtests(fileperm, waittime)
mode_t fileperm;
int waittime;
{
	int 	ret = 0;
	int	retval;
	int	pathok = 0;

	DBUG_LENTER("privtests",2);


	if (setprv(PRV_ACCESS) != 0)
	{
		xx_rpt(DELETION);
		in_rpt("setprv(PRV_ACCESS) failed in do_access()");
		DBUG_RETURN(-1);
	}
	else
		PATH_TRACE;

	(void) sprintf(acc_desc[0],
		"privileged process with effective uid = %d, gid = %d",
		(int)geteuid(), (int)getegid());

	dirmode = 0;
	filemode = 0;

	/* owner */
	/* Privileged execute on non-directory requires any x bit to be set,
	   so set it in an "unused" part of the mode */
	if (!targetisdir && (fileperm & S_IXUSR) != 0)
		filemode = S_IXOTH;
	tdir = pudir; tfile = pufile; tuid = geteuid(); tgid = getegid();
	retval = cppair(cr_target, NULLFN, waittime, NOK);
	if ( retval == PATH_OK)
	{
		PATH_TRACE;
	}
	else
	{
		if (retval != OK)
			DBUG_RETURN(-1);
	}

	targdesc(pudir, pufile, (int)geteuid(), (int)getegid(),
						0, (int)filemode);
	globok = 0;
	if ((*usrfunc)(pufile, acc_desc) != 0)
		ret = -1;
	else
		PATH_FUNC_TRACE;

	/* group */
	/* Privileged execute on non-directory requires any x bit to be set,
	   so set it in an "unused" part of the mode */
	if (!targetisdir && (fileperm & S_IXUSR) != 0)
		filemode = S_IXUSR;
	tdir = pgdir; tfile = pgfile; tuid = vguid1; tgid = getegid();
	retval = cppair(cr_target, NULLFN, waittime, NOK);
	if ( retval == PATH_OK)
	{
		PATH_TRACE;
	}
	else
	{
		if (retval != OK)
			DBUG_RETURN(-1);
	}
	targdesc(pgdir, pgfile, vguid1, (int)getegid(), 0, (int)filemode);
	globok = 0;
	if ((*usrfunc)(pgfile, acc_desc) != 0)
		ret = -1;
	else
		PATH_FUNC_TRACE;

	/* other */
	/* Privileged execute on non-directory requires any x bit to be set,
	   so set it in an "unused" part of the mode */
	if (!targetisdir && (fileperm & S_IXUSR) != 0)
		filemode = S_IXGRP;
	tdir = podir; tfile = pofile; tuid = vguid1; tgid = vggid1;
	retval = cppair(cr_target, NULLFN, waittime, NOK);
	if ( retval == PATH_OK)
	{
		PATH_TRACE;
	}
	else
	{
		if (retval != OK)
			DBUG_RETURN(-1);
	}

	targdesc(podir, pofile, vguid1, vggid1, 0, (int)filemode);
	globok = 0;
	if ((*usrfunc)(pofile, acc_desc) != 0)
		ret = -1;
	else
		PATH_FUNC_TRACE;

	PATH_FUNC_RPT(7);
	
	DBUG_RETURN(ret);
}


private void
gmodes(pathperm, fileperm, dmodep, fmodep)
mode_t pathperm, fileperm, *dmodep, *fmodep;
{
	/*
	 * set modes for group access in *dmodep, *fmodep equivalent
	 * to the owner access modes given in pathperm, fileperm
	 */

	*dmodep = 0;
	if ((pathperm & S_IRUSR) != 0)
		*dmodep |= S_IRGRP;
	if ((pathperm & S_IWUSR) != 0)
		*dmodep |= S_IWGRP;
	if ((pathperm & S_IXUSR) != 0)
		*dmodep |= S_IXGRP;

	*fmodep = 0;
	if ((fileperm & S_IRUSR) != 0)
		*fmodep |= S_IRGRP;
	if ((fileperm & S_IWUSR) != 0)
		*fmodep |= S_IWGRP;
	if ((fileperm & S_IXUSR) != 0)
		*fmodep |= S_IXGRP;
}


private void
omodes(pathperm, fileperm, dmodep, fmodep)
mode_t pathperm, fileperm, *dmodep, *fmodep;
{
	/*
	 * set modes for other access in *dmodep, *fmodep equivalent
	 * to the owner access modes given in pathperm, fileperm
	 */

	*dmodep = 0;
	if ((pathperm & S_IRUSR) != 0)
		*dmodep |= S_IROTH;
	if ((pathperm & S_IWUSR) != 0)
		*dmodep |= S_IWOTH;
	if ((pathperm & S_IXUSR) != 0)
		*dmodep |= S_IXOTH;

	*fmodep = 0;
	if ((fileperm & S_IRUSR) != 0)
		*fmodep |= S_IROTH;
	if ((fileperm & S_IWUSR) != 0)
		*fmodep |= S_IWOTH;
	if ((fileperm & S_IXUSR) != 0)
		*fmodep |= S_IXOTH;
}


private int
initgrps()
{
	/* Initialise supplementary group lists */

	int 	i;
	char 	tmpbuf[80];
	int	pathok = 0;

	DBUG_LENTER("initgrps",2);

	suppgrps = (gid_t *)malloc((size_t)(sizeof(gid_t) * ngroups_max));
	sgdir = (char **)malloc((size_t)(sizeof(char *) * ngroups));
	sgfile = (char **)malloc((size_t)(sizeof(char *) * ngroups));

	if (suppgrps == NULL || sgdir == NULL || sgfile == NULL)
	{
		xx_rpt(DELETION);
		in_rpt("could not malloc() %d bytes",
		    (int)((sizeof(gid_t) + 2 * sizeof(char *))
				* ngroups));
		ngroups = -1;
		DBUG_RETURN(-1);
	}
	else
		PATH_TRACE;

	if (getgroups(ngroups_max, suppgrps) != ngroups)
	{
		xx_rpt(DELETION);
#if TEST_XNFS
		in_rpt("getgroups() did not return minimum of client and server NGROUPS_MAX values and 16");
#else
		in_rpt("getgroups() did not return NGROUPS_MAX");
#endif
		in_rpt("- check supplementary groups for user vsx0");
		free((void *)suppgrps);
		free((void *)sgdir);
		free((void *)sgfile);
		ngroups = -1;
		DBUG_RETURN(-1);
	}
	else
		PATH_TRACE;

	for (i = 0; i < ngroups; i++)
	{
		if (i==0)
			PATH_TRACE;
		if (suppgrps[i] == vggid1)
		{
			xx_rpt(DELETION);
			in_rpt("group vsxg1 must not be in the supplementary group list for user vsx0");
			DBUG_RETURN(-1);
		}
		sgdir[i] = sgfile[i] = NULL;
	}

	for (i = 0; i < ngroups; i++)
	{
		if (i==0)
			PATH_TRACE;

		(void) sprintf(tmpbuf, "do_acc-sg.d%d", i);
		sgdir[i] = newstr(tmpbuf);

		(void) sprintf(tmpbuf, "do_acc-sg.d%d/do_acc-sg.t%d", i, i);
		sgfile[i] = newstr(tmpbuf);

		if (sgdir[i] == NULL || sgfile[i] == NULL)
		{
			xx_rpt(DELETION);
			in_rpt("malloc() failure");
			DBUG_RETURN(-1);
		}
	}

	PATH_FUNC_RPT(4);

	DBUG_RETURN(0);
}


private void
cr_target()
{
	int 	fd;
	int	pathok = 0;


	DBUG_LENTER("cr_target",2);

	/* Remove left-over ones here while we are privileged */
	(void) remove(tfile);
	(void) rmdir(tdir);

	(void) setgid(tgid);
	if (getegid() != tgid)
	{
		xx_rpt(DELETION);
		in_rpt("setgid(%d) did not set effective GID to %d",
			(int)tgid, (int)tgid);
		DBUG_EXIT(NOK);
	}
	else
		PATH_TRACE;

	(void) setuid((uid_t) tuid);
	if (geteuid() != tuid)
	{
		xx_rpt(DELETION);
		in_rpt("setuid(%d) did not set effective UID to %d",
			(int)tuid, (int)tuid);
		DBUG_EXIT(NOK);
	}
	else
		PATH_TRACE;

	/*
	 *  Create the directory with MOEDANY so that we can create the file
	 *  inside it OK. We chmod() the directory after the file is created.
	 */
	if (mkdir(tdir,MODEANY) != 0)
	{
		xx_rpt(DELETION);
		in_rpt("could not create directory \"%s\" - errno %d (%s)",
			tdir, errno, errname(errno));
		DBUG_EXIT(NOK);
	}
	else
		PATH_TRACE;

	/* Make sure ownership is right */
	if (chown(tdir, tuid, tgid) != 0)
	{
		xx_rpt(DELETION);
		in_rpt("chown() failed on directory \"%s\" - errno %d (%s)",
			tdir, errno, errname(errno));
		DBUG_EXIT(NOK);
	}
	else
		PATH_TRACE;

	if (targetisdir)
	{
		if (mkdir(tfile, filemode) == SYSERROR)
		{
			xx_rpt(DELETION);
			in_rpt("could not create directory \"%s\" - errno %d (%s)",
				tfile, errno, errname(errno));
			DBUG_EXIT(NOK);
		}
		else
			PATH_TRACE;
	}
	else
	{
		if ((fd = creat(tfile, filemode)) == SYSERROR)
		{
			xx_rpt(DELETION);
			in_rpt("could not create file \"%s\" - errno %d (%s)",
				tfile, errno, errname(errno));
			DBUG_EXIT(NOK);
		}
		else
			PATH_TRACE;

		(void) close(fd);
	}

	if (chmod(tdir, dirmode) == SYSERROR)
	{
		xx_rpt(DELETION);
		in_rpt("chmod(\"%s\", %#o) failed - errno %d (%s)",
			tdir, (int)dirmode, errno, errname(errno));
		DBUG_EXIT(NOK);
	}
	else
		PATH_TRACE;

	if (pathok==6)
		DBUG_EXIT(PATH_OK);
		
	DBUG_EXIT(OK);
}


private void
targdesc(dir, file, owner, group, dmode, fmode)
char	*dir, *file;
int	owner, group, dmode, fmode;
{
	(void) sprintf(acc_desc[1], "owner/group of %s and %s: %d/%d",
		dir, file, owner, group);

	(void) sprintf(acc_desc[2], "mode of %s: %#o, mode of %s: %#o",
		dir, dmode, file, fmode);
}


#if	TEST_XNFS

public void
cln_access()
{
	DBUG_LENTER("cln_access", 2);

	/* Remove files with user id = vguid0 */
	if(tet_fork(cln_norm_uid0, NULLFN, WAITTIME, OK) == -1)
	{
		DBUG_PRINTF("cln_access", 
			"Removal of normal user (vguid0=%d) files failed", 
				vguid0);
	}

	/* Remove files created with user id = vguid1 */
	if(tet_fork(cln_norm_uid1, NULLFN, WAITTIME, OK) == -1)
	{
		DBUG_PRINTF("cln_access", 
			"Removal of normal user (vguid1=%d) files failed", 
				vguid1);
	}

	/* Remove files created with user id = root for ANON_MAPPING */
	if(tet_fork(cln_anon_root, NULLFN, WAITTIME, OK) == -1)
	{
		DBUG_PRINTF("cln_access", 
			"Removal of files created with id=root and ANON_MAPPING set failed");
	}

	/* Remove files created with user id = vguid1 for ANON_MAPPING */
	if(tet_fork(cln_anon_uid1, NULLFN, WAITTIME, OK) == -1)
	{
		DBUG_PRINTF("cln_access",
			"Removal of files created with id=vguid1 (%d) and ANON_MAPPING set failed", vguid1);
	}


	/*
	 *  Remove priviledged files
	 */

	if(tet_fork(cln_privs, NULLFN, WAITTIME, OK) == -1)
	{
		DBUG_PRINTF("cln_access",
			"Removal of privileged files failed.");
	}


	/* Remove the test directory created on the set root file system */
	(void) rmdir(srootdir);

	DBUG_VOID_RETURN;
}


private void
cln_norm_uid0()
{
	DBUG_LENTER("cln_norm_uid0",2);
	
	(void) setuid((uid_t)vguid0);
	if (geteuid() != vguid0)
		DBUG_EXIT(NOK);

	/* Chmod directory then file */
	(void) chmod(nudir, MODEANY);
	(void) chmod(nufile, MODEANY);

	/* Remove file then directory */
	(void) remove(nufile);
	(void) rmdir(nudir);

	DBUG_EXIT(OK);
}


private void
cln_norm_uid1()
{
	int i;

	DBUG_LENTER("cln_norm_uid1",2);
	
	(void) setuid((uid_t)vguid1);
	if (geteuid() != vguid1)
		DBUG_EXIT(NOK);

	/* Chmod directory then file */
	(void) chmod(ngdir, MODEANY);
	(void) chmod(ngfile, MODEANY);

	/* Remove file then directory */
	(void) remove(ngfile);
	(void) rmdir(ngdir);

	/* Chmod directory then file */
	(void) chmod(nodir, MODEANY);
	(void) chmod(nofile, MODEANY);

	/* Remove file then directory */
	(void) remove(nofile);
	(void) rmdir(nodir);

	/* Chmod directory then file */
	for (i = 0; i < ngroups; i++)
	{
		if (sgdir[i] != NULL)
			(void) chmod(sgdir[i], MODEANY);
		if (sgfile[i] != NULL)
			(void) chmod(sgfile[i], MODEANY);
	}

	/* Remove file then directory */
	for (i = 0; i < ngroups; i++)
	{
		if (sgfile[i] != NULL)
			(void) remove(sgfile[i]);
		if (sgdir[i] != NULL)
			(void) rmdir(sgdir[i]);
	}


	DBUG_EXIT(OK);
}


private void
cln_anon_root()
{
	DBUG_LENTER("cln_anon_root",2);
	
	(void) setuid((uid_t)XNFS_MAPPED_UID);
	if (geteuid() != XNFS_MAPPED_UID)
		DBUG_EXIT(NOK);

	/* Chmod directory then file */
	(void) chmod(audir, MODEANY);
	(void) chmod(aufile, MODEANY);

	/* Remove file then directory */
	(void) remove(aufile);
	(void) rmdir(audir);

	DBUG_EXIT(OK);
}


private void
cln_anon_uid1()
{
	DBUG_LENTER("cln_anon_uid1",2);
	
	(void) setuid((uid_t)vguid1);
	if (geteuid() != vguid1)
		DBUG_EXIT(NOK);

	/* Chmod directory then file */
	(void) chmod(aodir, MODEANY);
	(void) chmod(aofile, MODEANY);

	/* Remove file then directory */
	(void) remove(aofile);
	(void) rmdir(aodir);

	DBUG_EXIT(OK);
}


private void
cln_privs()
{
	DBUG_LENTER("cln_privs", 2);

	/* Change to the directory that contains the priviledged files */
        if (chdir(srootdir) != 0)
        {
		DBUG_PRINTF("cln_access", 
			"could not change directory to (%s)", srootdir);
                DBUG_EXIT(NOK);
        }

	/* Remove all the files */
	(void) remove(pufile);
	(void) rmdir(pudir);
	(void) remove(pgfile);
	(void) rmdir(pgdir);
	(void) remove(pofile);
	(void) rmdir(podir);

	DBUG_EXIT(OK);
}


#else	/* TEST_XNFS */

public void
cln_access()
{
	int i;

	DBUG_LENTER("cln_access", 2);

	(void) remove(nufile);
	(void) rmdir(nudir);
	(void) remove(ngfile);
	(void) rmdir(ngdir);
	(void) remove(nofile);
	(void) rmdir(nodir);
	for (i = 0; i < ngroups; i++)
	{
		if (sgfile[i] != NULL)
			(void) remove(sgfile[i]);
		if (sgdir[i] != NULL)
			(void) rmdir(sgdir[i]);
	}

	/* Remove priviledged files */
	(void) remove(pufile);
	(void) rmdir(pudir);
	(void) remove(pgfile);
	(void) rmdir(pgdir);
	(void) remove(pofile);
	(void) rmdir(podir);

	DBUG_VOID_RETURN;
}

#endif	/* TEST_XNFS */
