/*
*      SCCS:   @(#)common/tsetlib(do_feio.c)	4.1 (01/21/97)  VSXgen release 1.4
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/


/***********************************************************************

NAME:		do_feio.c - arrange for stream EIO error

PROJECT:	VSX (X/OPEN Validation Suite)

MODULE CONTAINS:
	do_feio()
	do_fnoeio()
SYNOPSIS:
	void do_feio(type, reqfd, io, rpt)
		char	*type;
		int	reqfd;
		int	(*io)();
		void	(*rpt)();

	void do_fnoeio(type, reqfd, io, rpt)
		char	*type;
		int	reqfd;
		int	(*io)();
		void	(*rpt)();

DESCRIPTION:
	Do_feio() sets up a stream pointer such that the next operation
	of the given "type" will generate errno = EIO.  Valid "type"s are:

	 	"rb"	read with SIGTTIN blocked
	 	"ri"	read with SIGTTIN ignored
	 	"ro"	read from orphaned process group
	 	"w"	write (from orphaned pgrp)
	 	"f"	flush (from orphaned pgrp)

	The supplied function "io" is called to carry out this I/O operation.
	It is passed the (FILE *) stream pointer to be used.

	If "reqfd" is >= 0 it indicates the fd which should underly the
	stream.  If this is 0, 1 or 2 then the stream set up by do_feio()
	will be stdin/out/err respectively.

	Do_feio() passes the return value of (*io)(fp), and the errno value
	after the call to the supplied function "rpt" for reporting.

	N.B. as "io" and "rpt" are called in a child process, the calling
	process should not set any global/static variables in these functions
	and examine their values after the call to do_feio().

	Do_fnoeio() is identical to do_feio() except that it uses a
	non-controlling terminal, and therefore the operation should not
	generate errno = EIO.

CONDITIONAL COMPILE PARAMS:
SIDE EFFECTS:
RELATED DOCUMENTS:
SOURCE:
AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	30 May 1989
MODIFICATIONS:
		Geoff Clare, 1 Aug 1989
			Writing operations should not block or ignore SIGTTOU
			in order to get EIO.  Hence replace options "wb",
			"wi", "fb" and "fi" with just "w" and "f".
		
		Geoff Clare, 22 Aug 1989
			Create an orphaned grandchild instead of child, so
			the session leader is still alive.
		
		Geoff Clare, 23 Aug 1989
			Use fputc() instead of fputs() to put data in the
			buffer, because some versions of fputs() flush it.
		
		Geoff Clare, 11 Sep 1989
			More thorough signal initialisation in ch_setup().
		
		Geoff Clare, 14 Feb 1990
			Use termios_prep() instead of tty_prep().

		J.A.Nave	15/05/90
		Added resilience testing

		Geoff Clare, 7 June 1990
			UNSUPPORTED result if termios_prep() returns 1.

		Anthony Lee, 11 July 1990
			Retrofitted resillience macros.

		Stuart Boutell, 4 Dec 1990
			Added tet_setcontext() and tet_setblock() calls
			where fork() is called
		
		Geoff Clare, 14 Oct 1992
			Changes for XNFS.

		Geoff Clare, 10 Feb 1995
			Added do_fnoeio().

************************************************************************/

#include <std.h>

#if !TEST_XNFS
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <errno.h>
#include <signal.h>
#include <termios.h>
#include <dbug.h>
#include <report.h>
#include <timeout.h>
#include <tet_api.h>
#include <sysdep.h>

#define WAITTIME	(5 * SPEEDFACTOR)
#define NAP		2

extern	void	setsigs();
extern	void	(*vsignal())();
extern	char	*signame();
extern	char	*errname();
extern	int	globok;

private void	ch_eio();
private void	gch_eio();
private void	ch_setup();
private void	do_io();
private void	sig_orphan();

private	int		tty_fildes, loop_fildes;
private	struct termios	tty_termios, *tty_termios_p = &tty_termios;
private	struct termios	loop_termios, *loop_termios_p = &loop_termios;
private int		orphaned, signum, required_fd, cntrl_tty;
private char		*testtype;
private int		(*iofunc)();
private void		(*rptfunc)();


public void
do_feio(type, reqfd, io, rpt)
char	*type;		/* "rb", "ri", "ro", "w" or "f"		*/
int	reqfd;		/* required fd (0,1,2), else -1		*/
int	(*io)();	/* function to perform I/O of 'type'	*/
void	(*rpt)();	/* function to report results		*/
{
	DBUG_LENTER("do_feio", 2);
	DBUG_PRINTF("args", "do_feio(%s, %d, 0x%lx, 0x%lx)",
		type, reqfd, (long)io, (long)rpt);

	errno = 0;
	if (sysconf(_SC_JOB_CONTROL) == -1)
	{
		if (errno == 0)
		{
			xx_rpt(UNSUPPORTED);
			in_rpt("Job control is not supported");
		}
		else
		{
			xx_rpt(DELETION);
			in_rpt("sysconf(_SC_JOB_CONTROL) failed - errno %d (%s)",
				errno, errname(errno));
		}
		DBUG_VOID_RETURN;
	}

	cntrl_tty = TRUE;
	orphaned = (type[0] != 'r' || type[1] == 'o');
	if (type[0] == 'r')
		signum = SIGTTIN;
	else
		signum = SIGTTOU;
	testtype = type;
	required_fd = reqfd;
	iofunc = io;
	rptfunc = rpt;

	(void) cppair(ch_eio, NULLFN, 2 * WAITTIME, 0);

	DBUG_VOID_RETURN;
}

public void
do_fnoeio(type, reqfd, io, rpt)
char	*type;
int	reqfd;
int	(*io)();
void	(*rpt)();
{
	DBUG_LENTER("do_fnoeio", 2);
	DBUG_PRINTF("args", "do_fnoeio(%s, %d, 0x%lx, 0x%lx)",
		type, reqfd, (long)io, (long)rpt);

	cntrl_tty = FALSE;
	orphaned = (type[0] != 'r' || type[1] == 'o');
	if (type[0] == 'r')
		signum = SIGTTIN;
	else
		signum = SIGTTOU;
	testtype = type;
	required_fd = reqfd;
	iofunc = io;
	rptfunc = rpt;

	(void) cppair(ch_eio, NULLFN, 2 * WAITTIME, 0);

	DBUG_VOID_RETURN;
}

private void
ch_eio()
{
	/*
	 * Open and initialize TERMIOS_TTY and TERMIOS_LOOP.
	 * The loop terminal is needed to ensure disconnect behaviour
	 * cannot occur instead of the desired result.
	 */

	int	ret, pathok = 0;

	DBUG_LENTER("ch_eio", 2);

	if (cntrl_tty)
	{
		ret = termios_prep(&tty_fildes, tty_termios_p,
				&loop_fildes, loop_termios_p);
	}
	else
	{
		(void) setsid();
		ret = terms_prep(O_RDWR|O_NOCTTY, &tty_fildes, tty_termios_p,
				O_RDWR, &loop_fildes, loop_termios_p);
	}

	switch (ret)
	{
	case 0:
		PATH_TRACE;
		break;
	case 1:
		xx_rpt(UNSUPPORTED);
		in_rpt("VSX_TERMIOS_TTY/LOOP is set to \"unsup\"");
		DBUG_VOID_RETURN;
	default:
		xx_rpt(DELETION);
		if (alrm_flag > 0)
			in_rpt("open/initialise of VSX_TERMIOS_TTY and VSX_TERMIOS_LOOP timed out");
		else
			in_rpt("could not open/initialise VSX_TERMIOS_TTY and VSX_TERMIOS_LOOP");
		DBUG_VOID_RETURN;
	}

	tty_termios_p->c_lflag |= TOSTOP;

	if (tcsetattr(tty_fildes, TCSANOW, tty_termios_p) == SYSERROR)
	{
		xx_rpt(DELETION);
		in_rpt("could not set TOSTOP on VSX_TERMIOS_TTY");
		DBUG_VOID_RETURN;
	}

	if (orphaned)
	{
		(void) vsignal(SIGHUP, SIG_IGN);
		(void) cppair(gch_eio, NULLFN, WAITTIME, 0);
		(void) sleep(WAITTIME);	/* allow orphan to report */
	}
	else
		gch_eio();

	DBUG_VOID_RETURN;
}

private void
gch_eio()
{
	/* Create background process to do the I/O */

	pid_t pid;

	switch (pid = fork())
	{
	case SYSERROR :
		xx_rpt(DELETION);
		in_rpt("fork() failed - errno %d (%s)", errno, errname(errno));
		break;
	
	case 0 :
		tet_setcontext();
		setsigs(sig_orphan);
		ch_setup();
		if (orphaned)
			(void) sleep(NAP);
		do_io();
		_exit(0);	/* no fclose */
		break;
	
	default :
		tet_setblock();
		if (!orphaned)
		{
			errno = 0;
			if (waitpid(pid, (int *)0, 0) != pid)
			{
				xx_rpt(DELETION);
				in_rpt("waitpid() failed - errno %d (%s)",
					errno, errname(errno));
			}
		}
		break;
	}

	tet_setblock();
        (void) close(tty_fildes);

	return;
}

private void
ch_setup()
{
	/* set up background child process */

	struct sigaction sig;
	sigset_t set;

	int	pathok = 0;

	DBUG_LENTER("ch_setup", 2);

	if (setpgid((pid_t)0, (pid_t)0) == SYSERROR)
	{
		xx_rpt(DELETION);
		in_rpt("setpgid(0, 0) failed");
		DBUG_VOID_RETURN;
	}
	else
		PATH_TRACE;

	if (testtype[1] == 'i')
	{
		sig.sa_handler = SIG_IGN;
		sig.sa_flags = 0;
		(void) sigemptyset(&sig.sa_mask);
		if (sigaction(signum, &sig, NULLSA) == SYSERROR)
		{
			xx_rpt(DELETION);
			in_rpt("sigaction() failed when setting %s to be ignored",
				signame(signum));
			DBUG_VOID_RETURN;
		}
		else
			PATH_TRACE;
	}
	else
	{
		/* This won't have been done by setsigs() if signum
		   was already being ignored */
		sig.sa_handler = sig_orphan;
		sig.sa_flags = 0;
		(void) sigemptyset(&sig.sa_mask);
		if (sigaction(signum, &sig, NULLSA) == SYSERROR)
		{
			xx_rpt(DELETION);
			in_rpt("sigaction() failed when setting catching function for %s",
				signame(signum));
			DBUG_VOID_RETURN;
		}
		else
			PATH_TRACE;
	}

	if (testtype[1] == 'b')
	{
		(void) sigemptyset(&set);
		(void) sigaddset(&set, signum);
		if (sigprocmask(SIG_BLOCK, &set, (sigset_t *)0) == SYSERROR)
		{
			xx_rpt(DELETION);
			in_rpt("sigprocmask() failed when blocking %s",
				signame(signum));
			DBUG_VOID_RETURN;
		}
		else
			PATH_TRACE;
	}
	else
	{
		(void) sigemptyset(&set);
		(void) sigaddset(&set, signum);
		if (sigprocmask(SIG_UNBLOCK, &set, (sigset_t *)0) == SYSERROR)
		{
			xx_rpt(DELETION);
			in_rpt("sigprocmask() failed when unblocking %s",
				signame(signum));
			DBUG_VOID_RETURN;
		}
		else
			PATH_TRACE;
	}

	globok = 0;
	PATH_FUNC_RPT(3); 

	DBUG_VOID_RETURN;
}

private void
do_io()
{
	/* Set up stream to the terminal and call function to do the I/O */

	FILE	*tty_fp;
	int	savfd, ret, err;

	static char	buf[BUFSIZ];
	int	pathok = 0;


	DBUG_LENTER("do_io", 2);

	/* If a particular fd is required, make it a dup of tty_fildes */

	if (required_fd != -1)
	{
		savfd = dup(required_fd);
		if (savfd == SYSERROR)
		{
			xx_rpt(DELETION);
			in_rpt("dup() failed - errno %d (%s)",
				errno, errname(errno));
			DBUG_VOID_RETURN;
		}
		else
			PATH_TRACE;
		(void) close(required_fd);
		if (dup(tty_fildes) != required_fd)
		{
			(void) close(required_fd);
			(void) dup(savfd);
			(void) close(savfd);
			xx_rpt(DELETION);
			in_rpt("dup() did not return fd just closed");
			DBUG_VOID_RETURN;
		}
		else
			PATH_TRACE;

		switch(required_fd)
		{
		case STDIN_FILENO  : tty_fp = stdin;  break;
		case STDOUT_FILENO : tty_fp = stdout; break;
		case STDERR_FILENO : tty_fp = stderr; break;
		}
	}
	else
	{
		PATH_TRACE;
		tty_fp = fdopen(tty_fildes, (testtype[0] == 'r' ? "r" : "w"));
		if (tty_fp == NULL)
		{
			xx_rpt(DELETION);
			in_rpt("fdopen() failed on fd open to VSX_TERMIOS_TTY - errno %d (%s)",
				errno, errname(errno));
			DBUG_VOID_RETURN;
		}
		else
			PATH_TRACE;
	}

	if (testtype[0] == 'f')
	{
		setbuf(tty_fp, buf);
		(void) fputc('Z', tty_fp);
	}
	else if (testtype[0] == 'w')
		setbuf(tty_fp, (char *)NULL);

	SET_TIMEOUT(WAITTIME/2)

	errno = 0;
	globok = 0;
	ret = (*iofunc)(tty_fp);
	err = errno;

	CLEAR_ALARM

	if (required_fd != -1)
	{
		(void) close(required_fd);
		(void) dup(savfd);
	}

	/*
	 * This has to be delayed untill after savfd is restored
	 * Since we need stdout.
	 */

	PATH_FUNC_TRACE; 

	if (alrm_flag > 0)
	{
		xx_rpt(FAILURE);
		in_rpt("I/O timed out");
	}
	else
	{
		globok = 0;
		PATH_FUNC_RPT(3); 
		(*rptfunc)(ret, err);
	}

	/* Don't bother closing stream, just get out */
		DBUG_VOID_RETURN;
}

private void
sig_orphan(sig)
int sig;
{
	/* Signal handler for orphaned children */

	if (sig == signum)
	{
		xx_rpt(FAILURE);
		in_rpt("signal %d (%s) was received", sig, signame(sig));
	}
	else
	{
		xx_rpt(DELETION);
		in_rpt("deletion reason: caught unexpected signal %d (%s)",
			sig, signame(sig));
		vsxerror("received unexpected signal %d (%s)",
			sig, signame(sig));
		exit(0);
	}
}
#else /* !TEST_XNFS */
static int dummy;
#endif /* !TEST_XNFS */
