/*
*      SCCS:   @(#)common/tsetlib(do_ferr.c)	4.2 (01/07/99)  VSXgen release 1.4
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/


/***********************************************************************

NAME:		do_ferr.c - arrange for stream I/O error

PROJECT:	VSX (X/OPEN Validation Suite)

MODULE CONTAINS:
	do_ferr()
SYNOPSIS:
	int do_ferr(errnum, type, reqfd, func, ferr, gotsig)
		int		errnum;
		int		type;
		int		reqfd;
		int		(*func)();
		int		*ferr;
		int		*gotsig;

DESCRIPTION:
	Do_ferr() sets up a stream pointer such that the next operation
	of the given "type" ('r' = read, 'w' = write, 'f' = flush) should
	generate errno = "errnum".  The supplied function "func" is called
	to carry out this I/O operation.  It is passed two arguments:
	the (FILE *) stream pointer and the (int) number of bytes to be
	read/written (always <= 10).  For EAGAIN and EPIPE tests, the
	stream uses a pipe if errnum is > 0 or a FIFO if errnum is < 0.

	If "reqfd" is >= 0 it indicates the fd which should underly the
	stream.  If this is 0, 1 or 2 then the stream set up by do_ferr()
	will be stdin/out/err respectively.  (This fd is NOT restored by
	do_ferr() & so should not be used after the call).

	Do_ferr() returns the return value of "func", with errno restored
	to its value after the call to "func", and the result of a ferror()
	immediately after the call to "func" stored in the location pointed
	to by "ferr".  If "ferr" is NULL, then ferror() is not called.
	(This is needed for tests of fclose(), since calling ferror() after
	fclose() causes undefined behaviour.)  The location pointed to by
	"gotsig" is set to non-zero if a SIGPIPE arrived during an EPIPE
	test.

	N.B. as "func" may be called in a child process, the calling
	process should not set any global/static variables in "func" and
	examine their values after the call to do_ferr().

CONDITIONAL COMPILE PARAMS:
SIDE EFFECTS:
	If anything goes wrong during setting up the current test is
	marked for deletion using tet_delete().  Hence the first thing
	the calling function should do after do_ferr() returns is:
		if (dltd(tet_thistest)) {
			xx_rpt(DELETION);
			in_rpt("deletion reason: %s", reason(tet_thistest));
			DBUG_VOID_RETURN;
		}
RELATED DOCUMENTS:
SOURCE:
AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	3 Nov 1988
MODIFICATIONS:
		Geoff Clare, 10 July 1989
			Catch & check for SIGPIPE in EPIPE tests,
			instead of ignoring it.
		
		Geoff Clare, 26 Sep 1989
			Set signals to default action in child
			Use "WIF" macros for wait status
		
		Geoff Clare, 27 Feb 1990
			Modify pipe-filling strategy to take less time.

		J.A.Nave	03/05/90
			Added resiliance testing
		
		Geoff Clare, 5 June 1990
			Unsupported result in EFBIG tests if VSX_ULIMIT_BLKS
			is set to -1
		
		Geoff Clare, 13 June 1990
			errnum -EAGAIN, -EPIPE: use FIFO instead of pipe

		Anthony Lee, 11 July 1990
			Retrofitted resillience macro.

		David Sawyer
		Fri Jul 13 16:16:50 BST 1990
		Restructured resillience testing.

		Peter Lee, Unisoft Ltd.,
		Tue Jul 17 11:07:06 BST 1990
		Resillience alterations.

		Geoff Clare, 16 Oct 1990
			Use _exit() instead of exit() from child.

		Stuart Boutell, 4 Dec 1990
			Added tet_setcontext() and tet_setblock() calls
			where fork() is called

		Geoff Clare, 30 Mar 1992
			Add forward declaration of ch_func().

		Geoff Clare, 20 Oct 1992
			Temporary fix for problem with use of alrm_flag.

		Geoff Clare, 25 Oct 1994
			Use close() instead of fclose() on stdin/out/err.

		Geoff Clare, 7 Jan 1999
			Passing in ferr=NULL means don't call ferror().

		Geoff Clare, 12 April 2000
			If ferr is NULL, don't call fclose before returning.

************************************************************************/

#include	<std.h>
#include	<sys/types.h>	
#include	<unistd.h>
#include	<errno.h>
#include	<stdio.h>	
#include	<fcntl.h>	
#include	<sys/stat.h>	
#include	<sys/wait.h>	
#include	<string.h>
#include	<signal.h>
#include	<dbug.h>
#include	<report.h>
#include	<timeout.h>
#include	<stdlib.h>
#include <tet_api.h>
#include	<sysdep.h>

#define		WAITTIME	(5*SPEEDFACTOR)
#define		KILLWAIT	(2*SPEEDFACTOR)
#define		MODEANY		(S_IRWXU | S_IRWXG | S_IRWXO)
#define		ULMT_NBLK	1L
#define		BLK_SIZE	512

#define BADARGS		{ \
	vsxerror("errnum = %d, type = '%s' not catered for in do_ferr()", \
		errnum, pr_char(type)); \
	(void) tet_delete(tet_thistest, "bad call to do_ferr()"); \
	DBUG_RETURN(0); \
}

extern	char		*pr_char();
extern	char		*errname();
extern	char		*itoa();
extern	void		setsigs();
extern	void		(*vsignal())();
extern	long		setlimit();

extern	int		alrm_flag;	/* timeout from alarm occured	*/
extern	int		globok;

private	int		pipeorfifo();
private	int		setnonblock();
private	FILE		*getstream();
private	void		sig_catch();
private	void		error();
private	int		ch_func();

private	char		*tfile = "./do_ferr-t";
private	long		ulimblks;
private	int		pfd[2];
private	int		caught_sig;
private	char		buf[BLK_SIZE];	/* general purpose buffer */

public int
do_ferr(errnum, type, reqfd, func, ferr, gotsig)
int	errnum;		/* required errno value			*/
int	type;		/* 'r', 'w' or 'f'			*/
int	reqfd;		/* required fd (0,1,2), else -1		*/
int	(*func)();	/* function to perform I/O of 'type'	*/
int	*ferr;		/* location to store ferror() result, if not NULL */
int	*gotsig;	/* location to indicate if SIGPIPE received */
{
	FILE	*fp;
	char	*ulimaddr;
	long	nbyte;
	int	use_fifo = 0;
	int	fd;
	int	fret;
	int	errsav = 0;		/* temporary store for errno */
	int	i;
	int	pathok = 0;
	struct sigaction act, oact;

	DBUG_LENTER("do_ferr", 2);
	DBUG_PRINTF("args", "do_ferr(%d, '%c', 0x%lx)",
		errnum, type, (long)func);

	if (errnum == -EAGAIN || errnum == -EPIPE)
	{
		use_fifo = 1;
		errnum = -errnum;
	}

	if (errnum == EFBIG)
	{
		ulimaddr = tet_getvar("VSX_ULIMIT_BLKS");
		if (ulimaddr == NULL || *ulimaddr == '\0')
		{
			(void) tet_delete(tet_thistest,
				"VSX_ULIMIT_BLKS parameter not set");
			DBUG_RETURN(0);
		}
		else
			PATH_TRACE;

		ulimblks = atol(ulimaddr);
		if (ulimblks == -1)
		{
			xx_rpt(UNSUPPORTED);
			(void) tet_delete(tet_thistest, "VSX_ULIMIT_BLKS is set to -1");
			DBUG_RETURN(0);
		}
		else 
		{
			PATH_TRACE;
			if (ulimblks < ULMT_NBLK)
				ulimblks = ULMT_NBLK;
		}
	}
	else
	{
		PATH_TRACE;
		PATH_TRACE; /* to  balance path tracing */
	}


	pfd[0] = pfd[1] = fd = -1;
	fp = NULL;
	act.sa_handler = sig_catch;
	act.sa_flags = 0;
	(void) sigemptyset(&act.sa_mask);

	if (reqfd >= 0)
	{
		/* Close fd's as necessary, so the one used will be reqfd */
		if (errnum == EFBIG)
		{
			/* only one descriptor will be opened */
			(void) close(reqfd);
		}
		else
		{
			/* two descriptors will be opened to pipe/FIFO */
			(void) close(0);
			if (type != 'r')
				(void) close(reqfd);
		}
	}

	/* Set up stream pointer as required */
	switch (errnum)
	{
	case EAGAIN :
		/* read from empty/write to full non-blocking pipe/FIFO */
		/* FALLTHROUGH */

	case EINTR :
		/* read from empty/write to full blocking pipe */
		if (pipeorfifo(use_fifo, pfd) != 0)
			DBUG_RETURN(0);
		else
			PATH_TRACE;

		if (type == 'r')
			fd = pfd[0];
		else
			fd = pfd[1];

		errno = 0;
		if (type != 'r' || errnum == EAGAIN)
		{
			if (setnonblock(fd, TRUE) != 0)
				DBUG_RETURN(0);
			else
				PATH_TRACE;
		}
		else
			PATH_TRACE;

		if (type != 'r')
		{
			while (write(fd, buf, BLK_SIZE) > 0)
				;
			while (write(fd, "z", 1) > 0)
				;
			errno = 0;
			if (errnum == EINTR && setnonblock(fd, FALSE) != 0)
				DBUG_RETURN(0);
			else
				PATH_TRACE;
		}
		else
			PATH_TRACE;

		PATH_TRACE; /* so that each case has the same number */

		break;

	case EBADF :
		/* close underlying file descriptor */
		if (pipe(pfd) == SYSERROR)
		{
			error("pipe() failed", (char *)NULL);
			DBUG_RETURN(0);
		}
		else
			PATH_TRACE;

		if (type == 'r')
			fp = getstream(pfd[0], "r", reqfd);
		else
			fp = getstream(pfd[1], "w", reqfd);
		if (fp == NULL)
		{
			error("fdopen() failed on pipe fd", (char *)NULL);
			DBUG_RETURN(0);
		}
		else
			PATH_TRACE;

		if (close(pfd[0]) == SYSERROR || close(pfd[1]) == SYSERROR)
		{
			error("could not close fd underlying open stream",
				(char *)NULL);
			DBUG_RETURN(0);
		}
		else
			PATH_TRACE;

		PATH_TRACE; /* so that each case has the same number */

		break;

	case EPIPE:
		/* write to pipe/FIFO with no reader */
		if (type == 'r')
			BADARGS
		else
			PATH_TRACE;

		if (pipeorfifo(use_fifo, pfd) != 0)
			DBUG_RETURN(0);
		else
			PATH_TRACE;

		if (close(pfd[0]) == SYSERROR)
		{
			error("could not close reading side of pipe/FIFO",
				(char *)NULL);
			DBUG_RETURN(0);
		}
		else
			PATH_TRACE;

		if (sigaction(SIGPIPE, &act, &oact) == SYSERROR)	
		{
			error("sigaction() failed to set SIGPIPE to be caught",
				(char *)NULL);
			DBUG_RETURN(0);
		}
		else
			PATH_TRACE;

		fd = pfd[1];
		break;
		
	case EFBIG :
		fd = crfile(tfile, MODEANY, (int)geteuid(), (int)getegid());
		if (fd < 0)
		{
			error("could not creat test file: %s", tfile);
			DBUG_RETURN(0);
		}
		else
			PATH_TRACE;

		/* Write out just less than the ulimit */
		nbyte = ulimblks * BLK_SIZE - 5;
		i = 0;
		while (nbyte > 0)
		{
			int n;
			if (i++ == 0)
				PATH_TRACE;
			n = write(fd, buf,
			    (unsigned)(nbyte < BLK_SIZE ? nbyte : BLK_SIZE));
			if (n <= 0)
			{
				error("write error preparing test file: %s",
					tfile);
				DBUG_RETURN(0);
			}
			nbyte -= n;
		}

		PATH_TRACE;
		PATH_TRACE; /* so that each case has the same number */

		break;

	default :
		BADARGS
	}

	if (fd >= 0 && fp == NULL)
	{
		fp = getstream(fd, (type == 'r' ? "r" : "w"), reqfd);
		if (fp == NULL)
		{
			error("fdopen() failed on open fd", (char *)NULL);
			DBUG_RETURN(0);
		}
		else
			PATH_TRACE;
	}
	else
		PATH_TRACE;

	/* check required fd was produced */
	if (reqfd >= 0 && fileno(fp) != reqfd)
	{
		error("could not provide required fd (%d) on stream",
			itoa(reqfd));
		DBUG_RETURN(0);
	}
	else
		PATH_TRACE;

	if (type == 'f')
	{
		/* put some data on the stream to be flushed */
		if (fputs("1234567890", fp) == EOF)
		{
			error("fputs() failed", (char *)NULL);
			DBUG_RETURN(0);
		}
		else
			PATH_TRACE;
	}
	else 
	{
		PATH_TRACE;
		if (type == 'w')
		{
		/* make stream unbuffered so a write will physically occur */
		setbuf(fp, (char *)NULL);
		}
	}

	SET_TIMEOUT(2 * WAITTIME)

	globok = 0;
	if (errnum != EFBIG && errnum != EINTR)
	{
		caught_sig = 0;
		clearerr(fp);
		errno = 0;
		fret = (*func)(fp, 1);
		errsav = errno;
		if (ferr != NULL)
			*ferr = ferror(fp);
		*gotsig = (caught_sig == SIGPIPE);
	}
	else
	{
		fret = ch_func(errnum, func, fp, ferr);
		errsav = errno;
	}
	PATH_FUNC_TRACE; /* checking that globok is set in 'func' */

	CLEAR_ALARM
	
	/* Reset SIGPIPE action */
	if (errnum == EPIPE)
		(void) sigaction(SIGPIPE, &oact, NULLSA);

	/* See if time out was what was intended */
	if (alrm_flag > 0 && errnum != EINTR)
	{
		errno = NOERROR;
		error("test function timed out", (char *)NULL);
		DBUG_RETURN(0);
	}
	else
		PATH_TRACE;


	/* Clean up opened files */
	if (pfd[0] >= 0)
		(void) close(pfd[0]);
	if (pfd[1] >= 0)
		(void) close(pfd[1]);
	SET_TIMEOUT(WAITTIME/2)
	if (fp == stdin || fp == stdout || fp == stderr)
		(void) close(fileno(fp));
	else
	{
		/* If ferr is NULL, the stream may already have been closed */
		if (ferr != NULL)
			(void) fclose(fp);
	}
	CLEAR_ALARM
	(void) unlink(tfile);

	DBUG_PRINTF("return", "do_ferr = %d", fret);
	errno = errsav;

	PATH_FUNC_RPT(11); 

	DBUG_RETURN(fret);
}

private int
pipeorfifo(fifoflag, fds)
int fifoflag;
int fds[2];
{
	DBUG_LENTER("pipeorfifo", 2);

	if (fifoflag)
	{
		(void) remove(tfile);
		if (mkfifo(tfile, (mode_t)MODEANY) == SYSERROR)
		{
			error("mkfifo() failed", (char *)NULL);
			DBUG_RETURN(-1);
		}
		if ((fds[0] = open(tfile, O_RDONLY|O_NONBLOCK)) < 0)
		{
			error("open(FIFO, O_RDONLY|O_NONBLOCK) failed",
				(char *)NULL);
			DBUG_RETURN(-1);
		}
		if ((fds[1] = open(tfile, O_WRONLY|O_NONBLOCK)) < 0)
		{
			error("open(FIFO, O_WRONLY|O_NONBLOCK) failed",
				(char *)NULL);
			DBUG_RETURN(-1);
		}
		if (setnonblock(fds[0], FALSE) != 0)
			DBUG_RETURN(-1);
		if (setnonblock(fds[1], FALSE) != 0)
			DBUG_RETURN(-1);
	}
	else
	{
		if (pipe(fds) == SYSERROR)
		{
			error("pipe() failed", (char *)NULL);
			DBUG_RETURN(-1);
		}
	}

	DBUG_RETURN(0);
}

private int
setnonblock(fd, yesno)
int fd, yesno;
{
	int flags;

	DBUG_LENTER("setnonblock", 2);
	if ((flags = fcntl(fd, F_GETFL)) == SYSERROR)
	{
		error("fcntl(F_GETFL) failed", (char *)NULL);
		DBUG_RETURN(-1);
	}

	if (yesno)
		flags |= O_NONBLOCK;
	else
		flags &= ~O_NONBLOCK;

	if (fcntl(fd, F_SETFL, flags) == SYSERROR)
	{
		error("fcntl(F_SETFL, %s) failed", itoa(flags));
		DBUG_RETURN(-1);
	}

	DBUG_RETURN(0);
}

private int
ch_func(errnum, func, fp, ferr)
int errnum, (*func)(), *ferr;
FILE *fp;
{
	int	syn[2], status;
	pid_t	ch_pid;
	struct { int ret; int err; int ferr; int alrm; } results;
	int	pathok = 0;

	DBUG_LENTER("ch_func", 2);

	results.alrm = 0; /* temporary fix */

	if (pipe(syn) == SYSERROR)
	{
		error("pipe() failed", (char *)NULL);
		DBUG_RETURN(0);
	}
	else
		PATH_TRACE;

	switch (ch_pid = fork())
	{
	case -1:
		error("fork() failed", (char *)NULL);
		DBUG_RETURN(0);
	
	case 0:
		tet_setcontext();
		/* child */
		setsigs(SIG_DFL);

		(void) close(syn[0]);
		if (errnum == EFBIG)
		{
			if (setlimit(ulimblks) == SYSERROR)
				_exit(1);
		}
		else if (errnum == EINTR)
		{
			(void) vsignal(SIGALRM, alrm);
			(void) alarm(WAITTIME);
		}

		clearerr(fp);
		errno = 0;
		globok = 0;
		results.ret = (*func)(fp, 10);
		results.err = errno;
		if (ferr != NULL) /* NULL value means don't call ferror() */
			results.ferr = ferror(fp);
		results.alrm = alrm_flag;

		if (write(syn[1], (char *)&results, sizeof(results)) !=
							    sizeof(results))
			_exit(2);

		if (globok == 0)
			_exit(3);

		_exit(0);
		break;
	
	default:
		tet_setblock();
		/* parent */
		(void) close(syn[1]);
		if (waitpid(ch_pid, &status, 0) == SYSERROR)
		{
			if (alrm_flag > 0)
			{
				errno = NOERROR;
				error("child process timed out", (char *)NULL);
			}
			else
				error("waitpid() failed", (char *)NULL);
			(void) killchild(ch_pid, KILLWAIT);
			(void) close(syn[0]);
			DBUG_RETURN(0);
		}
		else
			PATH_TRACE;

		if (!WIFEXITED(status))
		{
			error("child wait status = %s", itoa(status));
			results.ret = 0;
		}
		else if ((status = WEXITSTATUS(status)) == 1)
		{
			error("setlimit(%s) failed", itoa((int)ulimblks));
			results.ret = 0;
		}
		else if (status == 2)
		{
			error("child could not write results to pipe",
				(char *)NULL);
			results.ret = 0;
		}
		else if (status != 0 && status != 3)
		{
			error("child exit code = %s", itoa(status));
			results.ret = 0;
		}
		else
		{
			if (status == 0)
				PATH_TRACE;

			errno = NOERROR;
			if (read(syn[0], (char *)&results, sizeof(results))
							!= sizeof(results))
			{
				error("could not read results from child",
					(char *)NULL);
				results.ret = 0;
			}
			else
				PATH_TRACE;
		}
		(void) close(syn[0]);
	}
	tet_setblock();
	PATH_FUNC_RPT(4); 
	alrm_flag = results.alrm;
	if (ferr != NULL)
		*ferr = results.ferr;
	errno = results.err;
	DBUG_RETURN(results.ret);
}

private FILE *
getstream(fd, type, reqfd)
int fd, reqfd;
char *type;
{
	/* If calling process has asked for specific fd, give it
	   the corresponding stream */

	if (reqfd == fd)
	{
		if (reqfd == fileno(stdin))
			return stdin;
		if (reqfd == fileno(stdout))
			return stdout;
		if (reqfd == fileno(stderr))
			return stderr;
	}
	
	return fdopen(fd, type);
}

private void
sig_catch(sig)
int sig;
{
	caught_sig = sig;
}

private void
error(msg, val)
char *msg, *val;
{
	/*
	 * tet_delete the current test, then cleanup
	 */

	char *end;
	int err = errno;

	(void) strcpy(buf, "error in do_ferr(): ");
	end = buf + strlen(buf);
	(void) sprintf(end, msg, val);
	if (err != NOERROR)
	{
		end = buf + strlen(buf);
		(void) sprintf(end, ", errno = %d(%s)", err, errname(err));
	}

	(void) tet_delete(tet_thistest, buf);

	/* Clean up opened files */
	if (pfd[0] >= 0)
		(void) close(pfd[0]);
	if (pfd[1] >= 0)
		(void) close(pfd[1]);

	(void) unlink(tfile);
}
