/*
*      SCCS:   @(#)common/tsetlib(mlrantest.c)	4.2 (03/06/98)  VSXgen release 1.4
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/


/***********************************************************************

NAME:		mlrantest.c - basic driver and random number generator
				for mathematical library tests.

PROJECT:	VSX (X/OPEN Validation Suite)

MODULE CONTAINS:
		mlrandom()
		randl()
		mlsetglobals()
		mlrantest()
		mlresults()
		mlreport()
SYNOPSIS:
		double mlrandom()
		double randl(x)
		       double x;
		void   mlsetglobals()
		void   mlrantest(f,g,type)
		       double (*f)(),(*g)();
		       int type;
		int    mlresults(factor)
		       double factor;
		void   mlreport(start, end)
		       double start, end;

DESCRIPTION:
	Mlrandom() generates a random number in the range 0 to 1

	Randl() takes the random number generated by mlrandom and
	logarithmically distributes this over the range 1 to exp(x).

	Mlsetglobals() sets a series of global variables which are
	used in the error calculations. These variables need to be
	reset before each test.

	Mlrantest() generates a series of random numbers and invokes
	the functions f() and g() which should produce the same results.
	The results of these functions are compared and the maximum
	relative difference and sum of the square of the relative
	difference is calculated. The global variable iterrs determines
	how many random numbers will be generated. The variables high
	and low determine the limits for random number generation.
	The variable type determines whether linear distribution (type=0)
	or logarithmic distribution of random numbers is required (type=1).

	Mlresults() calculates the Root Mean Square relative error and
	checks the maximum relative error and RMS relative error against
	values appropriate for the machine. (as determined by mlmachar()).

	Mlreport() reports details of the values calculated by mlresults()
	and mlrantest() if a failure has been notified.

SIDE EFFECTS:
RELATED DOCUMENTS:
AUTHOR:		Andrew Twigger, UniSoft Ltd.
DATE CREATED:	Wed Nov 26 15:04:11 BST 1986
MODIFICATIONS:
		Geoff Clare, 22 Nov 1989
			Change (high-low)/iterrs to high/iterrs-low/iterrs
			to avoid floating exception if high-low > DBL_MAX
		
		Geoff Clare, 7 Dec 1989
			Don't increment xstart on last iteration as
			rounding may cause overflow when high == DBL_MAX

		Geoff Clare, 29 Mar 1990
			Additional iteration with x=high, when type == 0
		
		Geoff Clare, 5 Mar 1992
			Guard against overflow in RMS error calculation
			and low-high interval stepping.  Handle range
			errors in called functions (*f)() and (*g)().

		Geoff Clare, 14 Oct 1992
			Changes for XNFS.

		Geoff Clare, 6 March 1998
			Rename random() to mlrandom().

************************************************************************/
#include <std.h>

#if !TEST_XNFS
#include <sys/types.h>
#include <dbug.h>
#include <math.h>
#include <errno.h>
#include <tet_api.h>
#include <sysdep.h>

extern int matherr();
extern int dsignif;
extern int expbase;
extern double mindble;
extern double maxdble;
double errmax;
double xerrmax;
double yerrmax;
double quaderr;
double maxdsig;
double maxdloss;
double acceptloss;
double accepterror;
double rmsdsig;
double rmsdloss;
double low;
double high;
double argy;
int iterrs;
int nlarge;
int nequal;
int nsmall;
int nerror;

/* This function uses three multiplicative congruence
     generators to provide approximately 48-bit random
     sequences from three 16-bit integer ones. The three
     sequences are:
	   ix := (ix * 171) mod 30269
	   iy := (iy * 172) mod 30307
     and   iz := (iz * 170) mod 30323
     Since 30269, 30307 and 30323 are primes, all sequences can be
     of maximal length (see Seminumerical Algorithms, D E Knuth,
     Addison Wesley 1969, p19). The simple steps above cannot
     be performed without overflow on a 16-bit machine. This
     is avoided by writing:
	   iy = k * 176 + r
     where    0 <= k <= 172
     and      0 <= r <= 175
     Then     172 * iy = k * 176 * 172 + r * 172
		       = k * 30272 + r * 172
		       = - k * 35 + r * 172 mod 30307
     Similarly
     with     iz = k * 178 + r
	      170 * iz = - k * 63 + r * 170 mod 30323.
     and with ix = k * 177 + r
	      171 * ix = - k * 2 + r * 171 mod 30269
     The values are now bounded for a 16-bit machine.
     The period is about 2.78E13.
     The fractional part of the sum of the three values as a
     fraction of the prime modulus gives the real random value. */


double
mlrandom()
{
	int ni,k,r;
	double x;
	static int ix=1;
	static int iy=10001;
	static int iz=4987;
	double	rval;

	DBUG_LENTER("mlrandom", 2);

	/* calculate k and r for ix */
	k = ix / 177;
	r = ix - k * 177;
	ni = - k - k + 171 * r;

	/* -342 <= ni <= 30096, so now reduce range */
	if( ni < 0 )
		ix = ni + 30269;
	else
		ix = ni;

	/* k and r for iy generator */
	k = iy / 176;
	r = iy - k * 176;
	ni = - 35 * k + r * 172;

	/* reduce range, -6020 <= ni <= 30100 */
	if( ni < 0 )
		iy = ni + 30307;
	else
		iy = ni;

	/* now the same for iz */
	k = iz / 178;
	r = iz - k * 178;
	ni = - 63 * k + r * 170;

	/* reduce range, -10710 <= ni <= 30090 */
	if( ni < 0 )
		iz = ni + 30323;
	else
		iz = ni;
	x = ( (double)ix /30269.0 + (double)iy /30307.0 + (double)iz /30323.0 );

	rval = x-floor(x);
	DBUG_PRINTF("return", "mlrandom = %f", rval);
	DBUG_RETURN(rval);
}  /*mlrandom*/

double
randl(x)
double x;

/*     returns pseudo random numbers logarithmically distributed
      over (1,exp(x)).	thus a*randl(ln(b/a)) is logarithmically
      distributed in (a,b).

      other subroutines required

	 exp(x) - the exponential routine

	 random - a function program returning random real
		  numbers uniformly distributed over (0,1).
								  */


{
	return(exp ( x * mlrandom() ));
}

void
mlrantest(f,g,type)
double (*f)(),(*g)();
int type;
{
	double xstart = low;
	double distance;
	double x,w,z;
	int i;

	DBUG_LENTER("mlrantest", 2);
	DBUG_PRINTF("args", "mlrantest(0x%lx, 0x%lx, %d)", (long)f, (long)g, type);
	
	distance = high/iterrs - low/iterrs;	/* NOT (high-low)/iterrs */
	if (type == 0)
		iterrs++;
	for( i=1 ;i<=iterrs ; i++ )
	{
		if (type == 0)
		{
			if (i < iterrs)
			{
				x=distance*mlrandom()+xstart;
				/* In case of overflow/underflow */
				if (x < low || x > high)
					x = low/2 + high/2;
			}
			else
				x=high;
		}
		else if (type == 1)
		{
			if (low < 0)
			{
				if (i%2 || high < 0)
					x = -randl(log(fabs(low)));
				else
					x = randl(log(high));
			}
			else
				x = randl(log(high)) + low;
		}
		errno = 0;
		z=(*g)(x);
		if (errno != ERANGE)
			w=(*f)(x)-z;
		if (errno == ERANGE)
			nerror++;
		else
		{
			if( w<0 )
				nsmall++;
			else
			{
				if( w>0 )
					nlarge++;
				else
					nequal++;
			}
			if( z > mindble || z < -mindble)
			{
				w=fabs(w/z);
				if( w>errmax )
				{
					errmax=w;
					xerrmax=x;
					yerrmax=argy;
				}
				if (w > sqrt(maxdble/expbase))
					w = sqrt(maxdble/expbase);
				z = nlarge+nequal+nsmall;
				z = (z-1)/z;
				quaderr = quaderr*z + w*w*(1-z);
			}
		}
		if (i < iterrs-1)
			xstart = xstart + distance;
	}
	DBUG_VOID_RETURN;
}

void
mlsetglobals()
{
	errmax = 0.0;
	xerrmax = 0.0;
	argy = 0.0;
	yerrmax = 0.0;
	quaderr = 0.0;
	nlarge = 0;
	nequal = 0;
	nsmall = 0;
	nerror = 0;
	return;
}

int
mlresults(factor)
double factor;
{
	int	count = nequal+nsmall+nlarge+nerror;
	int	failed = 0;
	double	albeta;
	double	loss;
	double	zero = 0.0;

	DBUG_LENTER("mlresults", 2);
	DBUG_PRINTF("args", "mlresults(%f)", factor);

	albeta = log((double) expbase);

	if (count == 0)
		count++;

	quaderr=sqrt(quaderr);

	/* The limit for loss in accuracy corresponds to 4/3(BASE
	     digit plus 2 bits) for the Maximum Relative Error and to
	     half this for the Root Mean Square error. Such limits are
	     only exceeded by routines which are believed to contain
	     numerically poor algorithms except for isolated argument
	     values.
	     Hence MRE <= 4, RMS <= 2 for a binary machine
	       and MRE <= 2, RMS <= 1 for a hexadecimal machine.
	     This is an NPL addition to the Cody and Waite tests. */
	
	/*  Rather than include the return value from ERANGE errors in
	    the loss calculations, they are counted separately and
	    checked against an allowed percentage. */

	loss = 4.0 * (1.0 + log(4.0)/albeta)/3.0;
	loss = loss * factor;
	acceptloss = loss;
	if( errmax != zero )
		maxdsig = log(fabs(errmax))/albeta;
	else
		maxdsig = -999.0;
	if( maxdsig + dsignif < zero )
		maxdloss = zero;
	else
		maxdloss = maxdsig + dsignif;
	if( maxdloss > loss )
		failed++;
	if( quaderr != zero )
		rmsdsig = log(fabs(quaderr))/albeta;
	else
		rmsdsig = -999.0;
	if( rmsdsig + dsignif < zero )
		rmsdloss = zero;
	else
		rmsdloss = rmsdsig + dsignif;
	if( rmsdloss > 0.5 * loss )
		failed++;
	accepterror = factor;
	if (100.0*nerror/count > accepterror)
		failed = 2;
	if (failed == 2)
	{
		DBUG_PRINTF("return", "mlresults = 1"); 
		DBUG_RETURN(1);
	}
	DBUG_PRINTF("return", "mlresults = 0"); 
	DBUG_RETURN(0);
}

void
mlreport(start, end)
double start, end;
{
	int	count = nequal+nsmall+nlarge+nerror;

	if (count == 0)
		count++;

	in_rpt("Random arguments were tested from the interval [%g, %g]", start, end);
	in_rpt("The result was too large %5d times", nlarge);
	in_rpt("                   equal %5d times", nequal );
	in_rpt("               too small %5d times", nsmall);
	if (nerror > 0)
		in_rpt("         an ERANGE error %5d times", nerror);
	if (yerrmax == 0.0)
	{
		in_rpt("The maximum relative error of %g occured for value %g",errmax, xerrmax);
	}
	else
	{
		in_rpt("The maximum relative error of %g occured for values %g, %g",errmax, xerrmax, yerrmax);
	}
	in_rpt("This gave a maximum loss of %.2g significant digits of base %d", maxdloss, expbase);
	in_rpt("The maximum acceptable loss is %.2g significant digits", acceptloss);
	in_rpt("The root-mean-square relative error is %g",quaderr);
	in_rpt("This gave an average loss of %.2g significant digits of base %d", rmsdloss, expbase);
	in_rpt("The maximum acceptable loss is %.2g significant digits", (acceptloss*0.5));
	if (nerror > 0)
	{
		in_rpt("The percentage of ERANGE errors is %.2g", 100.0*nerror/count);
		in_rpt("The maximum acceptable percentage is %.2g", accepterror);
	}
	return;
}
/* Routine to force inclusion of matherr routine */
mldummy()
{
	int x;
	(void) matherr(&x);
}
#else /* !TEST_XNFS */
static int dummy;
#endif /* !TEST_XNFS */
