/*
*      SCCS:   @(#)common/tsetlib(nametoolng.c)	4.1 (01/21/97)  VSXgen release 1.4
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/


/***********************************************************************

NAME:		nametoolng.c - generate path names for testing ENAMETOOLONG

PROJECT:	VSX (X/OPEN Validation Suite)

MODULE CONTAINS:
	pathtoolng()
	nametoolng()
	xnfs_nametoolng()
	clntoolng()
SYNOPSIS:
	char **pathtoolng(filenam)
		char *filenam;

	char **nametoolng()

	char *xnfs_nametoolng(casenum, spno)
		int casenum;
		int spno;

	void clntoolng()

DESCRIPTION:
 	Pathtoolng() creates two pathnames to the named file, of
	PATH_MAX and PATH_MAX+1 characters.  It returns a pointer to an
	array of two malloc()ed buffers, or NULL if an error occurred.
	If PATH_MAX > PCTS_PATH_MAX, or == -1 then the first name will be
	PCTS_PATH_MAX characters, and the second will be a NULL pointer
	to indicate the error part of the test is to be skipped.

	Nametoolng() creates two filenames of NAME_MAX and NAME_MAX+1
	bytes.  If _POSIX_NO_TRUNC is not in effect (i.e. exceeding
	NAME_MAX does not give an error) it reports UNSUPPORTED and
	returns NULL.  If _POSIX_NO_TRUNC is in effect, it returns a
	pointer to an array of two malloc()ed buffers, or NULL if an
	error occurred.

	When testing XNFS, nametoolng() is not used.  Instead
	xnfs_nametoolng() is used to obtain a single file name for use in
	a test.  The first argument is an integer from 1 to 4 indicating
	which of the four NAME_MAX/ENAMETOOLONG assertions is being tested.
	The second argument is a sync point number.  If xnfs_nametoolng()
	returns NULL, it has already reported an UNSUPPORTED or
	UNRESOLVED result, and the calling test should just return.
	Now that the server's NAME_MAX and NO_TRUNC are obtained from
	parameters, the sync point number is no longer used.

	Clntoolng() frees the path/name buffers and removes the temporary
	directories created by pathtoolng().

CONDITIONAL COMPILE PARAMS:
SIDE EFFECTS:
RELATED DOCUMENTS:
SOURCE:
AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	14 Nov 1988
MODIFICATIONS:
		Geoff Clare, 29 Mar 1990
			Changes for POSIX.3 draft 11
		
		Geoff Clare, 5 July 1990
			Make NAME_MAX tests consistent with PATH_MAX tests

		Anthony Lee, 11 July 1990
			Retrofitted resillience macros.

		Geoff Clare, 21 Feb 1991
			Allow for PATH_MAX to include null byte.

		Geoff Clare, 19 May 1992
			In XPG4 mode, PATH_MAX includes null byte.

		Geoff Clare, 17 Sept 1992
			Changes for testing XNFS.

		Geoff Clare, 14 Jan 1993
			New configurable header <rpcincs.h>

		Geoff Clare, 27 Oct 1993
			Implement interpretation XNFS/003/931001.

************************************************************************/

#include <std.h>
#if TEST_XNFS
#include <rpcincs.h>	/* must preceed any system headers */
#endif
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <dbug.h>
#include <report.h>
#include <pcts_max.h>
#if TEST_XNFS
#include <nfs.h>
#endif
#include <tet_api.h>
#include <sysdep.h>

#define	TOPDIR		"./ENAMETOOLONG"
#define	ALTDIR		"./ENAMETOOLONGX"
#define	ADDSTR		"/../ENAMETOOLONG"
#define	ADDLEN		(sizeof(ADDSTR) - 1)

extern char	*errname();

private char	*paths[2] = { NULL, NULL };
private char	*names[2] = { NULL, NULL };

static void
makepath(pathnam, filenam, len)
char *pathnam;
char *filenam;
int len;
{
	/* construct pathnam of len bytes which refers to filenam */

	char *ptr, *end;

	DBUG_LENTER("makepath", 2);
	DBUG_PRINTF("args", "makepath(0x%lx, \"%s\", %d)",
		pathnam, filenam, len);

	/* Adjust len to be length required before final "/../filenam" */
	len -= (strlen(filenam) + 4);

	if (((len - sizeof(TOPDIR) + 1) % ADDLEN) % 2 == 0)
	{
		(void) strcpy(pathnam, TOPDIR);
		ptr = pathnam + sizeof(TOPDIR)-1;
	}
	else
	{
		(void) strcpy(pathnam, ALTDIR);
		ptr = pathnam + sizeof(ALTDIR)-1;
	}
	end = &pathnam[len];
	for ( ; ptr < end - ADDLEN; ptr += ADDLEN)
		(void) strcpy(ptr, ADDSTR);

	while (ptr < end-1)
	{
		*ptr++ = '/';
		*ptr++ = '.';
	}
	(void) strcpy(ptr, "/../");
	(void) strcpy(&ptr[4], filenam);

	DBUG_PRINTF("makepath", "strlen %d", (int)strlen(pathnam));

	DBUG_VOID_RETURN;
}

char **
pathtoolng(filenam)
char *filenam;
{
	int	len1, len2;
	int	pathok = 0;

	DBUG_LENTER("pathtoolng", 2);

	globok = 0;
	errno = 0;
	len1 = pathconf(".", _PC_PATH_MAX);
	if (len1 == -1 && errno != 0)
	{
		xx_rpt(DELETION);
		in_rpt("pathconf(\".\", _PC_PATH_MAX) failed - errno %d (%s)",
			errno, errname(errno));
		DBUG_PRINTF("return", "pathtoolng = NULL");
		DBUG_RETURN((char **)NULL);
	}
	else if (len1 != -1 && len1 <= 0)
	{
		xx_rpt(DELETION);
		in_rpt("pathconf(\".\", _PC_PATH_MAX) returned %d", len1);
		DBUG_PRINTF("return", "pathtoolng = NULL");
		DBUG_RETURN((char **)NULL);
	}
	else
		PATH_TRACE;

	if (len1 == -1 || len1 > PCTS_PATH_MAX)
	{
		len1 = PCTS_PATH_MAX;
		len2 = -1;
	}
	else
	{
		/*
		 * POSIX.1 allows PATH_MAX either to include or exclude
		 * the null byte, so we have to test PATH_MAX-1 for success
		 * and PATH_MAX+1 for failure (both excluding the null)
		 * in POSIX and XPG3 modes.  XPG4 says PATH_MAX includes
		 * the null byte, so test PATH_MAX-1 for success and
		 * PATH_MAX for failure in XPG4 mode.
		 */

		len2 = len1;
#if !(TEST_XPG > 3)
		len2++;
#endif
		len1--;
	}

	paths[0] = (char *)malloc((size_t)(len1 + 1));
	if (len2 > 0)
		paths[1] = (char *)malloc((size_t)(len2 + 1));
	if (paths[0] == NULL || len2 > 0 && paths[1] == NULL)
	{
		xx_rpt(DELETION);
		in_rpt("malloc() failed in pathtoolng()");
		DBUG_PRINTF("return", "pathtoolng = NULL");
		DBUG_RETURN((char **)NULL);
	}
	else
		PATH_TRACE;

	if (vxmkdir(TOPDIR) == SYSERROR || vxmkdir(ALTDIR) == SYSERROR)
	{
		xx_rpt(DELETION);
		in_rpt("could not create directories \"%s\" and \"%s\" in pathtoolng()",
			TOPDIR, ALTDIR);
		DBUG_PRINTF("return", "pathtoolng = NULL");
		DBUG_RETURN((char **)NULL);
	}
	else
		PATH_TRACE;

	makepath(paths[0], filenam, len1);
	if (len2 > 0)
		makepath(paths[1], filenam, len2);

	PATH_FUNC_RPT(3); 

	DBUG_PRINTF("return", "pathtoolng = 0x%lx", (long)paths);
	DBUG_RETURN(paths);
}

static void
makename(filenam, len)
char *filenam;
int len;
{
	/* construct filename of len bytes */

	int i;

	for (i = 0; i < len; i++)
		filenam[i] = '0' + (i+1)%10;
	filenam[i] = '\0';
}

#if !TEST_XNFS
char **
nametoolng()
{
	int	len1, len2;
	int	pathok = 0;

	DBUG_LENTER("nametoolng", 2);

	globok = 0;

	errno = 0;
	if (pathconf(".", _PC_NO_TRUNC) < 0)
	{
		if (errno == 0)
		{
#ifndef TESTBED
			xx_rpt(UNSUPPORTED);
			in_rpt("_POSIX_NO_TRUNC is not in effect");
			DBUG_PRINTF("return", "nametoolng = NULL");
			DBUG_RETURN((char **)NULL);
#else
			PATH_TRACE;	/*	just for testing	*/
#endif
		}
		else
		{
			xx_rpt(DELETION);
			in_rpt("pathconf(\".\", _PC_NO_TRUNC) failed in nametoolng(), errno = %d (%s)",
				errno, errname(errno));
			DBUG_PRINTF("return", "nametoolng = NULL");
			DBUG_RETURN((char **)NULL);
		}
	}
	else
		PATH_TRACE;

	errno = 0;
	len1 = pathconf(".", _PC_NAME_MAX);
	if (len1 == -1 && errno != 0)
	{
		xx_rpt(DELETION);
		in_rpt("pathconf(\".\", _PC_NAME_MAX) failed - errno %d (%s)",
			errno, errname(errno));
		DBUG_PRINTF("return", "nametoolng = NULL");
		DBUG_RETURN((char **)NULL);
	}
	else if (len1 != -1 && len1 <= 0)
	{
		xx_rpt(DELETION);
		in_rpt("pathconf(\".\", _PC_NAME_MAX) returned %d", len1);
		DBUG_PRINTF("return", "nametoolng = NULL");
		DBUG_RETURN((char **)NULL);
	}
	else
		PATH_TRACE;

	if (len1 == -1 || len1 > PCTS_NAME_MAX)
	{
		len1 = PCTS_NAME_MAX;
		len2 = -1;
	}
	else
		len2 = len1 + 1;

	names[0] = (char *)malloc((size_t)(len1 + 1));
	if (len2 > 0)
		names[1] = (char *)malloc((size_t)(len2 + 1));
	if (names[0] == NULL || (len2 > 0 && names[1] == NULL))
	{
		xx_rpt(DELETION);
		in_rpt("malloc() failed in nametoolng()");
		DBUG_PRINTF("return", "nametoolng = NULL");
		DBUG_RETURN((char **)NULL);
	}
	else
		PATH_TRACE;

	makename(names[0], len1);
	if (len2 > 0)
		makename(names[1], len2);

	PATH_FUNC_RPT(3); 

	DBUG_PRINTF("return", "nametoolng = 0x%lx", (long)names);
	DBUG_RETURN(names);
}

#else /* if TEST_XNFS */

/* ARGSUSED */
char *
xnfs_nametoolng(casenum, spno)
int casenum;
int spno;
{
	long	name_max_c;
	long	name_max_s;
	int	no_trunc_c;
	int	no_trunc_s;
	int	len;
	char	*cp;
	int	pathok = 0;

	DBUG_LENTER("xnfs_nametoolng", 2);

	/* Obtain _POSIX_NO_TRUNC and NAME_MAX for client */

	errno = 0;
	if (pathconf(".", _PC_NO_TRUNC) < 0)
	{
		if (errno == 0)
		{
			no_trunc_c = 0;
			PATH_TRACE;
		}
		else
		{
			xx_rpt(DELETION);
			in_rpt("pathconf(\".\", _PC_NO_TRUNC) failed in xnfs_nametoolng(), errno = %d (%s)",
				errno, errname(errno));
			DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
			DBUG_RETURN((char *)NULL);
		}
	}
	else
	{
		no_trunc_c = 1;
		PATH_TRACE;
	}

	errno = 0;
	name_max_c = pathconf(".", _PC_NAME_MAX);
	if (name_max_c == -1 && errno != 0)
	{
		xx_rpt(DELETION);
		in_rpt("pathconf(\".\", _PC_NAME_MAX) failed - errno %d (%s)",
			errno, errname(errno));
		DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
		DBUG_RETURN((char *)NULL);
	}
	else if (name_max_c != -1 && name_max_c <= 0)
	{
		xx_rpt(DELETION);
		in_rpt("pathconf(\".\", _PC_NAME_MAX) returned %ld", name_max_c);
		DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
		DBUG_RETURN((char *)NULL);
	}
	else
		PATH_TRACE;

	if (casenum != 1)
	{
		/* Obtain _POSIX_NO_TRUNC and NAME_MAX for server */

		cp = tet_getvar("XNFS_SERVER_NO_TRUNC");
		if (cp == NULL || *cp == '\0')
		{
			xx_rpt(DELETION);
			in_rpt("parameter XNFS_SERVER_NO_TRUNC is not set");
			DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
			DBUG_RETURN((char *)NULL);
		}
		else
		{
			PATH_TRACE;
			no_trunc_s = (*cp == 'Y' || *cp == 'y');
		}

		cp = tet_getvar("XNFS_SERVER_NAME_MAX");
		if (cp == NULL || *cp == '\0')
		{
			xx_rpt(DELETION);
			in_rpt("parameter XNFS_SERVER_NAME_MAX is not set");
			DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
			DBUG_RETURN((char *)NULL);
		}
		else if ((name_max_s = atol(cp)) <= 0 && name_max_s != -1)
		{
			xx_rpt(DELETION);
			in_rpt("parameter XNFS_SERVER_NAME_MAX has an invalid value");
			DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
			DBUG_RETURN((char *)NULL);
		}
		else
			PATH_TRACE;
	}
	else
	{
		PATH_TRACE;
		PATH_TRACE;
	}

	switch (casenum)
	{
	case 1:
		/* _POSIX_NO_TRUNC is defined on the client and {NAME_MAX}
		    on the client is less than or equal to PCTS_NAME_MAX */

		if (!no_trunc_c)
		{
			xx_rpt(UNSUPPORTED);
			in_rpt("_POSIX_NO_TRUNC is not in effect for client");
			DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
			DBUG_RETURN((char *)NULL);
		}
		else if (name_max_c == -1 || name_max_c > PCTS_NAME_MAX)
		{
			xx_rpt(UNSUPPORTED);
			in_rpt("NAME_MAX on the client is > PCTS_NAME_MAX");
			DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
			DBUG_RETURN((char *)NULL);
		}
		else
		{
			len = name_max_c+1;
			PATH_TRACE;
		}
		break;

	case 2:
		/* _POSIX_NO_TRUNC is defined on the server, NAME_MAX on the
		    server is less than or equal to PCTS_NAME_MAX, NAME_MAX
		    on the server is less than NFS_MAXNAMLEN and NAME_MAX
		    on the server is less than NAME_MAX on the client */

		if (!no_trunc_s)
		{
			xx_rpt(UNSUPPORTED);
			in_rpt("_POSIX_NO_TRUNC is not in effect for server");
			DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
			DBUG_RETURN((char *)NULL);
		}
		else if (name_max_s == -1 || name_max_s > PCTS_NAME_MAX)
		{
			xx_rpt(UNSUPPORTED);
			in_rpt("NAME_MAX on the server is > PCTS_NAME_MAX");
			DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
			DBUG_RETURN((char *)NULL);
		}
		else if (name_max_s >= NFS_MAXNAMLEN)
		{
			xx_rpt(UNSUPPORTED);
			in_rpt("NAME_MAX on the server is >= NFS_MAXNAMLEN");
			DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
			DBUG_RETURN((char *)NULL);
		}
		else if (name_max_s >= name_max_c && name_max_c != -1)
		{
			xx_rpt(UNSUPPORTED);
			in_rpt("NAME_MAX on the server is >= NAME_MAX on the client");
			DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
			DBUG_RETURN((char *)NULL);
		}
		else
		{
			len = name_max_s+1;
			PATH_TRACE;
		}
		break;

	case 3:
		/* _POSIX_NO_TRUNC is defined on the client, NAME_MAX on the
		   client is greater than or equal to NFS_MAXNAMLEN,
		   _POSIX_NO_TRUNC is defined on the server and NAME_MAX
		   on the server is greater than or equal to NFS_MAXNAMLEN */

		if (!no_trunc_c)
		{
			xx_rpt(UNSUPPORTED);
			in_rpt("_POSIX_NO_TRUNC is not in effect for client");
			DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
			DBUG_RETURN((char *)NULL);
		}
		else if (name_max_c < NFS_MAXNAMLEN)
		{
			xx_rpt(UNSUPPORTED);
			in_rpt("NAME_MAX on the client is < NFS_MAXNAMLEN");
			DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
			DBUG_RETURN((char *)NULL);
		}
		else if (!no_trunc_s)
		{
			xx_rpt(UNSUPPORTED);
			in_rpt("_POSIX_NO_TRUNC is not in effect for server");
			DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
			DBUG_RETURN((char *)NULL);
		}
		else if (name_max_s < NFS_MAXNAMLEN)
		{
			xx_rpt(UNSUPPORTED);
			in_rpt("NAME_MAX on the server is < NFS_MAXNAMLEN");
			DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
			DBUG_RETURN((char *)NULL);
		}
		else
		{
			len = NFS_MAXNAMLEN;
			PATH_TRACE;
		}
		break;

	case 4:
		/* _POSIX_NO_TRUNC is defined on the client, NAME_MAX on the
		   client is greater than or equal to NFS_MAXNAMLEN and
		   _POSIX_NO_TRUNC is not defined on the server */

		if (!no_trunc_c)
		{
			xx_rpt(UNSUPPORTED);
			in_rpt("_POSIX_NO_TRUNC is not in effect for client");
			DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
			DBUG_RETURN((char *)NULL);
		}
		else if (name_max_c < NFS_MAXNAMLEN)
		{
			xx_rpt(UNSUPPORTED);
			in_rpt("NAME_MAX on the client is < NFS_MAXNAMLEN");
			DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
			DBUG_RETURN((char *)NULL);
		}
		else if (no_trunc_s)
		{
			xx_rpt(UNSUPPORTED);
			in_rpt("_POSIX_NO_TRUNC is in effect for server");
			DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
			DBUG_RETURN((char *)NULL);
		}
		else
		{
			len = NFS_MAXNAMLEN;
			PATH_TRACE;
		}
		break;

	default:
		xx_rpt(DELETION);
		in_rpt("invalid argument to xnfs_nametoolng(): %d", casenum);
		DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
		DBUG_RETURN((char *)NULL);
	}

	names[0] = (char *)malloc((size_t)(len + 1));
	if (names[0] == NULL)
	{
		xx_rpt(DELETION);
		in_rpt("malloc() failed in xnfs_nametoolng()");
		DBUG_PRINTF("return", "xnfs_nametoolng = NULL");
		DBUG_RETURN((char *)NULL);
	}
	else
		PATH_TRACE;

	makename(names[0], len);

	PATH_FUNC_RPT(6); 

	DBUG_PRINTF("return", "xnfs_nametoolng = 0x%lx", (long)names[0]);
	DBUG_RETURN(names[0]);
}
#endif /* TEST_XNFS */

void
clntoolng()
{
	if (paths[0] != NULL)
	{
		(void) rmdir(TOPDIR);
		(void) rmdir(ALTDIR);
		free((void *)paths[0]);
		paths[0] = NULL;
	}
	if (paths[1] != NULL)
	{
		free((void *)paths[1]);
		paths[1] = NULL;
	}
	if (names[0] != NULL)
	{
		free((void *)names[0]);
		names[0] = NULL;
	}
	if (names[1] != NULL)
	{
		free((void *)names[1]);
		names[1] = NULL;
	}
}
