/*
*      SCCS:   @(#)common/tsetlib(pathres.c)	4.1 (01/21/97)  VSXgen release 1.4
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/


/***********************************************************************

NAME:		pathres.c - generate names for testing pathname resolution
			    and portable filenames

PROJECT:	VSX (X/OPEN Validation Suite)

MODULE CONTAINS:
	pathresnam[]
	portfnam[]
	init_pathres()
	root_alias()
	cln_pathres()

SYNOPSIS:
	#include <pathres.h>

	char *pathresnam[];

	char *portfnam[];

	int init_pathres(opts)
		int opts;

	char *root_alias();

	void cln_pathres();

DESCRIPTION:
	Pathresnam[] contains a list of pairs of pathnames which
	should refer to the same file if pathname resolution
	is performed correctly.  I.e. pathresnam[0] and pathresnam[1]
	should be the same file, and so on.

	Portfnam[] contains a null-terminated list of filenames making
	use of the entire portable filename character set.  It includes
	names which will erroneously refer to the same file if filenames
	are not case sensitive.

	Init_pathres() initialises pathresnam[] and creates a
	directory hierarcy, then does a chdir() to the top directory
	of the hierarchy.  The "opts" argument consists of the
	bitwise OR of the following elements:

		PRES_FILES	paths refer to plain files
		PRES_DIRS	paths refer to directories
		PRES_NOROOT	exclude name pairs for "/"
		PRES_NOTRAIL	exclude names with trailing "/", "/." or "/.."
		PRES_NODOT	exclude names with trailing "/."
		PRES_NODOTDOT	exclude names with trailing "/.."
		PRES_CREAT	create the listed files/dirs

	One of PRES_FILES and PRES_DIRS must be specified, the others
	are optional.  If PRES_DIRS is specified, extra name pairs
	which apply only to directories are added to the list.  By
	default name pairs for "/" WILL be added UNLESS opts includes
	PRES_NOROOT.  Likewise for PRES_NOTRAIL, PRES_NODOT and
	PRES_NODOTDOT.

	Init_pathres() returns the number of pathname pairs, or -1
	if an error occurs.

	Root_alias() can be called directly to obtain a relative path
	to "/", for testsets which can't do the same with "/" as with
	the other pathname pairs.  It returns a pointer to static data,
	or NULL if an error occurs.  Root_alias() should only be used
	if PRES_NOROOT was specified when init_pathres() was called.

	Cln_pathres() does a chdir("..") and removes the directory
	hierarchy created by init_pathres().

CONDITIONAL COMPILE PARAMS:
SIDE EFFECTS:
	Current directory is changed by init_pathres() and restored
	by cln_pathres().

RELATED DOCUMENTS:
SOURCE:
AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	15 Mar 1989
MODIFICATIONS:
		Geoff Clare, 16 Nov 1989
			Cater for systems with NAME_MAX == PATH_MAX
		
		Geoff Clare, 17 Jan 1990
			Handle consecutive '/'s in current dir when
			forming relative path to root.
		
		Geoff Clare, 22 Mar 1990
			User supplied pathdepth() routine to obtain
			depth of paths beginning with "//".
		
		Geoff Clare, 5 July 1990
			Allow for unlimited NAME_MAX

		Geoff Clare, 8 May 1992
			Added "///path" case.
			Create directories for trailing slash cases.
			Exclude trailing slash cases if PRES_NOTRAIL is set.
			Added "dir/." and "dir/.." cases.

		Geoff Clare, 14 Oct 1992
			Changes for XNFS.

		Geoff Clare, 14 Jan 1993
			New configurable header <rpcincs.h>

		Geoff Clare, 27 Oct 1993
			Implement interpretation XNFS/003/931001.

************************************************************************/

#include <std.h>
#if TEST_XNFS
#include <rpcincs.h>	/* must preceed any system headers */
#endif
#include <dbug.h>
#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <limits.h>
#include <sys/stat.h>
#include <pathres.h>
#include <xnfs.h>
#if TEST_XNFS
#include <nfs.h>
#endif
#include <tet_api.h>
#include <sysdep.h>

#define NUMELT(a)	(sizeof(a)/sizeof(a[0]))
#define FIXEDPAIRS	8
#define MODEANY		(S_IRWXU|S_IRWXG|S_IRWXO)

extern char	*errname();

private int	set_cwd();
private int	name_max();
private int	cr_files();
private void	rm_files();

public char	*portfnam[] = {
	".ABCDEFGHI",
	"JKLMNOPQR_",
	"_STUVWXYZ.",
	".abcdefghi",
	"jklmnopqr_",
	"_stuvwxyz.",
	"0123456789-",
	NULL
};

private char	cwdpath[_POSIX_PATH_MAX+2] = "//";
private char	rootnam[_POSIX_PATH_MAX] = "";

public char	*pathresnam[] = {
	&cwdpath[2],	"F",
	cwdpath,	"F",		/* If you add or delete		*/
	"A/./F",	"A/F",		/* non-NULL entries, change	*/
	"./F",		"F",		/* FIXEDPAIRS above		*/
	"A/B/../F",	"A/F",
	"../CWD/F",	"F",
	"A//F",		"A/F",
	"A///F",	"A/F",
	NULL,		NULL,		/* space for dirresnam[]	*/
	NULL,		NULL,		/*   dotnam[] and dotdotnam[]	*/
	NULL,		NULL,		/*   (must have same no. 	*/
	NULL,		NULL,		/*    of elements)		*/
	NULL,		NULL,		/* space for 1st NAME_MAX pair	*/
	NULL,		NULL,		/* space for 2nd NAME_MAX pair	*/
	NULL,		NULL,		/* space for 1st root dir pair	*/
	NULL,		NULL,		/* space for 2nd root dir pair	*/
};

private char	*dirresnam[] = {	/* extra directory pairs */
	"A/D1/",	"A/D1",
	"A/D1//",	"A/D1",
};

private char	*dotnam[] = {		/* extra trailing "/." pairs */
	"A/D1/.",	"A/D1",
};

private char	*dotdotnam[] = {	/* extra trailing "/.." pairs */
	"A/D2/D/..",	"A/D2",
};

private char	*name[] = {		/* NAME_MAX pairs */
	NULL,		NULL,
	NULL,		NULL,
};

private int	npairs;

public char *
root_alias()
{
	/* work out a relative pathname for "/" of the form "../../.." */

	char *p;
	int dircnt;

	DBUG_LENTER("root_alias", 2);

	if (getcwd(rootnam, (int)sizeof(rootnam)) == NULL)
	{
		in_rpt("getcwd() failed in root_alias() - errno %d (%s)",
			errno, errname(errno));
		DBUG_PRINTF("return", "root_alias = NULL");
		DBUG_RETURN((char *)NULL);
	}

	DBUG_PRINTF("root_alias", "current dir = %s", rootnam);

	/* count directory levels in current dir path */

	if (rootnam[0] == '/' && rootnam[1] == '/' && rootnam[2] != '/')
	{
		/* path begins "//" - call user supplied routine */
		dircnt = pathdepth(rootnam);

		DBUG_PRINTF("root_alias", "pathdepth() returned %d", dircnt);
	}
	else
	{
		dircnt = 0;
		for (p = rootnam; *p != '\0'; p++)
		{
			if (*p == '/')
			{
				++dircnt;
				while (*(p+1) == '/')
					++p;
				if (*(p+1) == '\0')
					--dircnt;
			}
		
		}
	}
	dircnt--;

	/* construct relative path with same number of "/"s */
	for (p = rootnam; dircnt > 0; dircnt--)
	{
		*p++ = '.';
		*p++ = '.';
		*p++ = '/';
	}
	*p++ = '.';
	*p++ = '.';
	*p = '\0';

	DBUG_PRINTF("return", "root_alias = %s", rootnam);
	DBUG_RETURN(rootnam);
}

public int
init_pathres(opts)
int opts;
{
	char *p;
	int i;

	DBUG_LENTER("init_pathres", 2);
	DBUG_PRINTF("args", "init_pathres(0%o)", opts);

	npairs = FIXEDPAIRS;

	/* Create directory hierarchy (for fixed pairs) */
	if (vxmkdir(p = "CWD") != SUCCEED ||
	    vxmkdir(p = "CWD/A") != SUCCEED ||
	    vxmkdir(p = "CWD/A/B") != SUCCEED ||
	    vxmkdir(p = "CWD/A/D1") != SUCCEED ||
	    vxmkdir(p = "CWD/A/D2") != SUCCEED ||
	    vxmkdir(p = "CWD/A/D2/D") != SUCCEED)
	{
		in_rpt("could not create directory %s in init_pathres() - errno %d (%s)",
			p, errno, errname(errno));
		DBUG_PRINTF("return", "init_pathres = -1");
		DBUG_RETURN(-1);
	}

	/* Put paths starting with "/dir" and "///dir" into first two pairs */
	if (set_cwd() != 0)
	{
		DBUG_PRINTF("return", "init_pathres = -1");
		DBUG_RETURN(-1);
	}

	/* Change directory to top of new hierarchy */
	if (chdir("CWD") == SYSERROR)
	{
		in_rpt("chdir() failed in init_pathres() - errno %d (%s)",
			errno, errname(errno));
		DBUG_PRINTF("return", "init_pathres = -1");
		DBUG_RETURN(-1);
	}

	/* Add extra pairs for directories, if required */
	if ((opts & PRES_DIRS) != 0 && (opts & PRES_NOTRAIL) == 0)
	{
		for (i = 0; i < NUMELT(dirresnam); i++)
			pathresnam[2*npairs + i] = dirresnam[i];
		npairs += NUMELT(dirresnam)/2;
	}

	/* Add extra pairs for trailing "/.", if required */
	if ((opts & PRES_DIRS) != 0 && (opts & PRES_NOTRAIL) == 0
		&& (opts & PRES_NODOT) == 0)
	{
		for (i = 0; i < NUMELT(dotnam); i++)
			pathresnam[2*npairs + i] = dotnam[i];
		npairs += NUMELT(dotnam)/2;
	}

	/* Add extra pairs for trailing "/..", if required */
	if ((opts & PRES_DIRS) != 0 && (opts & PRES_NOTRAIL) == 0
		&& (opts & PRES_NODOTDOT) == 0)
	{
		for (i = 0; i < NUMELT(dotdotnam); i++)
			pathresnam[2*npairs + i] = dotdotnam[i];
		npairs += NUMELT(dotdotnam)/2;
	}

	/* Add a NAME_MAX pair, if required */
	if (name_max() != 0)
	{
		(void) chdir("..");
		DBUG_PRINTF("return", "init_pathres = -1");
		DBUG_RETURN(-1);
	}

	/* Create the test files/directories */
	if ((opts & PRES_CREAT) != 0)
	{
		if (cr_files((opts & PRES_DIRS) != 0) != 0)
		{
			(void) chdir("..");
			DBUG_PRINTF("return", "init_pathres = -1");
			DBUG_RETURN(-1);
		}
	}

	/* Add root directory pairs, if required */
	if ((opts & PRES_DIRS) != 0 && (opts & PRES_NOROOT) == 0)
	{
		if ((pathresnam[2*npairs] = root_alias()) == NULL)
		{
			(void) chdir("..");
			DBUG_PRINTF("return", "init_pathres = -1");
			DBUG_RETURN(-1);
		}
		pathresnam[2*npairs + 1] = "/";
		++npairs;
		pathresnam[2*npairs] = "///";
		pathresnam[2*npairs + 1] = "/";
		++npairs;
	}

	DBUG_PRINTF("return", "init_pathres = %d", npairs);
	DBUG_RETURN(npairs);
}

public void
cln_pathres()
{
	/* Remove directory hierarchy, change back to previous
	   current directory, and free malloc()ed buffers */

	DBUG_LENTER("cln_pathres", 2);

	rm_files();

	if (name[0] != NULL)
	{
		free((void *)name[0]);
		name[0] = NULL;
	}
	if (name[1] != NULL)
	{
		free((void *)name[1]);
		name[1] = NULL;
	}
	if (name[2] != NULL)
	{
		free((void *)name[2]);
		name[2] = NULL;
	}
	if (name[3] != NULL)
	{
		*strrchr(name[3], '/') = '\0';	/* can't fail */
		(void) rmdir(name[3]);
		free((void *)name[3]);
		name[3] = NULL;
	}

	if (chdir("..") != SYSERROR)
	{
		(void) rmdir("CWD/A/D2/D");
		(void) rmdir("CWD/A/D2");
		(void) rmdir("CWD/A/D1");
		(void) rmdir("CWD/A/B");
		(void) rmdir("CWD/A");
		(void) rmdir("CWD");
	}
	else
	{
		DBUG_PRINTF("cln_pathres", "chdir(\"..\") failed");
		(void) rmdir("A/D2/D");
		(void) rmdir("A/D2");
		(void) rmdir("A/D1");
		(void) rmdir("A/B");
		(void) rmdir("A");
	}

	DBUG_VOID_RETURN;
}

private int
set_cwd()
{
	char *cp;
	int cnt;

	/* Put absolute path to current directory preceded by "//" */
	/* into cwdpath[] */

	cp = &cwdpath[2];
	if (getcwd(cp, (int)sizeof(cwdpath) - 9) == NULL)
	{
		in_rpt("getcwd() failed in init_pathres() - errno %d (%s)",
			errno, errname(errno));
		return -1;
	}
#ifdef TESTBED
cp[1] = '/';
#endif
	if (cp[0] == '/' && cp[1] == '/' && cp[2] != '/')
	{
		/* path begins "//" - adjust using user-supplied depth */
		cnt = pathdepth(cp);

		cp = cwdpath + strlen(cwdpath) - 1;
		/* ignore trailing slashes */
		while (*cp == '/' && cp > cwdpath+2)
			cp--;
		/* count back to '/' which is at the level of the root */
		while (cnt > 0 && cp > cwdpath+2)
		{
			while (*cp != '/' && cp > cwdpath+2)
				cp--;
			if (--cnt > 0)
			{
				while (*cp == '/' && cp > cwdpath+2)
					cp--;
			}
		}

		/* move new path down to &cwdpath[2] */
		for (cnt = 2; *cp != '\0'; cnt++, cp++)
			cwdpath[cnt] = *cp;
		cwdpath[cnt] = '\0';
	}

	(void) strcat(cwdpath, "/CWD/F");

	return 0;
}

private int
name_max()
{
	/* Add pathname pairs for NAME_MAX testing */

	long	path_max;
	long	no_trunc_c;
	long	name_max_c;
	int	i, len;
#if TEST_XNFS
	long	no_trunc_s;
	long	name_max_s;
	char	*cp;
#endif

	/* Obtain _POSIX_NO_TRUNC and NAME_MAX for client */

	errno = 0;
	no_trunc_c = pathconf(".", _PC_NO_TRUNC);
	if (no_trunc_c < 0 && errno != 0)
	{
		in_rpt("pathconf(\".\", _PC_NO_TRUNC) failed in init_pathres(), errno = %d (%s)",
			errno, errname(errno));
		return -1;
	}

	errno = 0;
	name_max_c = pathconf(".", _PC_NAME_MAX);
	if (name_max_c < 0 && errno != 0)
	{
		in_rpt("pathconf(\".\", _PC_NAME_MAX) failed in init_pathres(), errno = %d (%s)",
			errno, errname(errno));
		return -1;
	}

#if !TEST_XNFS
	if (no_trunc_c >= 0 || name_max_c < 0)
	{
		/* _POSIX_NO_TRUNC is defined or NAME_MAX is unlimited:
		   test not required */
		return 0;
	}
	len = name_max_c;

#else /* if TEST_XNFS */

	/* Obtain _POSIX_NO_TRUNC and NAME_MAX for server */

	cp = tet_getvar("XNFS_SERVER_NO_TRUNC");
	if (cp == NULL || *cp == '\0')
	{
		in_rpt("parameter XNFS_SERVER_NO_TRUNC is not set");
		return -1;
	}
	no_trunc_s = (*cp == 'Y' || *cp == 'y');

	cp = tet_getvar("XNFS_SERVER_NAME_MAX");
	if (cp == NULL || *cp == '\0')
	{
		in_rpt("parameter XNFS_SERVER_NAME_MAX is not set");
		return -1;
	}
	else if ((name_max_s = atol(cp)) <= 0 && name_max_s != -1)
	{
		in_rpt("parameter XNFS_SERVER_NAME_MAX has an invalid value");
		return -1;
	}

	if (name_max_c < 0)
		name_max_c = NFS_MAXNAMLEN + 1;
	if (name_max_s < 0)
		name_max_s = NFS_MAXNAMLEN + 1;

	if (name_max_c <= name_max_s)
	{
		/* Truncation will occur on client if _POSIX_NO_TRUNC
		   is false and NAME_MAX <= NFS_MAXNAMLEN */

		if (no_trunc_c < 0 && name_max_c <= NFS_MAXNAMLEN)
			len = name_max_c;
		else
			return 0;
	}
	else
	{
		/* Truncation will occur on server if _POSIX_NO_TRUNC
		   is false and NAME_MAX < NFS_MAXNAMLEN */

		if (no_trunc_s < 0 && name_max_s < NFS_MAXNAMLEN)
			len = name_max_s;
		else
			return 0;
	}

#endif /* if TEST_XNFS */

	/* 1st pair: name[0] = "(NAME_MAX+1 chars)" */
	/*           name[1] = "(NAME_MAX chars)"  */

	if ((name[0] = (char *)malloc((size_t)(len+2))) == NULL ||
	    (name[1] = (char *)malloc((size_t)(len+2))) == NULL)
	{
		in_rpt("malloc(%u) failed in init_pathres()", len+2);
		return -1;
	}

	for (i = 0; i < len+1; i++)
		name[0][i] = '0' + (i+1)%10;
	name[0][i] = '\0';

	for (i = 0; i < len; i++)
		name[1][i] = '0' + (i+1)%10;
	name[1][i] = '\0';

	pathresnam[2*npairs] = name[0];
	pathresnam[2*npairs + 1] = name[1];
	++npairs;

	/* 2nd pair: name[2] = "A/(NAME_MAX+1 chars)/F" */
	/*           name[3] = "A/(NAME_MAX chars)/F" */

	/* If these names would exceed PATH_MAX, this is untestable */
	path_max = pathconf(".", _PC_PATH_MAX);
	if (path_max > 0 && len+4 > path_max)
		return 0;

	if ((name[2] = (char *)malloc((size_t)(len+6))) == NULL ||
	    (name[3] = (char *)malloc((size_t)(len+6))) == NULL)
	{
		in_rpt("malloc(%u) failed in init_pathres()", len+6);
		return -1;
	}

	name[2][0] = 'A';
	name[2][1] = '/';
	for (i = 2; i < len+3; i++)
		name[2][i] = '0' + (i-1)%10;
	name[2][i++] = '/';
	name[2][i++] = 'F';
	name[2][i] = '\0';

	name[3][0] = 'A';
	name[3][1] = '/';
	for (i = 2; i < len+2; i++)
		name[3][i] = '0' + (i-1)%10;
	name[3][i] = '\0';
	if (vxmkdir(name[3]) != SUCCEED)
	{
		in_rpt("could not create directory %s in init_pathres() - errno %d (%s)",
			name[3], errno, errname(errno));
		return -1;
	}
	name[3][i++] = '/';
	name[3][i++] = 'F';
	name[3][i] = '\0';

	pathresnam[2*npairs] = name[2];
	pathresnam[2*npairs + 1] = name[3];
	++npairs;

	return 0;
}

private int
cr_files(ifdirs)
int ifdirs;
{
	/*
	 * Create all the files/directories listed in pathresnam[].
	 * The second name of each pair is used as it is the simpler
	 * of the two.  Note that duplicate names in the list will not
	 * cause problems as vxmkdir() and crfile() both succeed if
	 * the dir/file already exists.
	 */
	
	char **namp;
	int i, fd, uid, gid;

	if (!ifdirs)
	{
		uid = getuid();
		gid = getgid();
	}

	for (namp = &pathresnam[1], i = 0; i < npairs; i++, namp += 2)
	{
		if (ifdirs)
		{
			if (vxmkdir(*namp) != SUCCEED)
			{
				in_rpt("could not create directory %s in init_pathres() - errno %d (%s)",
					*namp, errno, errname(errno));
				return -1;
			}
		}
		else
		{
			if ((fd = crfile(*namp, (mode_t)MODEANY, uid, gid)) == SYSERROR)
			{
				in_rpt("could not create file %s in init_pathres() - errno %d (%s)",
					*namp, errno, errname(errno));
				return -1;
			}
			(void) close(fd);
		}
	}

	return 0;
}

private void
rm_files()
{
	/* Remove all the files/directories in pathresnam[] */
	/* Skips name pairs for "/" provided the 2nd name is just "/" */
	
	char **namp;
	int i;

	for (namp = &pathresnam[1], i = 0; i < npairs; i++, namp += 2)
	{
		if (strcmp(*namp, "/") != 0)
			(void) remove(*namp);
	}
}
