/*
*      SCCS:  @(#)common/tsetlib (pipeclib.c)	4.1 (01/21/97)  VSXgen release 1.4
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/

/***********************************************************************

NAME:		pipeclib.c - pipe reporting library - child (reporting)
			process functions.

PROJECT:	VSX (X/OPEN Validation Suite)

MODULE CONTAINS:
		pipe_xs_rpt()		- like xs_rpt()
		pipe_xx_rpt()		- like xx_rpt()
		pipe_in_rpt()		- like in_rpt()
		pipe_end_rpt()		- called after reporting has finished
		pipe_path_xs_rpt()	- PATH_XS_RPT -with (number, pathok)
		pipe_devinode_rpt()	- pipe version of devinode_rpt()
		pipefd_xx_rpt()		- used here and in pipeplib.c
		pipefd_in_rpt()		- used here and in pipeplib.c
		pipefd_end_rpt()	- used here and in pipeplib.c
SYNOPSIS:
		void
		pipe_xs_rpt();

		void
		pipe_xx_rpt(int code);

		void
		pipe_in_rpt(char *line);

		void
		pipe_end_rpt();

		void
		pipe_path_xs_rpt(int expected, int pathok);

		void
		pipe_devinode_rpt(ino_t inode1, dev_t dev1,
				  ino_t inode2, dev_t dev2);
DESCRIPTION:
		These functions offer simple reporting where direct calls
		to the reporting functions cause problems. They replace the
		functions xs_rpt, xx_rpt, in_rpt and the macro PATH_XS_RPT in
		the child processes where normal reporting cannot be performed.
		The parent process must have initialised the pipes before
		use, and obviously must interrogate the pipe for the results!
		See pipeplib.c for details.
		The pipefd_*_rpt() functions are only intended to be called
		from within other pipe_*() functions.
		However, they must be public because they are called
		from pipeplib.c as well as from within this file.
METHODS:
		stdout is duped onto the write side of a pipe - the reporting
		messages are passed up the pipe to the child.
ASSUMES:
		Messages are not lengthy.  (the pipe could block, causing the
		child to time out - the pipe is currently not examined until
		the child has exited) 

CONDITIONAL COMPILE PARAMS:
SIDE EFFECTS:
		STDOUT cannot be used in the test.
RELATED DOCUMENTS:
SOURCE:
AUTHOR:		Stuart Boutell, UniSoft Ltd.
DATE CREATED:	Thu Nov 21 10:00:00 GMT 1990
MODIFICATIONS:
		Geoff Clare, 5 June 1992
			Change pipe_devinod_rpt to use %g and (double) in
			case dev_t or ino_t is double.

		Andrew Dingwall, UniSoft Ltd., September 1995
		More robust version - permits pipe_in_rpt(),
		pipe_xx_rpt() etc to be called any number
		of times and in any order.

************************************************************************/
#include	<std.h>
#include	<report.h>
#include	<sys/types.h>
#include	<unistd.h>
#include	<stdlib.h>
#include	<stdio.h>
#include	<errno.h>
#include	<ctype.h>
#include	<dbug.h>
#include	<string.h>
#include	<sysdep.h>
#include	"pipelib.h"

public	void	pipe_xx_rpt();
public	void	pipe_xs_rpt();
public	void	pipe_in_rpt();
public	void	pipe_end_rpt();
public	void	pipe_path_xs_rpt();
public	void	pipefd_xx_rpt();
public	void	pipefd_in_rpt();
public	void	pipefd_end_rpt();

private void pipefd_writemsg();
extern char *errname();

#ifdef	TEST_PIPE_REPORTING
private	char	*test_message = "<Test message from pipe_reporting subsystem>";
#endif

public void
pipe_xs_rpt()
{
	struct pipemsg msg;

	DBUG_LENTER("pipe_xs_rpt", 2);

	DBUG_PRINTF("pipe_xs_rpt", "REPORTING SUCCESS");

	msg.pm_mtype = PM_XS_RPT;
	msg.pm_dlen = 0;
	pipefd_writemsg(&msg, STDOUT_FILENO);

#ifdef	TEST_PIPE_REPORTING
	pipe_in_rpt(test_message);
#endif

	DBUG_VOID_RETURN;
}

public void
pipe_xx_rpt(c)
int c;
{
	pipefd_xx_rpt(c, STDOUT_FILENO);
}

public void
pipefd_xx_rpt(c, fd)
int c, fd;
{
	struct pipemsg msg;

	DBUG_LENTER("pipefd_xx_rpt", 2);

	DBUG_PRINTF("pipefd_xx_rpt", "REPORTING CODE %d (%c) to fd %d",
		c, isprint(c) ? c : '\0', fd);

	msg.pm_mtype = PM_XX_RPT;
	msg.pm_dlen = sizeof msg.pm_rescode;
	msg.pm_rescode = c;
	pipefd_writemsg(&msg, fd);

	DBUG_VOID_RETURN;
}

public void
pipe_end_rpt()
{
	pipefd_end_rpt(STDOUT_FILENO);
}

public void
pipefd_end_rpt(fd)
int fd;
{
	struct pipemsg msg;

	DBUG_LENTER("pipefd_end_rpt", 2);

	DBUG_PRINTF("pipefd_end_rpt","REPORTING ENDED");

	msg.pm_mtype = PM_END_RPT;
	msg.pm_dlen = 0;
	pipefd_writemsg(&msg, fd);

	DBUG_VOID_RETURN;
}

public void
pipe_in_rpt(s)
char *s;
{
	pipefd_in_rpt(s, STDOUT_FILENO);
}

public void
pipefd_in_rpt(s, fd)
char *s;
int fd;
{
	struct pipemsg msg;
	int len;

	DBUG_LENTER("pipe_in_rpt", 2);

	DBUG_PRINTF("pipe_in_rpt","REPORTING STRING : %s",s);

	msg.pm_mtype = PM_IN_RPT;
	len = strlen(s);

	do {
		msg.pm_dlen = MIN(len + 1, sizeof msg.pm_text);
		(void) sprintf(msg.pm_text, "%.*s", sizeof msg.pm_text - 1, s);
		pipefd_writemsg(&msg, fd);
		s += sizeof msg.pm_text - 1;
		len -= sizeof msg.pm_text - 1;
	} while (len > 0);

	DBUG_VOID_RETURN;
}

public void
pipe_path_xs_rpt(expected, got)
int expected, got;
{
	char rptbuf[256];

	DBUG_LENTER("pipe_path_xs_rpt", 2);

	DBUG_PRINTF("pipe_path_xs_rpt","REPORTING PATH SUCCESS");

	if (got == expected)
	{
		pipe_xs_rpt();
	}
	else
	{
		pipe_xx_rpt(DELETION);
		(void) sprintf(rptbuf,"deletion reason: path check value %d",
			got);
		pipe_in_rpt(rptbuf);
	}

	DBUG_VOID_RETURN;
}

public void
pipe_devinod_rpt(inode1, dev1, inode2, dev2)
ino_t	inode1;
dev_t	dev1;
ino_t	inode2;
dev_t	dev2;
{
	char	rptbuf[256];

	DBUG_LENTER("pipe_devinod_rpt", 2);

	(void) sprintf(rptbuf, "inode expected %g, device expected %g",
		(double)inode1, (double)dev1);
	pipe_in_rpt(rptbuf);
	(void) sprintf(rptbuf, "actual inode %g, actual device %g",
		(double)inode2, (double)dev2);
	pipe_in_rpt(rptbuf);

	DBUG_VOID_RETURN;
}

private void
pipefd_writemsg(mp, fd)
struct pipemsg *mp;
int fd;
{
	int len, n, err;

	DBUG_LENTER("pipefd_writemsg", 2);

	mp->pm_magic = PM_MAGIC;
	len = pipemsgsz(mp);

	DBUG_PRINTF("pipefd_writemsg", "write %d bytes to fd %d", len, fd);

	if ((n = write(fd, (char *) mp, (unsigned) len)) != len)
	{
		err = errno;
		if (n >= 0)
			DBUG_PRINTF("pipefd_writemsg",
				"write() failed, requested %d, returned %d",
					len, n);
		else
			DBUG_PRINTF("pipefd_writemsg",
				"write() failed, errno = %d (%s)",
				err, errname(err));

	}

	DBUG_VOID_RETURN;
}

