/*
*      SCCS:  @(#)common/tsetlib (pipeplib.c)	4.1 (01/21/97)  VSXgen release 1.4
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/

/***********************************************************************

NAME:		pipeplib.c - pipe reporting library used where normal
				reporting cannot be used.

PROJECT:	VSX (X/OPEN Validation Suite)

MODULE CONTAINS:
		pipe_parent() - reporting function
		pipe_child()  - child initialisation function.
		pipe_open()   - set up pipes - returns 1 on success, 0 on fail.
				Reports pipe failure for user.
		pipe_close()  - close pipes when finished.
		pipe_report() - parent pipe reporting for child processes.
SYNOPSIS:
		void
		pipe_parent();

		void
		pipe_child();

		int
		pipe_open();

		void
		pipe_close();
	
		void
		pipe_report(int code, char *message);
DESCRIPTION:
		These functions combine to provide a pipe reporting mechanism
		which can be used where normal reporting cannot - for example
		in testsets that deal with file descriptors often cannot use
		the tetapi reporting.
METHODS:
		Before the cppair, call pipe_open - if the return value is non
		zero, then the cppair may be called - otherwise the test will
		be deleted because the pipe open call failed.
		The child process of the cppair MUST call pipe_child()
		before using any of the pipe reporting functions. This
		closes STDOUT and dups the write side of the pipe to it.

		pipe_report() allows the child to perform simple reporting
		of a single line and the result code, without the requirement
		of using the pipeclib.c calls.

		AFTER THE CPPAIR HAS RETURNED, the test must call pipe_parent()
		to perform the reporting. A pipe_close() call will close the
		pipe silently.
ASSUMES:
		Messages are short. (Otherwise pipes could block)
CONDITIONAL COMPILE PARAMS:
SIDE EFFECTS:
		STDOUT cannot used by the child process.
RELATED DOCUMENTS:
SOURCE:

AUTHOR:		Stuart Boutell, UniSoft Ltd.
DATE CREATED:	Thu Nov 21 10:00:00 GMT 1990
MODIFICATIONS:
		Andrew Dingwall, UniSoft Ltd., September 1995
		More robust version - permits pipe_in_rpt(),
		pipe_xx_rpt() etc to be called any number
		of times and in any order.

************************************************************************/
#include	<std.h>
#include	<report.h>
#include	<sys/types.h>
#include	<errno.h>
#include	<unistd.h>
#include	<stdlib.h>
#include	<stdio.h>
#include	<fcntl.h>
#include	<string.h>
#include	<dbug.h>
#include	<sysdep.h>
#include	"pipelib.h"

#define		READ 0
#define		WRITE 1

extern	char	*errname();

private	int	pipe_fds[2];

private int pipefd_getmsg(), pipefd_read();
private char *pipe_msgtype();


public int
pipe_open()
{
	int r;
	DBUG_LENTER("pipe_open", 2);

	r = pipe(pipe_fds);
	if(r==SYSERROR)
	{
		int err = errno;
		xx_rpt(DELETION);
		in_rpt("pipe() call failed - errno %d(%s)", err, errname(err));
		DBUG_RETURN(0);
	}
	DBUG_RETURN(1);
}
	
public void
pipe_close()
{
	DBUG_LENTER("pipe_close", 2);
	(void) close(pipe_fds[READ]);
	(void) close(pipe_fds[WRITE]);
	DBUG_VOID_RETURN;
}

/*
**	pipe_parent() - read messages from child until either a
**		message of type PM_END_RPT is received or EOF is encountered
*/

public	void
pipe_parent()
{
	struct pipemsg msg;

	DBUG_LENTER("pipe_parent", 2);

	(void) close(pipe_fds[WRITE]);

	while (pipefd_getmsg(&msg, pipe_fds[READ]) > 0)
	{
		DBUG_PRINTF("pipe_parent", "received %s message",
			pipe_msgtype(msg.pm_mtype));
		switch (msg.pm_mtype)
		{
		case PM_XS_RPT:
			xs_rpt();
			break;
		case PM_XX_RPT:
			xx_rpt(msg.pm_rescode);
			break;
		case PM_IN_RPT:
			msg.pm_text[MIN(sizeof msg.pm_text, msg.pm_dlen) - 1] = '\0';
			in_rpt("%s", msg.pm_text);
			break;
		case PM_END_RPT:
			DBUG_VOID_RETURN;
			break;
		default:
			in_rpt("pipe_parent(): received unexpected message type %d", msg.pm_mtype);
			xx_rpt(DELETION);
			break;
		}
	}

	DBUG_VOID_RETURN;
}

/*
**	pipefd_getmsg() - read a single message from the child
**
**	return	1 if successful
**		0 for EOF
**	       -1 on error
*/

private int
pipefd_getmsg(mp, fd)
struct pipemsg *mp;
int fd;
{
	int count, n, rc;
	union {
		long magic;
		char mbyte[sizeof (long)];
	} zu;

	DBUG_LENTER("pipefd_getmsg", 2);

	/* read the message magic number */
	if ((rc = pipefd_read(fd, (char *) &mp->pm_magic, sizeof mp->pm_magic, 1)) <= 0)
		DBUG_RETURN(rc);

	/* make sure that we are in sync with sender - resync if not */
	for (count = 0; mp->pm_magic != PM_MAGIC; count++)
	{
		if (count == 0)
		{
			in_rpt("pipefd_getmsg(): out of sync, seeking - some messages from child may have been lost");
			xx_rpt(DELETION);
		}
		zu.magic = mp->pm_magic;
		for (n = 1; n < sizeof zu.mbyte; n++)
			zu.mbyte[n - 1] = zu.mbyte[n];
		if ((rc = pipefd_read(fd, &zu.mbyte[sizeof zu.mbyte - 1], 1, 1)) <= 0)
			DBUG_RETURN(rc);
		mp->pm_magic = zu.magic;
	}
	if (count)
		in_rpt("pipefd_getmsg(): resync after reading %d bytes", count);

	/* read the message header */
	if ((rc = pipefd_read(fd, (char *) &mp->pm_hdr, sizeof mp->pm_hdr, 0)) <= 0)
		DBUG_RETURN(rc);

	/* return now if there is no message data to read */
	if (mp->pm_dlen <= 0)
		DBUG_RETURN(1);

	/* here to read message data */
	if (mp->pm_dlen > sizeof mp->pm_data)
	{
		in_rpt("pipefd_getmsg(): preposterous pm_dlen = %d, message truncated", mp->pm_dlen);
		mp->pm_dlen = sizeof mp->pm_data;
	}
	rc = pipefd_read(fd, (char *) &mp->pm_data, mp->pm_dlen, 0);

	DBUG_RETURN(rc);
}

/*
**	pipefd_read() - read len bytes from the child into *buf
**
**	if eof_ok is false, EOF or a partial read is considered an error
**
**	return	1 if successful
**		0 on EOF or partial read
**	       -1 on error
*/

private int
pipefd_read(fd, buf, len, eof_ok)
int fd, len, eof_ok;
char *buf;
{
	int n, err;

	DBUG_LENTER("pipefd_read", 2);

	if ((n = read(fd, buf, (unsigned) len)) == len)
	{
		DBUG_RETURN(1);
	}
	else if (n == 0)
	{
		if (eof_ok)
			DBUG_RETURN(0);
		else
		{
			in_rpt("pipefd_read(): unexpected EOF");
			xx_rpt(DELETION);
			DBUG_RETURN(-1);
		}
	}
	else if (n < 0)
	{
		err = errno;
		in_rpt("pipefd_read(): read failed, errno = %d (%s)",
			err, errname(err));
		xx_rpt(DELETION);
		DBUG_RETURN(-1);
	}
	else
	{
		in_rpt("pipefd_read(): unexpected partial read: expected %d bytes, received %d", len, n);
		xx_rpt(DELETION);
		(void) memset(buf + n, '\0', (size_t) (len - n));
		DBUG_RETURN(eof_ok ? 0 : -1);
	}
}

private char *
pipe_msgtype(msgtype)
int msgtype;
{
	static char buf[64];

	switch (msgtype)
	{
	case PM_XS_RPT:
		return("PM_XS_RPT");
	case PM_XX_RPT:
		return("PM_XX_RPT");
	case PM_IN_RPT:
		return("PM_IN_RPT");
	case PM_END_RPT:
		return("PM_END_RPT");
	default:
		(void) sprintf(buf, "<unexpected type %d>", msgtype);
		return(buf);
	}
}

public	void
pipe_report(code,message)
int	code;
char	*message;
{
	DBUG_LENTER("pipe_report", 2);

	pipefd_in_rpt(message, pipe_fds[WRITE]);
	pipefd_xx_rpt(code, pipe_fds[WRITE]);
	pipefd_end_rpt(pipe_fds[WRITE]);

	DBUG_VOID_RETURN;
}

public	void
pipe_child()
{
	DBUG_LENTER("pipe_child", 2);

	(void) close(STDOUT_FILENO);

	if (fcntl(pipe_fds[WRITE], F_DUPFD, STDOUT_FILENO) != STDOUT_FILENO)
		pipe_report(DELETION, "dup on pipe_fds to stdout failed");

	DBUG_VOID_RETURN;
}

