/*
*      SCCS:   @(#)common/tsetlib(pure_run.c)	4.1 (01/21/97)  VSXgen release 1.4
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/


/***********************************************************************

NAME:		pure_run.c - run a pure executable

PROJECT:	VSX (X/OPEN Validation Suite)

MODULE CONTAINS:
	pure_copy()
	pure_run()

SYNOPSIS:
	int pure_copy(from, to)
		char	*from;
		char	*to;

	int pure_run(path, timeout, pid)
		char	*path;
		int	timeout
		int	*pid;

DESCRIPTION:
	  Pure_copy() copies the "from" to the file "to".  This is done
	with a system("cp from to"),  after the copy, the mod is changed
	to rwx for everyone.  If the system() returns a non-SUCCEED value
	or if chmod() fails then "to" is unlink()'ed and SYSERROR returned.
	If all goes well then SUCCEED is returned.

	  Pure_run() forks and runs the pure executable "path", and puts
	the process id of the child in *pid, SUCCEED is returned is
	all goes well.  The timeout value is passed to the child, 
	and it will wait a maximium of timeout seconds before exiting.
	This is done so that no processes will be left around.  If
	something goes wrong, the return value will indicate the
	error by using "exitcode.h" values.
	The possible failures are:
		A_NOMEM		malloc() failed
		A_OPENSYNC	opensync() failed
		A_FORKFAIL	fork() failed
		A_WAITSYNC	waitsync() failed

CONDITIONAL COMPILE PARAMS:
SIDE EFFECTS:
RELATED DOCUMENTS:
SOURCE:
AUTHOR:		Mike Rendell, UniSoft Ltd.
DATE CREATED:	Mon Oct  6 13:54:37 BST 1986
MODIFICATIONS:	Dave Cater, Fri Apr 28 14:59:54 BST 1989
		Make pure_copy() public for T.exec_X

		Anthony Lee, 16 July 1990
		Added resillience code.

		Geoff Clare, 25 June 1991
			VSX_PURE_FILE="unsup" for unsupported result.

		Geoff Clare, 14 Oct 1992
			Changes for XNFS.

************************************************************************/

#include	<std.h>

#if !TEST_XNFS
#include	<errno.h>
#include	<stdio.h>
#include	<dbug.h>
#include	<exitcodes.h>
#include	<synch.h>
#include	<limits.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<report.h>
#include	<string.h>
#include	<unistd.h>
#include	<stdlib.h>
#include <tet_api.h>
#include	<sysdep.h>

#define	WAITTIME	(SPEEDFACTOR * 5)
#define	SYNCSIZE	4
#define	VSX_PURE_FILE	"VSX_PURE_FILE"

extern	char	*newstr();
extern	char	*errname();
extern	char	*itoa();
extern	char	*newstr();
extern	int	waitsync();
extern	int	killchild();

extern	char	*vg_dbug;

public int	pure_copy();
public int	pure_run();


public	int
pure_copy(from, to)
	char	*from;
	char	*to;
{
	char	buf[_POSIX_PATH_MAX * 2 + 30];
	int 	pathok = 0; 

	DBUG_LENTER("pure_copy", 2);
	globok = 0;

	if (strlen(from) + strlen(to) > 2 * _POSIX_PATH_MAX)
	{
		vsxerror("path names too long in pure_run()");
		DBUG_RETURN(SYSERROR);
	}
	else
		PATH_TRACE; 

	/* generate the string "cp from to" */
	(void) strcat(strcat(strcat(strcat(strcpy(buf, "cp "), from), " "),
		to), "> /dev/null 2>&1");
	if (system(buf) != SUCCEED
		|| chmod(to, S_IRWXU|S_IRWXG|S_IRWXO) == SYSERROR)
	{
		(void) unlink(to);
		DBUG_RETURN(SYSERROR);
	}
	else
		PATH_TRACE;

	PATH_FUNC_RPT(2); 
		
	DBUG_RETURN(SUCCEED);
}


public int
pure_run(path, timeout, pid)
	char	*path;
	int	timeout;
	pid_t	*pid;
{
	char	*pure;
	char	*sfd;
	char	*stimeout;
	char	syncbuf[SYNCSIZE];
	synch	*syncb;
	int 	pathok = 0; 

	DBUG_LENTER("pure_run", 2);
	DBUG_PRINTF("args", "pure_run(%s, %d, %ld)", path, timeout, (long)*pid);

	if ((pure = tet_getvar(VSX_PURE_FILE)) == NULL || *pure == '\0')
	{
		xx_rpt(DELETION);
		in_rpt("parameter %s is not set", VSX_PURE_FILE);
		DBUG_PRINTF("return", "pure_run = %d", A_MISC);
		DBUG_RETURN (A_MISC);
	}
	else if (strncmp(pure, "unsup", (size_t)5) == 0)
	{
		xx_rpt(UNSUPPORTED);
		in_rpt("parameter %s is set to \"unsup\"", VSX_PURE_FILE);
		DBUG_PRINTF("return", "pure_run = %d", A_MISC);
		DBUG_RETURN (A_MISC);
	}
	else
		PATH_TRACE;
			
	if (pure_copy(pure, path) != SUCCEED)
	{
		xx_rpt(DELETION);
		in_rpt("deletion reason: could not copy %s to %s", pure, path);
		DBUG_PRINTF("return", "pure_run = %d", A_MISC);
		DBUG_RETURN (A_MISC);
	}
	else
		PATH_FUNC_TRACE;
		
	if ((syncb = opensync()) == (synch *) NULL)
	{
		xx_rpt(DELETION);
		in_rpt("deletion reason: could not opensync() channel");
		DBUG_PRINTF("return", "pure_run = %d", A_OPENSYNC);
		DBUG_RETURN (A_OPENSYNC);
	}
	else
		PATH_TRACE;
		
	if ((sfd = newstr(itoa(syncb[0][1]))) == (char *) NULL
		|| (stimeout = newstr(itoa(timeout))) == (char *) NULL)
	{
		vsxerror("ran out of memory (malloc() failed)");
		DBUG_PRINTF("return", "pure_run = %d", A_NOMEM);
		DBUG_RETURN (A_NOMEM);
	}
	else
		PATH_TRACE;
		
	(void) fflush(stdout);
	(void) fflush(stderr);
	switch (*pid = fork())
	{
	case SYSERROR:
		DBUG_PRINTF("return", "pure_run = %d", A_FORKFAIL);
		DBUG_RETURN (A_FORKFAIL);

	case 0:	/* child */
		{
			char *args[5];

 			args[0] = path;
 			args[1] = sfd;
 			args[2] = stimeout;
 			args[3] = vg_dbug;
 			args[4] = NULL;

			(void) execv(path, args);
			vsxerror("execv(%s, ...) failed with errno %d(%s)",
				path, errno, errname(errno));
			exit(0);
		}

		/* parent falls through */
	}
	if (waitsync(syncb[0][0], syncbuf, sizeof(syncbuf), (unsigned)timeout)
								== SYSERROR)
	{
		(void) closesync(syncb);
		(void) killchild(*pid, (unsigned)timeout);
		xx_rpt(DELETION);
		in_rpt("deletion reason: waitsync() failed waiting for pure executable");
		DBUG_PRINTF("return", "pure_run = %d", A_WAITSYNC);
		DBUG_RETURN (A_WAITSYNC);
	}
	else
		PATH_TRACE;
		
	(void) closesync(syncb);

	DBUG_PRINTF("return", "pure_run = %d", SUCCEED);

	PATH_FUNC_RPT(5);
	
	DBUG_RETURN (SUCCEED);
}
#else /* !TEST_XNFS */
static int dummy;
#endif /* !TEST_XNFS */
