/*
*      SCCS:  @(#)common/tsetlib(termios.c)	4.1 (01/21/97)  VSXgen release 1.4
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/


/***********************************************************************

NAME:		termios.c - terminal setup for devclass tests

PROJECT:	VSX (X/OPEN Validation Suite)

MODULE CONTAINS:

	terms_prep()
	tty_prep()
	loop_prep()
	tty_init()
	loop_init()
	termios_prep()
	baud_mask()

	PERASE, PKILL, PINTR, PQUIT, PSTART, PSTOP,
	PEOF, PEOL, PSUSP, PMIN, PTIME

SYNOPSIS:

	int terms_prep(tty_flags, tty_fildes_p, tty_termios_p,
			loop_flags, loop_fildes_p, loop_termios_p)
		int tty_flags;
		int *tty_fildes_p;
		struct termios *tty_termios_p;
		int loop_flags;
		int *loop_fildes_p;
		struct termios *loop_termios_p;

	int tty_prep(tty_fildes_p, tty_termios_p)
		int *tty_fildes_p;
		struct termios *tty_termios_p;

	int loop_prep(loop_fildes_p, loop_termios_p)
		int *loop_fildes_p;
		struct termios *loop_termios_p;

	int termios_prep(tty_fildes_p, tty_termios_p,
				loop_fildes_p, loop_termios_p)
		int *tty_fildes_p;
		struct termios *tty_termios_p;
		int *loop_fildes_p;
		struct termios *loop_termios_p;

	int tty_init(fildes, termios_p)
		int fildes;
		struct termios *termios_p;

	int loop_init(fildes, termios_p)
		int fildes;
		struct termios *termios_p;

	speed_t baud_mask(speed) 
		char *speed;

	extern cc_t PERASE, PKILL, PINTR, PQUIT, PSTART, PSTOP,
		    PEOF, PEOL, PSUSP, PMIN, PTIME;

DESCRIPTION:
	terms_prep() opens VSX_TERMIOS_TTY and VSX_TERMIOS_LOOP using
	the flags specified and returns a file descriptor and an
	initialised termios structure for each.  It first calls
	tty_prep() to open VSX_TERMIOS_TTY so that if the processes
	is a session leader it will become the controlling terminal.
	The opens are done with a timeout of WAITTIME seconds.
	If VSX_MASTER_TTY and VSX_MASTER_LOOP are set, two processes are
	forked to provide a software loopback between the master pty's.

	tty_prep() opens and initialises VSX_TERMIOS_TTY.  It uses openctl()
	but if setsid() has not been called before calling tty_prep() then
	the terminal will not become the controlling terminal.
	If VSX_MASTER_TTY is set, a user-supplied pseudo-terminal opener,
	openpty(), is used instead.

	loop_prep() opens and initialises VSX_TERMIOS_LOOP using the same
	procedure as tty_prep() but using plain open() instead of openctl().

	termios_prep() calls setsid() then terms_prep() with O_RDWR flags.

	These four functions return 0 for success, -1 for failure (with
	an error message reported via vsxerror()), or 1 if the general
	terminal interface is unsupported.

	Note that to ensure a loopback is set up, either terms_prep()
	or termios_prep() must be used.  If tty_prep() and loop_prep()
	are used individually, no software loopback will be set up when
	pseudo-terminals are being used.

	tty_init() resets a termios structure to the initial settings
	on VSX_TERMIOS_TTY.

	loop_init() resets a termios structure to the initial settings
	on VSX_TERMIOS_LOOP.

	These two functions return 0 for success or -1 for failure.


CONDITIONAL COMPILE PARAMS:
SIDE EFFECTS:
RELATED DOCUMENTS:
SOURCE:
AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	6 Dec 1988
MODIFICATIONS:
		Geoff Clare, 13 Sep 1989
			Ensure LOOP has same c_cflag values as TTY
		
		Geoff Clare, 26 Sep 1989
			Set global data_rate to allow baud-dependent timeouts
		
		Geoff Clare, 6 June 1990
			VSX_TERMIOS_TTY=unsup means general terminal
			interface is unsupported

		Geoff Clare, 16 Oct 1990
			Set CREAD in tty_init() and loop_init()

		Geoff Clare, 23 Jan 1991
			Only clear ISTRIP if CSIZE is CS8

		Peter Skan, 30 Sept 1991
			Add setprv(PRV_DEVICE) to tty_prep() and loop_prep()

		Geoff Clare, 15 Nov 1991
			Changes to support pseudo-terminal loopbacks.

		Geoff Clare, 16 Mar 1992
			Use O_NOCTTY when opening loop side.

		Geoff Clare, 17 July 1992
			Change child1 and child2 from int to pid_t.

		Geoff Clare, 14 Oct 1992
			Changes for XNFS.

		Geoff Clare, 17 Nov 1993
			Check for O_NOCTTY flag when calling openpty() .

		Geoff Clare, 20 Sept 1995
			Use configured control characters instead of
			hard coded values.

		Geoff Clare, 28 Aug 1996
			Get ppid of rdwr() processes before forking.

************************************************************************/

#include <std.h>

#if !TEST_XNFS
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <limits.h>
#include <termios.h>
#include <string.h>
#include <timeout.h>
#include <tet_api.h>
#include <setprv.h>
#include <ctrlcodes.h>
#include <sysdep.h>

#define	WAITTIME	(5 * SPEEDFACTOR)
#define	COMBITS		11
#define TTY_SIDE	1
#define LOOP_SIDE	2

extern	char	*errname();
extern	void	setsigs();

extern	pid_t	child1; /* these are killed by tet_cleanup() if > 0 */
extern	pid_t	child2;

public	int	data_rate = 9600 / COMBITS;



/*********************************
* Special character constants
*********************************/

public cc_t PERASE	= PSL0_DEL;
public cc_t PKILL	= PSL0_NAK;
public cc_t PINTR	= PSL0_ETX;
public cc_t PQUIT	= PSL0_FS;
public cc_t PSTART	= PSL0_DC1;
public cc_t PSTOP	= PSL0_DC3;
public cc_t PEOF	= PSL0_EOT;
public cc_t PEOL	= PSL0_VT;
public cc_t PSUSP	= PSL0_SUB;
public cc_t PMIN	= 0;
public cc_t PTIME	= 10;

private char	*termios_speed;
private int	tty_done = 0;
private struct termios tty_termios;
private int	loop_done = 0;
private struct termios loop_termios;
private int	tty_master_fd = -1;
private int	loop_master_fd = -1;


/*********************************
* Baud rate table containing
* mask and associated decimal
* value
*********************************/
private struct mask_table
{
	speed_t mask;
	char *symbol;
} masks[] =
{
	B50, "B50",
	B75, "B75",
	B110, "B110",
	B134, "B134",
	B150, "B150",
	B200, "B200",
	B300, "B300",
	B600, "B600",
	B1200, "B1200",
	B1800, "B1800",
	B2400, "B2400",
	B4800, "B4800",
	B9600, "B9600",
	B19200, "B19200",
	B38400, "B38400",
};

#define BAUD_CNT (sizeof(masks)/sizeof(struct mask_table))

/******************************************
* return baud rate mask for symbol string
******************************************/
public speed_t
baud_mask(speed) 
char *speed;
{ 
	int i;

	for (i = 0; i < BAUD_CNT; ++i)
	{
		if (strcmp(speed,masks[i].symbol) == 0)
		{
			data_rate = atoi(&speed[1]) / COMBITS;
			return(masks[i].mask);
		}
	}
	/* no match */
	vsxerror("Illegal value for VSX_TERMIOS_SPEED: %s", speed);
	return (B9600);	/* must match default data_rate */
}

private int
getspeed()
{
	termios_speed = tet_getvar("VSX_TERMIOS_SPEED");
	if (termios_speed == NULL || *termios_speed == '\0')
	{
		vsxerror("parameter VSX_TERMIOS_SPEED not set");
		return -1;
	}
	return 0;
}

private void
rdwr(fdin, fdout)
int fdin;
int fdout;
{
	/* read from fdin and write to fdout - called in child process */

	int n;
	char buf[_POSIX_MAX_INPUT/2];

	while ((n = read(fdin, buf, sizeof(buf))) >= 0)
	{
		/* On EOF write EOF character */
		if (n == 0)
		{
			if (ptygetattr(fdout, &tty_termios) != SYSERROR)
				buf[n++] = tty_termios.c_cc[VEOF];
			else
				buf[n++] = PEOF;
		}

		errno = 0;
		if (write(fdout, buf, (unsigned int)n) != n)
		{
			/* die quietly on EIO */
			if (errno != EIO)
				vsxerror("write error on VSX_MASTER_TTY or VSX_MASTER_LOOP - errno %d (%s)",
					errno, errname(errno));
			_exit(1);
		}
	}
	/* die quietly on EIO */
	if (errno != EIO)
		vsxerror("read error on VSX_MASTER_TTY or VSX_MASTER_LOOP - errno %d (%s)",
			errno, errname(errno));
	_exit(1);
}

private int
ptycon(tty_fildes, loop_fildes)
int tty_fildes;
int loop_fildes;
{
	/* connect two master pty's so slaves can be used as loopback */

	pid_t ppid;

	switch (child1 = fork())
	{
		case -1:
			vsxerror("fork() failed in terms_prep()");
			return -1;
		
		case 0:
			setsigs(SIG_IGN);
			(void) close(tty_fildes);
			(void) close(loop_fildes);
			rdwr(tty_master_fd, loop_master_fd);
			_exit(0);
	}

	switch (child2 = fork())
	{
		case -1:
			vsxerror("fork() failed in terms_prep()");
			(void) killchild(child1, 2*SPEEDFACTOR);
			child1 = 0;
			return -1;
		
		case 0:
			setsigs(SIG_IGN);
			(void) close(tty_fildes);
			(void) close(loop_fildes);
			rdwr(loop_master_fd, tty_master_fd);
			_exit(0);
	}

	/* rdwr processes may not be killed by tests, so this process
	   checks for when the parent dies, and kills the rdwr's */

	ppid = getpid();
	switch (fork())
	{

		case -1:
			vsxerror("fork() failed in terms_prep()");
			(void) killchild(child1, 2*SPEEDFACTOR);
			child1 = 0;
			(void) killchild(child2, 2*SPEEDFACTOR);
			child2 = 0;
			return -1;
		
		case 0:
			setsigs(SIG_IGN);
			(void) close(tty_fildes);
			(void) close(loop_fildes);

			/* When parent PID changes, parent has died */
			while (ppid == getppid())
				(void) sleep(2);

			(void) kill(child1, SIGKILL);
			(void) kill(child2, SIGKILL);

			_exit(0);
	}

	return 0;
}

private int
do_prep(side, oflags, fildes_p, termios_p)
int side;
int oflags;
int *fildes_p;
struct termios *termios_p;
{
	/* common function called by tty_prep() and loop_prep() */

	char *termios_tty, *master_tty;
	int flags, cntrl, *master_fdp;

	*fildes_p = -1;

	if (side == TTY_SIDE)
		termios_tty = tet_getvar("VSX_TERMIOS_TTY");
	else
		termios_tty = tet_getvar("VSX_TERMIOS_LOOP");
	if (termios_tty == NULL || *termios_tty == '\0')
	{
		if (side == TTY_SIDE)
			vsxerror("parameter VSX_TERMIOS_TTY not set");
		else
			vsxerror("parameter VSX_TERMIOS_LOOP not set");
		return -1;
	}
	if (strncmp(termios_tty, "unsup", (size_t)5) == 0)
		return 1;

	/* check/obtain necessary privilege */
	if (setprv(PRV_DEVICE) != 0)
	{
		vsxerror("setprv(PRV_DEVICE) failed");
		return -1;
	}

	if (side == TTY_SIDE)
	{
		master_tty = tet_getvar("VSX_MASTER_TTY");
		master_fdp = &tty_master_fd;
	}
	else
	{
		master_tty = tet_getvar("VSX_MASTER_LOOP");
		master_fdp = &loop_master_fd;
	}

	if (master_tty == NULL || *master_tty == '\0')
	{
		if (side == TTY_SIDE)
			*fildes_p = openctl(termios_tty, oflags | O_NONBLOCK);
		else
			*fildes_p = open(termios_tty,
				oflags | O_NONBLOCK | O_NOCTTY);
		if(*fildes_p < 0)
			return -1;
	}
	else
	{
		/* allow previous rdwr() processes to finish */
		(void) sleep(3);

		if (side == TTY_SIDE)
			cntrl = (oflags & O_NOCTTY) == 0;
		else
			cntrl = 0;

		/* Implementation-supplied pseudo-terminal opener */
		if (openpty(master_tty, termios_tty,
				master_fdp, fildes_p, cntrl) != 0)
		{
			vsxerror("openpty(\"%s\", \"%s\", mfdp, sfdp, %d) failed",
				master_tty, termios_tty, cntrl);
			if (*master_fdp >= 0)
			{
				(void) close(*master_fdp);
				*master_fdp = -1;
			}
			return -1;
		}
	}

	/* Initialize fields */
	if (side == TTY_SIDE)
	{
		if (tty_init(*fildes_p, termios_p) != 0)
			return -1;
	}
	else
	{
		if (loop_init(*fildes_p, termios_p) != 0)
			return -1;
	}
	if (tcsetattr(*fildes_p, TCSANOW, termios_p) == SYSERROR)
		return -1;

	if ((oflags & O_NONBLOCK) == 0)
	{
		/* Clear O_NONBLOCK flag after opening */
		if ((flags = fcntl (*fildes_p, F_GETFL)) == SYSERROR)
			return -1;
		if (fcntl (*fildes_p, F_SETFL, flags & ~O_NONBLOCK) == SYSERROR)
			return -1;
	}

	return 0;
}

/*******************************
*  Prepare tty's
********************************/
public int
terms_prep(tty_flags, tty_fildes_p, tty_termios_p,
	   loop_flags, loop_fildes_p, loop_termios_p)
int tty_flags;
int *tty_fildes_p;
struct termios *tty_termios_p;
int loop_flags;
int *loop_fildes_p;
struct termios *loop_termios_p;
{
	int ret;

	*tty_fildes_p = *loop_fildes_p = -1;

	/*
	 * TERMIOS_TTY must be the next tty opened so that if the process
	 * is a session leader it will become the controlling terminal
	 */

	SET_TIMEOUT(WAITTIME)
	ret = do_prep(TTY_SIDE, tty_flags, tty_fildes_p, tty_termios_p);
	CLEAR_ALARM
	if (ret != 0)
		return ret;

	SET_TIMEOUT(WAITTIME)
	ret = do_prep(LOOP_SIDE, loop_flags, loop_fildes_p, loop_termios_p);
	CLEAR_ALARM
	if (ret != 0)
		return ret;
	
	/* If pseudo-terminal looback required, connect master sides */
	if (tty_master_fd >= 0 && loop_master_fd >= 0)
		ret = ptycon(*tty_fildes_p, *loop_fildes_p);

	if (tty_master_fd >= 0)
	{
		(void) close(tty_master_fd);
		tty_master_fd = -1;
	}
	if (loop_master_fd >= 0)
	{
		(void) close(loop_master_fd);
		loop_master_fd = -1;
	}
	
	return ret;
}

public int
termios_prep (tty_fildes_p, tty_termios_p, loop_fildes_p, loop_termios_p)
int *tty_fildes_p;			/* TERMIOS_TTY file descriptor	*/
struct termios *tty_termios_p;		/* terminal attributes		*/
int *loop_fildes_p;			/* TERMIOS_LOOP file descriptor	*/
struct termios *loop_termios_p;		/* terminal attributes		*/
{
	(void) setsid();

	return terms_prep(O_RDWR, tty_fildes_p, tty_termios_p,
			  O_RDWR, loop_fildes_p, loop_termios_p);
}

/*************************
* TERMIOS_TTY
*************************/
public int
tty_prep (tty_fildes_p, tty_termios_p)
int *tty_fildes_p;			/* TERMIOS_TTY file descriptor	*/
struct termios *tty_termios_p;		/* terminal attributes		*/
{
	return do_prep(TTY_SIDE, O_RDWR, tty_fildes_p, tty_termios_p);
}

/*************************
* TERMIOS_LOOP
*************************/
public int
loop_prep (loop_fildes_p, loop_termios_p)
int *loop_fildes_p;			/* TERMIOS_LOOP file descriptor	*/
struct termios *loop_termios_p;		/* terminal attributes		*/
{
	return do_prep(LOOP_SIDE, O_RDWR, loop_fildes_p, loop_termios_p);
}

/**********************************************
* Save the initial termios state
* Modify struct termios to VSX values
* Modify struct termios for test values
**********************************************/
public int
tty_init(fildes, termios_p)
int fildes;
struct termios *termios_p;
{
	if (termios_speed == NULL && getspeed() < 0)
		return -1;

	/*************************
	* save initial state
	**************************/
	if (!tty_done)
	{
		if (tcgetattr(fildes, &tty_termios) == SYSERROR)
			return -1;
	}
	tty_done = 1;

	*termios_p = tty_termios;	/* ANSI C structure assignment	*/

	/*************************
	* set to VSX values
	**************************/
	(termios_p)->c_iflag &= ~(PARMRK|INLCR|INPCK|IGNCR|IXOFF|BRKINT|IXON);
	(termios_p)->c_iflag |= (IGNPAR|IGNBRK|ICRNL);
	if ((termios_p->c_cflag & CSIZE) == CS8)
		(termios_p)->c_iflag &= ~ISTRIP;
#ifdef _XOPEN_SOURCE
	(termios_p)->c_iflag &= ~(IUCLC|IXANY);
#endif
	(termios_p)->c_oflag &= ~OPOST;

	(termios_p)->c_cflag |= (CLOCAL|CREAD);

	(termios_p)->c_lflag &= ~(ECHO|ECHOE|ECHOK|ECHONL|TOSTOP|IEXTEN);
	(termios_p)->c_lflag |= (ISIG|ICANON|NOFLSH);
#ifdef _XOPEN_SOURCE
	(termios_p)->c_lflag &= ~XCASE;
#endif
	(termios_p)->c_cc[VMIN] = PMIN;
	(termios_p)->c_cc[VTIME] = PTIME;
	(termios_p)->c_cc[VINTR] = PINTR;
	(termios_p)->c_cc[VQUIT] = PQUIT;
	(termios_p)->c_cc[VERASE] = PERASE;
	(termios_p)->c_cc[VKILL] = PKILL;
	(termios_p)->c_cc[VEOF] = PEOF;
	(termios_p)->c_cc[VEOL] = PEOL;
	(termios_p)->c_cc[VSUSP] = PSUSP;
	(termios_p)->c_cc[VSTART] = PSTART;
	(termios_p)->c_cc[VSTOP] = PSTOP;

	if (cfsetispeed((termios_p),baud_mask(termios_speed)) == SYSERROR ||
	    cfsetospeed((termios_p),baud_mask(termios_speed)) == SYSERROR)
		return -1;

	return 0;
}

/**********************************
*  Initialize LOOP to noncanonical
***********************************/
public int
loop_init(fildes, termios_p)

int fildes;
struct termios *termios_p;
{
	static long NOTHING;		/* "disable" character		*/

	if (termios_speed == NULL && getspeed() < 0)
		return -1;

	/*************************
	* save initial state
	**************************/
	if (!loop_done)
	{
		if (tcgetattr(fildes, &loop_termios) == SYSERROR)
			return -1;
		NOTHING = fpathconf(fildes, _PC_VDISABLE);
		if (NOTHING == -1)
			NOTHING = 0;
	}
	loop_done = 1;

	*termios_p = loop_termios;	/* ANSI C structure assignment	*/

	/*************************
	* set VSX values
	**************************/
	(termios_p)->c_iflag &= ~(PARMRK|ICRNL|INLCR|INPCK|IGNCR|IXOFF|BRKINT|IXON);
	(termios_p)->c_iflag |= (IGNPAR|IGNBRK);
#ifdef _XOPEN_SOURCE
	(termios_p)->c_iflag &= ~(IUCLC|IXANY);
#endif
	(termios_p)->c_oflag &= ~OPOST;

	(termios_p)->c_cflag |= (CLOCAL|CREAD);
	/* make sure LOOP has similar enough c_cflag values to TTY
	   so they can talk to each other */
	if (tty_done)
	{
		termios_p->c_cflag &= ~(CSIZE|CSTOPB|PARENB|PARODD);
		termios_p->c_cflag |=
			(tty_termios.c_cflag & (CSIZE|CSTOPB|PARENB|PARODD));
	}

	if ((termios_p->c_cflag & CSIZE) == CS8)
		(termios_p)->c_iflag &= ~ISTRIP;

	(termios_p)->c_lflag &= ~(ECHO|ECHOE|ECHOK|ECHONL|ISIG|ICANON|TOSTOP|IEXTEN);
	(termios_p)->c_lflag |= (NOFLSH);
#ifdef _XOPEN_SOURCE
	(termios_p)->c_lflag &= ~XCASE;
#endif
	(termios_p)->c_cc[VINTR] = NOTHING;
	(termios_p)->c_cc[VQUIT] = NOTHING;
	(termios_p)->c_cc[VERASE] = NOTHING;
	(termios_p)->c_cc[VKILL] = NOTHING;
	(termios_p)->c_cc[VEOF] = NOTHING;
	(termios_p)->c_cc[VEOL] = NOTHING;
	(termios_p)->c_cc[VSUSP] = NOTHING;
	(termios_p)->c_cc[VMIN] = PMIN;
	(termios_p)->c_cc[VTIME] = PTIME;
	(termios_p)->c_cc[VSTART] = PSTART;
	(termios_p)->c_cc[VSTOP] = PSTOP;

	if (cfsetispeed((termios_p),baud_mask(termios_speed)) == SYSERROR ||
	    cfsetospeed((termios_p),baud_mask(termios_speed)) == SYSERROR)
		return -1;
	
	return 0;
}
#else /* !TEST_XNFS */
static int dummy;
#endif /* !TEST_XNFS */
