/*
*	SCCS:   @(#)common/tsetlib/tztest.c	4.1 (01/21/97)  VSXgen release 1.4
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/


/***********************************************************************

NAME:		tztest.c - test timezone names set by time functions

PROJECT:	VSX (X/OPEN Validation Suite)

MODULE CONTAINS:
	tztest()
SYNOPSIS:
	int tztest(funcnam, func)
		char *funcnam;
		char *(*func)();

DESCRIPTION:
		Tztest() tests that interfaces correctly set the
		standard and daylight savings timezone names.  The
		argument funcnam describes the interface being tested
		(e.g. "ctime()" or "strftime(%Z)").  The function
		pointed to by func should call the interface being
		tested and return the resulting timezone name specified
		by an argument to (*func)() which is 0 for std or 1 for
		dst.  Tztest() performs all test reporting (including
		calling xx_rpt() or xs_rpt()).  No result reporting is
		done by the calling test code.

CONDITIONAL COMPILE PARAMS:
SIDE EFFECTS:
		tztest() alters the environment variable TZ, and so
		should be called in a child process.
RELATED DOCUMENTS:
SOURCE:
AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	13 Jan 1992
MODIFICATIONS:
		Geoff Clare, 24 Apr 1992
			Use isgraph() instead of isprint().

		Geoff Clare, 25 June 1992
			Change POSIX mode to same as XPG4 mode.

		Geoff Clare, 14 Oct 1992
			Changes for XNFS.

************************************************************************/

#include	<std.h>

#if !TEST_XNFS
#include	<stdlib.h>
#include	<time.h>
#include	<sys/types.h>	
#include	<unistd.h>
#include	<errno.h>
#include	<limits.h>
#include	<string.h>
#include	<pcts_max.h>
#include	<dbug.h>
#include	<report.h>
#include <tet_api.h>
#include	<sysdep.h>

extern char	**environ;
extern char	*psl_to_asc();
extern char	*errname();

private int	testfail;

private int
set_tz(tzval)
char *tzval;
{
	char	*tz;
	static char	buf[2 * PCTS_TZNAME_MAX + 5];

	DBUG_LENTER("set_tz", 2);

	(void) strcpy(buf, "TZ=");
	(void) strcat(buf, tzval);
	(void) vputenv(&environ, buf);
	tz = getenv("TZ");
	if (tz == NULL || strcmp(tz, tzval) != 0)
	{
		if (testfail++ == 0)
			xx_rpt(DELETION);
		in_rpt("could not put \"TZ=%s\" in environment",
			psl_to_asc(tzval, 0));
		DBUG_RETURN(-1);
	}

	DBUG_RETURN(0);
}

public void
tztest(funcnam, func)
char *funcnam;
char *(*func)();
{
	long	tzname_max;
	int	c;
	char	*ret;
	char	std[PCTS_TZNAME_MAX+1], dst[PCTS_TZNAME_MAX+1];
	char	tz[2 * PCTS_TZNAME_MAX + 2];
	int	pathok = 0;
	
	DBUG_LENTER("tztest", 2);

	testfail = 0;

	if (set_tz("ABC-4:30:20") != 0)
	{
		DBUG_VOID_RETURN;
	}
	else
		PATH_TRACE;
		
	ret = (*func)(0);
	if ( strcmp ( ret, "ABC") != 0 )
	{
		if (testfail++ == 0)
			xx_rpt(FAILURE);
		in_rpt("with TZ=ABC-4:30:20 %s gave incorrect standard timezone name",
			funcnam);
		in_rpt("observed \"%s\", expected \"ABC\"", ret);
	}
	else
		PATH_TRACE;
		
	if (set_tz("JKL3:10PNM4:40") != 0)
	{
		DBUG_VOID_RETURN;
	}
	else
		PATH_TRACE;
			
	ret = (*func)(0);
	if ( strcmp ( ret, "JKL") != 0 )
	{
		if (testfail++ == 0)
			xx_rpt(FAILURE);
		in_rpt("with TZ=JKL3:10PNM4:40 %s gave incorrect standard timezone name",
			funcnam);
		in_rpt("observed \"%s\", expected \"JKL\"", ret);
	}
	else
		PATH_TRACE; 

	ret = (*func)(1);
	if ( strcmp ( ret, "PNM") != 0 )
	{
		if (testfail++ == 0)
			xx_rpt(FAILURE);
		in_rpt("with TZ=JKL3:10PNM4:40 %s gave incorrect daylight saving timezone name",
			funcnam);
		in_rpt("observed \"%s\", expected \"PNM\"", ret);
	}
	else
		PATH_TRACE; 

	/* skip remainder of test if basics don't work */
	if (testfail > 0)
	{
		DBUG_VOID_RETURN;
	}
	else
		PATH_TRACE;
		
	/* any three characters in "std" */
	for (c = 0; c <= UCHAR_MAX; c++)
	{
		/* skip disallowed characters */

#if TEST_XPG == 3
		/* XPG3 mode: skip non-alphabetics as in VSX3 */
		if (!isalpha(c))
		{
			PATH_TRACE; PATH_TRACE;
			continue;
		}
#else
		/* skip specifically disallowed characters */
		if (c == ',' || c == '-' || c == '+' || c == '\0' || isdigit(c))
		{
			PATH_TRACE; PATH_TRACE;
			continue;
		}

		/* skip non-graphic characters */
		if (!isgraph(c))
		{
			PATH_TRACE; PATH_TRACE;
			continue;
		}
#endif

		std[0] = std[1] = std[2] = c;
		std[3] = '\0';

		/* leading colon not allowed */
		if (c == ':')
			std[0] = 'A';
		
		(void) strcpy(tz, std);
		(void) strcat(tz, "0");
		if (set_tz(tz) != 0)
		{
			DBUG_VOID_RETURN;
		}
		else
			PATH_TRACE;
			
		ret = (*func)(0);
		if (strcmp(ret, std) != 0)
		{
			if (testfail++ == 0)
				xx_rpt(FAILURE);
			in_rpt("with TZ=\"%s\" %s gave incorrect standard timezone name",
				psl_to_asc(tz, 0), funcnam);
			in_rpt("observed \"%s\", expected \"%s\"",
				psl_to_asc(ret, 0), psl_to_asc(std, 0));
		}
		else
			PATH_TRACE; 
	}

	/* any three characters in "dst" */
	for (c = 0; c <= UCHAR_MAX; c++)
	{
		/* skip disallowed characters */

#if TEST_XPG == 3
		/* XPG3 mode: skip non-alphabetics as in VSX3 */
		if (!isalpha(c))
		{
			PATH_TRACE; PATH_TRACE;
			continue;
		}
#else
		/* skip specifically disallowed characters */
		if (c == ',' || c == '-' || c == '+' || c == '\0' || isdigit(c))
		{
			PATH_TRACE; PATH_TRACE;
			continue;
		}

		/* skip non-graphic characters */
		if (!isgraph(c))
		{
			PATH_TRACE; PATH_TRACE;
			continue;
		}
#endif

		dst[0] = dst[1] = dst[2] = c;
		dst[3] = '\0';

		/* leading colon not allowed */
		if (c == ':')
			dst[0] = 'A';
		
		(void) strcpy(tz, "XYZ-1:00:00");
		(void) strcat(tz, dst);
		if (set_tz(tz) != 0)
		{
			DBUG_VOID_RETURN;
		}
		else
			PATH_TRACE;

		ret = (*func)(1);
		if (strcmp(ret, dst) != 0)
		{
			if (testfail++ == 0)
				xx_rpt(FAILURE);
			in_rpt("with TZ=\"%s\" %s gave incorrect daylight saving timezone name",
				psl_to_asc(tz, 0), funcnam);
			in_rpt("observed \"%s\", expected \"%s\"",
				psl_to_asc(ret, 0), psl_to_asc(dst, 0));
		}
		else
			PATH_TRACE;
	}


	/* Very long names */

#if TEST_XPG == 3
	/* XPG3 mode: use fixed length -
	   maximum length of a portable filename seems a sensible choice */
	tzname_max = _POSIX_NAME_MAX;
	PATH_TRACE;
#else
	/* other modes: examine TZNAME_MAX */
	errno = 0;
	tzname_max = sysconf(_SC_TZNAME_MAX);
	if (tzname_max == -1 && errno == 0 || tzname_max > PCTS_TZNAME_MAX)
	{
		PATH_TRACE;
		tzname_max = PCTS_TZNAME_MAX;
	}
	else if (tzname_max < 0)
	{
		int err = errno;
		xx_rpt(DELETION);
		in_rpt("sysconf(_SC_TZNAME_MAX) failed - errno %d (%s)",
			err, errname(err));
		DBUG_VOID_RETURN;
	}
	else
		PATH_TRACE;
#endif
#ifdef TESTBED
	tzname_max++;
#endif

	for (c = 0; c < tzname_max; c++)
	{
		std[c] = 'A';
		dst[c] = 'B';
	}
	std[c] = dst[c] = '\0';

	(void) strcpy(tz, std);
	(void) strcat(tz, "0");
	(void) strcat(tz, dst);
	if (set_tz(tz) != 0)
	{
		DBUG_VOID_RETURN;
	}
	else
		PATH_TRACE;

	ret = (*func)(0);
	if (strcmp(ret, std) != 0)
	{
		if (testfail++ == 0)
			xx_rpt(FAILURE);
		in_rpt("with long zone names in TZ %s gave incorrect standard timezone name",
			funcnam);
		in_rpt("observed value: \"%s\"", psl_to_asc(ret, 0));
		in_rpt("expected value: \"%s\"", std);
	}
	else
		PATH_TRACE;

	ret = (*func)(1);
	if (strcmp(ret, dst) != 0)
	{
		if (testfail++ == 0)
			xx_rpt(FAILURE);
		in_rpt("with long zone names in TZ %s gave incorrect daylight saving timezone name",
			funcnam);
		in_rpt("observed value: \"%s\"", psl_to_asc(ret, 0));
		in_rpt("expected value: \"%s\"", dst);
	}
	else
		PATH_TRACE;

	if (testfail == 0)
		PATH_XS_RPT(6 + (2 + 2)*(UCHAR_MAX + 1) + 1 + 3);

	DBUG_VOID_RETURN;
}

#else /* !TEST_XNFS */
static int dummy;
#endif /* !TEST_XNFS */
