/*
 *      SCCS:   @(#)common/vlib(arc_extr.c)	4.3 (03/26/98)  VSXgen release 1.4
 *
 *	UniSoft Ltd., London, England
 *
 * (C) Copyright 1991 X/Open Company Limited
 * 
 * All rights reserved.  No part of this source code may be reproduced,
 * stored in a retrieval system, or transmitted, in any form or by any
 * means, electronic, mechanical, photocopying, recording or otherwise,
 * except as stated in the end-user licence agreement, without the prior
 * permission of the copyright owners.
 * 
 * X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
 * the UK and other countries.
 */

/***********************************************************************

NAME:		arc_extr.c - extract test case names out of a archive

PROJECT:	VSX (X/OPEN Validation Suite)

SYNOPSIS:
	int
	arc_extr(arcname, dirname, fnames)
		char	*arcname;
		char	*dirname;
		char	***fnames;

	int
	arc_print(prefix, stream, arcname, fname)
		char	*prefix;
		FILE	*stream;
		char	*arcname;
		char	*fname;

	char *
	arc_errname(arcname)
		char	*arcname;

DESCRIPTION:
	  Arc_extr() takes the name of a testset archive in arcname, and
	extracts the contents of it into directory dirname.  The names
	of the files are put into malloc()ed memory in fnames (without
	the directory name prefix).  The names in the archive must be of
	the form
		c[cd][0-9][0-9][CcLlEe][sw0-9].c
	all other names are ignored.  The names returned in fnames are
	garenteed to be of this form and sorted (using qsort()) by the
	number (i.e. 3rd and 4th characters in name).  The value
	returned is the number of file names in fnames. SYSERROR
	is returned if something goes wrong.

	  Arc_print() takes the name of a cs_arc archive in arcname, and
	outputs to stream the contents of file fname from the archive.
	There are no restrictions on the file name.
	The value returned is 0 for success, -2 if the archive file does
	not exist, or -1 for all other errors.

	  If arc_errname() is called after arc_extr() or arc_print() fails
	then a string containing the relevant error message is returned.

AUTHOR:		Mike Rendell, UniSoft Ltd.
DATE CREATED:	Mon Oct 27 12:35:29 GMT 1986

MODIFICATIONS:
		Wed Mar 16 14:46:40 WET 1988	Geoff Clare
		- addition of arc_print()
		Fri May 12 17:04:02 BST 1989	Dave Cater
		Fixed bus error. 3rd argument of qsort() was 1!

		Geoff Clare, 2 Nov 1989
			Provide for warnings from compilation failure

		Geoff Clare, 27 Jan 1992
			Change name_cmp() args to be compatible with qsort()

		Geoff Clare, 25-26 March 1998
			Allow for long names in arc_print().
			Handle long lines in print_file().

************************************************************************/

#include <std.h>
#include <stdlib.h>
#include <ts_arc.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <limits.h>
#include <dbug.h>
#include <sysdep.h>

#define NARCERROR		11
#define LINELEN			256

private int	extract_file();
private int	name_cmp();
private int	print_file();
private int	ok_ts_name();

private int	arcerrno;
private char *arc_errmsg[NARCERROR] = {
/* 0 */		"no error",
/* 1 */		"could not open archive %s",
/* 2 */		"error read()ing archive %s",
/* 3 */		"bad header format in archive %s",
/* 4 */		"ran out of memory for filenames reading %s",
/* 5 */		"bad header format in archive %s",
/* 6 */		"could not extract file from archive %s, open() failed",
/* 7 */		"format error in archive %s, could not extract file",
/* 8 */		"could not extract file from archive %s, write() failed",
/* 9 */		"could not extract file from archive %s, lseek() failed",
/* 10 */	"requested file not found in archive %s"
		};

public	int
arc_extr(arcname, dirname, fnames)
	char	*arcname;
	char	*dirname;
	char	***fnames;
{
	int		arc_fd;
	int		i, j, k;
	int		nfiles;
	char		**files;
	unsigned	*filesize;
	char		buf[_POSIX_PATH_MAX];
	char		num1, num2;


	DBUG_LENTER("arc_extr", 2);
	DBUG_PRINTF("args", "arc_extr(%s, %s, 0x%lx)", arcname, dirname,         				     (long) fnames);
	arcerrno = 0;
	*fnames = (char **) NULL;
	arc_fd = open(arcname, O_RDONLY, 0);
	if (arc_fd == SYSERROR)
	{
		arcerrno = 1;
		DBUG_PRINTF("return", "arc_extr = %d", SYSERROR);
		DBUG_RETURN(SYSERROR);
	}

	j = read(arc_fd, buf, TS_HEADSIZE);
	if (j == SYSERROR)
	{
		arcerrno = 2;
		(void) close(arc_fd);
		DBUG_PRINTF("return", "arc_extr = %d", SYSERROR);
		DBUG_RETURN(SYSERROR);
	}

	i = 0;
	if (j == TS_HEADSIZE
		&& strncmp(buf, TS_MAGIC, TS_MAGICSIZE) == 0
		&& buf[TS_MAGICSIZE + 3] == '\n')
	{
		for (nfiles = 0 ; i < 3 ; i = i + 1)
			if (!isdigit(buf[TS_MAGICSIZE + i]))
				break;
			else
				nfiles = (nfiles * 10)
					+ buf[TS_MAGICSIZE + i] - '0';
	}
	(void) fflush(stderr);
	if (i != 3)
	{
		arcerrno = 3;
		(void) close(arc_fd);
		DBUG_PRINTF("return", "arc_extr = %d", SYSERROR);
		DBUG_RETURN(SYSERROR);
	}
	if (nfiles == 0)
	{
		(void) close(arc_fd);
		DBUG_PRINTF("return", "arc_extr = %d", 0);
		DBUG_RETURN(0);
	}

	files = (char **) malloc((size_t) (nfiles * sizeof(char *)));
	filesize = (unsigned *) malloc((size_t) (nfiles * sizeof(unsigned)));
	if (files == (char **) NULL || filesize == (unsigned *) NULL)
	{
		arcerrno = 4;
		(void) close(arc_fd);
		DBUG_PRINTF("return", "arc_extr = %d", SYSERROR);
		DBUG_RETURN(SYSERROR);
	}

	for (i = 0 ; i < nfiles ; i = i + 1)
	{
		j = read(arc_fd, buf, TS_NAMELEN + TS_SEEKLEN + 1);
		(void) fflush(stderr);
		if (j == SYSERROR)
		{
			arcerrno = 2;
			break;
		}
		if (j != TS_NAMELEN + TS_SEEKLEN + 1)
		{
			arcerrno = 5;
			break;
		}
		for (j = 0 ; j < TS_NAMELEN ; j = j + 1)
			if (buf[j] == ' ')
				break;
		buf[j] = NUL;
		if (j != TS_FNAMELEN || !ok_ts_name(buf)) /* empty/bad name */
		{
			i--;
			nfiles--;
			continue;
		}
		files[i] = (char *) malloc((size_t) (TS_FNAMELEN + 1));
		if (files[i] == (char *) NULL)
		{
			arcerrno = 4;
			break;
		}
		(void) strcpy(files[i], buf);
		filesize[i] = 0;
		for (j = TS_NAMELEN ; j < TS_NAMELEN + TS_SEEKLEN && buf[j] == ' ' ;)
			j = j + 1;
		for ( ; j < TS_NAMELEN + TS_SEEKLEN ; j = j + 1)
			filesize[i] = filesize[i] * 10
				+ (buf[j] - '0');
	}
	if (i == nfiles)
	{
		*fnames = files;
		qsort((void *)files, (size_t)nfiles, sizeof(*files), name_cmp);
		for (i = 0 ; i < nfiles ; i = i + 1)
		{
			(void) strcat(strcat(strcpy(buf,
				dirname), "/"), files[i]);
			if (extract_file(buf, filesize[i], arc_fd) == SYSERROR)
				break;
		}
		/*
		 *	Remove multiple names in list (i.e. files that
		 *	are to be linked together)
		 */
		for (i = 0 ; i < nfiles ; i = i + 1)
		{
			if (files[i][TS_NLINK] == 's' ||
			    files[i][TS_NLINK] == 'w')
				continue;
			num1 = files[i][TS_NUMBER];
			num2 = files[i][TS_NUMBER+1];
			for (j = i ; j < nfiles ; j = j + 1)
			{
				if (files[j][TS_NLINK] == 's' ||
				    files[j][TS_NLINK] == 'w')
					break;
				if (files[j][TS_NUMBER] != num1 ||
					files[j][TS_NUMBER+1] != num2)
					break;
			}
			j = (j - i) - 1;
			nfiles -= j;
			for (k = i ; k < i + j ; k = k + 1)
				free((void *)files[k]);
			for (k = i ; k < nfiles ; k = k + 1)
				files[k] = files[k + j];
		}
	}

	free((void *) filesize);
	(void) close(arc_fd);

	if (i != nfiles)
	{
		DBUG_PRINTF("return", "arc_extr = %d", SYSERROR);
		DBUG_RETURN(SYSERROR);
	}
	DBUG_PRINTF("return", "arc_extr = %d", nfiles);
	DBUG_RETURN(nfiles);
}


private int
extract_file(name, fsize, arc_fd)
	char		*name;		/* name of file to extract */
	unsigned	fsize;		/* number of bytes in file */
	int		arc_fd;		/* archive file desc. */
{
	int		fd;
	int		nread;
	int		nwrite;
	unsigned	nbytes;
	char		buf[BUFSIZ];

	fd = open(name, O_WRONLY|O_CREAT|O_TRUNC, S_IRWXU|S_IRWXG|S_IRWXO);
	if (fd == SYSERROR)
		arcerrno = 6;
	else
		while (fsize != 0)
		{
			if (fsize > BUFSIZ)
				nbytes = BUFSIZ;
			else
				nbytes = fsize;
			nread = read(arc_fd, buf, nbytes);
			if (nread == SYSERROR)
			{
				arcerrno = 2;
				break;
			}
			if (nread == 0)
			{
				arcerrno = 7;
				break;
			}
			nwrite = write(fd, buf, (unsigned) nread);
			if (nwrite != nread)
			{
				arcerrno = 8;
				break;
			}
			fsize -= nread;
		}
	(void) close(fd);

	if (fsize != 0)
		return(SYSERROR);
	return(SUCCEED);
}


private int
name_cmp(a1, a2)
#if __STDC__
	const void	*a1;
	const void	*a2;
#else
	char	*a1;
	char	*a2;
#endif
{
	char	**s1 = (char **)a1;
	char	**s2 = (char **)a2;

	if (s1 == (char **) NULL || s2 == (char **) NULL)
	{
		return((int) (s1 - s2));
	}
	else if (*s1 == (char *) NULL || *s2 == (char *) NULL)
	{
		return((int) (*s1 - *s2));
	}
	return(strcmp(*s1, *s2));
}


private int
ok_ts_name(s)
	char	*s;
{
	if (s[0] == 'c'
		&& (s[TS_CLASS] == 'c'
			|| s[TS_CLASS] == 'd')
		&& isdigit(s[TS_NUMBER]) && isdigit(s[TS_NUMBER + 1])
		&& (s[TS_COMP] == 'c' || s[TS_COMP] == 'C'
			|| s[TS_COMP] == 'l' || s[TS_COMP] == 'L'
			|| s[TS_COMP] == 'e' || s[TS_COMP] == 'E')
		&& (s[TS_NLINK] == 's' || s[TS_NLINK] == 'w'
			|| isdigit(s[TS_NLINK])))
			return 1;
	return 0;
}

public	int
arc_print(prefix, stream, arcname, fname)
	char	*prefix;	/* prefix each line with this */
	FILE	*stream;
	char	*arcname;
	char	*fname;
{
	FILE		*arc_fp;
	int		i, j;
	int		nfiles;
	long		skipsize;
	unsigned int	filesize, tmpsize;
	char		buf[_POSIX_PATH_MAX];
	char		*entry, *sizestr;


	DBUG_LENTER("arc_print", 2);
	if (stream == (FILE*) NULL)
		DBUG_PRINTF("args", "arc_print((FILE *)NULL, %s, %s )"			           , arcname, fname);
	else
		DBUG_PRINTF("args", "arc_print((FILE *)[fd=%d], %s, %s)",                               fileno(stream), arcname, fname);
	arcerrno = 0;
	arc_fp = fopen(arcname, "r");
	if (arc_fp == (FILE *)NULL)
	{
		arcerrno = 1;
		if (errno == ENOENT)
		{
			DBUG_PRINTF("return", "arc_print = -2");
			DBUG_RETURN(-2);
		}
		else
		{
			DBUG_PRINTF("return", "arc_print = %d\t arcerrno = %d", -1, arcerrno);
			DBUG_RETURN(-1);
		}
	}

	if (fgets(buf, TS_HEADSIZE+1, arc_fp) == (char *)NULL)
	{
		arcerrno = 2;
		(void) fclose(arc_fp);
		DBUG_PRINTF("return", "arc_print = %d\t arcerrno = %d", -1, arcerrno);
		DBUG_RETURN(-1);
	}

	DBUG_PRINTF("arc_print", "buf[] :%s:", buf);

	if (strncmp(buf, TS_MAGIC, TS_MAGICSIZE) == 0
		&& buf[TS_MAGICSIZE + 3] == '\n')
	{
		DBUG_PRINTF("arc_print", "buf[] :%.*s:", TS_MAGICSIZE, buf);
		for (i = 0 ; i < 3 ; i++)
			if (!isdigit(buf[TS_MAGICSIZE + i]))
				break;
		DBUG_PRINTF("arc_print", "buf[TS_MAGICSIZE] :%s:", &buf[TS_MAGICSIZE]);

	}
	DBUG_PRINTF("arc_print", "i :%d:",i);
	(void) fflush(stderr);
	if (i != 3)
	{
		arcerrno = 3;
		(void) fclose(arc_fp);
		DBUG_PRINTF("return", "arc_print = %d\t arcerrno = %d", -1, arcerrno);
		DBUG_RETURN(-1);
	}

	DBUG_PRINTF("arc_print", "buf[TS_MAGICSIZE] :%s:", &buf[TS_MAGICSIZE]);

	if ((nfiles = atoi(&buf[TS_MAGICSIZE])) == 0)
	{
		arcerrno = 7;
		(void) fclose(arc_fp);
		DBUG_PRINTF("return", "arc_print = %d\t arcerrno = %d", -1, arcerrno);
		DBUG_RETURN(-1);
	}

	DBUG_PRINTF("arc_print", "nfiles :%d:", nfiles);

	skipsize = 0;
	filesize = 0;
	for (i = 0 ; i < nfiles ; i++)
	{
/*
		(void) fflush(stderr);
*/
		if (fgets(buf, (int)sizeof buf, arc_fp) == (char *)NULL)
		{
			arcerrno = 2;
			(void) fclose(arc_fp);
			DBUG_PRINTF("return", "arc_print = %d\t arcerrno = %d", -1, arcerrno);
			DBUG_RETURN(-1);
		}
		DBUG_PRINTF("arc_print", "buf :%s:", buf);

		if (filesize != 0)
			continue; /* skip entries after the one we want */

		entry = strtok(buf, " ");
		DBUG_PRINTF("arc_print", "entry :%s:", entry);
		sizestr = strtok((char *)NULL, " ");
		if (sizestr == NULL)
		{
			arcerrno = 5;
			(void) fclose(arc_fp);
			DBUG_PRINTF("return", "arc_print = %d\t arcerrno = %d", -1, arcerrno);
			DBUG_RETURN(-1);
		}

		tmpsize = atoi(sizestr);
		DBUG_PRINTF("arc_print", "tmpsize :%d:", tmpsize);

		if (strcmp(entry, fname) == 0)
			filesize = tmpsize;
		else
			skipsize += tmpsize;
	}

	DBUG_PRINTF("arc_print", "filesize :%d:", filesize);
	DBUG_PRINTF("arc_print", "skipsize :%d:", skipsize);

	if (filesize == 0)
		arcerrno = 10;
	else
	{
		if (skipsize > 0 &&
			fseek(arc_fp, skipsize, SEEK_CUR) == SYSERROR)
		{
			arcerrno = 9;
			filesize = 0;
		}
		else
			if (print_file(prefix, stream, filesize, arc_fp) == SYSERROR)
				filesize = 0;
	}

	(void) fclose(arc_fp);

	if (filesize == 0)
	{
		DBUG_PRINTF("return", "arc_print = %d\t arcerrno = %d", -1, arcerrno);
		DBUG_RETURN(-1);
	}

	DBUG_PRINTF("return", "arc_print = %d", SUCCEED);
	DBUG_RETURN(SUCCEED);
}

private int
print_file(prefix, stream, fsize, arc_fp)
	char		*prefix;	/* prefix each line with this */
	FILE		*stream;	/* destination for output */
	unsigned	fsize;		/* number of bytes in file */
	FILE		*arc_fp;	/* archive file pointer	*/
{
	int		nread;
	int		nwrite;
	unsigned	nbytes;
	char		buf[LINELEN+1];
	char		*ret;

	DBUG_LENTER("print_file()", 2);

	while (fsize != 0)
	{
		if ((ret = fgets(buf, LINELEN+1, arc_fp)) == (char *)NULL)
		{
			arcerrno = 2;
			(void) fclose(arc_fp);
			DBUG_PRINTF("return", "arc_print = %d\t arcerrno = %d", -1, arcerrno);
			DBUG_RETURN(-1);
		}
		if ((nread = strlen(buf)) == 0)
		{
			arcerrno = 7;
			break;
		}
		if (fputs(prefix, stream) == EOF)
		{
			arcerrno = 8;
			break;
		}
		DBUG_PRINTF("print_file",  "buf :%.*s:", nread, buf);
		if (fputs(buf, stream) == EOF)
		{
			arcerrno = 8;
			break;
		}
		fsize -= nread;
	}

	if (fsize != 0)
		DBUG_RETURN(SYSERROR);

	DBUG_RETURN(SUCCEED);
}

public char *
arc_errname(arcname)
	char	*arcname;
{
	static	char	buf[_POSIX_PATH_MAX + 50];
	char		*msg;

	DBUG_LENTER("arc_errname", 2);
	DBUG_PRINTF("args", "arc_errname(%s)", arcname);
	if (arcerrno > NARCERROR || arcerrno < 0)
		msg = "Unknown error";
	else
	{
		(void) sprintf(buf, arc_errmsg[arcerrno], arcname);
		msg = buf;
	}
	DBUG_PRINTF("return", "arc_errname = %s", msg);
	DBUG_RETURN(msg);
}
