/************************************************************************
 *
 *	SCCS:   @(#)common/vlib/dbug.c	4.1 (01/21/97)  VSXgen release 1.4
 *
 * (C) Copyright 1991 X/Open Company Limited
 *
 * All rights reserved.  No part of this source code may be reproduced,
 * stored in a retrieval system, or transmitted, in any form or by any
 * means, electronic, mechanical, photocopying, recording or otherwise,
 * except as stated in the end-user licence agreement, without the prior
 * permission of the copyright owners.
 *
 * X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
 * the UK and other countries.
 *
 ************************************************************************
 */

/*
 *  FILE
 *
 *	dbug.c   runtime support routines for dbug package
 *
 *  DESCRIPTION
 *
 *	These are the runtime support routines for the dbug package.
 *	The dbug package has two main components; the user include
 *	file containing various macro definitions, and the runtime
 *	support routines which are called from the macro expansions.
 *
 *	Externally visible functions in the runtime support module
 *	use the naming convention "db_func", thus they are unlikely
 *	to collide with user defined function names.
 *
 *  AUTHOR
 *
 *	Fred Fish
 *	(Currently at UniSoft Systems, Berkeley Ca.)
 *
 *
 *  MODIFICATION
 *
 *	Extensively modified for VSX.
 *
 *	Geoff Clare, 9 March 1988
 *	Altered functionality of 'f' option so that debugging/tracing
 *	is enabled while in OR BELOW any of the listed functions.
 *
 *	Steve Ratcliffe, 26 Feb 1990
 *	Do varargs.h/stdarg.h changover
 *
 *	Geoff Clare, 5 June 1990
 *	Change names from "_db_" to "db_" because "_db_" is in the
 *	namespace reserved for the implementation.
 *
 *	Dave Cater, 6 June 1990
 *	Add definition of globok
 *
 *	Geoff Clare, 17 Aug 1990
 *	Set close on exec flag when dbug file is opened
 *
 *	Geoff Clare, 4 Feb 1993
 *	Move sigsetjmp() call from DBUG_SETJMP macro into db_setjmp().
 *	Move siglongjmp() call from DBUG_LONGJMP macro into db_longjmp().
 *
 *	Geoff Clare, 22 June 1994
 *	Ensure sigsetjmp() is used in line with ANSI C constraints.
 *	Initialise db_fp to stderr at run time instead of compile time,
 *	as stderr might not be constant.
 *
 *	Geoff Clare, May 1996
 *	TenDRA printf format string checking on db_printf.
 */

#include <std.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#if TEST_ANSI
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <sysdep.h>

#ifdef __TenDRA__
#pragma TenDRA type db_printf_fmt for ... printf
#else
#define db_printf_fmt char *
#endif

/*
 *	Manifest constants that should not require any changes.
 */

#define EOS		'\000'	/* End Of String marker */

/*
 *	Manifest constants which may be "tuned" if desired.
 */

#define PRINTBUF	1024	/* Print buffer size */
#define INDENT		2	/* Indentation per trace level */
#define MAXDEPTH	200	/* Maximum trace depth default */

#define DBUG_DEST	stderr	/* Initial output destination, until db_push */

/*
 *	The following flags are used to determine which
 *	capabilities the user has enabled with the state
 *	push macro.
 */

#define TRACE_ON	000001	/* Trace enabled */
#define DEBUG_ON	000002	/* Debug enabled */
#define FILE_ON 	000004	/* File name print enabled */
#define LINE_ON		000010	/* Line number print enabled */
#define DEPTH_ON	000020	/* Function nest level print enabled */
#define PROCESS_ON	000040	/* Process name print enabled */
#define NUMBER_ON	000100	/* Number each line of output */

#define TRACING ((stack -> flags & TRACE_ON) != 0 && db_on > 0)
#define DEBUGGING ((stack -> flags & DEBUG_ON) != 0 && db_on > 0)
#define STREQ(a,b) (strcmp(a,b) == 0)

/*
 *	Typedefs to make things more obvious.
 */

#define VOID void		/* Can't use typedef for most compilers */
typedef int BOOLEAN;

/*
 *	Make it easy to change storage classes if necessary.
 */

#define LOCAL static		/* Names not needed by outside world */
#define IMPORT extern		/* Names defined externally */
#define EXPORT			/* Allocated here, available globally */
#define AUTO auto		/* Names to be allocated on stack */
#define REGISTER register	/* Names to be placed in registers */

/*
 *	One variable needed for VSX4 resilience upgrade.
 *	This may be modified/accessed in test sets and library functions 
 *	as required.
 */

EXPORT int globok;

/*
 *	Variables which are available externally but should only
 *	be accessed via the macro package facilities.
 */

EXPORT FILE *db_fp; /* can't init to DBUG_DEST (might not be a constant) */
EXPORT char *db_process = "dbug";	/* Pointer to process name; argv[0] */
EXPORT int db_on = 1;			/* > 0 when debugging/tracing on */
EXPORT int db_line = 0;		/* line number in source code */
EXPORT int db_lev = 1;		/* Level of debug being output */

/*
 *	The user may specify a list of functions to trace or 
 *	debug.  These lists are kept in a linear linked list,
 *	a very simple implementation.
 */

struct link {
    char *string;		/* Pointer to link's contents */
    struct link *next_link;	/* Pointer to the next link */
};

/*
 *	Debugging states can be pushed or popped off of a
 *	stack which is implemented as a linked list.  Note
 *	that the head of the list is the current state and the
 *	stack is pushed by adding a new state to the head of the
 *	list or popped by removing the first link.
 */

struct state {
    int flags;				/* Current state flags */
    int maxdepth;			/* Current maximum trace depth */
    int level;				/* Current function nesting level */
    int db_lev;				/* Required output level */
    FILE *out_file;			/* Current output stream */
    struct link *functions;		/* List of functions */
    struct link *keywords;		/* List of debug keywords */
    struct link *processes;		/* List of process names */
    struct state *next_state;		/* Next state in the list */
};

LOCAL struct state *stack = NULL;	/* Linked list of stacked states */

/*
 *	Local variables not seen by user.
 */

LOCAL int lineno = 0;		/* Current debugger output line number */
LOCAL char *func = "main";	/* Name of current user function */
LOCAL char *file = "?file";	/* Name of current user file */
LOCAL BOOLEAN init_done = FALSE;/* Set to TRUE when initialization done */

LOCAL int jmplevel;		/* Remember nesting level at setjmp () */
LOCAL char *jmpfunc;		/* Remember current function for setjmp */
LOCAL char *jmpfile;		/* Remember current file for setjmp */

LOCAL struct link *ListParse ();/* Parse a debug command string */
LOCAL char *StrDup ();		/* Make a fresh copy of a string */
LOCAL VOID OpenFile ();		/* Open debug output stream */
LOCAL VOID CloseFile ();	/* Close debug output stream */
LOCAL VOID PushState ();	/* Push current debug state */
LOCAL VOID ChangeOwner ();	/* Change file owner and group */
LOCAL VOID DoPrefix ();		/* Print line prefix prior to indentation */
LOCAL VOID FreeList ();		/* Free all memory held by a linked list */
LOCAL VOID Indent ();		/* Indent a line */
LOCAL BOOLEAN DoTrace ();	/* Test for tracing enabled */
LOCAL BOOLEAN InList ();	/* Test string for member of a list */
LOCAL BOOLEAN Writable ();	/* Test to see if file is writable */
LOCAL char *DbugMalloc ();	/* Allocate memory for runtime support */
LOCAL char *BaseName ();	/* Remove leading pathname components */

/*
 *	Miscellaneous printf format strings.
 */
 
#define MSG1 "%s: missing DBUG_RETURN or DBUG_VOID_RETURN macro in function \"%s\"\n"
#define MSG2 "%s: can't open debug output stream \"%s\": "
#define MSG3 "%s: can't close debug file: "
#define MSG4 "%s: debugger aborting because %s\n"
#define MSG5 "%s: can't change owner/group of \"%s\": "

/*
 *	Macros and defines for testing file accessibility under UNIX.
 */

#  define A_EXISTS	00		/* Test for file existance */
#  define A_EXECUTE	01		/* Test for execute permission */
#  define A_WRITE	02		/* Test for write access */
#  define A_READ	03		/* Test for read access */
#  define EXISTS(pathname) (access (pathname, A_EXISTS) == 0)
#  define WRITABLE(pathname) (access (pathname, A_WRITE) == 0)

/*
 *  FUNCTION
 *
 *	db_push    push current debugger state and set up new one
 *
 *  SYNOPSIS
 *
 *	VOID db_push (control)
 *	char *control;
 *
 *  DESCRIPTION
 *
 *	Given pointer to a debug control string in "control", pushes
 *	the current debug state, parses the control string, and sets
 *	up a new debug state.
 *
 *	The only attribute of the new state inherited from the previous
 *	state is the current function nesting level.  This can be
 *	overridden by using the "r" flag in the control string.
 *
 *	The debug control string is a sequence of colon separated fields
 *	as follows:
 *
 *		<field_1>:<field_2>:...:<field_N>
 *
 *	Each field consists of a mandatory flag character followed by
 *	an optional "," and comma separated list of modifiers:
 *
 *		flag[,modifier,modifier,...,modifier]
 *
 *	The currently recognized flag characters are:
 *
 *		d	Enable output from DBUG_PRINTF macros for
 *			for the current state.  May be followed
 *			by a list of keywords which selects output
 *			only for the DBUG macros with that keyword.
 *			A null list of keywords implies output for
 *			all macros.
 *
 *		f	Limit debugging and/or tracing to the
 *			list of named functions and the functions
 *			which they invoke.  Note that a null
 *			list will disable all functions.  The
 *			appropriate "d" or "t" flags must still
 *			be given, this flag only limits their
 *			actions if they are enabled.
 *
 *		F	Identify the source file name for each
 *			line of debug or trace output.
 *
 *		L	Identify the source file line number for
 *			each line of debug or trace output.
 *
 *		n	Print the current function nesting depth for
 *			each line of debug or trace output.
 *	
 *		N	Number each line of dbug output.
 *
 *		p	Limit debugger actions to specified processes.
 *			A process must be identified with the
 *			DBUG_PROCESS macro and match one in the list
 *			for debugger actions to occur.
 *
 *		P	Print the current process name and PID for each
 *			line of debug or trace output.
 *
 *		r	When pushing a new state, do not inherit
 *			the previous state's function nesting level.
 *			Useful when the output is to start at the
 *			left margin.
 *
 *		t	Enable function call/exit trace lines.
 *			May be followed by a list (containing only
 *			one modifier) giving a numeric maximum
 *			trace level, beyond which no output will
 *			occur for either debugging or tracing
 *			macros.  The default is a compile time
 *			option.
 *
 *		l	Specify level of debugging output required.
 *			Default is 1.
 *
 *	Some examples of debug control strings which might appear
 *	on a shell command line (the "-#" is typically used to
 *	introduce a control string to an application program) are:
 *
 *		-#d:t
 *		-#d:f,main,subr1:F:L:t,20
 *		-#d,input,output,files:n:l,2
 *
 *	For convenience, any leading "-#" is stripped off.
 *
 */

VOID db_push (control)
char *control;
{
    REGISTER char *scan;
    REGISTER struct link *temp;
    REGISTER int oflag_used = 0;

    if (db_fp == NULL)
	db_fp = DBUG_DEST;

    if (control && *control == '-') {
	if (*++control == '#') {
	    control++;
	}	
    }
    control = StrDup (control);
    PushState ();
    scan = strtok (control, ":");
    for (; scan != NULL; scan = strtok ((char *)NULL, ":")) {
	switch (*scan++) {
	    case 'd': 
		stack -> flags |= DEBUG_ON;
		if (*scan++ == ',') {
		    stack -> keywords = ListParse (scan);
		}
		break;
	    case 'f': 
		if (*scan++ == ',') {
		    stack -> functions = ListParse (scan);
		    db_on = 0;
		}
		break;
	    case 'F': 
		stack -> flags |= FILE_ON;
		break;
	    case 'L': 
		stack -> flags |= LINE_ON;
		break;
	    case 'n': 
		stack -> flags |= DEPTH_ON;
		break;
	    case 'N':
		stack -> flags |= NUMBER_ON;
		break;
	    case 'o':
		if (*scan++ == ',') {
		    temp = ListParse (scan);
		    OpenFile (temp -> string);
		    FreeList (temp);
		} else {
		    OpenFile ("-");
		}
    		oflag_used = 1;
		break;
	    case 'p':
		if (*scan++ == ',') {
		    stack -> processes = ListParse (scan);
		}
		break;
	    case 'P': 
		stack -> flags |= PROCESS_ON;
		break;
	    case 'r': 
		stack -> level = 0;
		break;
	    case 't': 
		stack -> flags |= TRACE_ON;
		if (*scan++ == ',') {
		    temp = ListParse (scan);
		    stack -> maxdepth = atoi (temp -> string);
		    FreeList (temp);
		}
		break;
	    case 'l': 
		if (*scan++ == ',') {
		    temp = ListParse (scan);
		    stack -> db_lev = atoi (temp -> string);
		    FreeList (temp);
		}
		break;
	}
    }
    free ((void *)control);
}


/*
 *  FUNCTION
 *
 *	db_pop    pop the debug stack
 *
 *  DESCRIPTION
 *
 *	Pops the debug stack, returning the debug state to its
 *	condition prior to the most recent db_push invocation.
 *	Note that the pop will fail if it would remove the last
 *	valid state from the stack.  This prevents user errors
 *	in the push/pop sequence from screwing up the debugger.
 *	Maybe there should be some kind of warning printed if the
 *	user tries to pop too many states.
 *
 */

VOID db_pop ()
{
    REGISTER struct state *discard;

    discard = stack;
    if (discard != NULL && discard -> next_state != NULL) {
	stack = discard -> next_state;
	db_fp = stack -> out_file;
	if (discard -> keywords != NULL) {
	    FreeList (discard -> keywords);
	}
	if (discard -> functions != NULL) {
	    FreeList (discard -> functions);
	}
	if (discard -> processes != NULL) {
	    FreeList (discard -> processes);
	}
	CloseFile (discard -> out_file);
	free ((void *) discard);
    }
}

/*
 *  FUNCTION
 *
 *	db_enter    process entry point to user function
 *
 *  SYNOPSIS
 *
 *	VOID db_enter (_func_, _file_, _line_, _sfunc_, _sfile_, _slevel_)
 *	char *_func_;		points to current function name
 *	char *_file_;		points to current file name
 *	int _line_;		called from source line number
 *	char **_sfunc_;		save previous _func_
 *	char **_sfile_;		save previous _file_
 *	int *_slevel_;		save previous nesting level
 *
 *  DESCRIPTION
 *
 *	Called at the beginning of each user function to tell
 *	the debugger that a new function has been entered.
 *	Note that the pointers to the previous user function
 *	name and previous user file name are stored on the
 *	caller's stack (this is why the ENTER macro must be
 *	the first "executable" code in a function, since it
 *	allocates these storage locations).  The previous nesting
 *	level is also stored on the callers stack for internal
 *	self consistency checks.
 *
 *	If the current function is in the function select list the
 *	debugging/tracing 'on' flag is incremented so that output
 *	will be enabled while below the current function.
 *
 *	Also prints a trace line if tracing is enabled and
 *	increments the current function nesting depth.
 *
 *	Note that this mechanism allows the debugger to know
 *	what the current user function is at all times, without
 *	maintaining an internal stack for the function names.
 *
 */

VOID db_enter (_func_, _file_, _line_, _sfunc_, _sfile_, _slevel_)
char *_func_;
char *_file_;
int _line_;
char **_sfunc_;
char **_sfile_;
int *_slevel_;
{
    if (!init_done) {
	db_push ("");
    }

    *_sfunc_ = func;
    *_sfile_ = file;
    func = _func_;
    file = BaseName (_file_);
    stack -> level += INDENT;
    *_slevel_ = stack -> level;

    /* Don't output anything if db_lev greater than requested */
    if (db_lev <= stack -> db_lev)
    {
	if (InList(stack -> functions, func))
	    ++ db_on;
	if (DoTrace ()) {
	    DoPrefix (_line_);
	    Indent (stack -> level);
	    (VOID) fprintf (db_fp, ">%10.10s\n", func);
	    (VOID) fflush (db_fp);
	}
    }
}

/*
 *  FUNCTION
 *
 *	db_return    process exit from user function
 *
 *  SYNOPSIS
 *
 *	VOID db_return (_line_, _sfunc_, _sfile_, _slevel_)
 *	int _line_;		current source line number
 *	char **_sfunc_;		where previous _func_ is to be retrieved
 *	char **_sfile_;		where previous _file_ is to be retrieved
 *	int *_slevel_;		where previous level was stashed
 *
 *  DESCRIPTION
 *
 *	Called just before user function executes an explicit or implicit
 *	return.  Prints a trace line if trace is enabled, adjusts the
 *	debugging/tracing 'on' flag if the current function is selected,
 *	decrements the current nesting level, and restores the current
 *	function and file names from the defunct function's stack.
 *
 */

VOID db_return (_line_, _sfunc_, _sfile_, _slevel_)
int _line_;
char **_sfunc_;
char **_sfile_;
int *_slevel_;
{
    if (!init_done) {
	db_push ("");
    }

    if (stack -> level != *_slevel_ && (TRACING || DEBUGGING)) {
	(VOID) fprintf (db_fp, MSG1, db_process, func);
	(VOID) fflush (db_fp);
    }

    /* If output level greater than requested, just reverse the
       things done in db_enter() */
    if (db_lev <= stack -> db_lev)
    {
	    if (stack -> level == *_slevel_ && DoTrace ()) {
		DoPrefix (_line_);
		Indent (stack -> level);
		(VOID) fprintf (db_fp, "<%10.10s\n", func);
		(VOID) fflush (db_fp);
	    }
	    if (InList(stack -> functions, func) && db_on > 0)
		-- db_on;
    }

    stack -> level = *_slevel_ - INDENT;
    func = *_sfunc_;
    file = *_sfile_;
}

/*
 *  FUNCTION
 *
 *	db_printf    handle print of debug lines
 *
 *  SYNOPSIS
 *
 *	VOID db_printf (keyword, format, arg1, ... argn)
 *	char *keyword,  *format;
 *
 *  DESCRIPTION
 *
 *	When invoked via the DBUG_PRINTF macro, tests the keyword
 *	to see if that macro has been selected for processing via
 *	the debugger control string, and if so, handles printing
 *	of the arguments via the format string.  The line number
 *	of the DBUG_PRINTF macro in the source is found in db_line.
 *
 *	Note that the format string SHOULD NOT include a terminating
 *	newline, this is supplied automatically.
 *
 */

/* split into db_vprintf and db_printf to allow use of db_vprintf
   in reporting functions */

#ifdef __TenDRA__
VOID db_vprintf (char *keyword, const char *format, va_list args)
#else
VOID db_vprintf (keyword, format, args)
char	*keyword;
char	*format;
va_list	args;
#endif
{
	int	_line_ = db_line;

	if (!init_done) {
	    db_push ("");
	}

	if (db_on <= 0 || db_lev > stack -> db_lev)
		return;

    	if (db_keyword (keyword)) {
		DoPrefix (_line_);
		if (TRACING) {
	    		Indent (stack -> level + INDENT);
		} else {
	    		(VOID) fprintf (db_fp, "%10.10s: ", func);
		}
		(VOID) fprintf (db_fp, "%10.10s: ", keyword);
		(VOID) vfprintf (db_fp, format, args);
		(VOID) fprintf (db_fp, "\n");
		(VOID) fflush (db_fp);
    	}

}

/*VARARGS2*/
#if TEST_ANSI
VOID db_printf (char *keyword, db_printf_fmt format, ...)
#else
VOID db_printf (keyword, format, va_alist)
char *keyword;
char *format;
va_dcl
#endif
{
	va_list	args;

	if (db_on <= 0 || db_lev > stack -> db_lev)
		return;

#if TEST_ANSI
	va_start(args, format);
#else
	va_start(args);
#endif

	db_vprintf (keyword, format, args);

	va_end(args);
}

/*
 *  FUNCTION
 *
 *	ListParse    parse list of modifiers in debug control string
 *
 *  SYNOPSIS
 *
 *	LOCAL struct link *ListParse (ctlp)
 *	char *ctlp;
 *
 *  DESCRIPTION
 *
 *	Given pointer to a comma separated list of strings in "cltp",
 *	parses the list, building a list and returning a pointer to it.
 *	The original comma separated list is destroyed in the process of
 *	building the linked list, thus it had better be a duplicate
 *	if it is important.
 *
 *	Note that since each link is added at the head of the list,
 *	the final list will be in "reverse order", which is not
 *	significant for our usage here.
 *
 */

LOCAL struct link *ListParse (ctlp)
char *ctlp;
{
    REGISTER char *start;
    REGISTER struct link *new;
    REGISTER struct link *head;

    head = NULL;
    while (*ctlp != EOS) {
	start = ctlp;
	while (*ctlp != EOS && *ctlp != ',') {
	    ctlp++;
	}
	if (*ctlp == ',') {
	    *ctlp++ = EOS;
	}
	new = (struct link *) DbugMalloc (sizeof (struct link));
	new -> string = StrDup (start);
	new -> next_link = head;
	head = new;
    }
    return (head);
}

/*
 *  FUNCTION
 *
 *	InList    test a given string for member of a given list
 *
 *  SYNOPSIS
 *
 *	LOCAL BOOLEAN InList (linkp, cp)
 *	struct link *linkp;
 *	char *cp;
 *
 *  DESCRIPTION
 *
 *	Tests the string pointed to by "cp" to determine if it is in
 *	the list pointed to by "linkp".  Linkp points to the first
 *	link in the list.  If linkp is NULL then the string is treated
 *	as if it is in the list (I.E all strings are in the null list).
 *	This may seem rather strange at first but leads to the desired
 *	operation if no list is given.  The net effect is that all
 *	strings will be accepted when there is no list, and when there
 *	is a list, only those strings in the list will be accepted.
 *
 */

LOCAL BOOLEAN InList (linkp, cp)
struct link *linkp;
char *cp;
{
    REGISTER struct link *scan;
    REGISTER BOOLEAN accept;

    if (linkp == NULL) {
	accept = TRUE;
    } else {
	accept = FALSE;
	for (scan = linkp; scan != NULL; scan = scan -> next_link) {
	    if (STREQ (scan -> string, cp)) {
		accept = TRUE;
		break;
	    }
	}
    }
    return (accept);
}

/*
 *  FUNCTION
 *
 *	PushState    push current state onto stack and set up new one
 *
 *  SYNOPSIS
 *
 *	LOCAL VOID PushState ()
 *
 *  DESCRIPTION
 *
 *	Pushes the current state on the state stack, and initializes
 *	a new state.  The only parameter inherited from the previous
 *	state is the function nesting level.  This action can be
 *	inhibited if desired, via the "r" flag.
 *
 *	The state stack is a linked list of states, with the new
 *	state added at the head.  This allows the stack to grow
 *	to the limits of memory if necessary.
 *
 */

LOCAL VOID PushState ()
{
    REGISTER struct state *new;

    new = (struct state *) DbugMalloc (sizeof (struct state));
    new -> flags = 0;
    new -> maxdepth = MAXDEPTH;
    if (stack != NULL) {
	new -> level = stack -> level;
    } else {
	new -> level = 0;
    }
    new -> db_lev = 1;
    new -> out_file = DBUG_DEST;
    new -> functions = NULL;
    new -> keywords = NULL;
    new -> processes = NULL;
    new -> next_state = stack;
    stack = new;
    init_done = TRUE;
}

/*
 *  FUNCTION
 *
 *	DoTrace    check to see if tracing is current enabled
 *
 *  SYNOPSIS
 *
 *	LOCAL BOOLEAN DoTrace ()
 *
 *  DESCRIPTION
 *
 *	Checks to see if tracing is enabled based on whether the
 *	user has specified tracing, the maximum trace depth has
 *	not yet been reached, and the current process is selected.
 *	Returns TRUE if	tracing is enabled, FALSE otherwise.
 *
 */

LOCAL BOOLEAN DoTrace ()
{
    if (TRACING)
	if (stack -> level / INDENT <= stack -> maxdepth)
	    if (InList (stack -> processes, db_process))
		return TRUE;

    return (FALSE);
}

/*
 *  FUNCTION
 *
 *	db_keyword    test keyword for member of keyword list
 *
 *  SYNOPSIS
 *
 *	BOOLEAN db_keyword (keyword)
 *	char *keyword;
 *
 *  DESCRIPTION
 *
 *	Test a keyword to determine if it is in the currently active
 *	keyword list.  As with the function list, a keyword is accepted
 *	if the list is null, otherwise it must match one of the list
 *	members.  When debugging is not on, no keywords are accepted.
 *	After the maximum trace level is exceeded, no keywords are
 *	accepted (this behavior subject to change).  Additionally,
 *	the current process must be accepted.
 *
 *	Returns TRUE if keyword accepted, FALSE otherwise.
 *
 */

BOOLEAN db_keyword (keyword)
char *keyword;
{
    if (!init_done)
	db_push ("");

    if (DEBUGGING)
	if (stack -> level / INDENT <= stack -> maxdepth)
	    if (InList (stack -> keywords, keyword))
		if (InList (stack -> processes, db_process))
		    return TRUE;

    return (FALSE);
}

/*
 *  FUNCTION
 *
 *	Indent    indent a line to the given indentation level
 *
 *  SYNOPSIS
 *
 *	LOCAL Indent (indent)
 *	int indent;
 *
 *  DESCRIPTION
 *
 *	Indent a line to the given level.  Note that this is
 *	a simple minded but portable implementation.
 *	There are better ways.
 *
 */

LOCAL VOID Indent (indent)
int indent;
{
    REGISTER int count;
    AUTO char buffer[PRINTBUF];

    for (count = 0; (count < (indent - INDENT)) && (count < (PRINTBUF - 1)); count++) {
	if ((count % INDENT) == 0) {
	    buffer[count] = '|';
	} else {
	    buffer[count] = ' ';
	}
    }
    buffer[count] = EOS;
    (VOID) fprintf (db_fp, buffer);
    /* (VOID) fflush (db_fp);   ** don't flush till end of line **/
}

/*
 *  FUNCTION
 *
 *	FreeList    free all memory associated with a linked list
 *
 *  SYNOPSIS
 *
 *	LOCAL FreeList (linkp)
 *	struct link *linkp;
 *
 *  DESCRIPTION
 *
 *	Given pointer to the head of a linked list, frees all
 *	memory held by the list and the members of the list.
 *
 */

LOCAL VOID FreeList (linkp)
struct link *linkp;
{
    REGISTER struct link *old;

    while (linkp != NULL) {
	old = linkp;
	linkp = linkp -> next_link;
	if (old -> string != NULL) {
	    free ((void *) old -> string);
	}
	free ((void *) old);
    }
}

/*
 *  FUNCTION
 *
 *	StrDup   make a duplicate of a string in new memory
 *
 *  SYNOPSIS
 *
 *	LOCAL char *StrDup (string)
 *	char *string;
 *
 *  DESCRIPTION
 *
 *	Given pointer to a string, allocates sufficient memory to make
 *	a duplicate copy, and copies the string to the newly allocated
 *	memory.  Failure to allocated sufficient memory is immediately
 *	fatal.
 *
 */

LOCAL char *StrDup (string)
char *string;
{
    REGISTER char *new;

    new = DbugMalloc ((int) strlen(string) + 1);
    (VOID) strcpy (new, string);
    return (new);
}

/*
 *  FUNCTION
 *
 *	DoPrefix    print debugger line prefix prior to indentation
 *
 *  SYNOPSIS
 *
 *	LOCAL DoPrefix (_line_)
 *	int _line_;
 *
 *  DESCRIPTION
 *
 *	Print prefix common to all debugger output lines, prior to
 *	doing indentation if necessary.  Print such information as
 *	current process name, current source file name and line number,
 *	and current function nesting depth.
 *
 *  MODIFICATION
 *	debug control prefix added for VSX project usage.
 *	June 6 1986  M.Ghezel-Ayagh
 *
 */
  
 
LOCAL VOID DoPrefix (_line_)
int _line_;
{
    lineno++;
    if (stack -> flags & NUMBER_ON) {
	(VOID) fprintf (db_fp, "%5.5d: ", lineno);
    }
    if (stack -> flags & PROCESS_ON) {
	(VOID) fprintf (db_fp, "%10.10s(%5.5d): ", db_process, (int)getpid());
    }
    if (stack -> flags & FILE_ON) {
	(VOID) fprintf (db_fp, "%14.14s: ", file);
    }
    if (stack -> flags & LINE_ON) {
	(VOID) fprintf (db_fp, "%5.5d: ", _line_);
    }
    if (stack -> flags & DEPTH_ON) {
	(VOID) fprintf (db_fp, "%4.4d: ", stack -> level / INDENT);
    }
    /* (VOID) fflush (db_fp);   ** don't flush till end of line **/
}

/*
 *  FUNCTION
 *
 *	OpenFile	open new output stream for debugger output
 *
 *  SYNOPSIS
 *
 *	LOCAL VOID OpenFile (name)
 *	char *name;
 *
 *  DESCRIPTION
 *
 *	Given name of a new file (or "-" for stdout) opens the file
 *	and sets the output stream to the new file.
 */

LOCAL VOID OpenFile (name)
char *name;
{
    REGISTER FILE *fp;
    REGISTER BOOLEAN newfile;
    REGISTER int flags;
    static char buffer[BUFSIZ];


    if (name != NULL) {
	if (strcmp (name, "-") == 0) {
	    db_fp = DBUG_DEST;
	    stack -> out_file = db_fp;
	} else {
	    if (!Writable (name)) {
		(VOID) fprintf (stderr, MSG2, db_process, name);
		perror ("");
		(VOID) fflush (stderr);
	    } else {
		if (EXISTS (name)) {
		    newfile = FALSE;
		} else {
		    newfile = TRUE;
		}
		fp = fopen (name, "a");
		if (fp == NULL) {
 		    (VOID) fprintf (stderr, MSG2, db_process, name);
		    perror ("");
		    (VOID) fflush (stderr);
		} else {
		    db_fp = fp;
		    stack -> out_file = fp;
		    if (newfile) {
			ChangeOwner (name);
		    }
		    /* Set fd to be closed across exec, otherwise it
		       will be open twice in new process */
		    flags = fcntl(fileno(db_fp), F_GETFD);
		    (void) fcntl(fileno(db_fp), F_SETFD, flags & FD_CLOEXEC);
		}
	    }
	}

        /* Ensure output is in append mode, and buffered */
	if ((flags = fcntl(fileno(db_fp), F_GETFL)) != -1)
		(void) fcntl(fileno(db_fp), F_SETFL, flags & O_APPEND);
	setbuf(db_fp, buffer);
    }
}

/*
 *  FUNCTION
 *
 *	CloseFile    close the debug output stream
 *
 *  SYNOPSIS
 *
 *	LOCAL VOID CloseFile (fp)
 *	FILE *fp;
 *
 *  DESCRIPTION
 *
 *	Closes the debug output stream unless it is standard output
 *	or standard error.
 *
 */

LOCAL VOID CloseFile (fp)
FILE *fp;
{
    if (fp != stderr && fp != stdout) {
	if (fclose (fp) == EOF) {
	    (VOID) fprintf (stderr, MSG3, db_process);
	    perror ("");
	    (VOID) fflush (stderr);
	}
    }
}

/*
 *  FUNCTION
 *
 *	DbugExit    print error message and exit
 *
 *  SYNOPSIS
 *
 *	LOCAL VOID DbugExit (why)
 *	char *why;
 *
 *  DESCRIPTION
 *
 *	Prints error message using current process name, the reason for
 *	aborting (typically out of memory), and exits with status 1.
 *	This should probably be changed to use a status code
 *	defined in the user's debugger include file.
 *
 */
 
LOCAL VOID DbugExit (why)
char *why;
{
    (VOID) fprintf (stderr, MSG4, db_process, why);
    (VOID) fflush (stderr);
    exit (1);
}

/*
 *  FUNCTION
 *
 *	DbugMalloc    allocate memory for debugger runtime support
 *
 *  SYNOPSIS
 *
 *	LOCAL char *DbugMalloc (size)
 *	int size;
 *
 *  DESCRIPTION
 *
 *	Allocate more memory for debugger runtime support functions.
 *	Failure to to allocate the requested number of bytes is
 *	immediately fatal to the current process.  This may be
 *	rather unfriendly behavior.  It might be better to simply
 *	print a warning message, freeze the current debugger state,
 *	and continue execution.
 *
 */
 
LOCAL char *DbugMalloc (size)
int size;
{
    register char *new;

    new = (char *) malloc ((size_t) size);
    if (new == NULL) {
	DbugExit ("out of memory");
    }
    return (new);
}

/*
 *  FUNCTION
 *
 *	BaseName    strip leading pathname components from name
 *
 *  SYNOPSIS
 *
 *	LOCAL char *BaseName (pathname)
 *	char *pathname;
 *
 *  DESCRIPTION
 *
 *	Given pointer to a complete pathname, locates the base file
 *	name at the end of the pathname and returns a pointer to
 *	it.
 *
 */

LOCAL char *BaseName (pathname)
char *pathname;
{
    register char *base;

    base = strrchr (pathname, '/');
    if (base++ == NULL) {
	base = pathname;
    }
    return (base);
}

/*
 *  FUNCTION
 *
 *	Writable    test to see if a pathname is writable/creatable
 *
 *  SYNOPSIS
 *
 *	LOCAL BOOLEAN Writable (pathname)
 *	char *pathname;
 *
 *  DESCRIPTION
 *
 *	Because the debugger might be linked in with a program that
 *	runs with the set-uid-bit (suid) set, we have to be careful
 *	about opening a user named file for debug output.  This consists
 *	of checking the file for write access with the real user id,
 *	or checking the directory where the file will be created.
 *
 *	Returns TRUE if the user would normally be allowed write or
 *	create access to the named file.  Returns FALSE otherwise.
 *
 */

LOCAL BOOLEAN Writable (pathname)
char *pathname;
{
    REGISTER BOOLEAN granted;
    REGISTER char *lastslash;

    granted = FALSE;
    if (EXISTS (pathname)) {
	if (WRITABLE (pathname)) {
	    granted = TRUE;
	}
    } else {
	lastslash = strrchr (pathname, '/');
	if (lastslash != NULL) {
	    *lastslash = EOS;
	} else {
	    pathname = ".";
	}
	if (WRITABLE (pathname)) {
	    granted = TRUE;
	}
	if (lastslash != NULL) {
	    *lastslash = '/';
	}
    }
    return (granted);
}

/*
 *  FUNCTION
 *
 *	ChangeOwner    change owner to real user for suid programs
 *
 *  SYNOPSIS
 *
 *	LOCAL VOID ChangeOwner (pathname)
 *
 *  DESCRIPTION
 *
 *	For unix systems, change the owner of the newly created debug
 *	file to the real owner and make it world writable.  This is
 *	strictly for the benefit of programs that are running with the
 *	set-user-id bit set.
 *
 *	Note that at this point, the fact that pathname represents
 *	a newly created file has already been established.  If the
 *	program that the debugger is linked to is not running with
 *	the suid bit set, then this operation is redundant (but
 *	harmless).
 *
 */

LOCAL VOID ChangeOwner (pathname)
char *pathname;
{
    (VOID) chmod(pathname, (mode_t)(S_IRUSR|S_IWUSR|S_IWGRP|S_IWOTH));
    if (chown (pathname, getuid (), getgid ()) == -1) {
	(VOID) fprintf (stderr, MSG5, db_process, pathname);
	perror ("");
	(VOID) fflush (stderr);
    }
}

/*
 *  FUNCTION
 *
 *	db_setjmp    save debugger environment
 *
 *  SYNOPSIS
 *
 *	EXPORT void db_setjmp ()
 *
 *  DESCRIPTION
 *
 *	Invoked as part of the user's DBUG_SETJMP macro to save
 *	the debugger environment in parallel with saving the user's
 *	environment.
 *
 */

EXPORT void db_setjmp ()
{
   if (stack != NULL)
   	jmplevel = stack -> level;
   else
	jmplevel = 0;
   jmpfunc = func;
   jmpfile = file;
}

/*
 *  FUNCTION
 *
 *	db_longjmp    restore previously saved debugger environment
 *
 *  SYNOPSIS
 *
 *	EXPORT void db_longjmp ()
 *
 *  DESCRIPTION
 *
 *	Invoked as part of the user's DBUG_LONGJMP macro to restore
 *	the debugger environment in parallel with restoring the user's
 *	previously saved environment.
 *
 */

EXPORT void db_longjmp ()
{
    stack -> level = jmplevel;
    if (jmpfunc) {
	func = jmpfunc;
    }
    if (jmpfile) {
	file = jmpfile;
    }
}
