/*
*      SCCS:   @(#)common/vlib(report.c)	4.1 (01/21/97)  VSXgen release 1.4
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/
/***********************************************************************

NAME:		report.c - results file output functions

PROJECT:	VSX (X/OPEN Validation Suite)

MODULE CONTAINS:
		er_rpt(),
		in_rpt(),
		xs_rpt(),
		xx_rpt()
SYNOPSIS:

	void er_rpt(reason)
		char *reason;
	
	void in_rpt(format [ , arg ] .... )
		char *format;

	void xs_rpt()

	void xx_rpt(result)
		char  result;

DESCRIPTION:
	These calls interface to the appropriate TET reporting calls.

	er_rpt() outputs a line consisting of the standard error
	control header followed by the string "reason".

	in_rpt() outputs additional information for test failure,
	test deletion and tests that require further inspection.

	xs_rpt() outputs a testset success line.
	xx_rpt() outputs a testset failure line.

AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	18 Oct 1990
MODIFICATIONS:
		Geoff Clare, 17 Dec 1991
			Rename Er_rpt, In_rpt, Xs_rpt and Xx_rpt to
			allow for case-insensitive linkers.

		Geoff Clare, 19 Aug 1994
			Change in_dbrpt() to call va_start(args) and
			va_end(args) around each use of args.

		Geoff Clare, 24 May 1996
			TenDRA printf format string checking on in_rpt.

		Geoff Clare, 20 Jan 1997
			Use tet_vprintf() if available.

************************************************************************/

#include	<std.h>
#include	<stdio.h>
#if TEST_ANSI
#include	<stdarg.h>
#else
#include	<varargs.h>
#endif
#include	<string.h>
#include	<errno.h>
#include	<dbug.h>
#include	<report.h>
#include	<tet_api.h>
#include	<sysdep.h>

#undef in_rpt
#undef er_rpt
#undef xs_rpt
#undef xx_rpt

/* extra TET result codes from the VSX tet_code file */
#define	VSX_TET_WARNING	101
#define	VSX_TET_FIP	102

/* tet_infoline() can only write 512 bytes - with the line prefix this
means you can normally only pass it about 480 without it truncating */
#define MAXTETINFO 470

private	char buf[2500];	/* this is vile, but the alternative is to do
			   something like a vfprintf() to /dev/null to
			   find the number of bytes in_rpt() will output */

void
er_rpt(reason)
	char *reason;
{
	(void) sprintf(buf, "ERROR: %s", reason);
	tet_infoline(buf);
}

#ifndef DBUG_OFF
void
er_dbrpt(reason)
	char *reason;
{
	if (db_on)
		db_printf("Error", "%s", reason);
	er_rpt(reason);
}
#endif

private void
#ifdef __TenDRA__
do_info(const char *format, va_list args)
#else
do_info(format, args)
char *format;
va_list args;
#endif
{
#ifdef TET_ER_OK /* defined by TETware, but not earlier TET versions */

	if (tet_vprintf(format, args) < 0)
	{
		vsxerror("tet_vprintf() failed - tet_errno %d", tet_errno);
		exit(1);
	}

#else /* not TETware */

	/* in_rpt() may well want to write more data than tet_infoline()
	   will accept, so split it up */

	char *cp, savchar = '\0';
	size_t len;

	(void) vsprintf(buf, format, args);

	cp = buf;
	while ((len = strlen(cp)) > 0)
	{
		if (len > MAXTETINFO)
		{
			savchar = cp[MAXTETINFO];
			cp[MAXTETINFO] = '\0';
			tet_infoline(cp);
			*(cp += MAXTETINFO) = savchar;
		}
		else
		{
			tet_infoline(cp);
			cp += len;
		}
	}
#endif /* not TETware */
}

void
#if TEST_ANSI
in_rpt(char *format, ...)
#else
/*VARARGS1*/
in_rpt(format, va_alist)
	char *format;
	va_dcl
#endif
{
	/* call as in_rpt(format, arg1, ... argn)	*/

	va_list	args;

#if TEST_ANSI
	va_start(args, format);
#else
	va_start(args);
#endif
	do_info(format, args);

	va_end(args);
}

#ifndef DBUG_OFF

void
#if TEST_ANSI
in_dbrpt(db_printf_fmt format, ...)
#else
/*VARARGS0*/
in_dbrpt(format, va_alist)
	char *format;
	va_dcl
#endif
{
	/* call as in_dbrpt(format, arg1, ... argn)	*/

	va_list	args;

	if (db_on)
	{
#if TEST_ANSI
		va_start(args, format);
#else
		va_start(args);
#endif
		db_vprintf("Info", format, args);
		va_end(args);
	}

#if TEST_ANSI
	va_start(args, format);
#else
	va_start(args);
#endif
	do_info(format, args);
	va_end(args);
}
#endif

void
xs_rpt()
{
	tet_result(TET_PASS);
}

#ifndef DBUG_OFF
void
xs_dbrpt()
{
	if (db_on)
		db_printf("Result", "Success");
	xs_rpt();
}
#endif

void
xx_rpt(result)
	char  result;
{
	int	tet_res = -1;
	int	saverrno = errno;

	switch(result)
	{
	case DELETION :	   tet_res = TET_UNRESOLVED;	break;
	case FAILURE :	   tet_res = TET_FAIL;		break;
	case NOTINUSE :	   tet_res = TET_NOTINUSE;	break;
	case UNSUPPORTED : tet_res = TET_UNSUPPORTED;	break;
	case UNTESTED :	   tet_res = TET_UNTESTED;	break;
	case CANCELLED :   tet_res = TET_UNINITIATED;	break;
	case RFI :	   tet_res = VSX_TET_FIP;	break;
	case WARNING :	   tet_res = VSX_TET_WARNING;	break;
	default :
		(void) sprintf(buf, "bad result code passed to xx_rpt(): %d",
			result);
		tet_infoline(buf);
	}

	if (tet_res != -1)
		tet_result(tet_res);

	/* Restore errno to allow: xx_rpt(FAILURE); in_rpt("%d", errno); */
	errno = saverrno;
}

#ifndef DBUG_OFF
void
xx_dbrpt(result)
	char  result;
{
	int		saverrno = errno;

	if (db_on)
		db_printf("Result", "Test Failure, Status: %c", result);
	xx_rpt(result);

	errno = saverrno;
}
#endif
