/*
*      SCCS:   @(#)common/vport(nonposix.c)	4.1 (01/21/97)  VSXgen release 1.3
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/
/***********************************************************************

NAME:		nonposix.c - non-POSIX interfaces used in POSIX testsets

PROJECT:	VSX (X/OPEN Validation Suite)

MODULE CONTAINS:
		getopt(), optind, optarg, opterr,
		isnan(),
		memchr(), memcmp(), memcpy(), memset(),
		popen(), pclose(),
		strtol(),
		system(),
		vprintf(), vfprintf(), vsprintf().
SYNOPSIS:
		as XPG3

DESCRIPTION:
	If VSX is built in a POSIX mode (i.e. with _XOPEN_SOURCE not
	defined) the suite cannot use any non-POSIX interfaces as they
	may not be available on the system.  Therefore portable versions
	of those interfaces which are used are provided here.

AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	24 May 1989
MODIFICATIONS:
		Geoff Clare, 13 Feb 1990
			Change execl("/bin/sh", ...) to execlp("sh", ...)
		
		Geoff Clare, 4 May 1990
			Don't supply ANSI functions if using ANSI compiler
		
		Geoff Clare, 25 May 1990
			Allow for unlimited OPEN_MAX

		Geoff Clare, 07 Dec 1990
			Fix malloc() problem in popen().

		Geoff Clare, 27 Jan 1992
			Fix use of NULL as an integer in getopt().

		Geoff Clare, 21 May 1992
			Add vprintf() et al.

************************************************************************/

#ifdef _XOPEN_SOURCE

static int dummy;

#else /* _XOPEN_SOURCE */

#include <std.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <locale.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/wait.h>
#if !(__STDC__ > 0)
#include <varargs.h>
#endif
#include <sysdep.h>

#undef	getopt
#undef	optind
#undef	optarg
#undef	opterr
#undef	isnan
#undef	memchr
#undef	memcmp
#undef	memcpy
#undef	memset
#undef	popen
#undef	pclose
#undef	strtol
#undef	system
#undef	vprintf
#undef	vfprintf
#undef	vsprintf

#if !(__STDC__ > 0)

/*
 *  Some compilers don't grok (void *) as a formal parameter, hence
 *  the use of (char *) instead for the memory functions.  This will
 *  blow up old version of "lint" which don't know that they are
 *  required to have the same representation.
 */

public void *
memchr(s, c, n)
char *s;
int c;
size_t n;
{
	unsigned char *sp = (unsigned char *)s;

	while (n-- > 0)
		if (*sp++ == (unsigned char)c)
			return (void *)(--sp);
	return NULL;
}

public int
memcmp(s1, s2, n)
char *s1, *s2;
size_t n;
{
	unsigned char *sp1 = (unsigned char *)s1;
	unsigned char *sp2 = (unsigned char *)s2;

	if (s1 != s2)
		while (n-- > 0)
			if (*sp1++ != *sp2++)
			{
				if (*--sp1 > *--sp2)
					return 1;
				else
					return -1;
			}
	return 0;
}

public void *
memcpy(s1, s2, n)
char *s1, *s2;
size_t n;
{
	char *os1 = s1;

	while (n-- > 0)
		*s1++ = *s2++;

	return (void *)os1;
}

public void *
memset(s, c, n)
char *s;
int c;
size_t n;
{
	unsigned char *sp = (unsigned char *)s;

	while (n-- > 0)
		*sp++ = c;

	return (void *)s;
}

#define DIGIT(x)	(isdigit(x) ? (x) - '0' : \
			islower(x) ? (x) + 10 - 'a' : (x) + 10 - 'A')
#define MBASE	('z' - 'a' + 1 + 10)

public long
strtol(str, ptr, base)
char *str;
char **ptr;
int base;
{
	char *oldloc;
	long val;
	int c;
	int xx, neg = 0;

	if (ptr != (char **)0)
		*ptr = str; /* in case no number is formed */
	if (base < 0 || base > MBASE)
	{
		errno = EINVAL;
		return 0;
	}

	/* switch to C locale temporarily for <ctype.h> functions */
	oldloc = setlocale(LC_CTYPE, "C");

	if (!isalnum(c = *str))
	{
		while (isspace(c))
			c = *++str;
		switch (c)
		{
		case '-':
			neg++;
		case '+': /* fall-through */
			c = *++str;
		}
	}
	if (base == 0)
	{
		if (c != '0')
			base = 10;
		else if (str[1] == 'x' || str[1] == 'X')
			base = 16;
		else
			base = 8;
	}
	/*
	 * for any base > 10, the digits incrementally following
	 *	9 are assumed to be "abc...z" or "ABC...Z"
	 */
	if (!isalnum(c) || (xx = DIGIT(c)) >= base)
	{
		(void) setlocale(LC_CTYPE, oldloc);
		return 0; /* no number formed */
	}
	if (base == 16 && c == '0' && isxdigit(str[2]) &&
	    (str[1] == 'x' || str[1] == 'X'))
		c = *(str += 2); /* skip over leading "0x" or "0X" */
	for (val = -DIGIT(c); isalnum(c = *++str) && (xx = DIGIT(c)) < base; )
		/* accumulate neg avoids surprises near LONG_MAX */
		val = base * val - xx;
	if (ptr != (char **)0)
		*ptr = str;
	(void) setlocale(LC_CTYPE, oldloc);
	return (neg ? val : -val);
}

public int
system(s)
char	*s;
{
	int	status, err = errno;
	pid_t	pid;
	sigset_t set, oset;

	if ((pid = fork()) == 0)
	{
		(void) execlp("sh", "sh", "-c", s, (char *)0);
		_exit(127);
	}
	else if (pid == SYSERROR)
		return SYSERROR;

	(void) sigemptyset(&set);
	(void) sigaddset(&set, SIGHUP);
	(void) sigaddset(&set, SIGINT);
	(void) sigaddset(&set, SIGQUIT);
	(void) sigprocmask(SIG_BLOCK, &set, &oset);

	if (waitpid(pid, &status, 0) != pid)
	{
		status = SYSERROR;
		err = errno;
	}

	(void) sigprocmask(SIG_SETMASK, &oset, (sigset_t *)NULL);
	if (sigpending(&set) != SYSERROR &&
	   (sigismember(&set, SIGHUP) ||
	    sigismember(&set, SIGINT) ||
	    sigismember(&set, SIGQUIT)))
	{
		(void) sigsuspend(&oset);
		err = errno;
	}

	errno = err;
	return status;
}

#endif /* __STDC__ */

#ifndef __linux__

public	int	opterr = 1;
public	int	optind = 1;
public	char	*optarg;

public int
getopt(argc, argv, opts)
int	argc;
char	**argv, *opts;
{
	static int sp = 1;
	int c;
	char *cp;

	if (sp == 1)
	{
		if (optind >= argc ||
		   argv[optind][0] != '-' || argv[optind][1] == '\0')
			return EOF;
		else if(strcmp(argv[optind], "--") == 0)
		{
			optind++;
			return EOF;
		}
	}

	c = argv[optind][sp];
	if (c == ':' || (cp=strchr(opts, c)) == NULL)
	{
		if (opterr)
			(void) fprintf(stderr, "%s: illegal option -- %c\n",
				argv[0], c);
		if (argv[optind][++sp] == '\0')
		{
			optind++;
			sp = 1;
		}
		return '?';
	}
	if (*++cp == ':')
	{
		if (argv[optind][sp+1] != '\0')
			optarg = &argv[optind++][sp+1];
		else if(++optind >= argc)
		{
			if (opterr)
				(void) fprintf(stderr,
				    "%s: option requires an argument -- %c\n",
				    argv[0], c);
			sp = 1;
			return '?';
		} else
			optarg = argv[optind++];
		sp = 1;
	}
	else
	{
		if (argv[optind][++sp] == '\0')
		{
			sp = 1;
			optind++;
		}
		optarg = NULL;
	}
	return c;
}

#endif /* __linux__ */

public int
isnan(x)
double x;
{
	return (x != x);
}

#ifndef __linux__

#define	tst(a, b) (*mode == 'r' ? (b) : (a))

static pid_t *popen_pid = NULL;

public FILE *
popen(cmd, mode)
char	*cmd, *mode;
{
	int	p[2];
	int	myside, yourside;
	pid_t	*poptr, pid;

	static long open_max;

	if (popen_pid == NULL)
	{
		errno = 0;
		open_max = sysconf(_SC_OPEN_MAX);
		if (open_max < 0 && errno == 0)
			open_max = 256;
		if (open_max < 0 ||
		   (popen_pid = (pid_t *)calloc((size_t) open_max,
		   sizeof(pid_t))) == NULL)
			return NULL;
	}

	if (pipe(p) < 0)
		return NULL;

	myside = tst(p[1], p[0]);
	yourside = tst(p[0], p[1]);

	if ((pid = fork()) == 0)
	{
		/* myside and yourside reverse roles in child */
		int	stdio;

		/* close all pipes from other popen's */
		poptr = &popen_pid[open_max];
		while (poptr > popen_pid)
		{
			if (*--poptr != 0)
				(void) close(poptr - popen_pid);
		}
		stdio = tst(0, 1);
		(void) close(myside);
		(void) close(stdio);
		(void) fcntl(yourside, F_DUPFD, stdio);
		(void) close(yourside);
		(void) execlp("sh", "sh", "-c", cmd, (char *)0);
		_exit(1);
	}
	else if (pid == -1)
		return NULL;

	popen_pid[myside] = pid;
	(void) close(yourside);

	return fdopen(myside, mode);
}

public int
pclose(ptr)
FILE	*ptr;
{
	int f, status, err = errno;
	pid_t p;
	sigset_t set, oset;

	if (popen_pid == NULL)
	{
		/* no popen() calls made */
		errno = ECHILD;
		return SYSERROR;
	}

	f = fileno(ptr);
	p = popen_pid[f];
	(void) fclose(ptr);
	popen_pid[f] = 0;

	if (p <= 0)
	{
		/* no child to wait for */
		errno = ECHILD;
		return SYSERROR;
	}

	(void) sigemptyset(&set);
	(void) sigaddset(&set, SIGHUP);
	(void) sigaddset(&set, SIGINT);
	(void) sigaddset(&set, SIGQUIT);
	(void) sigprocmask(SIG_BLOCK, &set, &oset);

	if (waitpid(p, &status, 0) != p)
	{
		status = SYSERROR;
		err = errno;
	}

	(void) sigprocmask(SIG_SETMASK, &oset, (sigset_t *)NULL);
	if (sigpending(&set) != SYSERROR &&
	   (sigismember(&set, SIGHUP) ||
	    sigismember(&set, SIGINT) ||
	    sigismember(&set, SIGQUIT)))
	{
		(void) sigsuspend(&oset);
		err = errno;
	}

	errno = err;
	return status;
}

#endif /* __linux__ */

#if !(__STDC__ > 0) && !ANSI_VPRINTF_FUNC 

#define PUTCH(c) { \
	if (buf != NULL) { \
		*buf++ = (c); \
		*buf = '\0'; \
		count++; \
	} else { \
		ret = putc((c), fp); \
		if (ret == EOF) \
			return (count > 0 ? count : -1); \
		else \
			count++; \
	} \
}

#define DOARG(var,type) { \
	argdone = 1; \
	var = va_arg(args, type); \
	*strp++ = *endp; \
	*strp = '\0'; \
	if (buf != NULL) { \
		ret = sprintf(buf, fstr, var); \
		count += ret; \
		buf += strlen(buf); \
	} else { \
		ret = fprintf(fp, fstr, var); \
		if (ret < 0) \
			return (count > 0 ? count : -1); \
		else \
			count += ret; \
	} \
}

private int
dovprnt(buf, fp, fmt, args)
char *buf;
FILE *fp;
char *fmt;
va_list args;
{
	/*
	 * if (buf != NULL) do a vsprintf(buf, fmt, args),
	 * otherwise do a vfprintf(fp, fmt, args).
	 */

#if LONG_DOUBLE_SUPPORTED
	long double	ld;
#endif
	double		d;
	char		*strp;
	char		*endp;
	char		*cp;
	long		l;
	unsigned long	ul;
	int		i;
	unsigned int	ui;
	int		shortflag;
	int		longflag;
	int		longdflag;
	int		ret;
	int		argdone;
	int		count = 0;
	char		fstr[256];

	if (buf != NULL)
		*buf = '\0';

	while (*fmt != '\0')
	{
		if (*fmt != '%')
		{
			PUTCH(*fmt)
			++fmt;
			continue;
		}

		shortflag = longflag = longdflag = 0;
		strp = fstr;
		*strp++ = '%';

		/*
		 * Find the next conversion character and use
		 * sprintf/fprintf to process the piece of format
		 * string from here up to the character.
		 */

		for (endp = fmt + 1; *endp != '\0'; endp++)
		{
			argdone = 0;
			switch (*endp)
			{
			case 'h':
				shortflag++;
				break;

			case 'l':
				longflag++;
				break;

			case 'L':
				longdflag++;
				break;

			case '*':
				i = va_arg(args, int);
				(void) sprintf(strp, "%d", i);
				strp += strlen(strp);
				break;

			case 'd':
			case 'i':
				if (longflag)
					DOARG(l, long)
				else
					DOARG(i, int)
				break;

			case 'o':
			case 'u':
			case 'x':
			case 'X':
				if (longflag)
					DOARG(ul, unsigned long)
				else
					DOARG(ui, unsigned int)
				break;

			case 'e':
			case 'E':
			case 'f':
			case 'g':
			case 'G':
#if LONG_DOUBLE_SUPPORTED
				if (longdflag)
					DOARG(ld, long double)
				else
#endif
					DOARG(d, double)
				break;

			case 'c':
				DOARG(i, int)
				break;

			case 's':
			case 'p':
				DOARG(cp, char *)
				break;

			case 'n':
				argdone = 1;
				if (longflag)
					*(va_arg(args, long int *)) = count;
				else if (shortflag)
					*(va_arg(args, short int *)) = count;
				else
					*(va_arg(args, int *)) = count;
				break;

			case '%':
				argdone = 1;
				PUTCH('%')
				break;

			default:
				*strp++ = *endp;
				break;
			}
			if (argdone)
				break;
		}
		if (*endp == '\0')
			return count;

		fmt = endp + 1;
	}

	return count;
}

public int
vfprintf(fp, fmt, ap)
FILE *fp;
char *fmt;
va_list ap;
{
	return dovprnt((char *)0, fp, fmt, ap);
}

public int
vprintf(fmt, ap)
char *fmt;
va_list ap;
{
	return dovprnt((char *)0, stdout, fmt, ap);
}

public int
vsprintf(buf, fmt, ap)
char *buf;
char *fmt;
va_list ap;
{
	return dovprnt(buf, (FILE *)0, fmt, ap);
}
#endif /* !(__STDC__ > 0) && !ANSI_VPRINTF_FUNC */

#endif /* _XOPEN_SOURCE */
