/*
*      SCCS:   @(#)common/vport(userintfES.c)	Rel 4.4.1 (06/09/97)
*
*	UniSoft Ltd., London, England
*/
/***********************************************************************

NAME:		userintfES.c - user-modifiable interface routines
		Example version for Unix System V Release 4.1ES.

PROJECT:	VSX (X/OPEN Validation Suite)

MODULE CONTAINS:
		mnt_rw()
		mnt_ro()
		unmnt()	
		setprv()
		unsetprv()
		prv_assign()
		setgrps()
		openctl()
		openpty()
		ptygetattr()
		sys_call()
		use_time()
		newroot()
		setlimit()
		pathdepth()
		ftok()
SYNOPSIS:
		int mnt_rw(spec, dir)
			char *spec;
			char *dir;

		int mnt_ro(spec, dir)
			char *spec;
			char *dir;

		int unmnt(spec, dir)
			char *spec;
			char *dir;

		int setprv(prvtype)
			int prvtype;

		int unsetprv(prvtype)
			int prvtype;

		int prv_assign(file, prvtypes)
			char *file;
			int *prvtypes;

		int setgrps(ngrps, grparray)
			int   ngrps;
			gid_t grparray[];
		
		int openctl(spec, oflag)
			char *spec;
			int  oflag;

		int openpty(master, slave, mfdp, sfdp, cntrl)
			char *master;
			char *slave;
			int *mfdp;
			int *sfdp;
			int cntrl;

		int ptygetattr(mfd, termios_p)
			int mfd;
			struct termios *termios_p;
		
		int sys_call()

		void use_time()

		int newroot(path)
			char *path;

		long setlimit(lim)
			long lim;
		
		int pathdepth(path)
			char *path;

		key_t ftok(path, id)
			char *path;
			int id;

DESCRIPTION:
	These user-supplied routines should be modified as necessary
	before the VSX test suite is installed.

	Descriptions of the individual interfaces may be found
	with the corresponding function code below.

AUTHOR:		Peter Skan, UniSoft Ltd.
DATE CREATED:	Oct 1991
MODIFICATIONS:

	Geoff Clare, 17 Jan 1992
		Changed to use filesystem independent mount calls.

	Geoff Clare, 3 June 1992
		Use O_NOCTTY when opening master side in openpty().
		In openpty() remember original length of dummy
		slave names for use in subsequent calls.

	Geoff Clare, 20 Aug 1992
		Add ftok().

	Geoff Clare, 3 June 1993
		Applied patch from Andrew Josey in err.4.3.1.010.

************************************************************************/

#include <limits.h>
#undef _POSIX_SOURCE	/* don't want namespace restrictions in here */
#undef _POSIX_C_SOURCE
#undef _XOPEN_SOURCE

#define _ALL_SOURCE	/* change to suitable feature test macro */

#include <std.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>
#include <setprv.h>
#include <sysdep.h>

#include <priv.h>
#include <sys/mount.h>
#include <sys/fstyp.h>

private priv_t	*mapprv();
private char	*privname();

/*
 *	Mnt_rw() mounts the file system specified by "spec" onto the
 *	directory "dir" for reading and writing.
 *	It returns 0 for success, -1 for failure.
 */

public int
mnt_rw(spec, dir)
char *spec;
char *dir;
{
	int ret;

	/* file system independent mount for sfs filesystem */
	/* for other file system types change "sfs" to say "s5", or "ufs" */
	ret = mount(spec, dir, MS_FSS, sysfs(GETFSIND,"sfs"), NULL, 0);
	if (ret == -1 && errno == EBUSY)
	{
		(void) unmnt(spec, dir);
		errno = 0;
		/* file system independent mount for sfs filesystem */
		ret = mount(spec, dir, MS_FSS, sysfs(GETFSIND,"sfs"), NULL, 0);
	}

	return ret;
}

/*
 *	Mnt_ro() mounts the file system specified by "spec" onto the
 *	directory "dir" for reading only.
 *	It returns 0 for success, -1 for failure.
 */

public int
mnt_ro(spec, dir)
char *spec;
char *dir;
{
	int ret;

	/* file system independent mount for sfs filesystem */
	ret = mount(spec, dir, MS_RDONLY|MS_FSS,
		sysfs(GETFSIND, "sfs"), NULL, 0);
	if (ret == -1 && errno == EBUSY)
	{
		(void) unmnt(spec, dir);
		errno = 0;
		/* file system independent mount for sfs filesystem */
		ret = mount(spec, dir, MS_RDONLY|MS_FSS,
			sysfs(GETFSIND, "sfs"), NULL, 0);
	}

	return ret;
}

/*
 *	Unmnt() unmounts the file system specified by "spec" from the
 *	directory "dir" where it has previously been mounted.
 *	It returns 0 for success, -1 for failure.
 */

/* ARGSUSED */
public int
unmnt(spec, dir)
char *spec;
char *dir;
{
	return umount(spec);
}

/*
 *	Setprv() provides the current process with "appropriate privileges".
 *	The argument specifies what privilege is being requested.
 *	It returns 0 for success, -1 for failure.
 */

public int
setprv(prvtype)
int prvtype;
{
	priv_t *privs;
	priv_t cprivs[NPRIVS * 2];
	int i, j, nprivs, ncprivs;

	if ((privs = mapprv(prvtype, &nprivs)) == (priv_t *)0) {
		(void) fprintf(stderr,
			"Error in vport/userintf: unknown privilege request (%d) in setprv()\n",
			prvtype);
		return -1;
	}

#ifdef DEBUG
	fprintf(stderr, "setprv(%d): setting privs:", prvtype);
	for (i = 0; i < nprivs; i++)
		fprintf(stderr, " %x", privs[i]);
	fprintf(stderr, "\n");
#endif

	/*
	 * Set the privileges in the working set. (This is not strictly
	 * required since they should already be set on process startup;
	 * could just do the check below)
	 */
	if (procpriv(SETPRV, privs, nprivs) == -1) {
		(void) fprintf(stderr, "procpriv(SETPRV) failed\n");
		return -1;
	}

	/*
	 * Check that the privileges are indeed set - this is necessary
	 * because SETPRV ignores (returning success) requests to set
	 * privileges which are not contained in the current maximum set
	 */
	if ((ncprivs = procpriv(CNTPRV, (priv_t *)0, 0)) == -1) {
		(void) fprintf(stderr, "procpriv(CNTPRV) failed\n");
		return -1;
	}

	if (privs[0] == ALLPRIVS_W) {
		/*
		 * special case: after setting all privileges we should
		 * now have NPRIVS privileges in the working set and NPRIVS
		 * in the maximum set.  This avoids having to check for
		 * every one individually.
		 */
		if (ncprivs == (NPRIVS * 2))
			return 0;
		return -1;
	}

	if (procpriv(GETPRV, cprivs, ncprivs) == -1) {
		(void) fprintf(stderr, "procpriv(GETPRV) failed\n");
		return -1;
	}

	/* check each requested privilege is in the current set */
	for (i = 0; i < nprivs; i++) {
		for (j = 0; j < ncprivs; j++)
			if (cprivs[j] == privs[i])
				break;
		if (j == ncprivs) {
			(void) fprintf(stderr,
				"setprv: privilege %x not set\n", privs[i]);
			return -1;
		}
	}

	return 0;
}

/*
 *	Unsetprv() removes the specified privilege from the current process.
 *	The argument specifies what privilege is to be removed.
 *	It returns 0 for success, -1 for failure.  (If the process already
 *	does not have the privilege, this is considered success).
 */

public int
unsetprv(prvtype)
int prvtype;
{
	priv_t *privs;
	priv_t cprivs[NPRIVS * 2];
	int i, j, nprivs, ncprivs;

	if ((privs = mapprv(prvtype, &nprivs)) == (priv_t *)0) {
		(void) fprintf(stderr,
			"Error in vport/userintf: unknown privilege request (%d) in unsetprv()\n",
			prvtype);
		return -1;
	}

#ifdef DEBUG
	fprintf(stderr, "unsetprv(%d): unsetting privs:", prvtype);
	for (i = 0; i < nprivs; i++)
		fprintf(stderr, " %x", privs[i]);
	fprintf(stderr, "\n");
#endif

	/* clear the privileges */
	if (procpriv(CLRPRV, privs, nprivs) == -1) {
		(void) fprintf(stderr, "procpriv(CLRPRV) failed\n");
		return -1;
	}

	/* check that the privileges have gone away */
	if ((ncprivs = procpriv(CNTPRV, (priv_t *)0, 0)) == -1) {
		(void) fprintf(stderr, "procpriv(CNTPRV) failed\n");
		return -1;
	}

	if (privs[0] == ALLPRIVS_W) {
		/*
		 * special case: after clearing all privileges we should
		 * now have zero privileges in the working set and NPRIVS
		 * in the maximum set (assuming that the process started
		 * with all privileges)
		 */
		if (ncprivs == NPRIVS)
			return 0;
		return -1;
	}

	if (procpriv(GETPRV, cprivs, ncprivs) == -1) {
		(void) fprintf(stderr, "procpriv(GETPRV) failed\n");
		return -1;
	}

	/* check each undesired privilege for absence from the current set */
	for (i = 0; i < nprivs; i++) {
		for (j = 0; j < ncprivs; j++)
			if (cprivs[j] == privs[i]) {
				(void) fprintf(stderr,
					"unsetprv: privilege %x not cleared\n",
					privs[i]);
				return -1;
			}
	}

	return 0;
}


/*
 *	Prv_assign() assigns "appropriate privileges" to an executable file.
 *	The arguments are the file name and a zero-terminated array of the
 *	privileges to be assigned.
 *	It returns 0 for success, -1 for failure.
 */

public int
prv_assign(file, prvtypes)
char *file;
int *prvtypes;
{
	int prvtype, nprivs, i;
	priv_t *privs;
	char cmd[NPRIVS * 10] = "/sbin/filepriv -f ";
	char buf[_POSIX_PATH_MAX];
	char *s;

	/*
	 * Build command line for filepriv(1M).
	 * Note that filepriv(2) cannot be used, as it does not change
	 * the Privilege Data File and thus privileges would vanish on
	 * a reboot.
	 */
	while ((prvtype = *prvtypes++) != 0) {
		if ((privs = mapprv(prvtype, &nprivs)) == (priv_t *)0) {
			(void) fprintf(stderr,
				"Error in vport/userintf: unknown privilege request (%d) in prv_assign()\n",
				prvtype);
			return -1;
		}

		for (i = 0; i < nprivs; i++) {
			if ((s = privname(privs[i])) == NULL) {
				(void) fprintf(stderr,
					"Error in vport/userintf: prv_assign(): cannot translate privilege 0x%x into ASCII form\n",
					privs[i]);
				return -1;
			}
				
			(void) strcat(cmd, s);
			(void) strcat(cmd, ",");
		}
	}

	(void) strcat(cmd, " ");

	/* "filepriv" must be given an absolute pathname */
	if (*file != '/') {
		(void) strcat(cmd, getcwd(buf, sizeof(buf)));
		(void) strcat(cmd, "/");
	}

	(void) strcat(cmd, file);
#ifdef DEBUG
	fprintf(stderr, "prv_assign(): cmd = %s\n", cmd);
#endif

	/* Execute the filepriv command */
	if (system(cmd) != 0)
		return -1;

	return 0;
}

/*
 * Map the VSX privilege "prvtype" into the corresponding set of
 * SVR4.1ES privileges.
 * Returns a pointer to a static priv_t array on success, NULL otherwise.
 * The number of array entries is returned in "count".
 */

#define addpriv(p)	privs[i++] = (p)

private priv_t *
mapprv(prvtype, count)
int prvtype;
int *count;
{
	static priv_t privs[NPRIVS];
	int i = 0;

	switch (prvtype) {

	case PRV_DEVICE :	/* access device files */
		addpriv(DEV_W);
		break;

	case PRV_SETID :	/* change to any uid/gid */
		addpriv(SETUID_W);
		break;

	case PRV_MOUNT :	/* mount and unmount file systems */
		addpriv(MOUNT_W);
		addpriv(OWNER_W);
		break;

	case PRV_LINKDIR :	/* create links to directories */
		addpriv(FILESYS_W);
		break;

	case PRV_ACCESS :	/* unrestricted access to files */
		addpriv(DACREAD_W);
		addpriv(DACWRITE_W);
		addpriv(MACREAD_W);
		addpriv(MACWRITE_W);
		break;

	case PRV_CHOWN :	/* change owner/group/mode of ANY file */
		addpriv(OWNER_W);
		addpriv(DACREAD_W);
		addpriv(DACWRITE_W);
		addpriv(MACREAD_W);
		addpriv(MACWRITE_W);
		addpriv(COMPAT_W);
		addpriv(FILESYS_W);
		addpriv(SETPLEVEL_W);
		addpriv(SETFLEVEL_W);
		addpriv(FSYSRANGE_W);
		addpriv(MULTIDIR_W);
		break;

	case PRV_SETGRPS :	/* set supplementary group IDs */
		addpriv(SETUID_W);
		break;

	case PRV_NEWROOT :	/* set root directory of process */
		addpriv(FILESYS_W);
		break;

	case PRV_NICE :		/* change process priority */
		addpriv(TSHAR_W);
		break;

	case PRV_KILL :		/* send signal to any process */
		addpriv(OWNER_W);
		break;

	case PRV_IPC :		/* allow unrestricted IPC access */
		addpriv(OWNER_W);
		addpriv(SYSOPS_W);
		break;

	case PRV_ULIMIT :	/* increase file size limit */
		addpriv(SYSOPS_W);
		break;

	case PRV_ASSIGN :	/* assign privileges to an executable file */
		addpriv(ALLPRIVS_W);
		break;
	
	default :
		return (priv_t *)0;
	}

	*count = i;
	return (privs);
}

/*
 * Convert the SVR4.1ES privilege "priv" into its corresponding name,
 * as used by the utility filepriv(1M).
 * Return name or NULL on failure.
 */
private char *
privname(priv)
priv_t priv;
{
	switch(priv) {

	case OWNER_W:
		return("owner");

	case AUDIT_W:
		return("audit");

	case COMPAT_W:
		return("compat");

	case DACREAD_W:
		return("dacread");

	case DACWRITE_W:
		return("dacwrite");

	case DEV_W:
		return("dev");

	case FILESYS_W:
		return("filesys");

	case MACREAD_W:
		return("macread");

	case MACWRITE_W:
		return("macwrite");

	case MOUNT_W:
		return("mount");

	case MULTIDIR_W:
		return("multidir");

	case SETPLEVEL_W:
		return("setplevel");

	case SETSPRIV_W:
		return("setspriv");

	case SETUID_W:
		return("setuid");

	case SYSOPS_W:
		return("sysops");

	case SETUPRIV_W:
		return("setupriv");

	case DRIVER_W:
		return("driver");

	case RTIME_W:
		return("rtime");

	case MACUPGRADE_W:
		return("macupgrade");

	case FSYSRANGE_W:
		return("fsysrange");

	case SETFLEVEL_W:
		return("setflevel");

	case AUDITWR_W:
		return("auditwr");

	case TSHAR_W:
		return("tshar");

	case PLOCK_W:
		return("plock");

	case ALLPRIVS_W:
		return("allprivs");

	default:
		return(NULL);
	}
}


/*
 *	Setgrps() sets the current groups list to the group id's in
 *	"grparray".  This array should contain "ngrps" entries.
 *	Setgrps() is only called after setprv(PRV_SETGRPS) since setting
 *	groups is usually a privileged operation.
 *	It returns 0 for success, -1 for failure.
 *
 *	If the implementation does not support setting the supplementary
 *	groups list, then setgrps() must return -1 with errno set to ENOSYS.
 */

public int
setgrps(ngrps, grparray)
int   ngrps;
gid_t grparray[];
{
	/* If setting supplementary groups is not supported replace
	   the line below with:

		errno = ENOSYS;
		return -1;
	*/

	return setgroups(ngrps, grparray);

	/* If setgroups() takes an array of integers of a different
	   size to gid_t, the gid_t array will need to be copied.
	   In this case, remove the line above and the #if/#endif
	   around the following, and alter the typedef as necesary:
	 */

#if 0
	typedef int GID_T;
	GID_T *gcopy, gdummy;
	int i, rval;

	if (ngrps > 0)
	{
		gcopy = (GID_T *)malloc((size_t)(ngrps * sizeof(GID_T)));
		if (gcopy == NULL)
			return -1;
	}
	else
		gcopy = &gdummy;

	for (i=0; i<ngrps; i++)
		gcopy[i] = grparray[i];

	rval = setgroups(ngrps, gcopy);

	if (ngrps > 0)
		free((void *)gcopy);

	return rval;
#endif /* 0 */
}

/*
 *	Openctl() opens the terminal device "spec" with the specified flags.
 *	This then becomes the controlling terminal for the current process.
 *	If it cannot become the controlling terminal, openctl() must still
 *	open the device as an ordinary terminal.
 *	Openctl() returns the open file descriptor to the terminal,
 *	or -1 on failure.
 */
		
public int
openctl(spec, oflag)
char *spec;
int  oflag;
{
	/* Note: setsid() will have been called before this routine */

	int fd;

	/* allow for modem line delay if there has just been a close() */
	(void) sleep(1);

	fd = open(spec, oflag, 0);

	/* Make sure failure is not just because the device cannot
	   become the controlling terminal */
	
	if (fd < 0 && errno != EINTR)
		fd = open(spec, oflag|O_NOCTTY, 0);

	return fd;
}

/*
 *	Openpty() opens the master and slave sides of a pseudo-terminal.
 *	Both device names are supplied, but if master pty's are obtained
 *	from a clone device the slave name is a dummy which must be
 *	overwritten with the real device name.  If cntrl is true
 *	the slave must be opened as a controlling terminal by calling
 *	openctl() instead of plain open().  Both devices must be opened
 *	for reading and writing.  On many systems master devices can
 *	only be opened once, giving EBUSY for example on subsequent
 *	opens.  If openpty() encounters a busy master device, it must
 *	simply open the slave.
 *
 *	The open file descriptors are returned via the pointers mfdp and
 *	sfdp.  Openpty() returns 0 for success or -1 for failure.
 *
 *	This routine is only called if the VSX_MASTER_TTY and
 *	VSX_MASTER_LOOP parameters are set.
 */

#define CLONE_MPTY	/* uncomment if master pty's are obtained from a
		 	   clone device (e.g. /dev/ptmx) */
#define STREAMS_PTY	/* uncomment if pty's use STREAMS */

#ifdef STREAMS_PTY
#include <sys/stropts.h>
#endif

public int
openpty(master, slave, mfdp, sfdp, cntrl)
char *master;
char *slave;
int *mfdp;
int *sfdp;
int cntrl;
{
	*mfdp = open(master, O_RDWR|O_NOCTTY);
	if (*mfdp < 0 && errno != EBUSY)
		return -1;

#ifdef CLONE_MPTY
	if (*mfdp >= 0)
	{
		char *newslave;
		size_t len;
		static struct { char *ptr; size_t len; } slavelen[2];
		extern char *ptsname();

		(void) grantpt(*mfdp);   /* change permission of slave */
		(void) unlockpt(*mfdp);  /* unlock slave */
		newslave=ptsname(*mfdp); /* get a slave */
		if (slavelen[0].ptr == NULL)
		{
			/* remember length of original dummy name
			   to be used if called again */
			slavelen[0].ptr = slave;
			slavelen[0].len = strlen(slave);
		}
		else if (slave != slavelen[0].ptr && slavelen[1].ptr == NULL)
		{
			/* remember length of original dummy name
			   to be used if called again */
			slavelen[1].ptr = slave;
			slavelen[1].len = strlen(slave);
		}
		if (slave == slavelen[0].ptr)
			len = slavelen[0].len;
		else
			len = slavelen[1].len;
		if (strlen(newslave) > len)
		{
			(void) fprintf(stderr,
				"Error in vport/userintf: dummy slave name too short to take \"%s\" in openpty()\n",
				newslave);
			return -1;
		}
		(void) strcpy(slave, newslave);
	}
#endif

	if (cntrl)
		*sfdp = openctl(slave, O_RDWR);
	else
		*sfdp = open(slave, O_RDWR|O_NOCTTY);

	if (*sfdp < 0)
		return -1;

#ifdef STREAMS_PTY
	if (*mfdp >= 0)
	{
		(void) ioctl(*sfdp, I_PUSH, "ptem");   /* pseudo tty emulator */
		(void) ioctl(*sfdp, I_PUSH, "ldterm"); /* line discipline */
	}
#endif
	
	return 0;
}

/*
 *	Ptygetattr() obtains terminal attributes for the slave pseudo-
 *	terminal corresponding to the master device addressed by mfd.
 *	It returns 0 for success, -1 for failure.
 */

public int
ptygetattr(mfd, termios_p)
int mfd;
struct termios *termios_p;
{
#ifdef STREAMS_PTY
	extern char *ptsname();
	char *slave;
	int ret, sfd = -1;
	
	/* Obtain slave device name and call tcgetattr() on the slave */
	slave = ptsname(mfd);
	if (slave && (sfd = open(slave, O_RDWR)) >= 0)
		ret = tcgetattr(sfd, termios_p);
	else
		ret = -1;
	if (sfd >= 0)
		(void) close(sfd);
	return ret;
#else
	/* This code assumes tcgetattr() on the master returns the
	   settings of the slave (true for BSD-style pseudo-terminals) */
	return tcgetattr(mfd, termios_p);
#endif
}

/*
 *	Sys_call() makes one or more calls to an interface which is
 *	known to be a system call.  I.e. after several calls the system
 *	CPU time for the process will have increased measurably.
 *	It returns 0 for success, -1 for failure.
 */

public int
sys_call()
{
	return access("/", F_OK);
}

/*
 *	Use_time() performs processing which will accumulate
 *	user CPU time.  The number of iterations may be	adjusted
 *	if necessary.
 */

public void
use_time()
{
	int i, dummy = 0;

	for (i = 0; i < 100; i++)
		dummy = dummy + i / (dummy + 1) * (dummy % 3);
}

/*
 *	Newroot() causes "path" to become the root directory for the
 *	current process.
 *	Newroot() is only called after setprv(PRV_NEWROOT) since setting
 *	the root directory is usually a privileged operation.
 *	It returns 0 for success, -1 for failure.
 */

public int
newroot(path)
char *path;
{
	return chroot(path);
}

/*
 *	Setlimit() is used to reduce the output file size limit for 
 *	the process.  The argument is the value to be set in units of
 *	512 byte blocks, and will always be >= VSX_ULIMIT_BLKS.
 *	Setlimit() returns the new limit, or -1L for failure.
 */

#ifndef MAKE_DEPEND	/* don't want dependency in Makefile */
#include <ulimit.h>
#endif

public long
setlimit(lim)
long lim;
{
	return ulimit(UL_SETFSIZE, lim);
}

/*
 *	Pathdepth() is used to determine the depth of pathnames
 *	beginning with "//" below the root directory of the process.
 *	The return value is such that a path equivalent to "/usr"
 *	gives a value of 1, one equivalent to "/usr/bin" gives 2, etc.
 *	The default code supplied is suitable for systems which use
 *	"//" to refer to a directory level above the root, i.e. when
 *	the root directory is referred to by "//somename".
 */

public int
pathdepth(path)
char *path;
{
	char *p;
	int dircnt = 0;

	/* count '/' characters after the initial "//" */
	for (p = &path[2]; *p != '\0'; p++)
	{
		if (*p == '/')
		{
			++dircnt;
			/* skip multiple '/'s */
			while (*(p+1) == '/')
				++p;
			/* watch out for trailing '/' */
			if (*(p+1) == '\0')
				--dircnt;
		}
	
	}

	return dircnt;
}

/*
 *	Ftok returns a key based on the arguments path and id, for use
 *	in msgget(), semget() and shmget() calls.
 *	Ftok() is normally supplied with IPC implementations.  Only
 *	uncomment the #define below if you do not already have ftok()
 *	on your system, or if your ftok() has different arguments or
 *	return type.  The example version assumes key_t is a 32 bit
 *	integral type.
 *
 *	This function is only used in XPG4 mode.
 */

/* #define NO_FTOK */

#ifdef NO_FTOK
public key_t
ftok(path, id)
char *path;
int id;
{
        struct stat stb;

	if (stat(path, &stb) == -1)
		return (key_t)-1;

	return (key_t)((key_t)id << 24 |
		       (key_t)stb.st_dev << 16 | (key_t)stb.st_ino);
}
#endif
