/*
* Copyright (c) 1997 X/Open Company Ltd., A member of The Open Group
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* Motif, OSF/1 and UNIX are registered trademarks and X/Open,
* the "X Device" and The Open Group are trademarks of The Open Group.
*
* Project: LSB-OS
*
* File: SRC/common/vsrtlib/aio.c
*
* Modifications:
* $Log: aio.c,v $
* Revision 1.2  2001/10/02 11:59:05  ajosey
* clean up the various header comment blocks for lsb-os
*
* Revision 1.1  2001/09/28 08:21:02  cyeoh
* Move of lsb-os specific libraries to lsb-os test set
*
* Revision 1.1  2001/06/01 02:00:15  cyeoh
* Merges for lsb-os test suite
*
*
*/

#include <vsrt/vsrt.h>
#include <aio.h>
#include <termios.h>

#define VSRT_NBAUDS 15

static speed_t bauds[VSRT_NBAUDS] = {
	B38400,
	B19200,
	B9600,
	B4800,
	B2400,
	B1800,
	B1200,
	B600,
	B300,
	B200,
	B150,
	B134,
	B110,
	B75,
	B50
};

static int term_settings();

/*determine if a type of file is to be tried. set in startup*/
/*pipe support is determined dynamically*/
/*set by test code*/
extern int	vsrt_aio_test_with_tty;
extern int	vsrt_aio_test_with_file;
extern int	vsrt_aio_test_with_pty;

/*each piece of a test increments on of these.  must be zeroed at test start*/
extern int	vsrt_aio_test_pass;
extern int	vsrt_aio_test_unsup;
extern int	vsrt_aio_test_fail;

/*instances of things to test on from tetexec, loaded in startup*/
extern char	vsrt_aio_tty1[];
extern char	vsrt_aio_tty2[];
extern char	vsrt_aio_ptym[];
extern char	vsrt_aio_testfile[];

/* aio tests are generally run multiple times, using some combination
of files, pipes, ttys, and ptys

A test purpose will calls one of these vsrt_test routines to run a
given test function with the appropriate files and devices.

These routines call vsrt_forkatest to actually run the function in
forked process 
*/

void vsrt_aio_test_allbutfile(func)
int (*func)(int, int);
{

	vsrt_aio_test_pipe(func);
	vsrt_aio_test_pty(func);
	vsrt_aio_test_tty(func);
}

void vsrt_aio_test_all(func)
int (*func)(int, int);
{

	vsrt_aio_test_file(func);
	vsrt_aio_test_pipe(func);
	vsrt_aio_test_pty(func);
	vsrt_aio_test_tty(func);
}

void vsrt_aio_test_allbutfile_write(func)
int (*func)(int, int);
{

	vsrt_aio_test_pipe(func);
	vsrt_aio_test_pty_write(func);
	vsrt_aio_test_tty(func);
}

void vsrt_aio_test_all_write(func)
int (*func)(int, int);
{

	vsrt_aio_test_file(func);
	vsrt_aio_test_pipe(func);
	vsrt_aio_test_pty_write(func);
	vsrt_aio_test_tty(func);
}

void vsrt_aio_test_pipe(func)
int (*func)(int, int);
{
	int	retval;
	int	fildes[2];

	DBUG_ENTER("vsrt_test_pipe");

	in_rpt("> test with pipe");

	if (pipe(fildes) == -1) {
		vsrt_sysfail("pipe");
		vsrt_aio_test_fail++;
		DBUG_VOID_RETURN
	}

	/*see if AIO is supported on pipes*/
	if (fpathconf(fildes[0], _PC_ASYNC_IO) == -1) {
		vsrt_aio_test_unsup++;
		in_rpt("testing not supported on pipes");
		(void)close(fildes[0]);
		(void)close(fildes[1]);
		DBUG_VOID_RETURN
	}

	retval = vsrt_forkatest(func, fildes[0], fildes[1]);

	if (retval == -1)
		vsrt_aio_test_fail++;
	if (retval == 1)
		vsrt_aio_test_unsup++;
	if (retval == 0)
		vsrt_aio_test_pass++;

	(void)close(fildes[0]);
	(void)close(fildes[1]);

	DBUG_VOID_RETURN
}

void vsrt_aio_test_file(func)
int (*func)(int, int);
{
	int	retval;
	int	fd;

	DBUG_ENTER("vsrt_test_file");

	in_rpt("> test with file");

	if (vsrt_aio_test_with_file == 0) {
		vsrt_aio_test_unsup++;
		in_rpt("file testing not configured");
		DBUG_VOID_RETURN
	}

	(void)unlink(vsrt_aio_testfile);
	if ((fd = open(vsrt_aio_testfile, O_RDWR|O_CREAT, VSRT_PROT_ALL)) == -1) {
		vsrt_sysfail("open");
		vsrt_aio_test_fail++;
		DBUG_VOID_RETURN
	}

	retval = vsrt_forkatest(func, fd, fd);
	if (retval == -1)
		vsrt_aio_test_fail++;
	if (retval == 1)
		vsrt_aio_test_unsup++;
	if (retval == 0)
		vsrt_aio_test_pass++;

	(void)close(fd);
	(void)unlink(vsrt_aio_testfile);

	DBUG_VOID_RETURN
}

/*we need to flip the side the slave is on for write tests*/
void vsrt_aio_test_pty_write(func)
int (*func)(int, int);
{
	int	retval;
	int	readfd, writefd;

	DBUG_ENTER("vsrt_test_pty");

	in_rpt("> test with pty");

	if (vsrt_aio_test_with_pty == 0) {
		vsrt_aio_test_unsup++;
		in_rpt("pty testing not configured");
		DBUG_VOID_RETURN
	}

	retval = vsrt_init_ptys(vsrt_aio_ptym, &writefd, &readfd);
	if (retval == -1) {
		vsrt_aio_test_fail++;
		DBUG_VOID_RETURN
	}
	if (retval == 1) {
		vsrt_aio_test_unsup++;
		DBUG_VOID_RETURN
	}

	retval = vsrt_forkatest(func, readfd, writefd);
	if (retval == -1)
		vsrt_aio_test_fail++;
	if (retval == 1)
		vsrt_aio_test_unsup++;
	if (retval == 0)
		vsrt_aio_test_pass++;

	tcflush(writefd, TCIOFLUSH);

	(void)close(readfd);
	(void)close(writefd);

	DBUG_VOID_RETURN
}

void vsrt_aio_test_pty(func)
int (*func)(int, int);
{
	int	retval;
	int	readfd, writefd;

	DBUG_ENTER("vsrt_test_pty");

	in_rpt("> test with pty");

	if (vsrt_aio_test_with_pty == 0) {
		vsrt_aio_test_unsup++;
		in_rpt("pty testing not configured");
		DBUG_VOID_RETURN
	}

	retval = vsrt_init_ptys(vsrt_aio_ptym, &readfd, &writefd);
	if (retval == -1) {
		vsrt_aio_test_fail++;
		DBUG_VOID_RETURN
	}
	if (retval == 1) {
		vsrt_aio_test_unsup++;
		DBUG_VOID_RETURN
	}

	retval = vsrt_forkatest(func, readfd, writefd);
	if (retval == -1)
		vsrt_aio_test_fail++;
	if (retval == 1)
		vsrt_aio_test_unsup++;
	if (retval == 0)
		vsrt_aio_test_pass++;

	tcflush(readfd, TCIOFLUSH);

	(void)close(readfd);
	(void)close(writefd);

	DBUG_VOID_RETURN
}

void vsrt_aio_test_tty(func)
int (*func)(int, int);
{
	int	retval;
	int	readfd, writefd;

	DBUG_ENTER("vsrt_test_tty");

	in_rpt("> test with tty");

	if (vsrt_aio_test_with_tty == 0) {
		vsrt_aio_test_unsup++;
		in_rpt("tty testing not configured");
		DBUG_VOID_RETURN
	}

	retval = vsrt_init_ttys(vsrt_aio_tty1, vsrt_aio_tty2, &readfd, &writefd);
	if (retval == -1) {
		vsrt_aio_test_fail++;
		DBUG_VOID_RETURN
	}
	if (retval == 1) {
		vsrt_aio_test_unsup++;
		DBUG_VOID_RETURN
	}

	retval = vsrt_forkatest(func, readfd, writefd);
	if (retval == -1)
		vsrt_aio_test_fail++;
	if (retval == 1)
		vsrt_aio_test_unsup++;
	if (retval == 0)
		vsrt_aio_test_pass++;

	tcflush(readfd, TCIOFLUSH);
	tcflush(writefd, TCIOFLUSH);

	(void)close(readfd);
	(void)close(writefd);

	DBUG_VOID_RETURN
}


/*initialize ptys for use in aio testing*/
int vsrt_init_ptys(ptym, read_fd, write_fd) 
char	*ptym;
int	*read_fd;
int	*write_fd;
{
	char	*snm;

	DBUG_ENTER("vsrt_init_ptys");

#ifdef _XOPEN_SOURCE
	*write_fd = open(ptym, O_RDWR|O_NOCTTY);
	if (*write_fd == -1) {
		vsrt_sysfail("open of master pty");
		DBUG_RETURN(-1);
	}

	if (grantpt(*write_fd) == -1) {
		vsrt_sysfail("grantpt");
		(void)close(*write_fd);
		DBUG_RETURN(-1);
	}

	if (unlockpt(*write_fd) == -1) {
		vsrt_sysfail("unlockpt");
		(void)close(*write_fd);
		DBUG_RETURN(-1);
	}

	snm = (char *)ptsname(*write_fd);
	*read_fd = open(snm , O_RDWR);
	if (*read_fd == -1) {
		vsrt_sysfail("open of slave pty");
		(void)close(*write_fd);
		DBUG_RETURN(-1);
	}

	if (term_settings(*read_fd) == -1) {
		(void)close(*write_fd);
		(void)close(*read_fd);
		DBUG_RETURN(-1);
	}

	DBUG_RETURN(0);

#else
	DBUG_RETURN(1);
#endif

}

/*initialize ttys for use in aio testing*/
int vsrt_init_ttys(termiostty, termiosloop, read_fd, write_fd) 
char	*termiostty;
char	*termiosloop;
int	*read_fd;
int	*write_fd;
{
	int	retry;

	DBUG_ENTER("vsrt_init_ttys");

	*write_fd = open(termiostty,  O_RDWR|O_NOCTTY);
	if (*write_fd == -1) {
		vsrt_sysfail("open of tty");
		DBUG_RETURN(-1);
	}
	/*some systems return EAGAIN or EBUSY for a while - dunno why*/
	retry = 0;
	while ((*write_fd == -1) && ((errno == EAGAIN) || (errno == EBUSY)) && (retry < 5)) {
		sleep(2);
		*write_fd = open(termiostty,  O_RDWR|O_NOCTTY);
		retry++;
	}

	if (*write_fd == -1) {
		vsrt_sysfail("open of loop tty");
		(void)close(*write_fd);
		DBUG_RETURN(-1);
	}

	*read_fd = open (termiosloop, O_RDWR|O_NOCTTY);
	/*some systems return EAGAIN or EBUSY for a while - dunno why*/
	retry = 0;
	while ((*read_fd == -1) && ((errno == EAGAIN) || (errno == EBUSY)) && (retry < 5)) {
		sleep(2);
		*read_fd = open (termiosloop, O_RDWR|O_NOCTTY,0);
		retry++;
	}

	if (*read_fd == -1) {
		vsrt_sysfail("open of loop tty");
		(void)close(*write_fd);
		DBUG_RETURN(-1);
	}

	if ((term_settings(*read_fd) == -1) || (term_settings(*write_fd)) == -1) {
		(void)close(*read_fd);
		(void)close(*write_fd);
		DBUG_RETURN(-1);
	}
	
	DBUG_RETURN(0);

}

static int term_settings(fildes)
int	fildes;
{
	int	i;
	static struct termios termios;
	struct termios *termios_p;

	DBUG_ENTER("term_settings");

	if (tcgetattr(fildes, &termios) == -1) {
		vsrt_sysfail("tcgetattr");
		DBUG_RETURN(-1);
	}

	termios_p = &termios;

	/*set to testing values*/
	(termios_p)->c_iflag |=  (IGNPAR|IGNBRK|ICRNL|IXON|IXOFF);
	(termios_p)->c_oflag &=  ~OPOST;
	(termios_p)->c_cflag |=  (CLOCAL|CREAD);
	(termios_p)->c_cflag &=  ~CSIZE;
	(termios_p)->c_cflag |=  CS8;
	(termios_p)->c_lflag |=  (ISIG|ICANON);

	(termios_p)->c_cc[VMIN]	 = 0;
	(termios_p)->c_cc[VTIME]	= 0;

	for (i = 0; i < VSRT_NBAUDS; i++)
		if (cfsetispeed(termios_p, bauds[i]) == 0)
			break;

	if (i == VSRT_NBAUDS) {
		in_rpt("cfsetispeed failed for all possible baud rate values");
		xx_rpt(FAILURE);
	}

	for (i = 0; i < VSRT_NBAUDS; i++)
		if (cfsetospeed(termios_p, bauds[i]) == 0)
			break;

	if (i == VSRT_NBAUDS) {
		in_rpt("cfsetospeed failed for all possible baud rate values");
		xx_rpt(FAILURE);
		DBUG_RETURN(-1);
	}

	if (tcsetattr(fildes, TCSANOW, &termios) == -1) {
		vsrt_sysfail("tcsetattr");
		DBUG_RETURN(-1);
	}

	DBUG_RETURN(0);
}

/*run a test in a forked process*/
int vsrt_forkatest(func, readfd, writefd)
int	(*func)(int, int);
int readfd;
int writefd;
{
	pid_t	pid;
	int	status;
	int	retval;

	DBUG_ENTER("vsrt_forkatest");

	VSRT_FORK(pid);
		if(setpgid(getpid(), getpid())) {
			vsrt_sysfail("setpgid");
			exit(-1);
		}
		retval = func(readfd, writefd);
		exit(retval);
	}
	tet_setblock();
	status = vsrt_wait_for(pid, VSRT_AIO_TO);
	if (status == -1)
		DBUG_RETURN(-1);
	else if (!WIFEXITED(status)) {
		if (WIFSIGNALED(status) != 0)
			in_rpt("Child process exited due to signal %d",  WTERMSIG(status));
		else
			in_rpt("Child process exited abnormally with exit status %d",  WEXITSTATUS(status));
		xx_rpt(FAILURE);
		DBUG_RETURN(-1);
	}

	if (WEXITSTATUS(status) == 1)
		DBUG_RETURN(1);
	if (WEXITSTATUS(status) != 0)
		DBUG_RETURN(-1);
	else
		DBUG_RETURN(0);
}
