/*
* Copyright (c) 1997 X/Open Company Ltd., A member of The Open Group
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* Motif, OSF/1 and UNIX are registered trademarks and X/Open,
* the "X Device" and The Open Group are trademarks of The Open Group.
*
* Project: LSB-OS
*
* File: SRC/common/vsrtlib/event.c
*
* Modifications:
* $Log: event.c,v $
* Revision 1.2  2001/10/02 11:59:05  ajosey
* clean up the various header comment blocks for lsb-os
*
* Revision 1.1  2001/09/28 08:21:02  cyeoh
* Move of lsb-os specific libraries to lsb-os test set
*
* Revision 1.1  2001/06/01 02:00:15  cyeoh
* Merges for lsb-os test suite
*
*
*/

#include <vsrt/vsrt.h>
#include <sys/ipc.h>
#include <sys/sem.h>

extern     char    *errname(int);

#define	NUM_EVENTS	8

static	int		semid = -1;	/* semaphore key*/

#define	SEM_PERM	(S_IRWXU | S_IRWXG | S_IRWXO)


/* Allocates Semaphores*/
int vsrt_alloc_sem(void)
{
	if (semid != -1)
		return(0);

	semid = semget(IPC_PRIVATE, NUM_EVENTS, IPC_CREAT | SEM_PERM);
	if (semid == -1) {
		in_rpt("vsrt_alloc_sem: semget failed, errno = %s", errname(errno));
		xx_rpt(FAILURE);
		return(-1);
	}

	return(0);
}


void vsrt_free_sem(void)
{
	if (semid != -1) {
		(void) semctl(semid, 0, IPC_RMID);
		semid = -1;
	}
}

/*
**	called with an event number in "event" and a time to wait 't'
**
**	if t is positive an alarm is set for time t to interrupt the semaphore
**	call which will block on the event being posted
**
**	if t is 0 we return immediately either 0 if the event
**	was there or -2 if it is not
**
**	if t is negative we'll block on the event forever
**
*/

int vsrt_wait_event(int event, int t)
{
	struct sembuf mysembuf;
	int save_errno, rval, rval2;

	if (semid == -1) {
		in_rpt("vsrt_wait_event: Semaphores used before initialization");
		xx_rpt(FAILURE);
		return(-1);
	}

	if (event >= NUM_EVENTS) {
		in_rpt("vsrt_wait_event: event was: %d, expected less than %d", event, NUM_EVENTS);
		xx_rpt(FAILURE);
		return(-1);
	}

	mysembuf.sem_num = event;
	/*-1 means wait for it to go to 1 and then set it to 0*/
	mysembuf.sem_op = -1;
	/*block on it unless t said not to*/
	if (t != 0)
		mysembuf.sem_flg = 0;
	else
		mysembuf.sem_flg = IPC_NOWAIT;

	/*limit length of blocking if requested*/
	if (t > 0) {
		got_sigalrm = 0;
		sigemptyset(&(Sigaction.sa_mask));
		Sigaction.sa_handler = vsrt_lsah;
		Sigaction.sa_flags = 0;
		rval = sigaction(SIGALRM,&Sigaction,(struct sigaction *)NULL);
		if (rval == -1) {
			in_rpt("vsrt_wait_event: sigaction failed, errno = %s (%s), errno, errname(errno)");
			xx_rpt(FAILURE);
			return -1;
		}
		(void) alarm((unsigned int)t);
	}

	rval = semop(semid, &mysembuf, 1);
	while ((rval == -1) && (errno == EINTR))
		rval = semop(semid, &mysembuf, 1);

	/*protect errno from calls to stop alarm*/
	save_errno = errno;

	/*stop alarm*/
	if (t > 0) {
		unused_time = alarm(0);
		sigemptyset(&(Sigaction.sa_mask));
		Sigaction.sa_handler = SIG_DFL;
		Sigaction.sa_flags = 0;
		rval2 = sigaction(SIGALRM,&Sigaction,(struct sigaction *)NULL);
		if (rval2 == -1) {
			in_rpt("vsrt_wait_event: sigaction failed, errno = %s (%s), errno, errname(errno)");
			xx_rpt(FAILURE);
			return -1;
		}
		if (got_sigalrm != 0) {
			in_rpt("vsrt_wait_event: Timed out waiting for completion.");
			xx_rpt(FAILURE);
			return -1;
		}
	}

	/*any fail is an error if we requested blocking*/
	if (mysembuf.sem_flg != IPC_NOWAIT) {
		if (rval == -1) {
			in_rpt("vsrt_wait_event: semop failed, errno = %s", errname(save_errno));
			xx_rpt(FAILURE);
			return(-1);
		}
	}

	/*if we did not request blocking EAGAIN means we would have blocked*/
	/*treat this special*/
	else {
		if (rval == -1) {
			if (save_errno == EAGAIN) {
				return(-2);
			}
			else {
				in_rpt("vsrt_wait_event: semop failed, errno = %s", errname(save_errno));
				xx_rpt(FAILURE);
				return(-1);
			}
		}
	}
	return 0;
}

int vsrt_post_event(int event)
{
	struct sembuf mysembuf;
	int	rval;

	if (semid == -1) {
		tet_infoline("vsrt_post_event: Semaphores used before initialization");
		xx_rpt(FAILURE);
		return(-1);
	}
	if (event >= NUM_EVENTS) {
		in_rpt("vsrt_post_event: event was: %d, expected less than %d", event, NUM_EVENTS);
		xx_rpt(FAILURE);
		return(-1);
	}

	mysembuf.sem_num = event;
	/*positive means we're setting it, to 1*/
	mysembuf.sem_op = 1;
	mysembuf.sem_flg = 0;

	rval = semop(semid, &mysembuf, 1);
	if (rval == -1) {
		in_rpt("vsrt_post_event: semop failed, errno = %s", errname(errno));
		xx_rpt(FAILURE);
		return(-1);
	}
	return 0;
}

int vsrt_get_event(int event)
{
	int rval;

	if (semid == -1) {
		tet_infoline("vsrt_get_event: semaphores used before initialization");
		xx_rpt(FAILURE);
		return(-1);
	}
	if (event >= NUM_EVENTS) {
		in_rpt("vsrt_get_event: event was: %d, expected less than %d", event, NUM_EVENTS);
		return(-1);
	}

	/*read it out*/
	rval = semctl(semid, event, GETVAL);
	if (rval == -1) {
		in_rpt("vsrt_get_event: semctl failed, errno = %s", errname(errno));
		xx_rpt(FAILURE);
		return(-1);
	}
	return (rval);
}
