/*
* Copyright (c) 1997 X/Open Company Ltd., A member of The Open Group
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* Motif, OSF/1 and UNIX are registered trademarks and X/Open,
* the "X Device" and The Open Group are trademarks of The Open Group.
*
* Project: LSB-OS
*
* File: SRC/common/vsrtlib/mkmmap.c
*
* Modifications:
* $Log: mkmmap.c,v $
* Revision 1.2  2001/10/02 11:59:05  ajosey
* clean up the various header comment blocks for lsb-os
*
* Revision 1.1  2001/09/28 08:21:02  cyeoh
* Move of lsb-os specific libraries to lsb-os test set
*
* Revision 1.1  2001/06/01 02:00:15  cyeoh
* Merges for lsb-os test suite
*
*
*/

/*
This file contains support code to tests for  mmap, etc.

It contains the following functions:
	vsrt_create_exec_file
	vsrt_invalid_sig_fn
	vsrt_map3
	vsrt_mmapsig_fn
	vsrt_mmap_setup
	vsrt_mprotect_setup
	vsrt_open_file
*/

#include <vsrt/vsrt.h>
#include <setjmp.h>
#include <sys/mman.h>

int	vsrt_got_sigsegv = 0;
int	vsrt_got_sigbus = 0;
int	vsrt_got_sig = 0;
int	vsrt_setjmp_called = 0;
unsigned long	vsrt_pgsz;
unsigned long	vsrt_ipgsz;
sigjmp_buf	vsrt_env;

/*
** Signal lists
*/
typedef struct {
	char	name[10];
	int	signal;
	int	flags;
} vsrt_siglst;

static vsrt_siglst	vsrt_signals[] = {
	{ "SIGFPE",	SIGFPE,		0 },
	{ "SIGILL",	SIGILL,		0 },
};
#define	VSRT_NUM_SIGNALS	(sizeof(vsrt_signals) / sizeof(vsrt_siglst))

extern  char    *errname();

/*
** Function: void vsrt_invalid_sig_fn(int s)
**
** Description:
**	This function checks for any invalid signal that was
**	generated. The argument int is compared against the 
**	list of all the signals for the match.
**
** Inputs:
**	int s	->  signal number
**	
** Returns:
**	Nothing at all...
**	
*/

static void vsrt_invalid_sig_fn();
static void vsrt_invalid_sig_fn(s)
int s;
{
	int i;

	for (i = 0; i < VSRT_NUM_SIGNALS; i++) {
		if (s == vsrt_signals[i].signal) {
			in_rpt("Invalid signal: %d(%s) received", s, vsrt_signals[i].name);
			xx_rpt(FAILURE);
			return;
		}
	}
}


/*
** Function: void vsrt_mmapsig_fn(int s)
**
** Description:
**
**	This function updates the global counters for the SIGBUS and
**	the SIGSEGV signals. For all other signals the vsrt_got_sig variable
**	is incremented.
**
** Inputs:
**	int s	->  signal number
**	
** Returns:
**	Nothing at all...
**	
*/

static void vsrt_mmapsig_fn();
static void vsrt_mmapsig_fn(s)
int	s;
{
	if (s == SIGSEGV)
		vsrt_got_sigsegv++;
	else if (s == SIGBUS)
		vsrt_got_sigbus++;
	vsrt_got_sig++;
	if (vsrt_setjmp_called) {
		vsrt_setjmp_called = 0;
		siglongjmp(vsrt_env, 1);
	}
}



/*
** Function: int vsrt_mmap_setup()
**
** Description:
**
**	This function forms the setup() routine for all the
**	mmap() tests. The function essentially creates a
**	text file, updates the file and then maps it in 
**	memory.
**
** Inputs:
**	None
**	
** Returns:
**	int:
**		0 - Success
**		-1- Failure
**	
*/

int vsrt_mmap_setup(void) {

	int i;

	if ((vsrt_pgsz = sysconf(_SC_PAGESIZE)) == -1) {
		in_rpt("sysconf() failed, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return -1;
	}

	vsrt_ipgsz = vsrt_pgsz/sizeof(int);

	vsrt_got_sigsegv = vsrt_got_sigbus = vsrt_got_sig = 0;

	for (i = 0; i < VSRT_NUM_SIGNALS; i++) {
		if (vsrt_signal(vsrt_signals[i].signal, &vsrt_invalid_sig_fn) == -1) {
			return -1;
		}
	}

	if (vsrt_signal(SIGSEGV, vsrt_mmapsig_fn) == -1)
		return -1;
	if (vsrt_signal(SIGBUS, vsrt_mmapsig_fn) == -1)
		return -1;

	return 0;
}


/*
** Function: int vsrt_open_file(char *s, int flag, mode_t mode, npages)
**
** Description:
**
**	This function creates and opens a text file and updates
**	it with data. The file descriptor is then returned to
**	the caller. This file descriptor is used for mapping.
**
** Inputs:
**	char *s	->  File to be opened
**	int flag	->  Flags to be passed for open
**	mode_t mode	->  Permission modes for open
**	int npages	->  Number of pages of data to write to the file
**	
** Returns:
**	int	-> File descriptor
**	
*/

int vsrt_open_file(s, flag, mode, npages)
char	*s;
int	flag;
mode_t	mode;
int	npages;
{
	int		fd;
	int		*buf, i, j;


	unlink(s);
	if ((fd = open(s, (O_RDWR|O_CREAT|O_TRUNC), (mode_t)(S_IRWXU|S_IRWXG|S_IRWXO))) == -1) {
		in_rpt("open() failed, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return -1;
	}

	if ((buf = (int *)malloc(sizeof(int) * vsrt_ipgsz * npages)) == NULL) {
		in_rpt("malloc() failed, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return -1;
	}

	for (j = 0; j < vsrt_ipgsz*npages; j += sizeof(buf)/sizeof(buf[0])) {
		for (i = 0; i < sizeof(buf)/sizeof(buf[0]); i++)
			buf[i] = i + j;

		if (write(fd, (void*)buf, sizeof(buf)) != sizeof(buf)) {
			in_rpt("write() failed, errno = %d (%s)", errno, errname(errno));
			xx_rpt(FAILURE);
			return -1;
		}
	}

	(void)close(fd);

	if (chmod(s, mode) == -1) {
		in_rpt("chmod() failed, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return -1;
	}

	if ((fd = open(s, flag)) == -1) {
		in_rpt("open() failed, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return -1;
	}

	return (fd);
}

/*
** Function: int vsrt_open_shm_np(char *s, int flag, mode_t mode, int npages)
**
** Description:
**
**	This function creates and opens a shared memory segment and updates
**	it with data. The file descriptor is then returned to
**	the caller. This file descriptor is used for mapping.
**
** Inputs:
**	char *s	->  File to be opened
**	int flag	->  Flags to be passed for open
**	mode_t mode	->  Permission modes for open
**	int npages	->  Number of pages of data to write to the segment
**	
** Returns:
**	int	-> File descriptor
**	
*/

int vsrt_open_shm_np(s, flag, mode, npages)
char	*s;
int	flag;
mode_t	mode;
int	npages;
{
	int	fd;
	int	*buf, j;
	int	err;

	(void)shm_unlink(s);
	if ((fd = shm_open(s, (O_RDWR|O_CREAT|O_TRUNC), (mode_t)(S_IRWXU|S_IRWXG|S_IRWXO))) == -1) {
		in_rpt("vsrt_open_shm_np: shm_open() failed for \"%s\", errno = %d (%s)", s, errno, errname(errno));
		xx_rpt(FAILURE);
		return -1;
	}

	if (ftruncate(fd, vsrt_pgsz*npages) == -1) {
		in_rpt("vsrt_open_shm_np: ftruncate() failed for \"%s\", errno = %d (%s)", s, errno, errname(errno));
		xx_rpt(FAILURE);
		return -1;
	}

	if ((buf = (int *)mmap(0, vsrt_pgsz*npages, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0)) == (int *)MAP_FAILED) {
		err = errno;
		in_rpt("vsrt_open_shm_np: mmap failed, errno = %d (%s)", err, errname(err));
		xx_rpt(FAILURE);
		return -1;
	}

	for (j = 0; j < npages*vsrt_ipgsz; j++)
			buf[j] = j;

	(void)msync((void *)buf, vsrt_pgsz, MS_SYNC);

	if (fchmod(fd, mode) == -1) {
		in_rpt("vsrt_open_shm_np: fchmod() failed, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return -1;
	}

	if (munmap((void *)buf, vsrt_pgsz) == -1) {
		err = errno;
		in_rpt("vsrt_open_shm_np: munmap failed, errno = %d (%s)", err, errname(err));
		xx_rpt(FAILURE);
		return -1;
	}

	(void) close(fd);

	if ((fd = shm_open(s, flag, mode)) == -1) {
		in_rpt("vsrt_open_shm_np: second shm_open() failed for \"%s\", s, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return -1;
	}

	return (fd);
}

/*
** Function: int vsrt_open_shm(char *s, int flag, mode_t mode)
**
** Description:
**
**	Same as vsrt_open_shm_np(), but with a fixed number of pages (2).
**
** Inputs:
**	char *s	->  File to be opened
**	int flag	->  Flags to be passed for open
**	mode_t mode	->  Permission modes for open
**	
** Returns:
**	int	-> File descriptor
**	
*/

int vsrt_open_shm(s, flag, mode)
char	*s;
int	flag;
mode_t	mode;
{
	return vsrt_open_shm_np(s, flag, mode, 2);
}
	
int vsrt_fn_for_map(void)
{
	return (99);
}

/*
** Function: int vsrt_create_exec_shm(char *s, int flag, mode_t mode, struct vsrt_exec_resource *inp)
**
** Description:
**
**	This function creates an executable shared memory object. This
**	object contains the binary image of a routine. When the object
**	is mapped and the execution control passed to the mapped
**	location the routine executes. The file descriptor to this
**	file is returned by the function in the exec_resource 
**	struct.
**
** Inputs:
**	char *s		->  Object to be opened
**	int flag	->  Flags for file open
**	mode_t mode	->  Permission modes
**	struct exec_resource -> Opened file descriptor
**	
** Returns:
**	int	-> 0 - Success
**		   -1 - Fail
**	
*/

int vsrt_create_exec_shm(s, flag, mode, inp)
char *s;
int flag;
mode_t mode;
struct vsrt_exec_resource *inp;
{
	int	fd, fd1;
	char	*addr, *addr1, *buf;
	int	size, err;
	int	success;
	ssize_t	ret;

	inp->fd = -1;

	unlink("vsrt_tmp");
	if ((fd = open("vsrt_tmp", (O_RDWR|O_CREAT|O_TRUNC), (mode_t)(S_IRWXU|S_IRWXG|S_IRWXO))) == -1) {
		in_rpt("open() failed, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return(-1);
	}


	buf = (char *)vsrt_fn_for_map;

	addr = buf - ((long)buf % vsrt_pgsz);

	if (addr > (char *)0) {
		if (addr >= (char *)vsrt_pgsz) {
			addr -= vsrt_pgsz;
			size = 3;
		}
		else
			size = 2;
	}
	else {
		addr = buf;
		size = 2;
	}

	errno = success = 0;
	while (!success) {
		if (write(fd, (void*)addr, (size_t)(size*vsrt_pgsz)) != (size*vsrt_pgsz)) {
			if ((errno != 0) && (size > 1)) {
				if (size == 3) {
					addr += vsrt_pgsz;
					size--;
				}
				else if (size == 2)
					size--;
			}
			else {
				in_rpt("vsrt_create_exec_shm: write() failed, errno = %d (%s)", errno, errname(errno));
				xx_rpt(FAILURE);
				return(-1);
			}
		}
		else
			success = 1;
		errno = 0;
	}

	(void)lseek(fd, 0, SEEK_SET);

	shm_unlink(s);
	if ((fd1 = shm_open(s, (O_RDWR|O_CREAT|O_TRUNC), (mode_t)(S_IRWXU|S_IRWXG|S_IRWXO))) == -1) {
		in_rpt("vsrt_create_ exec_shm: shm_open() failed for \"%s\", errno = %d (%s)", s, errno, errname(errno));
		xx_rpt(FAILURE);
		return(-1);
	}
	if (ftruncate(fd1, vsrt_pgsz*3) == -1) {
		in_rpt("vsrt_create_exec_shm: ftruncate() failed for \"%s\", errno = %d (%s)", s, errno, errname(errno));
		xx_rpt(FAILURE);
		return -1;
	}
	errno = 0;
	if ((addr1 = (char *)mmap(0, vsrt_pgsz*size, PROT_READ|PROT_WRITE, MAP_SHARED, fd1, 0)) == (char *)MAP_FAILED) {
		err = errno;
		in_rpt("vsrt_create_exec_shm: mmap failed, errno = %d (%s)", err, errname(err));
		xx_rpt(FAILURE);
		return -1;
	}

	if ((ret = read(fd, (void*)addr1, (size_t)(size*vsrt_pgsz))) == -1) {
		in_rpt("vsrt_create_exec_shm: read() failed, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return(-1);
	}

	(void)close(fd);
	(void)unlink("vsrt_tmp");

	(void)msync((void *)addr1, vsrt_pgsz*size, MS_SYNC);

	/* 
	 *	set the signal counters to zero. writing invalid pages
	 *	might have generated signals. The test should be carried
	 *	out after resetting them.
	 */
	vsrt_got_sigsegv = vsrt_got_sigbus = vsrt_got_sig = 0;

	if (fchmod(fd1, mode) == -1) {
		in_rpt("vsrt_create_exec_shm: fchmod() failed, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return(-1);
	}

	if (munmap((void *)addr1, vsrt_pgsz*3) == -1) {
		err = errno;
		in_rpt("vsrt_create_exec_shm: munmap failed, errno = %d (%s)", err, errname(err));
		xx_rpt(FAILURE);
		return -1;
	}
	(void)close(fd1);
	if ((fd1 = shm_open(s, flag, mode)) == -1) {
		in_rpt("vsrt_create_exec_shm: second shm_open() failed for \"%s\", s, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return(-1);
	}


	inp->fd = fd1;
	inp->offset = (off_t)(buf - addr);
	return (0);
}
/*
** Function: int vsrt_create_exec_file(char *s, int flag, mode_t mode, struct vsrt_exec_resource *inp)
**
** Description:
**
**	This function creates an executable file. This file 
**	contains the binary image of a routine. When the file
**	is mapped and the execution control passed to the mapped
**	location the routine executes. The file descriptor to this
**	file is returned by the function in the exec_resource 
**	struct.
**
** Inputs:
**	char *s	->  File to be opened
**	int flag	->  Flags for file open
**	mode_t mode	->  Permission modes
**	struct exec_resource -> Opened file descriptor
**	
** Returns:
**	int	-> 0 - Success
**		   -1 - Fail
**	
*/

int vsrt_create_exec_file(s, flag, mode, inp)
char *s;
int flag;
mode_t mode;
struct vsrt_exec_resource *inp;
{
	int		fd;
	char		*addr, *buf;
	int		size, success;

	inp->fd = -1;

	unlink(s);
	if ((fd = open(s, (O_RDWR|O_CREAT|O_TRUNC), (mode_t)(S_IRWXU|S_IRWXG|S_IRWXO))) == -1) {
		in_rpt("open() failed, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return(-1);
	}

	buf = (char *)vsrt_fn_for_map;

	addr = buf - ((long)buf % vsrt_pgsz);

	if (addr > (char *)0) {
		if (addr >= (char *)vsrt_pgsz) {
			addr -= vsrt_pgsz;
			size = 3;
		}
		else
			size = 2;
	}
	else {
		addr = buf;
		size = 2;
	}

	errno = success = 0;
	while (!success)
	{
		if (write(fd, (void*)addr, (size_t)(size*vsrt_pgsz)) != (size*vsrt_pgsz)) {
			if ((errno != 0) && (size > 1)) {
				if (size == 3) {
					addr += vsrt_pgsz;
					size--;
				}
				else if (size == 2)
					size--;
			}
			else {
				in_rpt("write() failed, errno = %d (%s)", errno, errname(errno));
				xx_rpt(FAILURE);
				return(-1);
			}
		}
		else
			success = 1;
		errno = 0;
	}
	close(fd);

	/* 
	 *	set the signal counters to zero. writing invalid pages
	 *	might have generated signals. The test should be carried
	 *	out after resetting them.
	 */
	vsrt_got_sigsegv = vsrt_got_sigbus = vsrt_got_sig = 0;

	if (chmod(s, mode) == -1) {
		in_rpt("chmod() failed, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return(-1);
	}

	if ((fd = open(s, flag)) == -1) {
		in_rpt("open() failed, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return(-1);
	}

	inp->fd = fd;
	inp->offset = (off_t)(buf - addr);
	return (0);
}


/*
** Function: void * vsrt_map3(int fd, int prot1, int flags1, int prot2, int flags2, int prot3, int flags3)
**
** Description:
**	Map a file into three consecutive areas of memory with different
**	sharing and permission attributes.  Return the address of the first.
**
** Inputs:
**	int fd -> file descriptor of the file to be mapped.
**	int prot1 -> Protection modes for first section
**	int flags1 -> Map flags for the first section
**	int prot2 -> Protection modes for second section
**	int flags2 -> Map flags for the second section
**	int prot3 -> Protection modes for third section
**	int flags3 -> Map flags for the third section
**	
** Returns:
**	-1 on error
**	address of first area mapped on success
*/

void *vsrt_map3(fd, prot1, flags1, prot2, flags2, prot3, flags3)
int	fd;
int	prot1;
int	flags1;
int	prot2;
int	flags2;
int	prot3;
int	flags3;
{
	void	*fixad, *ask_addr;
	void	*err_return;

	err_return = (void *)MAP_FAILED;

	if ((ask_addr = (void *)mmap((void *)0, (size_t)(3*vsrt_pgsz), (PROT_NONE), MAP_SHARED, fd, 0)) == err_return) {
		in_rpt("mmap() failed, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return ((void *)-1);
	}
	if (munmap(ask_addr, (size_t)(3*vsrt_pgsz)) == -1) {
		in_rpt("munmap() failed, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return ((void *)-1);
	}

	if ((fixad = (void *)mmap(ask_addr, (size_t)vsrt_pgsz, prot1, (flags1|MAP_FIXED), fd, 0)) == err_return) {
		in_rpt("mmap() failed, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return ((void *)-1);
	}

	if (fixad != ask_addr) {
		in_rpt("mmap() with MAP_FIXED didn't put segment at proper address");
		in_rpt("Expected: 0x%p  Received: 0x%p", fixad, ask_addr);
		xx_rpt(FAILURE);
		return ((void *)-1);
	}

	if ((fixad = (void *)mmap((void *)((char *)ask_addr+vsrt_pgsz),(size_t)vsrt_pgsz,prot2, (flags2|MAP_FIXED), fd, vsrt_pgsz)) == err_return) {
		in_rpt("mmap() failed, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return ((void *)-1);
	}

	if (fixad != (void *)((char *)ask_addr+vsrt_pgsz)) {
		in_rpt("mmap() with MAP_FIXED didn't put segment at proper address");
		in_rpt("Expected: 0x%p  Received: 0x%p", fixad, ask_addr);
		xx_rpt(FAILURE);
		return ((void *)-1);
	}

	if ((fixad = (void *)mmap((void *)((char *)ask_addr+(2*vsrt_pgsz)),(size_t)vsrt_pgsz,prot3, (flags3|MAP_FIXED), fd, 2*vsrt_pgsz)) == err_return) {
		in_rpt("mmap() failed, errno = %d (%s)", errno, errname(errno));
		xx_rpt(FAILURE);
		return ((void *)-1);
	}

	if (fixad != (void *)((char *)ask_addr+(2*vsrt_pgsz))) {
		in_rpt("mmap() with MAP_FIXED didn't put segment at proper address");
		in_rpt("Expected: 0x%p  Received: 0x%p", fixad, ask_addr);
		xx_rpt(FAILURE);
		return ((void *)-1);
	}

	return ((void *)ask_addr);
}
