
/*************************************************************************
*
* Copyright (c) 1997 X/Open Company Ltd., A member of The Open Group
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* Motif (R) and UNIX (R) are registered trademarks and 'IT DialTone' (TM),
* 'The Open Group' (TM) and the 'X Device' (TM) are trademarks of
* The Open Group.
*
*
*	Written by Mount Bonnell Inc. Austin, Texas, USA
*	@(#)vsth_fw.c	1.38
*
*	History:
*	Changed pthread_join to vsth_pthread_join and pthread_detach to
*	vsth_pthread_detach to allow TET API to reap only unreaped threads.
*
*************************************************************************/


/******	Common frameworks for:
	re-entrancy testing
	cancellation point testing
	EINTR testing						******/

#include	<errno.h>
#if	( TEST_XPG <= 4 )
#include	<sys/types.h>	/* needed for pthread types in POSIX 96 mode */
#endif
#include	<pthread.h>
#include	<sched.h>
#include	<signal.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<sys/stat.h>
#include	<sys/types.h>
#include	<time.h>
#include	<unistd.h>

#include	"tet_api.h"
#include	"vsth/vsth.h"
#include	"vsth/vsth_fw.h"

/******	Common framework for re-entrancy testing in VSTH.

	The difficulty in re-entrancy testing is to set up a set of
	conditions which guarantee that the function-under-test (FUT)
	is actually exercised by more than one thread at the same
	time. In the absence of access to the source code of the
	implementation under test, the best we can do is to adopt a
	strategy which makes it *likely* that the necessary test
	conditions have been established and provide a means for
	measuring the probability. The FUT is called by multiple
	threads a maximum of VSTH_RT_TRIES times total. An entry
	and exit counter is incremented just prior to entry to the
	FUT and just following exit from the FUT. The counters are
	protected by mutexes. A hit counter is incremented each
	time a thread finds the exit counter not equal to the entry
	counter, which is an indication that another thread has
	also entered the FUT. There is, of course, a small probability
	that the exit counter is inspected between the time that a
	second thread has incremented the entry counter but before the
	second thread has entered the FUT. We minimize the chances of
	this happening by testing until the hit count is reasonably
	high (VSTH_HITS_MIN) and give up only after a large number
	of tries (VSTH_TRIES_MAX). Playing with these numbers should
	allow us to increase the probability of achieving appropriate
	test conditions to an arbitrary level.			******/



int			Vsth_rt_cnt;
int			Vsth_wc_init;

pthread_mutex_t		Vsth_mx;	/* for test function use if needed */
char			Vsth_buf1[VSTH_BUFLEN];
char			Vsth_buf2[VSTH_BUFLEN];

static pthread_mutex_t	Rt_mx;
static pthread_cond_t	Rt_cv;
static int		Rt_flag;
static int		Rt_tries;
static int		Rt_hits;
static int		Rt_enter;
static int		Rt_exit;
static pthread_t	Rt_tids[VSTH_SWARM];
static int		Rt_hitsmin;
static int		Rt_triesmax;

static void *rt( void *arg );

void
vsth_rt( VSTH_PRT prt, int *errors, int *fails )
{
	int			cnt;
	int			rval;
	char			*p;
	void			*tval;
	pthread_attr_t		attr;

#ifdef	_POSIX_THREAD_PRIORITY_SCHEDULING
	struct sched_param	params;
#endif

	/* limit on the number of times we actually achieve reentrant conditions  */
	p = tet_getvar("VSTH_RTHITS_MIN");
	if (( p != NULL ) && ( *p ))
	{
		Rt_hitsmin = atoi(p);
	}
	else
	{
		vsth_badconfig("VSTH_RTHITS_MIN", errors);
		return;
	}

	/* limit on the number of tries to achieve Rt_hitsmin */
	p = tet_getvar("VSTH_RTTRIES_MAX");
	if (( p != NULL ) && ( *p ))
	{
		Rt_triesmax = atoi(p);
	}
	else
	{
		vsth_badconfig("VSTH_RTTRIES_MAX", errors);
		return;
	}

	rval = pthread_mutex_init( &Vsth_mx, NULL );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_mutex_init", 0, rval, errors);
		return;
	}

	rval = pthread_mutex_init( &Rt_mx, NULL );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_mutex_init", 0, rval, errors);
		return;
	}

	rval = pthread_cond_init( &Rt_cv, NULL );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_cond_init", 0, rval, errors);
		(void) pthread_mutex_destroy( &Rt_mx );
		return;
	}

	rval = pthread_attr_init( &attr );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_attr_init", 0, rval, errors);
		(void) pthread_cond_destroy( &Rt_cv );
		(void) pthread_mutex_destroy( &Rt_mx );
		return;
	}

	rval = pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_JOINABLE );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_attr_setdetachstate", 0, rval, errors);
		(void) pthread_attr_destroy( &attr );
		(void) pthread_cond_destroy( &Rt_cv );
		(void) pthread_mutex_destroy( &Rt_mx );
		return;
	}

#ifdef	_POSIX_THREAD_PRIORITY_SCHEDULING

#ifndef SCHEDBUGS

	vsth_getprio( SCHED_RR, NULL, &params.sched_priority, errors );
	if ( *errors )
	{
		tet_infoline("vsth_getprio unexpectedly failed");
		(void) pthread_attr_destroy( &attr );
		(void) pthread_cond_destroy( &Rt_cv );
		(void) pthread_mutex_destroy( &Rt_mx );
		return;
	}

	rval = pthread_attr_setschedpolicy( &attr, SCHED_RR );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_attr_setschedpolicy", 0, rval, errors);
		(void) pthread_attr_destroy( &attr );
		(void) pthread_cond_destroy( &Rt_cv );
		(void) pthread_mutex_destroy( &Rt_mx );
		return;
	}

	rval = pthread_attr_setschedparam( &attr, &params );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_attr_setschedparam", 0, rval, errors);
		(void) pthread_attr_destroy( &attr );
		(void) pthread_cond_destroy( &Rt_cv );
		(void) pthread_mutex_destroy( &Rt_mx );
		return;
	}

	rval = pthread_attr_setinheritsched( &attr, PTHREAD_EXPLICIT_SCHED );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_attr_setinheritsched", 0, rval, errors);
		(void) pthread_attr_destroy( &attr );
		(void) pthread_cond_destroy( &Rt_cv );
		(void) pthread_mutex_destroy( &Rt_mx );
		return;
	}

	rval = pthread_attr_setscope( &attr, PTHREAD_SCOPE_PROCESS );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_attr_setscope", 0, rval, errors);
		(void) pthread_attr_destroy( &attr );
		(void) pthread_cond_destroy( &Rt_cv );
		(void) pthread_mutex_destroy( &Rt_mx );
		return;
	}

#endif	/* SCHEDBUGS */

#endif	/* _POSIX_THREAD_PRIORITY_SCHEDULING */

	Rt_flag		= VSTH_SWARM;		/* for a barrier wait */
	Rt_tries	= 0;
	Rt_hits		= 0;
	Rt_enter	= 0;
	Rt_exit		= 0;

	/* If a setup function exists, run it */
	if ( prt->sfunc != NULL )
	{
		rval = (*prt->sfunc)();
		if ( rval != 0 )
		{
			vsth_surprise("Setup function", 0, rval, errors);
		}
	}

	/* spin a swarm of threads and then join them */
	for ( cnt = 0; cnt < VSTH_SWARM; ++cnt )
	{
		rval = vsth_pthread_create(	&(Rt_tids[cnt]),
						&attr,
						rt,
						(void *) prt );
		if ( rval != 0 )
		{
			vsth_surprise("vsth_pthread_create", 0, rval, errors);
		}
	}

	for ( cnt = 0; cnt < VSTH_SWARM; ++cnt )
	{
		rval = vsth_pthread_join( Rt_tids[cnt], &tval );
		if ( rval == 0 )
		{
			if ( tval == (void *) 1 )
			{
				++*errors;
			}
			else if ( tval == (void *) 2 )
			{
				++*fails;
			}
		}
		else
		{
			vsth_surprise("vsth_pthread_join", 0, rval, errors);
		}
	}

	/* If a cleanup/evaluation function exists, execute it */
	if ( prt->efunc != NULL )
	{
		rval = (*prt->efunc)();
		if ( rval != 0 )
		{
			vsth_surprise("Cleanup function", 0, rval, errors);
		}
	}

	if (Vsth_debug)
	{
		tet_printf("Rt_tries: %d, Rt_hits %d", Rt_tries, Rt_hits);
		tet_printf("Rt_enter: %d, Rt_exit %d", Rt_enter, Rt_exit);
	}

	(void) pthread_attr_destroy( &attr );
	(void) pthread_cond_destroy( &Rt_cv );
	(void) pthread_mutex_destroy( &Rt_mx );
	(void) pthread_mutex_destroy( &Vsth_mx );
}


/******	Common thread routine for re-entrancy testing.
	Return 0 if ok
	Return 1 if an error occurred
	Return 2 if the FUT returned an error				******/

static void *
rt( void *arg )
{
	int		rval;
	int		nerrs;
	int		nfails;
	int		done;
	int		eval;
	void		*tval;
	VSTH_PRT	prt;
	
	nerrs		= 0;
	nfails		= 0;
	done		= 0;
	tval		= (void *) 1;
	prt		= (VSTH_PRT) arg;


	(void) vsth_bwait( &Rt_cv, &Rt_mx, &Rt_flag );

	while ( ! done )
	{
		rval = pthread_mutex_lock( &Rt_mx );
		if ( rval != 0 )
		{
			vsth_surprise("pthread_mutex_lock", 0, rval, &nerrs);
			break;
		}

		++Rt_enter;

		rval = pthread_mutex_unlock( &Rt_mx );
		if ( rval != 0 )
		{
			vsth_surprise("pthread_mutex_unlock", 0, rval, &nerrs);
			break;
		}

		/* call the FUT wrapper and break on an error */
		rval = (*prt->tfunc)();
		if ( rval == 2 )
		{
			++nfails;
			break;
		}
		else if ( rval )
		{
			++nerrs;
			break;
		}

		rval = pthread_mutex_lock( &Rt_mx );
		if ( rval != 0 )
		{
			vsth_surprise("pthread_mutex_lock", 0, rval, &nerrs);
			break;
		}

		/*	If Rt_enter and Rt_exit don't have the same value, then
		 *	another thread is between the two checkpoints
		 */

		if ( Rt_enter != ++Rt_exit)
			++Rt_hits;

		++Rt_tries;

		if (	( Rt_tries >= Rt_triesmax )	||
			( Rt_hits >= Rt_hitsmin )	)
		{
			/* break on limit condition */
			done = 1;
		}

		rval = pthread_mutex_unlock( &Rt_mx );
		if ( rval != 0 )
		{
			vsth_surprise("pthread_mutex_lock", 0, rval, &nerrs);
			break;
		}
	}

	if ( nfails )
		tval = (void *) 2;
	else if (( nerrs == 0 ) && ( nfails == 0 ))
		tval = NULL;

	pthread_exit(tval);

	/* NOTREACHED */
	return(tval);
}

/******	Common framework for cancellation point testing in VSTH.
	Call the setup function. Spawn a thread. The thread
	calls pthread_cancel(pthread_self()) then calls the function
	under test. It then calls pthread_exit() with an argument
	!= PTHREAD_CANCELED. The thread is joined, the return value
	examined and the cleanup function called.		******/


static void	*cpt( void *arg );
static int	cpt_check( char *fname, int flag, void *tval );

void
vsth_cpt( VSTH_PRT prt, int flag, int *errors, int *fails )
{
	int		rval;
	void		*tval;
	pthread_t	tid;

	if ( prt->tfunc != NULL )
	{
		/* setup the test conditions for a call to the FUT */

		rval = 0;

		if ( prt->sfunc )
		{
			rval = (*prt->sfunc)();
		}
	
		if ( rval == 0 )
		{
			rval = vsth_pthread_create( &tid, NULL, cpt, (void *) prt );
			if ( rval == 0 )
			{
				rval = vsth_pthread_join( tid, &tval );
				if ( rval == 0 )
				{
					rval = cpt_check( prt->fname, flag, tval );
					if ( rval != 0 )
					{
						++*fails;
					}
				}
				else
				{
					(void) pthread_cancel( tid );
					(void) vsth_pthread_detach( tid );
					vsth_surprise("pthread_join", 0, rval, errors);
				}
			}
			else
			{
				vsth_surprise("vsth_pthread_create", 0, rval, errors);
			}
		}
		else
		{
			++*errors;
			tet_infoline("The setup function failed");
		}

		/* cleanup the test conditions */
		if ( prt->efunc )
		{
			(void) (*prt->efunc)();
		}
	}
}

/* return 0 if ok, else 1 */
static int
cpt_check( char *fname, int flag, void *tval )
{
	int	rval;

	rval = 1;

	if ( flag < 0 )
	{
		/* must not */
		if ( tval == NULL )
		{
			rval = 0;
		}
	}
	else if ( flag == 0 )
	{
		/* may */
		if ( tval == (void *) PTHREAD_CANCELED )
		{
			tet_printf("%s contains a CP", fname ? fname : "NULL");
		}
		else
		{
			tet_printf("%s doesn't contain a CP", fname ? fname : "NULL");
		}

		rval = 0;
	}
	else
	{
		/* must */
		if (tval == (void *) PTHREAD_CANCELED)
		{
			rval = 0;
		}
	}

	return(rval);
}

/******	Common thread routine for cancellation point testing
	Return NULL if ok					******/

static void *
cpt( void *arg )
{
	int		rval;
	int		nerrs;
	int		nfails;
	void		*tval;
	VSTH_PRT	prt;
	
	nerrs	= 0;
	nfails	= 0;
	tval	= (void *) 1;
	prt	= (VSTH_PRT) arg;

	rval = pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	if ( rval != 0 )
	{
		vsth_surprise("pthread_setcancelstate", 0, rval, &nerrs);
	}

	rval = pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
	if ( rval != 0 )
	{
		vsth_surprise("pthread_setcanceltype", 0, rval, &nerrs);
	}

	rval = pthread_cancel(pthread_self());
	if ( rval != 0 )
	{
		vsth_surprise("pthread_cancel", 0, rval, &nerrs);
	}

	if ( prt && prt->tfunc )
	{
		(*prt->tfunc)();
	}

	if ( nfails )
	{
		tval = (void *) 2;
	}
	else if (( nerrs == 0 ) && ( nfails == 0 ))
	{
		tval = NULL;
	}

	pthread_exit(tval);

	/* NOTREACHED */
	return(tval);
}

/******	Common framework for EINTR testing of thread functions.
	Create two threads and synchronize them on a point
	just prior to function under test invocation. One thread
	signal storms the other while it invokes the FUT. 	******/

static int		Et_sigcnt;
static int		Et_nsigs;
static int		Et_flag;
static pthread_t	*Et_ptid;
static pthread_mutex_t	Et_mx;
static pthread_cond_t	Et_cv;

static int		et_ssigs( sigset_t *oset, struct sigaction *osa );
static int		et_rsigs( sigset_t *oset, struct sigaction *osa );
static void		*et( void *arg );
static void		et_sigterm( int sig );

void
vsth_et( VSTH_PRT prt, int *errors, int *fails )
{
	int			rval;
	int			cnt;
	sigset_t		oset;
	struct sigaction	osa;
	pthread_attr_t		attr;
	pthread_t		tid1;
	pthread_t		tid2;
	void			*tval;

/*
#ifdef	_POSIX_THREAD_PRIORITY_SCHEDULING
	struct sched_param	params;
#endif
*/

	if ( prt->tfunc == NULL )
	{
		++*errors;
		tet_infoline("NULL function under test");
		return;
	}

	rval = pthread_mutex_init( &Et_mx, NULL );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_mutex_init", 0, rval, errors);
		return;
	}

	rval = pthread_cond_init( &Et_cv, NULL );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_mutex_init", 0, rval, errors);
		(void) pthread_mutex_destroy( &Et_mx );
		return;
	}

	rval = pthread_attr_init( &attr );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_attr_init", 0, rval, errors);
		(void) pthread_mutex_destroy( &Et_mx );
		(void) pthread_cond_destroy( &Et_cv );
		return;
	}

/*
#ifdef	_POSIX_THREAD_PRIORITY_SCHEDULING

	vsth_getprio( SCHED_RR, NULL, &params.sched_priority, errors );
	if ( *errors )
	{
		(void) pthread_attr_destroy( &attr );
		(void) pthread_mutex_destroy( &Et_mx );
		(void) pthread_cond_destroy( &Et_cv );
		return;
	}

	rval = pthread_attr_setschedpolicy( &attr, SCHED_RR );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_attr_setschedpolicy", 0, rval, errors);
		(void) pthread_attr_destroy( &attr );
		(void) pthread_mutex_destroy( &Et_mx );
		(void) pthread_cond_destroy( &Et_cv );
		return;
	}

	rval = pthread_attr_setschedparam( &attr, &params );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_attr_setschedparam", 0, rval, errors);
		(void) pthread_attr_destroy( &attr );
		(void) pthread_mutex_destroy( &Et_mx );
		(void) pthread_cond_destroy( &Et_cv );
		return;
	}

	rval = pthread_attr_setinheritsched( &attr, PTHREAD_EXPLICIT_SCHED );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_attr_setinheritsched", 0, rval, errors);
		(void) pthread_attr_destroy( &attr );
		(void) pthread_mutex_destroy( &Et_mx );
		(void) pthread_cond_destroy( &Et_cv );
		return;
	}

#endif
*/

	/****** block all signals in the main thread and
		install a handler for SIGTERM			******/
	rval = et_ssigs( &oset, &osa );
	if ( rval != 0 )
	{
		vsth_surprise("et_ssigs", 0, rval, errors);
		(void) pthread_attr_destroy( &attr );
		(void) pthread_mutex_destroy( &Et_mx );
		(void) pthread_cond_destroy( &Et_cv );
		return;
	}

	Et_nsigs	= 0;
	Et_sigcnt	= 0;

	for ( cnt = 0; cnt < VSTH_ETTRIES_MAX; ++cnt )
	{
		Et_flag = 0;
		Et_ptid	= NULL;

		/* create the threads */
		rval = vsth_pthread_create( &tid1, &attr, et, (void *) prt );
		if ( rval != 0 )
		{
			vsth_surprise("vsth_pthread_create", 0, rval, errors);
			(void) pthread_attr_destroy( &attr );
			(void) pthread_mutex_destroy( &Et_mx );
			(void) pthread_cond_destroy( &Et_cv );
			(void) et_rsigs( &oset, &osa );
			return;
		}
		rval = vsth_pthread_create( &tid2, &attr, et, (void *) prt );
		if ( rval != 0 )
		{
			vsth_surprise("vsth_pthread_create", 0, rval, errors);
			(void) pthread_cancel( tid1 );
			(void) vsth_pthread_join( tid1, NULL );
			(void) pthread_attr_destroy( &attr );
			(void) pthread_attr_destroy( &attr );
			(void) pthread_mutex_destroy( &Et_mx );
			(void) et_rsigs( &oset, &osa );
			return;
		}
	
		/* join tid1 first so tid2 gets ESRCH from pthread_kill() */
		rval = vsth_pthread_join( tid1, &tval );
		if ( rval == 0 )
		{
			if ( tval == (void *) 2 )
			{
				vsth_surprise("Thread 1 exitval", 0, (int) tval, fails);
			}
			else if ( tval )
			{
				vsth_surprise("Thread 1 exitval", 0, (int) tval, errors);
			}
		}
		else
		{
			vsth_surprise("pthread_join", 0, rval, errors);
		}
	
		rval = vsth_pthread_join( tid2, &tval );
		if ( rval == 0 )
		{
			if ( tval == (void *) 2 )
			{
				vsth_surprise("Thread 2 exitval", 0, (int) tval, fails);
			}
			else if ( tval )
			{
				vsth_surprise("Thread 2 exitval", 0, (int) tval, errors);
			}
		}
		else
		{
			vsth_surprise("pthread_join", 0, rval, errors);
		}

		if ( *errors || *fails )
			break;
	}

	tet_printf("Thread 1 delivered %d SIGTERM signals", Et_nsigs);
	tet_printf("Thread 2 caught %d SIGTERM signals", Et_sigcnt);

	(void) pthread_attr_destroy( &attr );
	(void) pthread_mutex_destroy( &Et_mx );
	(void) pthread_cond_destroy( &Et_cv );
	(void) et_rsigs( &oset, &osa );
}

/******	Common thread routines for EINTR testing.
	thread 1 signal storms thread 2 with SIGTERM signals after synchronizing.
	thread2 executes the setup function, synchronizes, unblocks SIGTERM,
	executes the function under test, blocks SIGTERM then executes the
	cleanup function.						******/

static void *
et( void *arg )
{

	int		rval;
	int		cnt;
	int		nerrs;
	int		nfails;
	int		futval;
	void		*tval;
	pthread_t	mytid;
	sigset_t	set;
	VSTH_PRT	pet;

	nerrs	= 0;
	nfails	= 0;
	tval	= (void *) 1;
	mytid	= pthread_self();
	pet	= (VSTH_PRT) arg;

	rval = pthread_mutex_lock( &Et_mx );
	if ( rval == 0 )
	{
		if ( Et_flag == 0 )
		{
			++Et_flag;

			while ( Et_flag != 2 )
			{
				rval = pthread_cond_wait( &Et_cv, &Et_mx );
				if ( rval != 0 )
				{
					vsth_surprise(	"pthread_cond_wait",
							0,
							rval,
							&nerrs );
				}
			}
	
			rval = pthread_mutex_unlock( &Et_mx );
			if ( rval != 0 )
			{
				vsth_surprise("pthread_mutex_unlock", 0, rval, &nerrs);
			}
	
			/* signal storm the other thread till ESRCH or we hit a limit */
			cnt = 0;
			while (( cnt++ < VSTH_ETSIGS_MAX ) && ( rval != ESRCH ))
			{
				rval = pthread_kill( *Et_ptid, SIGTERM );
				if (( rval != 0 ) && ( rval != ESRCH ))
				{
					vsth_surprise(	"pthread_kill",
							0,
							rval,
							&nerrs );
				}
				++Et_nsigs;
			}
		}
		else
		{
			Et_ptid	= &mytid;
	
			/* if there is a setup function, call it */
			if ( pet->sfunc != NULL )
			{
				rval = (*pet->sfunc)();
				if ( rval != 0 )
				{
					vsth_surprise(	"setup function",
							0,
							rval,
							&nerrs );
				}
			}
	
			rval = sigemptyset( &set );
			if ( rval != 0 )
			{
				vsth_surprise("sigemptyset", 0, rval, &nerrs);
			}
	
			rval = sigaddset( &set, SIGTERM  );
			if ( rval != 0 )
			{
				vsth_surprise("sigaddset", 0, rval, &nerrs);
			}
	
			rval = pthread_sigmask( SIG_UNBLOCK, &set, NULL );
			if ( rval != 0 )
			{
				vsth_surprise("pthread_sigmask", 0, rval, &nerrs);
			}
	
			++Et_flag;
	
			rval = pthread_mutex_unlock( &Et_mx );
			if ( rval != 0 )
			{
				vsth_surprise("pthread_mutex_unlock", 0, rval, &nerrs);
			}

			rval = pthread_cond_signal( &Et_cv );
			if ( rval != 0 )
			{
				vsth_surprise("pthread_cond_signal", 0, rval, &nerrs);
			}
	
			/* just for good measure */
			rval = pthread_kill( mytid, SIGTERM );
			if ( rval != 0 )
			{
				vsth_surprise("pthread_kill", 0, rval, &nerrs);
			}

			/* we're signal stormed through this section - exercise the fut */
			futval = (*pet->tfunc)();
	
			/* block signals */
			rval = pthread_sigmask( SIG_BLOCK, &set, NULL );
			if ( rval != 0 )
			{
				vsth_surprise("pthread_sigmask", 0, rval, &nerrs);
			}
	
			if ( futval != 0 )
			{
				vsth_surprise("test function", 0, futval, &nfails);
			}
	
			/* if there is a cleanup function, call it */
			if ( pet->efunc != NULL )
			{
				rval = (*pet->efunc)();
				if ( rval != 0 )
				{
					vsth_surprise("cleanup function", 0, rval, &nerrs);
				}
			}
		}
	}
	else
	{
		vsth_surprise("pthread_mutex_lock", 0, rval, &nerrs);
	}

	if ( nfails )
		tval = (void *) 2;
	else if (( nerrs == 0 ) && ( nfails == 0 ))
		tval = NULL;

	pthread_exit(tval);

	/* NOTREACHED */
	return(tval);
}

/* block all signals and install a handler for SIGTERM */
static int
et_ssigs(sigset_t *oset, struct sigaction *osa )
{
	int			rval;
	sigset_t		nset;
	struct sigaction	nsa;
	
	/* block SIGTERM in the main thread */
	rval = sigemptyset( &nset );
	if ( rval != 0 )
	{
		vsth_surprise("sigemptyset", 0, rval, NULL);
		return(rval);
	}

	rval = sigaddset( &nset, SIGTERM );
	if ( rval != 0 )
	{
		vsth_surprise("sigaddset", 0, rval, NULL);
		return(rval);
	}

	rval = pthread_sigmask( SIG_BLOCK, &nset, oset );
	if ( rval == 0 )
	{

		/* establish signal handlers for signal storm sigs */
		rval = sigfillset( &(nsa.sa_mask) );
		if ( rval == 0 )
		{
			nsa.sa_flags	= 0;
			nsa.sa_handler	= et_sigterm;
			rval = sigaction( SIGTERM, &nsa, osa );
			if ( rval != 0 )
			{
				vsth_surprise("sigaction", 0, rval, NULL);
			}
		}
		else
		{
			vsth_surprise("sigfillset", 0, rval, NULL);
		}
	}
	else
	{
		vsth_surprise("pthread_sigmask", 0, rval, NULL);
	}

	return(rval);
}

/* restore signal behavior */
static int
et_rsigs( sigset_t *oset, struct sigaction *osa )
{
	int	rval;

	/* restore signal behavior */
	rval = pthread_sigmask( SIG_SETMASK, oset, NULL );
	if ( rval == 0 )
	{
		rval = sigaction( SIGTERM, osa, NULL );
	}

	return(rval);
}

/* handler for SIGTERM */
static void
et_sigterm( int sig )
{
	++Et_sigcnt;
}

void
vsth_ec()
{
	int	rval;
	rval = pthread_mutex_lock(&Vsth_mx);
	if ( rval != 0 )
	{
		tet_printf("pthread_mutex_lock failed %d", rval);
	}
}

void
vsth_xc()
{
	int	rval;
	rval = pthread_mutex_unlock(&Vsth_mx);
	if ( rval != 0 )
	{
		tet_printf("pthread_mutex_unlock failed %d", rval);
	}
}

/******	Framework for FILE * tests in PTHR.os/all/General test 2	******/

static int		Fs_flag;
static pthread_mutex_t	Fs_mx;
static pthread_cond_t	Fs_cv;

static VSTH_TSET	Fs_tset[VSTH_SWARM];

static void		*fst( void *arg );
static int		cpt_check( char *fname, int flag, void *tval );

void
vsth_fst( VSTH_PRT prt, int *errors, int *fails )
{
	int		rval;
	int		cnt;
	int		priomin;
	void		*tval;
	VSTH_PTSET	pts;

	vsth_getprio(SCHED_RR, &priomin, NULL, errors);
	if ( *errors )
	{
		return;
	}

	if ( prt->tfunc != NULL )
	{
		rval = pthread_mutex_init( &Fs_mx, NULL );
		if ( rval != 0 )
		{
			vsth_surprise("pthread_mutex_init", 0, rval, errors);
			return;
		}

		rval = pthread_cond_init( &Fs_cv, NULL );
		if ( rval != 0 )
		{
			vsth_surprise("pthread_cond_init", 0, rval, errors);
			(void) pthread_mutex_destroy( &Fs_mx );
			return;
		}

		/* setup the test conditions for a call to the FUT */
		if ( prt->sfunc )
		{
			rval = (*prt->sfunc)();
			if ( rval )
			{
				++*errors;
			}
		}
	
		pts = &Fs_tset[0];
		for ( cnt = 0; cnt < VSTH_SWARM; ++cnt )
		{
			
			pts->policy		= SCHED_RR;
			pts->priority		= priomin;
			pts->scope		= PTHREAD_SCOPE_PROCESS;
			pts->detachstate	= PTHREAD_CREATE_JOINABLE;
			pts->func		= fst;
			pts->arg		= prt;
			pts->status		= -1;

			++pts;
		}

		Fs_flag = VSTH_SWARM;

		rval = vsth_ts_create( &Fs_tset[0], VSTH_SWARM, 0, errors, fails );
		if ( rval )
		{
			vsth_ts_cancel( &Fs_tset[0], VSTH_SWARM, 1 );
			(void) pthread_mutex_destroy( &Fs_mx );
			(void) pthread_cond_destroy( &Fs_cv );
			return;
		}

		rval = vsth_ts_join( &Fs_tset[0], VSTH_SWARM, errors, fails );
		if ( rval )
		{
			vsth_surprise("vsth_ts_join", 0, rval, errors);
		}

		/* evaluate and cleanup the test conditions */
		if ( prt->efunc )
		{
			rval = (*prt->efunc)();
			if ( rval == 2 )
			{
				++*fails;
			}
			else if ( rval )
			{
				++*errors;
			}
		}

		(void) pthread_mutex_destroy( &Fs_mx );
		(void) pthread_cond_destroy( &Fs_cv );
	}
}


/******	Common thread routine for FILE * and file locking testing
	Return NULL if ok					******/

static void *
fst( void *arg )
{
	int		rval;
	int		nerrs;
	int		nfails;
	void		*tval;
	VSTH_PRT	prt;
	
	nerrs	= 0;
	nfails	= 0;

	tval	= (void *) 1;
	prt	= (VSTH_PRT) arg;

	vsth_bwait( &Fs_cv, &Fs_mx, &Fs_flag );

	if ( prt && prt->tfunc )
	{
		rval = (*prt->tfunc)();
		if ( rval == 2 )
		{
			++nfails;
		}
		else if ( rval )
		{
			++nerrs;
		}
	}

	if ( nfails )
	{
		tval = (void *) 2;
	}
	else if (( nerrs == 0 ) && ( nfails == 0 ))
	{
		tval = NULL;
	}

	pthread_exit(tval);

	/* NOTREACHED */
	return(tval);
}

