
/*************************************************************************
*
* Copyright (c) 1997 X/Open Company Ltd., A member of The Open Group
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* Motif (R) and UNIX (R) are registered trademarks and 'IT DialTone' (TM),
* 'The Open Group' (TM) and the 'X Device' (TM) are trademarks of
* The Open Group.
*
*
*	Written by Mount Bonnell Inc. Austin, Texas, USA
*	@(#)vsth_gen.c	1.48
*
*	History:
*	Added vsth_pthread_detach as a wrapper for tet_pthread_detach.
*	Changed vsth_pthread_join to call tet_pthread_join instead of 
*	pthread_join.
*
*************************************************************************/

#include <errno.h>
#include <limits.h>
#if	( TEST_XPG <= 4 )	/* needed for pthread types in POSIX96 mode */
#include <sys/types.h>
#endif
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

#include "tet_api.h"
#include "vsth/vsth.h"

int	Vsth_debug	= 0;
int	Vsth_timeout	= VSTH_DEFAULT_TIMEOUT;

void
vsth_error( char *s, int rval, int *error )
{
	if ( s != NULL )
	{
		tet_printf( "%s failed %d", s, rval );
	}
	if ( error != NULL )
	{
		++*error;
	}
}

void
vsth_surprise( char *msg, int eval, int rval, int *fail )
{
	tet_printf("%s, expected %d, got %d", ( msg != NULL ) ? msg : "", eval, rval);
	if ( fail != NULL )
	{
		++*fail;
	}
}

void
vsth_badconfig( char *var, int *errors )
{
	if ( var != NULL )
	{
		tet_printf("The configuration variable %s is not set", var);
	}
	if ( errors != NULL )
	{
		++*errors;
	}
}



/*	initialize a pthread_attr_t	*/
int
vsth_attr_init( pthread_attr_t *attr,
		int detachstate,
		int policy,
		int priority,
		int inheritsched,
		int scope )
{
	int			rval;
	struct	sched_param	params;

	rval = pthread_attr_init( attr );
	if ( rval )
	{
		vsth_surprise("pthread_attr_init", 0, rval, NULL );
		return(rval);
	}

	rval = pthread_attr_setdetachstate( attr, detachstate );
	if ( rval )
	{
		vsth_surprise("pthread_attr_setdetachstate", 0, rval, NULL );
		return(rval);
	}

#ifdef	_POSIX_THREAD_PRIORITY_SCHEDULING

#ifndef	SCHEDBUGS
	rval = pthread_attr_setschedpolicy( attr, policy );
	if ( rval )
	{
		vsth_surprise("pthread_attr_setschedpolicy", 0, rval, NULL );
		return(rval);
	}

	params.sched_priority = priority;
	rval = pthread_attr_setschedparam( attr, &params );
	if ( rval )
	{
		vsth_surprise("pthread_attr_setschedparam", 0, rval, NULL );
		return(rval);
	}

	rval = pthread_attr_setinheritsched( attr, inheritsched );
	if ( rval )
	{
		vsth_surprise("pthread_attr_setinheritsched", 0, rval, NULL );
		return(rval);
	}

	rval = pthread_attr_setscope( attr, scope );
	if ( rval )
	{
		vsth_surprise("pthread_attr_setscope", 0, rval, NULL );
		return(rval);
	}
#endif	/* SCHEDBUGS */

#endif	/* _POSIX_THREAD_PRIORITY_SCHEDULING */

	return(0);

}

/*	methods for sync points	*/
int
vsth_sp_init( VSTH_PSP psp )
{
	int			rval;
	int			protocol;
	pthread_condattr_t	ca;
	pthread_mutexattr_t	ma;
	pthread_t		mytid;

	rval = 1;

	if ( psp )
	{
		rval = pthread_condattr_init( &ca );
		if ( rval )
		{
			vsth_surprise("pthread_condattr_init", 0, rval, NULL);
			return(rval);
		}

#ifdef	_POSIX_THREAD_PROCESS_SHARED
		rval = pthread_condattr_setpshared( &ca, PTHREAD_PROCESS_SHARED );
		if ( rval )
		{
			vsth_surprise("pthread_condattr_setpshared", 0, rval, NULL);
			(void) pthread_condattr_destroy(&ca);
			return(rval);
		}
#endif

		/* initialize the condition variable */
		rval = pthread_cond_init( &psp->c, &ca );
		if (rval)
		{
			vsth_surprise("pthread_cond_init", 0, rval, NULL);
			(void) pthread_condattr_destroy( &ca );
			return(rval);
		}
	
		(void) pthread_condattr_destroy( &ca );
	
		rval = pthread_mutexattr_init( &ma );
		if (rval)
		{
			vsth_surprise("pthread_mutexattr_init", 0, rval, NULL);
			(void) pthread_cond_destroy( &psp->c );
			return(rval);
		}

#ifdef	_POSIX_THREAD_PROCESS_SHARED

		rval = pthread_mutexattr_setpshared( &ma, PTHREAD_PROCESS_SHARED );
		if (rval)
		{
			vsth_surprise("pthread_mutexattr_setpshared", 0, rval, NULL);
			(void) pthread_cond_destroy( &psp->c );
			(void) pthread_mutexattr_destroy( &ma );
			return(rval);
		}

#endif	/* _POSIX_THREAD_PROCESS_SHARED */

#if	defined(_POSIX_THREAD_PRIO_INHERIT) || \
	defined(_POSIX_THREAD_PRIO_PROTECT)

		rval = pthread_mutexattr_setprotocol( &ma, PTHREAD_PRIO_NONE );
		if (rval)
		{
			vsth_surprise("pthread_mutexattr_setprotocol", 0, rval, NULL);
			(void) pthread_condattr_destroy( &ca );
			(void) pthread_mutexattr_destroy( &ma );
			return(rval);
		}

#endif

		/* initialize the mutex */
		rval = pthread_mutex_init( &psp->m, &ma );
		if (rval)
		{
			vsth_surprise("pthread_mutex_init", 0, rval, NULL);
			(void) pthread_cond_destroy( &psp->c );
			(void) pthread_mutexattr_destroy( &ma );
			return(rval);
		}
	
		(void) pthread_mutexattr_destroy( &ma );
	
		rval = vsth_sp_clear( psp );
	}

	return(rval);
}

int
vsth_sp_destroy( VSTH_PSP psp )
{
	int	rval;
	pthread_t	mytid;

	rval = 1;

	if ( psp )
	{
		/* destroy the condition variable */
		rval = pthread_cond_destroy( &psp->c );
		if ( rval == 0 )
		{
			rval = pthread_mutex_destroy( &psp->m );
			if ( rval )
			{
				vsth_surprise("pthread_mutex_destroy", 0, rval, NULL);
			}
		}
		else
		{
			vsth_surprise("pthread_cond_destroy", 0, rval, NULL);
		}
	}

	return(rval);
}

int
vsth_sp_signal( VSTH_PSP psp, long val )
{
	int		rval;
	int		retval;
	pthread_t	mytid;

	retval = 0;

	if ( psp )
	{
		rval = pthread_mutex_lock( &psp->m );
		if ( rval )
		{
			retval += rval;
			vsth_surprise("pthread_mutex_lock", 0, rval, NULL);
		}

		psp->p = val;

		rval = pthread_cond_signal( &psp->c );
		if ( rval )
		{
			retval += rval;
			vsth_surprise("pthread_cond_signal", 0, rval, NULL);
		}

		rval = pthread_mutex_unlock( &psp->m );
		if ( rval )
		{
			retval += rval;
			vsth_surprise("pthread_mutex_unlock", 0, rval, NULL);
		}
	}
	else
	{
		retval = 1;
	}

	return(retval);
}

int
vsth_sp_bcast( VSTH_PSP psp, long val )
{
	int	rval;
	pthread_t	mytid;

	if ( Vsth_debug )
	{
		mytid = pthread_self();
		tet_printf("%d vsth_sp_bcast enter: %d", (int) mytid, val);
	}

	rval = 0;

	if ( psp )
	{
		rval += pthread_mutex_lock( &psp->m );
		if ( Vsth_debug )
		{
			tet_printf("%d pthread_mutex_lock %d", (int) mytid, rval);
		}

		psp->p = val;

		rval += pthread_mutex_unlock( &psp->m );
		if ( Vsth_debug )
		{
			tet_printf("%d pthread_mutex_unlock %d", (int) mytid, rval);
		}

		rval += pthread_cond_broadcast( &psp->c );
		if ( Vsth_debug )
		{
			tet_printf("%d pthread_cond_broadcast %d", (int) mytid, rval);
		}
	}
	else
	{
		rval = 1;
	}

	if ( Vsth_debug )
	{
		tet_printf("%d vsth_sp_bcast exit: %d", (int) (int) mytid, rval);
	}

	return(rval);
}

/* enter with psp->m locked or all bets are off */
int
vsth_sp_wait( VSTH_PSP psp, long val )
{
	int	rval;
	pthread_t	mytid;

	if ( Vsth_debug )
	{
		mytid = pthread_self();
		tet_printf("%d vsth_sp_wait enter: %d", (int) mytid, val);
	}

	rval = 0;

	if ( psp )
	{
		while ( psp->p != val )
		{
			rval += pthread_cond_wait( &psp->c, &psp->m );
			if ( Vsth_debug )
			{
				tet_printf("%d pthread_cond_wait %d", (int) mytid, rval);
			}
			if ( rval != 0 )
			{
				vsth_surprise("vsth_sp_wait: pthread_cond_wait", 0, rval, NULL);
			}
		}
		rval += pthread_mutex_unlock( &psp->m );
		if ( Vsth_debug )
		{
			tet_printf("%d pthread_mutex_unlock %d", (int) mytid, rval);
		}
	}

	if ( Vsth_debug )
	{
		tet_printf("%d vsth_sp_wait exit: %d", (int) mytid, rval);
	}

	return(rval);
}

vsth_sp_set( VSTH_PSP psp, long val )
{
	int	rval;
	pthread_t	mytid;

	if ( Vsth_debug )
	{
		mytid = pthread_self();
		tet_printf("%d vsth_sp_set enter: %d", (int) mytid, val);
	}

	rval = 1;

	if ( psp )
	{
		rval = pthread_mutex_lock( &psp->m );
		if ( Vsth_debug )
		{
			tet_printf("%d pthread_mutex_lock %d", (int) mytid, rval);
		}

		if ( rval == 0 )
		{
			psp->p = val;
			rval = pthread_mutex_unlock( &psp->m );
			if ( Vsth_debug )
			{
				tet_printf("%d pthread_mutex_unlock %d", (int) mytid, rval);
			}
		}
	}

	if ( Vsth_debug )
	{
		tet_printf("%d vsth_sp_set exit: %d", (int) mytid, rval);
	}

	return(rval);
}

int
vsth_sp_clear( VSTH_PSP psp )
{
	int	rval;
	pthread_t	mytid;

	if ( Vsth_debug )
	{
		mytid = pthread_self();
		tet_printf("%d vsth_sp_clear enter:", (int) mytid);
	}

	rval = 1;

	if ( psp )
	{
		( void ) pthread_mutex_unlock( &psp->m );
		psp->p	= 0L;
		rval	= 0;
	}

	if ( Vsth_debug )
	{
		tet_printf("%d vsth_sp_clear exit: %d", (int) mytid, rval);
	}

	return(rval);
}

/* a wrapper to pthread_create() */
int
vsth_pthread_create(	pthread_t *ptid,
			pthread_attr_t *pa,
			void * (*start)(void *),
			void *arg )
{
	int	rval;

	if ( Vsth_debug )
	{
		tet_printf("vsth_pthread_create enter: %p", ptid );
	}

	rval = tet_pthread_create( ptid, pa, start, arg, VSTH_TIMEOUT );

	if ( Vsth_debug )
	{
		tet_printf("vsth_pthread_create exit: %d", rval);
		if ( rval == 0 )
			tet_printf("\ttid == %x", (int) *ptid);
	}

	return(rval);
}

int
vsth_pthread_join( pthread_t tid, void **val)
{
	int	rval;

	if ( Vsth_debug )
	{
		tet_printf("vsth_pthread_join enter: %x", (int) tid );
	}

	rval = tet_pthread_join( tid, val );

	if ( Vsth_debug )
	{
		tet_printf("vsth_pthread_join exit: %d", rval );
	}

	return(rval);
}

int
vsth_pthread_detach( pthread_t tid)
{
	int	rval;

	if ( Vsth_debug )
	{
		tet_printf("vsth_pthread_detach enter: %x", (int) tid );
	}

	rval = tet_pthread_detach( tid );

	if ( Vsth_debug )
	{
		tet_printf("vsth_pthread_detach exit: %d", rval );
	}

	return(rval);
}

int
vsth_pthread_cancel( pthread_t tid )
{
	int	rval;
	pthread_t	mytid;

	if ( Vsth_debug )
	{
		mytid = pthread_self();
		tet_printf("%d vsth_pthread_cancel enter: %p", (int) mytid, (int) tid );
	}

	rval = pthread_cancel( tid );

	if ( Vsth_debug )
	{
		tet_printf("%d vsth_pthread_cancel exit: %p %d",
				(int) mytid, (int) tid, rval );
	}

	return(rval);
}

int
vsth_pthread_mutex_lock( pthread_mutex_t *pm )
{
	int	rval;

	if ( Vsth_debug )
	{
		tet_printf("vsth_pthread_mutex_lock enter: %p", pm);
	}

	rval = pthread_mutex_lock( pm );

	if ( Vsth_debug )
	{
		tet_printf("vsth_pthread_mutex_lock exit: %p %d", pm, rval);
	}

	return(rval);
}

int
vsth_pthread_mutex_unlock( pthread_mutex_t *pm )
{

	int	rval;

	if ( Vsth_debug )
	{
		tet_printf("vsth_pthread_mutex_unlock enter: %p", pm);
	}

	rval = pthread_mutex_unlock( pm );

	if ( Vsth_debug )
	{
		tet_printf("vsth_pthread_mutex_unlock exit: %p %d", pm, rval);
	}

	return(rval);
}

void
vsth_rr_get_interval( pid_t pid, struct timespec *interval, int *errors)
{

	int	rval;
	char	*p;

#ifdef	_POSIX_PRIORITY_SCHEDULING
	rval = sched_rr_get_interval(pid, interval);
	if ( rval != 0 )
	{
		if (errors)
			++*errors;
		tet_printf("sched_rr_get_interval unexpectedly failed, errno == %d", errno);
	}
#else

	p = tet_getvar("VSTH_RR_INTERVAL_SECS");
	if ( p && *p && isdigit(*p) )
	{
		if ( interval )
			interval->tv_sec = atoi(p);
	}
	else
	{
		vsth_badconfig("VSTH_RR_INTERVAL_SECS", errors);
	}

	p = tet_getvar("VSTH_RR_INTERVAL_NSECS");
	if ( p && *p && isdigit(*p) )
	{
		if ( interval )
			interval->tv_nsec = atoi(p);
	}
	else
	{
		vsth_badconfig("VSTH_RR_INTERVAL_NSECS", errors);
	}

#endif

}

void
vsth_getprio( int policy, int *min, int *max, int *errors)
{
	char	*p;
	char	*mincv;
	char	*maxcv;

#if	defined(_POSIX_PRIORITY_SCHEDULING)

#ifdef	sched_get_priority_min
#undef	sched_get_priority_min
#endif

#ifdef	sched_get_priority_max
#undef	sched_get_priority_max
#endif

	if ( Vsth_debug )
	{
		tet_infoline("vsth_getprio: using sched functions");
	}

	if (	( policy == SCHED_FIFO )	||
		( policy == SCHED_RR )		||
		( policy == SCHED_OTHER))
	{
		if ( min != NULL )
		{
			*min = sched_get_priority_min(policy);
		}
		if ( max != NULL )
		{
			*max = sched_get_priority_max(policy);
		}
	}
	else
	{
		++*errors;
	}

#else

	mincv = NULL;
	maxcv = NULL;

	if ( Vsth_debug )
	{
		tet_infoline("vsth_getprio: using tetexec.cfg values");
	}

	if ( policy == SCHED_FIFO )
	{
		mincv = "VSTH_FIFO_MIN_PRIORITY";
		maxcv = "VSTH_FIFO_MAX_PRIORITY";
	}
	else if ( policy == SCHED_RR )
	{
		mincv = "VSTH_RR_MIN_PRIORITY";
		maxcv = "VSTH_RR_MAX_PRIORITY";
	}
	else if ( policy == SCHED_OTHER )
	{
		mincv = "VSTH_OTHER_MIN_PRIORITY";
		maxcv = "VSTH_OTHER_MAX_PRIORITY";
	}
	if (( mincv != NULL ) && ( maxcv != NULL ))
	{
		p = tet_getvar(mincv);
		if ( p != NULL )
		{
			if ( min != NULL )
			{
				*min = atoi(p);
			}
		}
		else
		{
			vsth_badconfig(mincv, errors);
		}

		p = tet_getvar(maxcv);
		if ( p != NULL )
		{
			if ( max != NULL )
			{
				*max = atoi(p);
			}
		}
		else
		{
			vsth_badconfig(maxcv, errors);
		}
	}

#endif	/* _POSIX_PRIORITY_SCHEDULING || SCHEDBUGS */

	if ( Vsth_debug )
	{

		char	buf[256];

		if ( policy == SCHED_RR )
			p = "SCHED_RR";
		else if ( policy == SCHED_FIFO )
			p = "SCHED_FIFO";
		else if ( policy == SCHED_OTHER )
			p = "SCHED_OTHER";
		else
			p = "UNKNOWN";

		(void) memset((void *) &buf[0], 0, 256);
		(void) sprintf(&buf[0], "vsth_getprio: %s policy, ", p);
		if ( min )
		{
			(void) sprintf(&buf[0] + (int) strlen(buf), "priomin == %d", *min);
		}
		if ( max )
		{
			(void) sprintf(&buf[0] + (int) strlen(buf), "priomax == %d", *max);
		}

		tet_printf("%s", &buf[0]);
	}

	return;
}

int
vsth_newthread(	int		policy,
		int		priority,
		int		scope,
		int		detachstate,
		void		*(*func)(void *),
		void		*arg,
		pthread_t	*ptid,
		int		*errors,
		int		*fails )
{
	int			rval;
	struct sched_param	params;
	pthread_attr_t		attr;

	params.sched_priority = priority;

	rval = pthread_attr_init( &attr );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_attr_init", 0, rval, errors);
		return(rval);
	}

#ifdef	_POSIX_THREAD_PRIORITY_SCHEDULING

	rval = pthread_attr_setinheritsched( &attr, PTHREAD_EXPLICIT_SCHED );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_attr_setinheritsched", 0, rval, errors);
		(void) pthread_attr_destroy( &attr );
		return(rval);
	}

	rval = pthread_attr_setschedpolicy( &attr, policy );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_attr_setschedpolicy", 0, rval, errors);
		(void) pthread_attr_destroy( &attr );
		return(rval);
	}

	rval = pthread_attr_setscope( &attr, scope );
	if ( rval != 0 )
	{
		if ( rval == ENOTSUP )
		{
			int	newscope;

			newscope = ( scope == PTHREAD_SCOPE_SYSTEM ? PTHREAD_SCOPE_PROCESS : PTHREAD_SCOPE_SYSTEM );

			tet_printf("The system does not support the %s scope, using %s",
				scope == PTHREAD_SCOPE_SYSTEM ? "PTHREAD_SCOPE_SYSTEM" : "PTHREAD_SCOPE_PROCESS",
				newscope == PTHREAD_SCOPE_SYSTEM ? "PTHREAD_SCOPE_SYSTEM" : "PTHREAD_SCOPE_PROCESS" );

			rval = pthread_attr_setscope( &attr, newscope );
		}

		if ( rval )
		{
			vsth_surprise("pthread_attr_setscope", 0, rval, errors);
			(void) pthread_attr_destroy( &attr );
			return(rval);
		}
	}

#endif	/* _POSIX_THREAD_PRIORITY_SCHEDULING */

	rval = pthread_attr_setdetachstate( &attr, detachstate );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_attr_setdetachstate", 0, rval, errors);
		(void) pthread_attr_destroy( &attr );
		return(rval);
	}

	rval = pthread_attr_setschedparam( &attr, &params );
	if ( rval != 0 )
	{
		vsth_surprise("pthread_attr_setschedparam", 0, rval, errors);
		(void) pthread_attr_destroy( &attr );
		return(rval);
	}

	rval = vsth_pthread_create( ptid, &attr, func, arg );
	if ( rval != 0 )
	{
		vsth_surprise("vsth_pthread_create", 0, rval, errors);
		(void) pthread_attr_destroy( &attr );
		return(rval);
	}

	(void) pthread_attr_destroy( &attr );
	return(rval);
}

int
vsth_spa_init( VSTH_PSP psp, int nsp )
{
	int		rval;
	int		cnt;
	VSTH_PSP	p;

	rval	= 1;
	p	= psp;

	if ( psp && ( nsp > 0 ))
	{
		for ( cnt = 0; cnt < nsp; ++cnt )
		{
			rval = vsth_sp_init( p++ );
			if ( ( rval != 0 ) && cnt )
			{
				vsth_spa_destroy( psp, cnt );
			}
			rval = 0;
		}
	}
	return(rval);
}

int
vsth_spa_destroy( VSTH_PSP psp, int nsp )
{
	int		rval;
	int		cnt;

	rval	= 1;

	for ( cnt = 0; cnt < nsp; ++cnt )
	{
		rval += vsth_sp_destroy( psp++ );
	}

	return(rval);
}

int
vsth_ts_create(	VSTH_PTSET pts, int nthreads, int syncflag, int *errors, int *fails )
{
	int		rval;
	int		nerrs;
	int		cnt;

	nerrs	= 0;

	if ( Vsth_debug )
	{
		tet_printf("vsth_ts_create enter: nthreads %d", nthreads);
	}

	for ( cnt = 0; cnt < nthreads; ++cnt )
	{
		if ( syncflag == VSTH_SYNC )
		{
			rval = pthread_mutex_lock( &pts->sp.m );
			if ( rval != 0 )
			{
				vsth_surprise("pthread_mutex_lock", 0, rval, &nerrs);
			}
		}
		pts->status = vsth_newthread(	pts->policy,
						pts->priority,
						pts->scope,
						pts->detachstate,
						pts->func,
						pts->arg,
						&pts->tid,
						errors,
						fails );
		if ( pts->status )
		{
			vsth_surprise("vsth_newthread", 0, pts->status, &nerrs);
		}

		if ( syncflag == VSTH_SYNC )
		{
			rval = vsth_sp_wait( &pts->sp, (long) pts->tid );
			if ( rval != 0 )
			{
				vsth_surprise("vsth_sp_wait", 0, rval, &nerrs);
			}
		}
		++pts;
	}

	return( nerrs );
}

void
vsth_ts_cancel( VSTH_PTSET pts, int nthreads, int detachflag )
{
	int	cnt;

	for ( cnt = 0; cnt < nthreads; ++cnt )
	{
		if ( pts->status == 0 )
		{
			(void) pthread_cancel( pts->tid );
			if ( detachflag )
			{
				(void) vsth_pthread_detach( pts->tid );
				pts->status = -1;
			}
		}
		++pts;
	}
}

int
vsth_ts_join( VSTH_PTSET pts, int nthreads, int *errors, int *fails )
{
	int	rval;
	int	cnt;
	int	val;

	rval	= 0;

	for ( cnt = 0; cnt < nthreads; ++cnt, ++pts )
	{
		if ( pts->detachstate != PTHREAD_CREATE_JOINABLE )
			continue;

		if ( pts->status == 0 )
		{
			val = vsth_pthread_join( pts->tid, &pts->exitval );
			if ( val == 0 )
			{
				if ( pts->exitval == (void *) 1 )
				{
					++*errors;
				}
				else if ( pts->exitval == (void *) 2 )
				{
					++*fails;
				}
			}
			else
			{
				(void) pthread_cancel( pts->tid );
				(void) vsth_pthread_detach( pts->tid );
				rval += val;
			}
		}
		else
		{
			++*errors;
			if ( Vsth_debug )
			{
				tet_printf("Bad status %d", pts->status);
			}
		}
	}

	return( rval );
}

vsth_ts_sync( VSTH_PTSET pts, int nthreads, int *errors, int *fails )
{
	int		rval;
	int		cnt;

	rval	= 0;

	for ( cnt = 0; cnt < nthreads; ++cnt )
	{
		rval = pthread_mutex_lock( &pts->sp.m );
		if ( rval == 0 )
		{
			rval = vsth_sp_wait( &pts->sp, (long) pts->tid );
		}
		++pts;
	}
	return(rval);
}

void
vsth_weave(	VSTH_PTSET	ptset,
		int		nthreads,
		void		*(*func)(void *),
		pthread_mutex_t	*pmx,
		pthread_cond_t	*pcv,
		int		policy,
		int		prio,
		int		prioflag,
		int		*errors,
		int		*fails )
{
	int		rval;
	int		cnt;
	VSTH_PTSET	pts;

	if ( pcv )
	{
		rval = pthread_cond_init( pcv, NULL );
		if ( rval != 0 )
		{
			vsth_surprise("pthread_cond_init", 0, rval, errors);
			return;
		}
	}

	if ( pmx )
	{
		rval = pthread_mutex_init( pmx, NULL );
		if ( rval != 0 )
		{
			vsth_surprise("pthread_mutex_init", 0, rval, errors);
			if ( pcv )
			{
				(void) pthread_cond_destroy( pcv );
			}
			return;
		}
	}

	for ( cnt = 0, pts = ptset; cnt < nthreads; ++cnt, ++pts )
	{
		pts->policy	= policy;

		if ( prioflag == 0 )
			pts->priority	= prio;
		else if ( prioflag < 0 )
			pts->priority	= --prio;
		else if ( prioflag > 0 )
			pts->priority	= ++prio;

		pts->scope	= PTHREAD_SCOPE_PROCESS;
		pts->detachstate= PTHREAD_CREATE_JOINABLE;
		pts->func	= func;
		pts->arg	= (void *) cnt;
		pts->status	= -1;
	}

	rval = vsth_ts_create( ptset, nthreads, VSTH_NOSYNC, errors, fails );
	if ( rval != 0 )
	{
		vsth_surprise("vsth_ts_create", 0, rval, errors);
	}

	rval = vsth_ts_join( ptset, nthreads, errors, fails );
	if ( rval != 0 )
	{
		vsth_surprise("vsth_ts_join", 0, rval, errors);
	}

	if ( pmx )
	{
		(void) pthread_mutex_destroy( pmx );
	}
	if ( pcv )
	{
		(void) pthread_cond_destroy( pcv );
	}
}


/* return 0 if flag becomes >= val within a timeout of tout seconds, else 1 */
int
vsth_spinwait( pthread_mutex_t *pmx, int *flag, long val, int tout, int *errors )
{
	int	rval;
	int	timedout;
	int	breakflag;
	time_t	endtime;

	rval		= 0;
	timedout	= 0;
	endtime 	= time(NULL) + tout;

	breakflag = 0;
	while ( ! breakflag )
	{
		rval += pthread_mutex_lock( pmx );

		if ( *flag >= val )
			breakflag = 1;

		rval += pthread_mutex_unlock( pmx );

		if ( time(NULL) > endtime )
		{
			breakflag = 1;
			timedout = 1;
		}
	}
	if (( rval != 0 ) && errors )
		++*errors;
	return(timedout);
}


/*	Common startup function for all tests.
	Some implementations of threads (notably linuxthreads from
	Xavier Leroy) use signals. Unless we disable the tcm
	signal handling feature, tcc reports unexpected signals
	and interrupts normal test case execution.		******/
void
vsth_startup(char *testname)
{
	char *s;


	s = tet_getvar("VSTH_SIG_IGN");
	if ( s != NULL )
	{
		int		signum;
		int		breakflag;
		char		*p;
		char		*q;
		struct sigaction act;

		sigfillset(&act.sa_mask);
		act.sa_handler	= SIG_IGN;
		act.sa_flags	= 0;

		p = s;
		q = s;
		breakflag = 0;
		while ( ! breakflag )
		{
			while ( *q && isdigit(*q) )
				++q;
			if (! *q )
				breakflag = 1;
			*q = '\0';
			signum = atoi(p);
			if ( signum > 0 )
			{
				tet_printf(
					"Setting signal action for signal %d",
					signum );
				(void) sigaction(signum, &act, NULL);
			}


			p = q + 1;
			q = p;
		}
		tet_nosigreset = 1;
	}

	s = tet_getvar("VSTH_DEBUG");
	if ( s != NULL )
	{
		if ( isdigit(*s) )
		{
			Vsth_debug = atoi(s);
		}
		else if (( *s == 'Y' ) || ( *s == 'y' ))
		{
			Vsth_debug = 1;
		}
	}

	s = tet_getvar("VSTH_TIMEOUT");
	if (( s != NULL ) && ( isdigit(*s) ))
	{
		Vsth_timeout = atoi(s);
	}

	if ( Vsth_debug )
	{
		tet_printf("Vsth_timeout is %d", Vsth_timeout);
	}


	/* to improve concurrency, if available */
	vsth_setconcurrency();

}

/******	Common cleanup function for all VSTH tests.
	Any cleanup code common to all tests belongs here	******/
void
vsth_cleanup(char *testname)
{
}

int
vsth_bwait( pthread_cond_t *pcv, pthread_mutex_t *pmx, int *flag )
{
	int	rval;
	int	nerrs;

	nerrs	= 0;

	rval = pthread_mutex_lock( pmx );
	if ( rval == 0 )
	{
		--*flag;

		if ( *flag )
		{
			while ( *flag )
			{
				rval = pthread_cond_wait( pcv, pmx );
				if ( rval != 0 )
				{
					vsth_surprise("pthread_cond_wait", 0, rval, &nerrs);
				}
			}
		}
		else
		{
			/* last thread in wakes the rest */
			rval = pthread_cond_broadcast( pcv );
			if ( rval != 0 )
			{
				vsth_surprise("pthread_cond_broadcast", 0, rval, &nerrs);
			}
		}

		rval = pthread_mutex_unlock( pmx );
		if ( rval != 0 )
		{
			vsth_surprise("pthread_mutex_unlock", 0, rval, &nerrs);
		}
	}

	rval = 1;
	if ( nerrs == 0 )
		rval = 0;

	return(rval);
}

/* the null thread - exit with arg */
void *
vsth_null_thread( void *arg )
{
	pthread_exit(arg);

	/* NOTREACHED */
	return(NULL);
}

/* a stub for implementations which don't yet support sched_yield */
int
vsth_sched_yield()
{
	int	rval;

#ifdef	NOYIELD

#ifdef	_POSIX_THREAD_PRIORITY_SCHEDULING

	int			policy;
	struct sched_param	params;

	rval = pthread_getschedparam( pthread_self(), &policy, &params );
	if ( rval == 0 )
	{
		rval = pthread_setschedparam( pthread_self(), policy, &params );
		if (  rval != 0 )
		{
			vsth_surprise("pthread_setschedparam", 0, rval, NULL);
		}
	}
	else
	{
		vsth_surprise("pthread_getschedparam", 0, rval, NULL);
	}
#else

	rval = 0;

	(void) sleep(1);

#endif

#else

	rval = sched_yield();

#endif	/* ! NOYIELD */

	return(rval);
}

