/*
*      SCCS:   @(#)common/vtools/hdrallow.c	4.4 (07/07/98)  VSXgen release 1.4
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/

/***********************************************************************

NAME:		hdrallow.c - rules for allowed symbols in headers

PROJECT:	VSX (X/OPEN Validation Suite)
MODULE CONTAINS:
		main()
SYNOPSIS:
		extern char *headerfile;

		int hdrallow(name, macrodef)
			char *name;
			int macrodef;

DESCRIPTION:
	Hdrallow() returns a non-zero value if the symbols "name"
	is allowed to be visible in the header file specified by
	the global variable "headerfile" by virtue of a special
	rule.

	It is used by the hdranal(vtools) and hdrdefs(vtools) programs.

AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	01 Feb 1990
MODIFICATIONS:
		Geoff Clare, 13 Nov 1991
			Add reserved classes for glob.h, regex.h,
			wchar.h and wordexp.h

		Geoff Clare, 21 Jan 1992
			Align with XSH4 D10 p45.

		Geoff Clare, 16 Apr 1992
			Align with XSH4 CoRev draft + resolved objections.

		Geoff Clare, 19 May 1992
			Allow "ipc_.*" in <sys/ipc.h>
			Allow symbols from <sys/ipc.h> in <sys/msg.h>, etc.

		Geoff Clare, 27 Aug 1992
			Align with interpretation VWG/009/063092.

		Geoff Clare, 8 Dec 1993
			Allow <limits.h> symbols in <wchar.h>

		Geoff Clare, 13 Feb 1995
			Distinguish between macros and non-macros in
			reserved prefix allowances.

		Geoff Clare, 13 Jan 1997
			Added support for UNIX95, UNIX98 and POSIX96 modes.
			Move body of hdrallow() into a new recursive
			function, chkallow().

		Geoff Clare, 1 July 1997
			Added dlfcn.h RTLD_ prefix (for XSH5 corrigenda)

		Geoff Clare, 6 Feb 1998
			Allow uc_ in <signal.h> and ss_ in <ucontext.h>
			in UNIX98 mode (for XSH5 corrigenda).

		Geoff Clare, 7 July 1998
			Allow SIG[A-Z].* in <signal.h> for POSIX modes (as
			macros only)

************************************************************************/

#include <std.h>
#include <string.h>
#include <ctype.h>
#include <sysdep.h>

#define MAXDEPTH 20
#define CHKPREFIX(name, pfx) (strncmp(name, pfx, sizeof(pfx)-1) == 0)

extern char *headerfile;

private int chkallow();

int
hdrallow(name, macrodef)
char *name;
int macrodef;
{
	return chkallow(headerfile, name, macrodef, 0);
}

private int
chkallow(hdrname, name, macrodef, depth)
char *hdrname, *name;
int macrodef, depth;
{
	/* No DBUG routines used because hdranal program doesn't
	   initialise debugging (it is mainly GNU code) */

	char *endp;

	/* For identifiers with file scope, all names beginning with "_" are
	   allowed.  For macros (macrodef != 0) only names beginning with
	   "_" and an upper case letter or another "_" are allowed. */

	if (macrodef)
	{
		if (name[0] == '_' && (name[1] == '_' || isupper(name[1])))
			return 1;
	}
	else
	{
		if (name[0] == '_')
			return 1;
	}

	/* All symbols ending with "_t" are allowed */

	for (endp = name; *endp != '\0'; endp++)
		;

	if (endp - name > 2 && endp[-2] == '_' && endp[-1] == 't')
		return 1;

	/* Individual header files allow classes of names */

	/* There are recursive calls where headers allow symbols from
	   other headers, so guard against infinite recursion */

	if (depth > MAXDEPTH)
		return 0;
	else if (strcmp(hdrname, "aio.h") == 0)
	{
		/* Allow "[al]io_.*" */

		if ((name[0] == 'a' || name[0] == 'l') &&
		    name[1] == 'i' && name[2] == 'o' && name[3] == '_')
			return 1;

		/* Allow macro "[AL]IO_.*" */

		if (macrodef && (name[0] == 'A' || name[0] == 'L') &&
		    name[1] == 'I' && name[2] == 'O' && name[3] == '_')
			return 1;
#if TEST_XPG >= 500
		/* Allow (any) "[AL]IO_.*" */

		if ((name[0] == 'A' || name[0] == 'L') &&
		    name[1] == 'I' && name[2] == 'O' && name[3] == '_')
			return 1;

		/* Allow anything from <fcntl.h>, <signal.h>,
		   <sys/types.h> and <time.h> */

		if (chkallow("fcntl.h", name, macrodef, depth+1) ||
		    chkallow("signal.h", name, macrodef, depth+1) ||
		    chkallow("sys/types.h", name, macrodef, depth+1) ||
		    chkallow("time.h", name, macrodef, depth+1))
			return 1;
#endif
	}
#if TEST_XPG == 4 && TEST_UNIX
	else if (strcmp(hdrname, "arpa/inet.h") == 0)
	{
		/* Allow "in_.*" */

		if (name[0] == 'i' && name[1] == 'n' && name[2] == '_')
			return 1;

		/* Allow "inet_.*" */

		if (name[0] == 'i' && name[1] == 'n' && name[2] == 'e' &&
		    name[3] == 't' && name[4] == '_')
			return 1;
	}
#endif
	else if (strcmp(hdrname, "ctype.h") == 0)
	{
		/* Allow "(is|to)[[:lower:]].*" */

		if ((name[0] == 'i' && name[1] == 's') ||
		    (name[0] == 't' && name[1] == 'o'))
		{
			if (islower(name[2]))
				return 1;
		}
	}
	else if (strcmp(hdrname, "dirent.h") == 0)
	{
		/* Allow "d_.*" */

		if (name[0] == 'd' && name[1] == '_')
			return 1;
	}
	else if (strcmp(hdrname, "dlfcn.h") == 0)
	{
		/* Allow macro "RTLD_.*" */

		if (macrodef && CHKPREFIX(name, "RTLD_"))
			return 1;
	}
	else if (strcmp(hdrname, "errno.h") == 0)
	{
#ifdef _XOPEN_SOURCE
		/* Allow "E.*" */

		if (name[0] == 'E')
			return 1;
#else
		/* Allow macro "E[[:digit:][:upper:]].*" */

		if (macrodef && name[0] == 'E')
		{
			if (isdigit(name[1]) || isupper(name[1]))
				return 1;
		}
#endif
	}
	else if (strcmp(hdrname, "fcntl.h") == 0)
	{
		/* Allow "l_.*" */

		if (name[0] == 'l' && name[1] == '_')
			return 1;

		/* Allow macro "[FOS]_.*" */

		if (macrodef && (name[0] == 'F' || name[0] == 'O' ||
		     name[0] == 'S') && name[1] == '_')
			return 1;

#ifdef _XOPEN_SOURCE
		/* Allow anything from <sys/stat.h> and <unistd.h> */

		if (chkallow("sys/stat.h", name, macrodef, depth+1) ||
		    chkallow("unistd.h", name, macrodef, depth+1))
			return 1;
#endif
#if TEST_XPG >= 500
		/* Allow anything from <sys/types.h> */

		if (chkallow("sys/types.h", name, macrodef, depth+1))
			return 1;
#endif
	}
	else if (strcmp(hdrname, "fmtmsg.h") == 0)
	{
		/* Allow macro "MM_.*" */

		if (macrodef &&
		    name[0] == 'M' && name[1] == 'M' && name[2] == '_')
			return 1;
	}
	else if (strcmp(hdrname, "fnmatch.h") == 0)
	{
		/* Allow macro "FNM_.*" */

		if (macrodef && name[0] == 'F' && name[1] == 'N' &&
		    name[2] == 'M' && name[3] == '_')
			return 1;
	}
	else if (strcmp(hdrname, "ftw.h") == 0)
	{
		/* Allow macro "FTW.*" */

		if (macrodef && name[0] == 'F' && name[1] == 'T' &&
		    name[2] == 'W')
			return 1;

		/* Allow anything from <sys/stat.h> */

		if (chkallow("sys/stat.h", name, macrodef, depth+1))
			return 1;
	}
	else if (strcmp(hdrname, "glob.h") == 0)
	{
		/* Allow macro "GLOB_.*" */

		if (macrodef && name[0] == 'G' && name[1] == 'L' &&
		    name[2] == 'O' && name[3] == 'B' && name[4] == '_')
			return 1;

		/* Allow "gl_.*" */

		if (name[0] == 'g' && name[1] == 'l' && name[2] == '_')
			return 1;
	}
	else if (strcmp(hdrname, "grp.h") == 0)
	{
		/* Allow "gr_.*" */

		if (name[0] == 'g' && name[1] == 'r' && name[2] == '_')
			return 1;
	}
	else if (strcmp(hdrname, "langinfo.h") == 0)
	{
#ifdef _XOPEN_SOURCE
		/* Allow anything from <nl_types> */

		if (chkallow("nl_types.h", name, macrodef, depth+1))
			return 1;
#endif
	}
	else if (strcmp(hdrname, "limits.h") == 0)
	{
		/* Allow ".*_MAX" */

		if (endp - name > 4 && endp[-4] == '_' && endp[-3] == 'M'
				    && endp[-2] == 'A' && endp[-1] == 'X')
			return 1;
	}
	else if (strcmp(hdrname, "locale.h") == 0)
	{
		/* Allow (macro only, in POSIX modes) "LC_[[:upper:]].*" */

		if (
#ifndef _XOPEN_SOURCE
		    macrodef &&
#endif
		    name[0] == 'L' && name[1] == 'C' && name[2] == '_' &&
		    isupper(name[3]))
			return 1;
	}
	else if (strcmp(hdrname, "mqueue.h") == 0)
	{
		/* Allow "mq_.*" */

		if (name[0] == 'm' && name[1] == 'q' && name[2] == '_')
			return 1;

		/* Allow macro "MQ_.*" */

		if (macrodef &&
		    name[0] == 'M' && name[1] == 'Q' && name[2] == '_')
			return 1;

#if TEST_XPG >= 500
		/* Allow (any) "MQ_.*" */

		if (name[0] == 'M' && name[1] == 'Q' && name[2] == '_')
			return 1;

		/* Allow anything from <fcntl.h>, <signal.h>,
		   <sys/types.h> and <time.h> */

		if (chkallow("fcntl.h", name, macrodef, depth+1) ||
		    chkallow("signal.h", name, macrodef, depth+1) ||
		    chkallow("sys/types.h", name, macrodef, depth+1) ||
		    chkallow("time.h", name, macrodef, depth+1))
			return 1;
#endif
	}
	else if (strcmp(hdrname, "ndbm.h") == 0)
	{
		/* Allow "dbm_.*" */

		if (name[0] == 'd' && name[1] == 'b' && name[2] == 'm' &&
		    name[3] == '_')
			return 1;

		/* Allow macro "DBM_.*" */

		if (macrodef &&
		    name[0] == 'D' && name[1] == 'B' && name[2] == 'M' &&
		    name[3] == '_')
			return 1;
	}
#if TEST_UNIX && TEST_XPG == 4
	else if (strcmp(hdrname, "netdb.h") == 0)
	{
		/* Allow "[hnps]_.*" */

		if ((name[0] == 'h' || name[0] == 'n' || name[0] == 'p' ||
		     name[0] == 's') && name[1] == '_')
			return 1;

		/* Allow macro "NO_.*" */

		if (macrodef &&
		    name[0] == 'N' && name[1] == 'O' && name[2] == '_')
			return 1;
	}
#endif
#if TEST_UNIX && TEST_XPG == 4
	else if (strcmp(hdrname, "netinet/in.h") == 0)
	{
		/* Allow "in_.*", "s_.*" and "sin_.*" */

		if ((name[0] == 'i' && name[1] == 'n' && name[2] == '_') ||
		    (name[0] == 's' && name[1] == '_') ||
		    (name[0] == 's' && name[1] == 'i' &&
				name[2] == 'n' && name[3] == '_'))
			return 1;

		/* Allow macro "I[NP]_.*" */

		if (macrodef && name[0] == 'I' &&
		    (name[1] == 'N' || name[1] == 'P') && name[2] == '_')
			return 1;

		/* Allow macro "IMPLINK_.*", "INADDR_.*", "IPPORT_.*",
		   "IPPROTO_.*" and "SOCK_.*" */

		if (macrodef &&
		    (CHKPREFIX(name, "IMPLINK_") ||
		     CHKPREFIX(name, "INADDR_") ||
		     CHKPREFIX(name, "IPPORT_") ||
		     CHKPREFIX(name, "IPPROTO_") ||
		     CHKPREFIX(name, "SOCK_")))
			return 1;
	}
#endif
	else if (strcmp(hdrname, "nl_types.h") == 0)
	{
		/* Allow macro "NL_.*" */

		if (macrodef &&
		    name[0] == 'N' && name[1] == 'L' && name[2] == '_')
			return 1;
	}
	else if (strcmp(hdrname, "poll.h") == 0)
	{
		/* Allow "p[dhs]_.*" */

		if (name[0] == 'p' &&
		    (name[1] == 'd' || name[1] == 'h' || name[1] == 's') &&
		    name[2] == '_')
			return 1;

		/* Allow macro "POLL.*" */

		if (macrodef && name[0] == 'P' && name[1] == 'O' &&
		    name[2] == 'L' && name[3] == 'L')
			return 1;
	}
	else if (strcmp(hdrname, "pthread.h") == 0)
	{
		/* Allow "pthread_.*" */

		if (CHKPREFIX(name, "pthread_"))
			return 1;

		/* Allow macro "PTHREAD_.*" */

		if (macrodef && CHKPREFIX(name, "PTHREAD_"))
			return 1;

#if TEST_XPG >= 500
		/* Allow (any) "PTHREAD_.*" */

		if (CHKPREFIX(name, "PTHREAD_"))
			return 1;

		/* Allow anything from <sched.h>
		   (and <time.h> - via <sched.h>) */

		if (chkallow("sched.h", name, macrodef, depth+1))
			return 1;
#endif
	}
	else if (strcmp(hdrname, "pwd.h") == 0)
	{
		/* Allow "pw_.*" */

		if (name[0] == 'p' && name[1] == 'w' && name[2] == '_')
			return 1;
	}
#if TEST_UNIX && TEST_XPG == 4
	else if (strcmp(hdrname, "re_comp.h") == 0)
	{
		/* Allow macro "REG_.*" */

		if (macrodef && name[0] == 'R' && name[1] == 'E' &&
		    name[2] == 'G' && name[3] == '_')
			return 1;
		
		/* Allow "re_.*" */

		if (name[0] == 'r' && name[1] == 'e' && name[2] == '_')
			return 1;
	}
#endif
	else if (strcmp(hdrname, "regex.h") == 0)
	{
		/* Allow macro "REG_.*" */

		if (macrodef && name[0] == 'R' && name[1] == 'E' &&
		    name[2] == 'G' && name[3] == '_')
			return 1;
		
		/* Allow "re_.*" and "rm_.*" */

		if (name[0] == 'r' && name[1] == 'e' && name[2] == '_')
			return 1;
		if (name[0] == 'r' && name[1] == 'm' && name[2] == '_')
			return 1;
	}
	else if (strcmp(hdrname, "sched.h") == 0)
	{
		/* Allow "sched_.*" */

		if (CHKPREFIX(name, "sched_"))
			return 1;

		/* Allow macro "SCHED_.*" */

		if (macrodef && CHKPREFIX(name, "SCHED_"))
			return 1;

		/* Allow anything from <time.h> */

		if (chkallow("time.h", name, macrodef, depth+1))
			return 1;

#if TEST_XPG >= 500
		/* Allow (any) "SCHED_.*" */

		if (CHKPREFIX(name, "SCHED_"))
			return 1;
#endif
	}
	else if (strcmp(hdrname, "semaphore.h") == 0)
	{
		/* Allow "sem_.*" */

		if (name[0] == 's' && name[1] == 'e' && name[2] == 'm' &&
		    name[3] == '_')
			return 1;

		/* Allow macro "SEM_.*" */

		if (macrodef && name[0] == 'S' && name[1] == 'E' &&
		    name[2] == 'M' && name[3] == '_')
			return 1;

#if TEST_XPG >= 500
		/* Allow (any) "SEM_.*" */

		if (name[0] == 'S' && name[1] == 'E' && name[2] == 'M' &&
		    name[3] == '_')
			return 1;

		/* Allow anything from <fcntl.h> and <sys/types.h> */

		if (chkallow("fcntl.h", name, macrodef, depth+1) ||
		    chkallow("sys/types.h", name, macrodef, depth+1))
			return 1;
#endif
	}
	else if (strcmp(hdrname, "signal.h") == 0)
	{
		/* Allow macro "SIG_.*" */

		if (macrodef && name[0] == 'S' && name[1] == 'I' &&
		    name[2] == 'G' && name[3] == '_')
			return 1;

		/* Allow "sa_.*" and macro "SA_.*" */

		if ((name[0] == 'S' && name[1] == 'A' && macrodef ||
		     name[0] == 's' && name[1] == 'a') && name[2] == '_')
			return 1;

		/* Allow (macro only, in POSIX modes) "SIG[A-Z].*" */

		if (
#ifndef _XOPEN_SOURCE
		    macrodef &&
#endif
		    name[0] == 'S' && name[1] == 'I' && name[2] == 'G' &&
		    isupper(name[3]))
			return 1;

#ifdef _XOPEN_SOURCE

		/* Allow "SIG_[A-Z].*" */

		if (name[0] == 'S' && name[1] == 'I' && name[2] == 'G' &&
		    name[3] == '_' && isupper(name[4]))
			return 1;
#endif
#if _POSIX_C_SOURCE >= 199506 || TEST_UNIX

		/* Allow "si_.*" */

		if (name[0] == 's' && name[1] == 'i' && name[2] == '_')
			return 1;

		/* Allow macro "SI_.*" */

		if (macrodef &&
		    name[0] == 'S' && name[1] == 'I' && name[2] == '_')
			return 1;
#endif
#if TEST_UNIX
		/* Allow "s[sv]_.*" */

		if (name[0] == 's' && (name[1] == 's' || name[1] == 'v') &&
		    name[2] == '_')
			return 1;

		/* Allow macro "S[SV]_.*" */

		if (macrodef &&
		    name[0] == 'S' && (name[1] == 'S' || name[1] == 'V') &&
		    name[2] == '_')
			return 1;

		/* Allow macro "BUS_.*", "CLD_.*", "FPE_.*" and "ILL_.*" */

		if (macrodef &&
		    ((name[0] == 'B' && name[1] == 'U' && name[2] == 'S') ||
		     (name[0] == 'C' && name[1] == 'L' && name[2] == 'D') ||
		     (name[0] == 'F' && name[1] == 'P' && name[2] == 'E') ||
		     (name[0] == 'I' && name[1] == 'L' && name[2] == 'L')) &&
		    name[3] == '_')
			return 1;

		/* Allow macro "POLL_.*", "SEGV_.*" and "TRAP_.*" */

		if (macrodef &&
		    ((name[0] == 'P' && name[1] == 'O' &&
		      name[2] == 'L' && name[3] == 'L') ||
		     (name[0] == 'S' && name[1] == 'E' &&
		      name[2] == 'G' && name[3] == 'V') ||
		     (name[0] == 'T' && name[1] == 'R' &&
		      name[2] == 'A' && name[3] == 'P')) &&
		    name[4] == '_')
			return 1;
#endif
#if _POSIX_C_SOURCE >= 199506 || TEST_XPG >= 500

		/* Allow "sigev_.*" and "sival_.*" */

		if (CHKPREFIX(name, "sigev_") || CHKPREFIX(name, "sival_"))
			return 1;

		/* Allow macro "SIGEV_.*" */

		if (macrodef && CHKPREFIX(name, "SIGEV_"))
			return 1;
#endif
#if TEST_XPG >= 500
		/* Allow (any) "SI_.*" */

		if (name[0] == 'S' && name[1] == 'I' && name[2] == '_')
			return 1;

		/* Allow (any) "SIGEV_.*" */

		if (CHKPREFIX(name, "SIGEV_"))
			return 1;

		/* Allow "uc_.*" */

		if (name[0] == 'u' && name[1] == 'c' && name[2] == '_')
			return 1;
#endif
	}
	else if (strcmp(hdrname, "stdio.h") == 0)
	{
#ifdef _XOPEN_SOURCE
		/* Allow anything from <stddef.h> */

		if (chkallow("stddef.h", name, macrodef, depth+1))
			return 1;
#endif
	}
	else if (strcmp(hdrname, "stdlib.h") == 0)
	{
		/* Allow "str[[:lower:]].*" */

		if (name[0] == 's' && name[1] == 't' && name[2] == 'r')
		{
			if (islower(name[3]))
				return 1;
		}

#ifdef _XOPEN_SOURCE
		/* Allow anything from <stddef.h>, <limits.h>,
		   <math.h> and <sys/wait.h> */

		if (chkallow("stddef.h", name, macrodef, depth+1) ||
		    chkallow("limits.h", name, macrodef, depth+1) ||
		    chkallow("math.h", name, macrodef, depth+1) ||
		    chkallow("sys/wait.h", name, macrodef, depth+1))
			return 1;
#endif
	}
	else if (strcmp(hdrname, "string.h") == 0)
	{
		/* Allow "(str|mem|wcs)[[:lower:]].*" */

		if ((name[0] == 's' && name[1] == 't' && name[2] == 'r') ||
		    (name[0] == 'm' && name[1] == 'e' && name[2] == 'm') ||
		    (name[0] == 'w' && name[1] == 'c' && name[2] == 's'))
		{
			if (islower(name[3]))
				return 1;
		}

#ifdef _XOPEN_SOURCE
		/* Allow anything from <stddef.h> */

		if (chkallow("stddef.h", name, macrodef, depth+1))
			return 1;
#endif
	}
	else if (strcmp(hdrname, "stropts.h") == 0)
	{
		/* Allow "bi_.*", "ic_.*", "l_.*", "sl_.*" and "str_.*" */

		if (name[0] == 'b' && name[1] == 'i' && name[2] == '_')
			return 1;
		if (name[0] == 'i' && name[1] == 'c' && name[2] == '_')
			return 1;
		if (name[0] == 'l' && name[1] == '_')
			return 1;
		if (name[0] == 's' && name[1] == 'l' && name[2] == '_')
			return 1;
		if (name[0] == 's' && name[1] == 't' && name[2] == 'r' &&
		    name[3] == '_')
			return 1;
		
		/* Allow macro "FLUSH[A-Z].*", "MUXID_R[A-Z].*"
		   and "SND[A-Z].*" */

		if (macrodef && name[0] == 'F' && name[1] == 'L' &&
		    name[2] == 'U' && name[3] == 'S' && name[4] == 'H' &&
		    isupper(name[5]))
			return 1;
		if (macrodef && name[0] == 'M' && name[1] == 'U' &&
		    name[2] == 'X' && name[3] == 'I' && name[4] == 'D' &&
		    name[5] == '_' && name[6] == 'R' &&
		    isupper(name[7]))
			return 1;
		if (macrodef && name[0] == 'S' && name[1] == 'N' &&
		    name[2] == 'D' && isupper(name[3]))
			return 1;

		/* Allow macro "[IMS]_.*" */

		if (macrodef &&
		    (name[0] == 'I' || name[0] == 'M' || name[0] == 'S') &&
		    name[1] == '_')
			return 1;

		/* Allow macro "STR.*" */

		if (macrodef &&
		    name[0] == 'S' && name[1] == 'T' && name[2] == 'R')
			return 1;

		/* Allow anything from <unistd.h> */

		if (chkallow("unistd.h", name, macrodef, depth+1))
			return 1;
	}
	else if (strcmp(hdrname, "syslog.h") == 0)
	{
		/* Allow macro "LOG_.*" */

		if (macrodef && name[0] == 'L' && name[1] == 'O' &&
		    name[2] == 'G' && name[3] == '_')
			return 1;
	}
	else if (strcmp(hdrname, "sys/ipc.h") == 0)
	{
		/* Allow macro "IPC_.*" */

		if (macrodef && name[0] == 'I' && name[1] == 'P' &&
		    name[2] == 'C' && name[3] == '_')
			return 1;

		/* Allow "ipc_.*" */

		if (name[0] == 'i' && name[1] == 'p' && name[2] == 'c' &&
		    name[3] == '_')
			return 1;
	}
	else if (strcmp(hdrname, "sys/mman.h") == 0)
	{
		/* Allow macro "PROT_.*", "MAP_.*" and "MS_.*" */

		if (macrodef &&
		    name[0] == 'P' && name[1] == 'R' && name[2] == 'O' &&
		    name[3] == 'T' && name[4] == '_')
			return 1;
		if (macrodef &&
		    name[0] == 'M' && name[1] == 'A' && name[2] == 'P' &&
		    name[3] == '_')
			return 1;
		if (macrodef &&
		    name[0] == 'M' && name[1] == 'S' && name[2] == '_')
			return 1;

#if _POSIX_C_SOURCE >= 199506 || TEST_XPG >= 500

		/* Allow "shm_.*" */

		if (name[0] == 's' && name[1] == 'h' && name[2] == 'm' &&
		    name[3] == '_')
			return 1;

		/* Allow macro "MCL_.*" */

		if (macrodef &&
		    name[0] == 'M' && name[1] == 'C' && name[2] == 'L' &&
		    name[3] == '_')
			return 1;
#endif
#if TEST_XPG >= 500
		/* Allow (any) "PROT_.*", "MAP_.*", "MS_.*" and "MCL_.*" */

		if (name[0] == 'P' && name[1] == 'R' && name[2] == 'O' &&
		    name[3] == 'T' && name[4] == '_')
			return 1;
		if (name[0] == 'M' && name[1] == 'A' && name[2] == 'P' &&
		    name[3] == '_')
			return 1;
		if (name[0] == 'M' && name[1] == 'S' && name[2] == '_')
			return 1;
		if (name[0] == 'M' && name[1] == 'C' && name[2] == 'L' &&
		    name[3] == '_')
			return 1;
#endif
	}
	else if (strcmp(hdrname, "sys/msg.h") == 0)
	{
#if TEST_UNIX
		/* Allow "msg.*" */

		if (name[0] == 'm' && name[1] == 's' && name[2] == 'g')
			return 1;
#else
		/* Allow "msg[_[:lower:]].*" */

		if (name[0] == 'm' && name[1] == 's' && name[2] == 'g' &&
		    (name[3] == '_' || islower(name[3])))
			return 1;
#endif

		/* Allow macro "MSG[[:upper:]].*" and "MSG_[[:upper:]].*" */

		if (macrodef && name[0] == 'M' && name[1] == 'S' &&
		    name[2] == 'G' && isupper(name[3]))
			return 1;

		if (macrodef && name[0] == 'M' && name[1] == 'S' &&
		    name[2] == 'G' && name[3] == '_' && isupper(name[4]))
			return 1;

		/* Allow anything from <sys/ipc.h> */

		if (chkallow("sys/ipc.h", name, macrodef, depth+1))
			return 1;
	}
	else if (strcmp(hdrname, "sys/resource.h") == 0)
	{
		/* Allow "rlim_.*" and "ru_.*" */

		if (name[0] == 'r' && name[1] == 'l' && name[2] == 'i' &&
		    name[3] == 'm' && name[4] == '_')
			return 1;
		if (name[0] == 'r' && name[1] == 'u' && name[2] == '_')
			return 1;

		/* Allow macro "PRIO_.*", "RLIM_.*", "RLIMIT_.*"
		   and "RUSAGE_.*" */

		if (macrodef &&
		    CHKPREFIX(name, "PRIO_") || CHKPREFIX(name, "RLIM_") ||
		    CHKPREFIX(name, "RLIMIT_") || CHKPREFIX(name, "RUSAGE_"))
			return 1;

		/* Allow anything from <sys/time.h> */

		if (chkallow("sys/time.h", name, macrodef, depth+1))
			return 1;
	}
	else if (strcmp(hdrname, "sys/sem.h") == 0)
	{
#if TEST_UNIX
		/* Allow "sem.*" */

		if (name[0] == 's' && name[1] == 'e' && name[2] == 'm')
			return 1;
#else
		/* Allow "sem[_[:lower:]].*" */

		if (name[0] == 's' && name[1] == 'e' && name[2] == 'm' &&
		    (name[3] == '_' || islower(name[3])))
			return 1;
#endif

		/* Allow macro "SEM_.*" */

		if (macrodef && name[0] == 'S' && name[1] == 'E' &&
		    name[2] == 'M' && name[3] == '_')
			return 1;

		/* Allow anything from <sys/ipc.h> */

		if (chkallow("sys/ipc.h", name, macrodef, depth+1))
			return 1;
	}
	else if (strcmp(hdrname, "sys/shm.h") == 0)
	{
#if TEST_UNIX
		/* Allow "shm.*" */

		if (name[0] == 's' && name[1] == 'h' && name[2] == 'm')
			return 1;
#else
		/* Allow "shm_.*" */

		if (name[0] == 's' && name[1] == 'h' && name[2] == 'm' &&
		    name[3] == '_')
			return 1;
#endif

		/* Allow macro "SHM[[:upper:]]" and "SHM_[[:upper:]].*" */

		if (macrodef && name[0] == 'S' && name[1] == 'H' &&
		    name[2] == 'M' && isupper(name[3]))
			return 1;

		if (macrodef && name[0] == 'S' && name[1] == 'H' &&
		    name[2] == 'M' && name[3] == '_' && isupper(name[4]))
			return 1;

		/* Allow anything from <sys/ipc.h> */

		if (chkallow("sys/ipc.h", name, macrodef, depth+1))
			return 1;
	}
	else if (strcmp(hdrname, "sys/socket.h") == 0)
	{
		/* Allow macro "AF_.*", "MSG_.*", "PF_.*" and "SO.*" */

		if (macrodef &&
		    name[0] == 'A' && name[1] == 'F' && name[2] == '_')
			return 1;
		if (macrodef &&
		    name[0] == 'M' && name[1] == 'S' && name[2] == 'G' &&
		    name[3] == '_')
			return 1;
		if (macrodef &&
		    name[0] == 'P' && name[1] == 'F' && name[2] == '_')
			return 1;
		if (macrodef &&
		    name[0] == 'S' && name[1] == 'O')
			return 1;

#if TEST_UNIX && TEST_XPG == 4

		/* Allow "sa_.*" and "if_.*" */

		if (((name[0] == 's' && name[1] == 'a') ||
		     (name[0] == 'i' && name[1] == 'f')) &&
		    name[2] == '_')
			return 1;

		/* Allow "ifc_.*" and "msg_.*" */

		if (((name[0] == 'i' && name[1] == 'f' && name[2] == 'c') ||
		     (name[0] == 'm' && name[1] == 's' && name[2] == 'g')) &&
		    name[3] == '_')
			return 1;

		/* Allow "ifru_.*", "infu_.*", "ifra_.*" and "cmsg_.*" */

		if (((name[0] == 'i' && name[1] == 'f' &&
		      name[2] == 'r' && name[3] == 'u') ||
		     (name[0] == 'i' && name[1] == 'n' &&
		      name[2] == 'f' && name[3] == 'u') ||
		     (name[0] == 'i' && name[1] == 'f' &&
		      name[2] == 'r' && name[3] == 'a') ||
		     (name[0] == 'c' && name[1] == 'm' &&
		      name[2] == 's' && name[3] == 'g')) &&
		    name[4] == '_')
			return 1;
#endif
	}
	else if (strcmp(hdrname, "sys/stat.h") == 0)
	{
		/* Allow "st_.*" and macro "S_.*" */

		if ((name[0] == 's' && name[1] == 't' && name[2] == '_') ||
		    (macrodef && name[0] == 'S' && name[1] == '_'))
			return 1;
	}
	else if (strcmp(hdrname, "sys/statvfs.h") == 0)
	{
		/* Allow "f_.*" and macro "ST_.*" */

		if ((name[0] == 'f' && name[1] == '_') ||
		    (macrodef && name[0] == 'S' && name[1] == 'T' &&
		     name[2] == '_'))
			return 1;
	}
	else if (strcmp(hdrname, "sys/time.h") == 0)
	{
		/* Allow "fds_.*", "it_.*", "tv_.*" and "FD_.*" */

		if (name[0] == 'f' && name[1] == 'd' && name[2] == 's' &&
		    name[3] == '_')
			return 1;
		if (name[0] == 'i' && name[1] == 't' && name[2] == '_')
			return 1;
		if (name[0] == 't' && name[1] == 'v' && name[2] == '_')
			return 1;
		if (name[0] == 'F' && name[1] == 'D' && name[2] == '_')
			return 1;

		/* Allow macro "ITIMER_.*" */

		if (macrodef && CHKPREFIX(name, "ITIMER_"))
			return 1;
	}
	else if (strcmp(hdrname, "sys/times.h") == 0)
	{
		/* Allow "tms_.*" */

		if (name[0] == 't' && name[1] == 'm' && name[2] == 's' &&
		    name[3] == '_')
			return 1;
	}
	else if (strcmp(hdrname, "sys/uio.h") == 0)
	{
		/* Allow "iov_.*" */

		if (name[0] == 'i' && name[1] == 'o' && name[2] == 'v' &&
		    name[3] == '_')
			return 1;

		/* Allow macro "IOV_.*" */

		if (macrodef &&
		    name[0] == 'I' && name[1] == 'O' && name[2] == 'V' &&
		    name[3] == '_')
			return 1;
	}
#if TEST_UNIX && TEST_XPG == 4
	else if (strcmp(hdrname, "sys/un.h") == 0)
	{
		/* Allow "sun_.*" */

		if (name[0] == 's' && name[1] == 'u' && name[2] == 'n' &&
		    name[3] == '_')
			return 1;
	}
#endif
	else if (strcmp(hdrname, "sys/utsname.h") == 0)
	{
#ifdef _XOPEN_SOURCE
		/* Allow "uts_.*" */

		if (name[0] == 'u' && name[1] == 't' && name[2] == 's' &&
		    name[3] == '_')
			return 1;
#endif
	}
	else if (strcmp(hdrname, "sys/wait.h") == 0)
	{
#if TEST_UNIX
		/* Allow "W[A-Z].*" */

		if (name[0] == 'W' && isupper(name[1]))
			return 1;

		/* Allow "si_.*" and macro "SI_.*" */

		if ((name[0] == 'S' && name[1] == 'I' && macrodef ||
		     name[0] == 's' && name[1] == 'i') && name[2] == '_')
			return 1;

		/* Allow "P_.*" */

		if (name[0] == 'P' && name[1] == '_')
			return 1;

		/* Allow anything from <signal.h> and <sys/resource.h> */

		if (chkallow("signal.h", name, macrodef, depth+1) ||
		    chkallow("sys/resource.h", name, macrodef, depth+1))
			return 1;
#endif
	}
	else if (strcmp(hdrname, "termios.h") == 0)
	{
		/* Allow "c_.*" and macro "TC.*" */

		if ((name[0] == 'c' && name[1] == '_') ||
		    (macrodef && name[0] == 'T' && name[1] == 'C'))
			return 1;

		/* Allow macro "[VIO].*" and "B[[:digit:]].*" */

		if (macrodef && (name[0] == 'V' || name[0] == 'I' ||
		    name[0] == 'O' || (name[0] == 'B' && isdigit(name[1]))))
			return 1;
	}
	else if (strcmp(hdrname, "time.h") == 0)
	{
#ifdef _XOPEN_SOURCE
		/* Allow "tm_.*" */

		if (name[0] == 't' && name[1] == 'm' && name[2] == '_')
			return 1;
#endif
#if _POSIX_C_SOURCE >= 199506 || TEST_XPG >= 500

		/* Allow "it_.*" and "tv_.*" */

		if (name[0] == 'i' && name[1] == 't' && name[2] == '_')
			return 1;
		if (name[0] == 't' && name[1] == 'v' && name[2] == '_')
			return 1;

		/* Allow "clock_.*" and "timer_.*" */

		if (CHKPREFIX(name, "clock_") || CHKPREFIX(name, "timer_"))
			return 1;

		/* Allow macro "CLOCK_.*" and "TIMER_.*" */

		if (macrodef &&
		    CHKPREFIX(name, "CLOCK_") || CHKPREFIX(name, "TIMER_"))
			return 1;
#endif
#if TEST_XPG >= 500
		/* Allow (any) "CLOCK_.*" and "TIMER_.*" */
		if (CHKPREFIX(name, "CLOCK_") || CHKPREFIX(name, "TIMER_"))
			return 1;
#endif
	}
	else if (strcmp(hdrname, "ucontext.h") == 0)
	{
		/* Allow "uc_.*" */

		if (name[0] == 'u' && name[1] == 'c' && name[2] == '_')
			return 1;

		/* Allow "ss_.*" */

		if (name[0] == 's' && name[1] == 's' && name[2] == '_')
			return 1;
	}
	else if (strcmp(hdrname, "ulimit.h") == 0)
	{
#if TEST_UNIX
		/* Allow "UL_.*" */

		if (name[0] == 'U' && name[1] == 'L' && name[2] == '_')
			return 1;
#endif
	}
	else if (strcmp(hdrname, "utime.h") == 0)
	{
#ifdef _XOPEN_SOURCE
		/* Allow "utim_.*" */

		if (name[0] == 'u' && name[1] == 't' && name[2] == 'i' &&
		    name[3] == 'm' && name[4] == '_')
			return 1;
#endif
	}
	else if (strcmp(hdrname, "utmpx.h") == 0)
	{
		/* Allow "ut_.*" */

		if (name[0] == 'u' && name[1] == 't' && name[2] == '_')
			return 1;

		/* Allow ".*_LVL", ".*_TIME" and ".*_PROCESS" */

		if (endp - name > 4 &&
		    endp[-4] == '_' && endp[-3] == 'L' &&
		    endp[-2] == 'V' && endp[-1] == 'L')
			return 1;
		if (endp - name > 5 &&
		    endp[-5] == '_' && endp[-4] == 'T' &&
		    endp[-3] == 'I' && endp[-2] == 'M' && endp[-1] == 'E')
			return 1;
		if (endp - name > 8 &&
		    endp[-8] == '_' && endp[-7] == 'P' &&
		    endp[-6] == 'R' && endp[-5] == 'O' &&
		    endp[-4] == 'C' && endp[-3] == 'E' &&
		    endp[-2] == 'S' && endp[-1] == 'S')
			return 1;

#if TEST_XPG >= 500
		/* Allow anything from <sys/time.h> */

		if (chkallow("sys/time.h", name, macrodef, depth+1))
			return 1;
#endif
	}
	else if (strcmp(hdrname, "wchar.h") == 0)
	{
		/* Allow anything from <ctype.h>, <stdio.h>, <stdarg.h>,
		   <stdlib.h>, <string.h>, <stddef.h> and <time.h> */

		if (chkallow("ctype.h", name, macrodef, depth+1) ||
		    chkallow("stdio.h", name, macrodef, depth+1) ||
		    chkallow("stdarg.h", name, macrodef, depth+1) ||
		    chkallow("stdlib.h", name, macrodef, depth+1) ||
		    chkallow("string.h", name, macrodef, depth+1) ||
		    chkallow("stddef.h", name, macrodef, depth+1) ||
		    chkallow("time.h", name, macrodef, depth+1))
			return 1;
#if TEST_XPG == 4
		/* Allow anything from <limits.h> */

		if (chkallow("limits.h", name, macrodef, depth+1))
			return 1;
#endif
	}
	else if (strcmp(hdrname, "wctype.h") == 0)
	{
#if TEST_XPG >= 500
		/* Allow anything from <wchar.h> */

		if (chkallow("wchar.h", name, macrodef, depth+1))
			return 1;
#endif
	}
	else if (strcmp(hdrname, "wordexp.h") == 0)
	{
		/* Allow macro "WRDE_.*" */

		if (macrodef && name[0] == 'W' && name[1] == 'R' &&
		    name[2] == 'D' && name[3] == 'E' && name[4] == '_')
			return 1;

		/* Allow "we_.*" */

		if (name[0] == 'w' && name[1] == 'e' && name[2] == '_')
			return 1;
	}

	return 0;
}
