/*
*      SCCS:   @(#)common/vtools/hdrdefs.c	4.3 (10/15/97)  VSXgen release 1.4
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/

static	char sccsid[] = "@(#)common/vtools/hdrdefs VSXgen release 1.4";
char	*copyright[] = {
		"(C) Copyright 1991 X/Open Company Limited",
		"All Rights Reserved."
};

/***********************************************************************

NAME:		hdrdefs.c - main entry point for hdrdefs(vtools)

PROJECT:	VSX (X/OPEN Validation Suite)
MODULE CONTAINS:
		main()
SYNOPSIS:
		hdrdefs headerfile [searchdir ...]

DESCRIPTION:
	Hdrdefs reads a system header file {header}.h and produces two
	output files: {header}.h.D and {header}.h.C for use in namespace
	pollution header tests.  {header}.h.D is a modified version
	of the header file with any #include's it makes expanded out.
	{header}.h.C contains a call to a check routine chkallow() for
	each #define'd symbol that is not in a reserved class of names,
	within an #ifdef/#endif.

	Hdrdefs follows #include lines in the header file, expanding
	them into the output.  If an include file cannot be found or
	opened a line of the form
		#define VSXdef<x> "error message"
	is placed at the start of the expanded #include file in
	{header}.h.D and a call to errallow(VSXdef<x>) is placed in
	{header}.h.C within a #ifdef VSXdef<x>.  This means an error
	will only be reported if the bad #include was not #ifdef'ed out
	in the header.  If no "missing" includes are encountered, then
	the expanded-out file is not needed, and is overwritten with
	a simple #include of {header}.h.

	The (void) routines chkallow() and errallow() must be defined
	before {header}.h.C is #include'd.

	If one or more "searchdir" arguments are given, they are searched
	for header files instead of /usr/include.

	If the environment variable HDRDEFS_NOEXPAND is set, hdrdefs
	will not attempt to write the expanded version of {header}.h.D
	at all.  In this case hdrdefs will report an error if it
	encounters a missing nested include.

RELATED DOCUMENTS:
	hdrdefs(vtools) - full manual entry

AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	23 Oct 1989
MODIFICATIONS:
		Geoff Clare, 01 Feb 1990
			Use new hdrallow() routine.

		Geoff Clare, 21 Feb 1990
			Guard against infinite recursion if header
			files include each other

		Geoff Clare, 1 May 1990
			Write comment line to output files

		Geoff Clare, 22 Aug 1990
			Handle multi-line comment starting on a #include line

		Geoff Clare, 15 July 1991
			Revise format of ".C" and ".D" files to cope with
			symbols being #undef'ed later in the header.
			Add predefs list.

		Geoff Clare, 27 Aug 1992
			Align with interpretation VWG/009/063092.

		Geoff Clare, 15 Oct 1997
			Expanded-out version of ".D" file not needed if
			no "missing" includes encountered.

		Geoff Clare, 7 April 2000
			Added support for #include_next

************************************************************************/

#include	<std.h>
#include	<stdlib.h>
#include	<stdio.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<unistd.h>
#include	<limits.h>
#include	<string.h>
#include	<dbug.h>
#include	<sysdep.h>

#define		LINELEN	512

extern	void	set_pname();	/* set program name		*/
extern	char	*get_pname();	/* get program name		*/

public	char	*headerfile;

/* symbols commonly pre-defined by compilers */
private	char	*predefs[] = { "unix", "gcos", "pdp11", "UNIX", NULL };

private	char	*defaultdirs[] = { "/usr/include", NULL };
private	char	**searchdirs = defaultdirs;
private	char	**currdir;
private char	Dfile[_POSIX_NAME_MAX+5];
private char	Cfile[_POSIX_NAME_MAX+5];
private FILE	*Dfp;
private FILE	*Cfp;
private	int	defcount = 0;
private	int	noexpand = 0;

private void	checkdirs();
private void	openfiles();
private void	rewrite_Dfile();
private int	procfile();
private void	procdef();
private void	procinc();
private FILE	*openhdr();
private int	already_open();
private char	*skipspace();

int
main(argc, argv)
int	argc;
char	**argv;
{
	int rc = 0;

	DBUG_PROCESS(argv[0]);

	set_pname(argv[0]);

	if (argc > 1 && argv[1][0] == '-')
	{
		DBUG_PUSH(argv[1]+2);
		argv++;
		argc--;
	}

	if (argc <= 1)
	{
		(void) fprintf(stderr, "usage: %s headerfile [searchdir ...]\n",
			get_pname());
		exit(2);
	}
	headerfile = argv[1];	/* for hdrallow() */

	if (argc > 2)
		searchdirs = &argv[2];

	noexpand = (getenv("HDRDEFS_NOEXPAND") != NULL);

	checkdirs(searchdirs);

	openfiles(headerfile);

	currdir = searchdirs;

	if (procfile(headerfile, FALSE) != 0)
	{
		vsxerror("could not find/open header \"%s\"", headerfile);
		exit(FAIL);
	}

	if (defcount == 0)
		rewrite_Dfile(headerfile);

	if (ferror(Dfp) || fflush(Dfp) != 0)
	{
		vsxerror("write error on \"%s\"", Dfile);
		rc = FAIL;
	}
	if (ferror(Cfp) || fflush(Cfp) != 0)
	{
		vsxerror("write error on \"%s\"", Cfile);
		rc = FAIL;
	}

	exit(rc);

	/* NOTREACHED */
}

private void
checkdirs(dirlist)
char **dirlist;
{
	/* Check include directories have search permission */

	char **dirp;
	int errors = 0;

	DBUG_ENTER("checkdirs");

	for (dirp = dirlist; *dirp != NULL; dirp++)
	{
		if (access(*dirp, X_OK) != 0)
		{
			vsxerror("no search access on \"%s\"", *dirp);
			errors++;
		}
	}

	if (errors > 0)
		DBUG_EXIT(FAIL);
	
	DBUG_VOID_RETURN;
}

private void
openfiles(hdrfile)
char *hdrfile;
{
	/* Open the ".D" and ".C" output files */

	char **sp;
	char *comment = "/* this comment prevents empty include files */\n";

	DBUG_ENTER("openfiles");

	if (strlen(hdrfile) > sizeof(Cfile) - 3)
	{
		vsxerror("name too long: \"%s\"", hdrfile);
		DBUG_EXIT(FAIL);
	}

	(void) strcpy(Cfile, hdrfile);
	(void) strcpy(Dfile, hdrfile);
	(void) strcat(Cfile, ".C");
	(void) strcat(Dfile, ".D");

	if ((Dfp = fopen(Dfile, "w")) == NULL)
	{
		vsxerror("cannot open \"%s\" for writing", Dfile);
		DBUG_EXIT(FAIL);
	}
	(void) fputs(comment, Dfp);

	if ((Cfp = fopen(Cfile, "w")) == NULL)
	{
		vsxerror("cannot open \"%s\" for writing", Cfile);
		DBUG_EXIT(FAIL);
	}
	for (sp = predefs; *sp != NULL; sp++)
	{
		if (fprintf(Cfp, "#ifdef %s\n\tchkallow(\"%s\");\n#endif\n\n",
			*sp, *sp) <= 0)
		{
			vsxerror("write error on \"%s\"", Cfile);
			DBUG_EXIT(FAIL);
		}
	}

	DBUG_VOID_RETURN;
}

private void
rewrite_Dfile(hdrfile)
char *hdrfile;
{
	/*
	 * The expanded-out .h.D file is not needed, so overwrite it
	 * with a simple #include of the target header.
	 *
	 * This avoids wasting lots of disc space (and compiler CPU
	 * time) on systems where the expansion produces very large
	 * files.
	 */

	DBUG_ENTER("rewrite_Dfile");

	(void) fclose(Dfp);
	if ((Dfp = fopen(Dfile, "w")) == NULL)
	{
		vsxerror("cannot reopen \"%s\" for writing", Dfile);
		DBUG_EXIT(FAIL);
	}

	(void) fprintf(Dfp, "#include <%s>\n", hdrfile);
	/* (write errors are picked up on fflush() later) */

	DBUG_VOID_RETURN;
}

private int
procfile(hdrfile, dodot)
char *hdrfile;
int dodot;
{
	/*
	 * Read through a header file, copying all lines except '#include's.
	 */

	FILE	*fp;
	char	*cp;
	char	**sav_currdir;
	char	line[LINELEN];

	DBUG_ENTER("procfile");

	DBUG_PRINTF("args", "procfile(%s, %d)", hdrfile, dodot);

	if ((fp = openhdr(hdrfile, dodot)) == NULL)
		DBUG_RETURN(-1);
	
	if (already_open(fileno(fp)))
	{
		DBUG_PRINTF("procfile", "file %s already open: ignored",
			hdrfile);
		(void) fclose(fp);
		DBUG_RETURN(0);
	}

	while(fgets(line, LINELEN, fp) != NULL)
	{
		cp = skipspace(line);
		if (*cp == '#')
		{
			cp = skipspace(cp+1);
			if (strncmp(cp, "define", (size_t)6) == 0)
				procdef(skipspace(cp+6));
			else if (strncmp(cp, "include_next", (size_t)12) == 0)
			{
				sav_currdir = currdir;
				currdir++;
				procinc(skipspace(cp+12), line);
				currdir = sav_currdir;
				continue; /* don't write out the #include_next */
			}
			else if (strncmp(cp, "include", (size_t)7) == 0)
			{
				sav_currdir = currdir;
				currdir = searchdirs;
				procinc(skipspace(cp+7), line);
				currdir = sav_currdir;
				continue; /* don't write out the #include */
			}
		}
		if (!noexpand && fputs(line, Dfp) < 0)
		{
			vsxerror("write error on \"%s\"", Dfile);
			DBUG_EXIT(FAIL);
		}
	}

	if (ferror(fp))
	{
		vsxerror("read error on \"%s\"", hdrfile);
		DBUG_EXIT(FAIL);
	}

	(void) fclose(fp);

	DBUG_RETURN(0);
}

private FILE *
openhdr(hdrfile, dodot)
char *hdrfile;
int dodot;
{
	/* Find and open a header file */
	/* dodot indicates whether to search "." first */

	/* The search starts from wherever currdir points to
	   in searchdirs.  Set currdir=searchdirs before calling
	   the function to search from the beginning. */

	FILE	*fp;
	char	pathname[_POSIX_PATH_MAX];

	DBUG_ENTER("openhdr");

	if (dodot && (fp = fopen(hdrfile, "r")) != NULL)
		DBUG_RETURN(fp);

	for ( ; *currdir != NULL; currdir++)
	{
		(void) sprintf(pathname, "%s/%s", *currdir, hdrfile);
		if ((fp = fopen(pathname, "r")) != NULL)
			DBUG_RETURN(fp);
	}

	DBUG_RETURN((FILE *)NULL);
}

private int
already_open(curfd)
int curfd;
{
	/*
	 * If the header file just opened on curfd is already open,
	 * ignore it (otherwise infinite recursion will result)
	 */

	int	fd;
	dev_t	curdev;
	ino_t	curino;
	struct stat stbuf;

	DBUG_ENTER("already_open");

	if (fstat(curfd, &stbuf) == SYSERROR)
		DBUG_RETURN(0);

	curdev = stbuf.st_dev;
	curino = stbuf.st_ino;

	for (fd = 0; fd < curfd; fd++)
	{
		if (fstat(fd, &stbuf) != SYSERROR &&
		    curdev == stbuf.st_dev && curino == stbuf.st_ino)
			DBUG_RETURN(1);
	}

	DBUG_RETURN(0);
}

private void
procdef(symbol)
char *symbol;
{
	/* process a '#define' line */

	char *cp, savchr;

	DBUG_ENTER("procdef");

	/* find the end of the symbol */
	for (cp = symbol; *cp != ' ' && *cp != '\t' && *cp != '\n' &&
			*cp != '(' && *cp != '\0'; cp++)
		;
	
	savchr = *cp;
	*cp = '\0';

	/* If it is allowed by a special rule, don't output it */
	if (hdrallow(symbol, 1))
	{
		*cp = savchr;
		DBUG_VOID_RETURN;
	}

	if (fprintf(Cfp, "#ifdef %s\n\tchkallow(\"%s\");\n#endif\n\n",
		symbol, symbol) <= 0)
	{
		vsxerror("write error on \"%s\"", Cfile);
		DBUG_EXIT(FAIL);
	}
	*cp = savchr;

	DBUG_VOID_RETURN;
}

private void
procinc(filespec, line)
char *filespec;
char *line;
{
	/* process a '#include' or '#include_next' line */

	char *cp, endchr = filespec[0];
	int dodot = TRUE;

	DBUG_ENTER("procinc");

	if (filespec[0] == '<')
	{
		endchr = '>';
		dodot = FALSE;
	}
	
	/* find end of name */
	for (cp = &filespec[1]; *cp != endchr && *cp != '\n' && *cp != '\0';
			cp++)
		;
	
	if (*cp != endchr)
	{
		/* A badly formed #include line is probably inside a
		   comment.  If not, then presumably it is never seen by
		   the normal compiler.  In any case just write it out. */
		if (!noexpand && fputs(line, Dfp) < 0)
		{
			vsxerror("write error on \"%s\"", Dfile);
			DBUG_EXIT(FAIL);
		}
		DBUG_VOID_RETURN;
	}
	*cp = '\0';

	/* recursive call to procfile() to process the new file */
	if (procfile(&filespec[1], dodot) != 0)
	{
		if (noexpand)
		{
			vsxerror("could not find/open nested header: %s",
				&filespec[1]);
			DBUG_EXIT(FAIL);
		}

		defcount++;
		if (fprintf(Dfp, "#define VSXdef%d \"%s: %s %s\"\n",
			defcount, get_pname(),
			"could not find/open header", &filespec[1] ) <= 0)
		{
			vsxerror("write error on \"%s\"", Dfile);
			DBUG_EXIT(FAIL);
		}
		if (fprintf(Cfp,
			"#ifdef VSXdef%d\n\terrallow(VSXdef%d);\n#endif\n\n",
			defcount, defcount) <= 0)
		{
			vsxerror("write error on \"%s\"", Cfile);
			DBUG_EXIT(FAIL);
		}
	}

	/* If the rest of the #include line starts a comment, write
	   it out on a separate line */
	
	for (++cp; *cp != '\n' && *cp != '\0'; cp++)
	{
		if (*cp == '/' && *(cp+1) == '*')
		{
			if (!noexpand && fputs(cp, Dfp) < 0)
			{
				vsxerror("write error on \"%s\"", Dfile);
				DBUG_EXIT(FAIL);
			}
			break;
		}
	}

	DBUG_VOID_RETURN;
}

private char *
skipspace(cp)
char *cp;
{
	for ( ; *cp == ' ' || *cp == '\t'; cp++)
		;
	return cp;
}
