/*
*      SCCS:  @(#)  ANSI.os/streamio/fopen/enfile.c	Rel 4.4.1     (06/09/97) 
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/

#ifndef lint
#define MAIN
#ifdef	MAIN
static	char sccsid[] = "@(#)ANSI.os/streamio/fopen - enfile.c Rel 4.4.1 (06/09/97)";
char	*copyright[] = {
		"(C) Copyright 1991 X/Open Company Limited",
		"All Rights Reserved."
};
#endif	/* MAIN	*/
#endif  /* lint	*/

/***********************************************************************

NAME:		enfile.c - used by T.fopen and other testsets to test
		the ENFILE condition.

PROJECT:	VSX (X/OPEN Validation Suite)

MODULE CONTAINS:
		tet_main()

SYNOPSIS:
		int main(argc, argv)
			int argc;
			char *argv[];

DESCRIPTION:
	  This source tests the ENFILE condition if possible for the
	  interfaces fopen, freopen and tmpfile.

	  This should be setuid/setgid to VSX_UID1/VSX_GID1.
	  usage:
		testname file uid gid testtype vg_ident vg_dbug

	  where "file" is the file/directory to be opened.
	  testtype is the function to test:
		f -- fopen()
		F -- freopen() 
		t -- tmpfile()

METHODS:
ASSUMES:
CONDITIONAL COMPILE PARAMS:
	This source uses the following conditional defines to separate
	out the dependencies.
		FOPEN_TEST -- fopen()
		FREOPEN_TEST -- freopen()
		TMPFILE_TEST -- tmpfile()

	Note that there may be more than one table that can be full to
	give ENFILE.  For instance the file table and the inode table
	in tradition implementations.  The latter may be tested in the
	tmpfile() test if it is smaller than the file table.

SIDE EFFECTS:
RELATED DOCUMENTS:
SOURCE:

AUTHOR:		Steve Ratcliffe, UniSoft Ltd.
DATE CREATED:	05 Jan 1990
MODIFICATIONS:
		Geoff Clare, 25 May 1990
			Allow for unlimited OPEN_MAX

		David Sawyer
		Fri Jul 27 16:53:59 BST 1990
		Added resilience code.

		Stuart Boutell, 5 Nov 1990
			Eliminated vg_env
			Eliminated vg_file

		Stuart Boutell,26 Nov 1990
			Cured bug in PIPE_TEST - i/2 changed to i
			in pipe() open call. Also used INC to decrement
			the close loop in the child.
	
		Stuart Boutell, 30 Nov 1990
			added close_before_report() function to try
			and ensure some file descriptors are open before
			reporting is attempted.

		Stuart Boutell, 4 Dec 1990
			Added tet_setcontext() and tet_setblock() calls
			where fork() is called

		Neil Moses, 28 May 1992
			Check that on error fopen() does not
			create/truncate the file.

		Neil Moses, 1 June 1992
			Check that on error freopen() does not
			create/truncate the file.
		
		Geoff Clare, 22 July 1992
			Use both OPEN_MAX and FOPEN_MAX when deciding
			how many streams can be opened per process.
			Remove redundant code.

		Dave Cater, 13th November 1992
			XNFS enhancements

************************************************************************/

#include <dbug.h>
#include <std.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <report.h>
#include <errno.h>
#include <signal.h>
#include <timeout.h>
#include <sys/wait.h>
#include <exitcodes.h>
#include <pcts_max.h>
#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <tet_api.h>
#include <xnfs.h>
#include <sysdep.h>

#ifdef UNDEF_MACROS
#undef fopen
#undef freopen
#undef tmpfile
#endif


#define	WAITTIME	(30 * SPEEDFACTOR)

public	char	*vg_ident =	"<no ident>";
public	char	*vg_dbug =	(char *) NULL;

extern	char	**environ;

extern	char	*newstr();
extern	char	*itoa();
extern	char	*errname();
extern	void	set_pname();
extern	void	setsigs();
extern	void	(*vsignal())();
extern	void	valrpt();

private int 	testfail;
private pid_t 	childpid;

private	void	close_before_report();
private	int	openfiles();
private void	termchild();

#define	INC 	1

#ifdef FOPEN_TEST

private void	chk_fopen();
private int 	tfail;
private char	testfunc = 'f';
typedef	FILE	*desctype;

#define open1(path, i) \
	if ((openedfiles[i] = fopen(path, "r")) == NULL) \
		testfail = 1;

#define close1(i) fclose(openedfiles[i])

#endif

#ifdef TMPFILE_TEST

private char	testfunc = 't';
typedef	FILE	*desctype;

#define open1(path, i) \
	if ((openedfiles[i] = tmpfile()) == NULL) \
		testfail = 1;

#define close1(i) fclose(openedfiles[i])

#endif

#ifdef FREOPEN_TEST

private void	chk_freopen();
private int 	tfail;
private char	testfunc = 'F';
typedef	FILE	*desctype;

/*
 * Although freopen() closes one file and then reopens another one.
 * it is still possible to get ENFILE if the file that is closed
 * has been dup()'ed.
 */
#define open1(path, i) { \
	FILE	*fp; \
	int 	fd; \
		if ((fd = dup(0)) == SYSERROR) \
			testfail = 2; \
		if ((fp = fdopen(fd, "r")) == NULL) \
			testfail = 2; \
		if ((openedfiles[i] = freopen(path, "r", fp)) == NULL) \
			testfail = 1; \
	}

#define close1(i)	fclose(openedfiles[i])

#endif

desctype	*openedfiles;
private	int 	sys_open_max;

public	int
tet_main(argc, argv)
int	argc;
char	*argv[];
{
	uid_t uid;
	gid_t gid;
	char	*somp;

	set_pname(argv[0]);

	if (argc < 5 || argc > 7)
	{
		xx_rpt(DELETION);
		in_rpt("argument count wrong");
		return OK;
	}
	if (argc >= 7)
		vg_dbug = argv[6];
	if (argc >= 6)
		vg_ident = argv[5];

	if (vg_dbug != NULL)
	{
		DBUG_PROCESS(vg_ident);
		DBUG_PUSH(vg_dbug+2);
	}

	/*
	 * This argument is left over from an earlier version
	 * it still provides a useful sanity check now.
	 */
	if (!argv[4] || testfunc != *argv[4]) {
		xx_rpt(DELETION);
		in_rpt("Miss-match between calling test and %s", argv[0]);
		in_rpt("Subtest probably incorrectly compiled");
		return OK;
	}

	DBUG_PRINTF("ENFILEtest", "testfunc=%c", testfunc);

	if ((somp = tet_getvar("VSX_SYS_OPEN_MAX")) == NULL || *somp == '\0') {

		xx_rpt(DELETION);
		in_rpt("deletion reason: VSX_SYS_OPEN_MAX not set in environment");

		return OK;

	} else {
		sys_open_max = atoi(somp);
	}

	if (sys_open_max == -1) {
		xx_rpt(UNTESTED);
		in_rpt("Parameter VSX_SYS_OPEN_MAX set to -1");
		return OK;
	}

	if ((uid = atoi(argv[2])) != geteuid() ||
		(gid = atoi(argv[3])) != getegid())
	{
		xx_rpt(DELETION);
		if (uid != geteuid())
			in_rpt("deletion reason: uid is not equal to VSX_UID1");
		else
			in_rpt("deletion reason: gid is not equal to VSX_GID1");
	}
	else
	{
		(void) setuid(uid);
		(void) setgid(gid);
		return openfiles(argv[1]);
	}

	return OK;
}


private int
openfiles(path)
char	*path;
{
	pid_t 	retval;
	int 	status;
	int 	i;
	int 	err;
	int 	open_max;
	static	int 	nopen = 0;
	int	pathok = 0;

	DBUG_ENTER("openfiles");
	DBUG_PRINTF("openfiles", "nopen=%d", nopen);

	testfail = 0;

	errno = 0;
	open_max = sysconf(_SC_OPEN_MAX);

	if (open_max < 0)
	{
		if (errno == 0)
		{
			PATH_TRACE;
			open_max = 256;
		}
		else
		{
			xx_rpt(DELETION);
			in_rpt("sysconf(_SC_OPEN_MAX) failed in openfiles");
			DBUG_RETURN(OK);
		}
	}
	else
		PATH_TRACE;
	/*
	 * Subtract the number of descriptors already open from total to open
	 */
	open_max -= n_open_fds();
	if (open_max <= 0) {
		testfail++;
		xx_rpt(DELETION);
		in_rpt("No available file descriptors for test");
	}
	else
		PATH_TRACE;

	/*
	 * The test is known to start off with 3 streams open
	 */
	if (open_max > FOPEN_MAX - 3)
		open_max = FOPEN_MAX - 3;

	DBUG_PRINTF("openfiles", "sys_open_max=%d, open_max=%d", sys_open_max, open_max);

	openedfiles = (desctype*)malloc(sizeof(openedfiles[0])*(size_t)((open_max>sys_open_max)? sys_open_max: open_max));

	if (openedfiles == (desctype*) 0)
	{
		DBUG_RETURN(A_NOMEM);
	}
	else
		PATH_TRACE;

	testfail = 0;

	for (i = 0; nopen < sys_open_max && i < open_max && !testfail; i += INC)
	{
		if (i==0)
			PATH_TRACE;

		open1(path, i);

		if (testfail == 1) {
			DBUG_PRINTF("openfiles", "nopen=%d failed with %s", nopen, errname(errno));
			if (errno == ENFILE) {
				close_before_report(i);
#ifdef FOPEN_TEST
				i = 0;
				tfail = 0;

				globok = 0;
				chk_fopen(path, 't', &i);
				PATH_FUNC_TRACE;

				close_before_report(i);
				i = 0;

				(void) unlink(path);
				globok = 0;
				chk_fopen(path, 'c', &i);
				PATH_FUNC_TRACE;

				if(!tfail)
				{
					close_before_report(i);
					PATH_XS_RPT(6);
				}
#endif
#ifdef FREOPEN_TEST
				i = 0;
				tfail = 0;

				globok = 0;
				chk_freopen(path, 't', &i);
				PATH_FUNC_TRACE;

				close_before_report(i);
				i = 0;

				(void) unlink(path);
				globok = 0;
				chk_freopen(path, 'c', &i);
				PATH_FUNC_TRACE;

				if(!tfail)
				{
					close_before_report(i);
					PATH_XS_RPT(6);
				}
#endif
#ifdef TMPFILE_TEST
				PATH_XS_RPT(4);
#endif
			} else {
				err = errno;
				close_before_report(i);
				xx_rpt(FAILURE);
				valrpt(0, 0, ENFILE, err);
			}
		} else {
			nopen += INC;
		}
	}

	if (nopen >= sys_open_max) {

		/*
		 * We got to the claimed size of the file table
		 * but didn't get ENFILE.  Therefore fail the test
		 */
		if (!testfail++) {
			close_before_report(i);
			xx_rpt(FAILURE);
			in_rpt("ENFILE not encountered after opening %d file%s", nopen, (nopen==1)? "": "s");
		}
	}

	if (testfail)
		DBUG_RETURN(OK);

	/*
	 * We didn't manage to provoke ENFILE so try a new process
	 */
	switch (childpid = fork()) {
	case 0:	/* Child */
		tet_setcontext();

		setsigs(SIG_DFL);

		/*
		 * Close the files that we have inherited from the parent
		 */
		for (i-=INC ; i >= 0; i-=INC) {
			(void) close1(i);
		}
		exit(openfiles(path));

	case SYSERROR:	/* Fork failed */
		/*
		 * Ran out of children -- give up.
		 * Can't say anything.
		 */
		if (errno == EAGAIN) {
			close_before_report(i);
			xx_rpt(UNTESTED);
			in_rpt("No more child processes after opening %d files", nopen);
		} else {
			/* * Fork failed for spurious reason
			 */
			err = errno;
			close_before_report(i);
			xx_rpt(DELETION);
			in_rpt("Fork failed errno %d (%s)", err, errname(err));
		}
		break;

	default:
		tet_setblock();
		(void) vsignal(SIGTERM, termchild);

		SET_TIMEOUT(WAITTIME);
		retval = waitpid(childpid, &status, 0);
		CLEAR_ALARM;

		if (retval == SYSERROR) {

			err = errno;
			close_before_report(i);
			xx_rpt(DELETION);
			in_rpt("waitpid in openfiles() failed errno %d (%s)", err, errname(err));

			/*
			 * Try to get rid of all the children,  wait
			 * for a long time
			 */
			(void) kill(childpid, SIGTERM);
			SET_TIMEOUT(5*WAITTIME);
			retval = waitpid(childpid, &status, 0);
			CLEAR_ALARM;

		} else if (WIFEXITED(status))
			DBUG_RETURN(WEXITSTATUS(status));
		else {
			close_before_report(i);
			xx_rpt(DELETION);
			in_rpt("Unexpected wait status %#x", status);
		}
	}
	tet_setblock();

	DBUG_RETURN(OK);
}

private void
termchild()
{
int 	status;

	(void) kill(childpid, SIGTERM);
	(void) waitpid(childpid, &status, 0);
	/* exit() is not reentrent in POSIX though it is in XPG */
	_exit(OK);
}

private	void
close_before_report(i)
int	i;
{
	int	a;
	a=i;
	for (a-=INC ; a >= 0; a-=INC)
	{
		(void) close1(a);
	}
}

#if defined(FOPEN_TEST)
private void
chk_fopen(path, type, ind)
char	*path;
char	type;
int	*ind;
{
	struct	stat buf;
	FILE	*fp;
	int	i, err;
	int	pathok = 0;

	DBUG_ENTER("chk_fopen");

	errno = 0;

	/* Need to ensure that ENFILE is generated for fopen() call */
	for(i = 0; i < sys_open_max; i++, (*ind)++)
	{
		if(i == 0)
			PATH_TRACE;

		if((fp = openedfiles[(*ind)] = fopen(path, "w")) == (FILE *)NULL
		&& errno == ENFILE)
			break;
		else if(fp == (FILE *)NULL && errno != ENFILE)
		{
			err = errno;
			close_before_report(*ind);

			if(!tfail++)
				xx_rpt(DELETION);

			in_rpt("fopen(\"%s\",\"w\") did not set errno as expected", path);
			in_rpt("Expected errno %d (ENFILE)", ENFILE);
			in_rpt("Actual errno %d (%s)", err, errname(err));
			DBUG_VOID_RETURN;
		}

		/* If truncating need to write to file */
		if(type == 't')
		{
			if(fputs("1", fp) == SYSERROR)
			{
				err = errno;
				close_before_report(*ind);
				if(!tfail)
					xx_rpt(DELETION);
				in_rpt("deletion reason: fputs() failed - errno %d (%s)", err, errname(err));
				DBUG_VOID_RETURN;
			}

			(void) fflush(fp);

			continue;
		}

		/* If creating need to remove file */
		if(type == 'c')
		{
			if(unlink(path) == SYSERROR)
			{
				err = errno;
				close_before_report(*ind);
				if(!tfail)
					xx_rpt(DELETION);
				in_rpt("deletion reason: unlink(\"%s\") failed - errno %d (%s)", path, err, errname(err));
				DBUG_VOID_RETURN;
			}

			continue;
		}
	}

	if (i >= sys_open_max)
		DBUG_VOID_RETURN
	else
		PATH_TRACE;

	/* Check that file is not truncated */
	if(type == 't')
	{
		if(stat(path, &buf) < 0)
		{
			err = errno;
			close_before_report(*ind);
			if(!tfail)
				xx_rpt(DELETION);
			in_rpt("stat(\"%s\", buf) failed - errno %d (%s)",
				path, err, errname(err));
			DBUG_VOID_RETURN;
		}
		else
			PATH_TRACE;

		if(buf.st_size == 0)
		{
			close_before_report(*ind);
			if(!tfail++)
				xx_rpt(FAILURE);
			in_rpt("fopen(\"%s\",\"w\") on error ENFILE truncated file", path);
			DBUG_VOID_RETURN;
		}
		else
			PATH_TRACE;
	}
	else
	{
		PATH_TRACE;
		PATH_TRACE;
	}

	/* Check that file is not created */
	if(type == 'c')
	{
		if(eaccess(path, F_OK) != SYSERROR)
		{
			close_before_report(*ind);
			if(!tfail++)
				xx_rpt(FAILURE);
			in_rpt("fopen(\"%s\",\"w\") on error ENFILE created file", path);
			DBUG_VOID_RETURN;
		}
		else
		{
			PATH_TRACE;
			PATH_TRACE;
		}
	}
	else
	{
		PATH_TRACE;
		PATH_TRACE;
	}

	PATH_FUNC_RPT(6);

	DBUG_VOID_RETURN;
}
#endif

#if defined(FREOPEN_TEST)
private void
chk_freopen(path, type, ind)
char	*path;
char	type;
int	*ind;
{
	struct	stat buf;
	FILE	*fp;
	int	i, fd, err;
	int	pathok = 0;

	DBUG_ENTER("chk_fopen");

	errno = 0;

	/* Need to ensure that ENFILE is generated for fopen() call */
	for(i = 0; i < sys_open_max; i++, (*ind)++)
	{
		if(i == 0)
			PATH_TRACE;

		if ((fd = dup(0)) == SYSERROR)
		{
			err = errno;
			close_before_report(*ind);
			if(!tfail++)
				xx_rpt(DELETION);
			in_rpt("deletion reason: dup() failed - errno %d (%s)", err, errname(err));
			DBUG_VOID_RETURN;
		}

		if ((fp = fdopen(fd, "w")) == NULL)
		{
			err = errno;
			close_before_report(*ind);
			if(!tfail++)
				xx_rpt(DELETION);
			in_rpt("deletion reason: fdopen() failed - errno %d (%s)", err, errname(err));
			DBUG_VOID_RETURN;
		}

		if ((fp = openedfiles[*ind] = freopen(path, "w", fp)) == NULL)
			break;
		else if(fp == NULL && errno != ENFILE)
		{
			err = errno;
			close_before_report(*ind);
			if(!tfail++)
				xx_rpt(DELETION);
			in_rpt("freopen(\"%s\",\"w\", fp) did not set errno as expected", path);
			in_rpt("Expected errno %d (ENFILE)", ENFILE);
			in_rpt("Actual errno %d (%s)", err, errname(err));
			DBUG_VOID_RETURN;
		}

		/* If truncating need to write to file */
		if(type == 't')
		{
			if(fputs("1", fp) == SYSERROR)
			{
				err = errno;
				close_before_report(*ind);
				if(!tfail++)
					xx_rpt(DELETION);
				in_rpt("deletion reason: fputs() failed - errno %d (%s)", err, errname(err));
				DBUG_VOID_RETURN;
			}

			(void) fflush(fp);

			continue;
		}

		/* If creating need to remove file */
		if(type == 'c')
		{
			if(unlink(path) == SYSERROR)
			{
				err = errno;
				close_before_report(*ind);
				if(!tfail++)
					xx_rpt(DELETION);
				in_rpt("deletion reason: unlink(\"%s\") failed - errno %d (%s)", path, err, errname(err));
				DBUG_VOID_RETURN;
			}

			continue;
		}
	}

	if (i >= sys_open_max)
		DBUG_VOID_RETURN
	else
		PATH_TRACE;

	/* Check that file is not truncated */
	if(type == 't')
	{
		if(stat(path, &buf) < 0)
		{
			err = errno;
			close_before_report(*ind);
			if(!tfail++)
				xx_rpt(DELETION);
			in_rpt("stat(\"%s\", buf) failed - errno %d (%s)",
				path, err, errname(err));
			DBUG_VOID_RETURN;
		}
		else
			PATH_TRACE;

		if(buf.st_size == 0)
		{
			close_before_report(*ind);
			if(!tfail++)
				xx_rpt(FAILURE);
			in_rpt("freopen(\"%s\",\"w\", fp) on error ENFILE truncated file", path);
			DBUG_VOID_RETURN;
		}
		else
			PATH_TRACE;
	}
	else
	{
		PATH_TRACE;
		PATH_TRACE;
	}

	/* Check that file is not created */
	if(type == 'c')
	{
		if(eaccess(path, F_OK) != SYSERROR)
		{
			close_before_report(*ind);
			if(!tfail++)
				xx_rpt(FAILURE);
			in_rpt("freopen(\"%s\",\"w\", fp) on error ENFILE created file", path);
			DBUG_VOID_RETURN;
		}
		else
		{
			PATH_TRACE;
			PATH_TRACE;
		}
	}
	else
	{
		PATH_TRACE;
		PATH_TRACE;
	}

	PATH_FUNC_RPT(6);

	DBUG_VOID_RETURN;
}
#endif
