/*
 * Program used with vsyslog test case. Sets up syslogd for testing vsyslog().
 */
#include <std.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <dbug.h>
#include <report.h>
#include <tet_api.h>

#if !defined(FALSE)
# define FALSE 0
#endif
#if !defined(TRUE)
# define TRUE 1
#endif

/* Size of buffer used read/write file contents */
#define	CPBUFSIZE 8192

#define LOGD_STARTTIME	10
#define LINELEN		256
#define PATHBUFSZ	1024

/* Global static data */
static char conffile[] = "/etc/syslog.conf";
static char saveconf[] = "./saved.syslog.conf";
static char pidfile[] = "/var/run/syslogd.pid";

/* Local prototypes */
static pid_t getsyslogdpid(void);
static int setup_syslogd(char *debuglogfile);
static int cleanup_syslogd(void);

private int filecopy(char *src, char *dest, int force);
/*
 * Get process ID of syslogd. Uses /etc/syslog.pid.
 */
static pid_t
getsyslogdpid(void)
{
	FILE *fp;
	char buf[20];
	long id;

	DBUG_ENTER("getsyslogpid");

	/* Open file containing PID of syslogd */
	fp = fopen(pidfile, "r");
	if (fp == NULL)
		DBUG_RETURN((pid_t)-1);

	/* Read PID from file */
	if (fgets(buf, sizeof(buf), fp) != NULL)
		id = atol(buf);
	else
		id = -1;

	pclose(fp);

	/* Verify PID is valid and process exists */
	if (id > 0 && kill((pid_t)id, 0) == 0)
		DBUG_RETURN((pid_t)id);

	DBUG_RETURN((pid_t)-1);
}

/*
 * Setup syslogd. Starts it if it's not running, and adds a line to log all
 * user debug output to debuglogfile.
 */
static int
setup_syslogd(char *debuglogfile)
{
	char log[PATHBUFSZ];
	int len;
	pid_t pid;
	int i;
	FILE *sfp;
	FILE *dfp;
	char buf[LINELEN + 1];
	int err;

	DBUG_ENTER("setup_syslogd");

	/* If debuglogfile is a relative path, prepend the current directory */
	if (*debuglogfile != '/')
	{
		len = (int)sizeof(log) - (int)strlen(debuglogfile);
		if (len < 0 || getcwd(log, len) == NULL)
		{
			in_rpt("Error generating absolute path for %s",
				debuglogfile);
			DBUG_RETURN(1);
		}

		strcat(log, "/");
		strcat(log, debuglogfile);
	}

	/* See if syslogd is running */
	pid = getsyslogdpid();
	if (pid == -1)
	{
		/* Syslogd isn't running, so start it now */
		system("/usr/sbin/syslogd >/dev/console 2>&1 &");

		/* Wait for syslogd to start */
		for (i = 0; i < LOGD_STARTTIME; i++)
		{
			sleep(1);

			pid = getsyslogdpid();
			if (pid != 0)
				break;
		}

		if (pid == -1)
		{
			in_rpt("Failed to start syslogd");
			DBUG_RETURN(1);
		}
	}

	/* Save conf file to enable later replacement */
	if (filecopy(conffile, saveconf, 1) != 0)
	{
		in_rpt("Error saving %s", conffile);
		DBUG_RETURN(1);
	}

	/* Copy saved conf file back to original location, adding in an extra
	 * line to enable us to read test output.
	 */
	sfp = fopen(saveconf, "r");
	if (sfp == NULL)
	{
		in_rpt("Error opening %s for reading", saveconf);
		DBUG_RETURN(1);
	}

	dfp = fopen(conffile, "w");
	if (dfp == NULL)
	{
		in_rpt("Error opening %s for writing", conffile);
		fclose(sfp);
		DBUG_RETURN(1);
	}

	fprintf(dfp, "# Temporary modification for LSB Test Suite\n");
	fprintf(dfp, "user.debug			%s.0\n", log);
	fprintf(dfp, "local1.*			%s.1\n", log);
	fprintf(dfp, "local2.*			%s.2\n", log);
	fprintf(dfp, "user.=warning			%s.3\n", log);
	fprintf(dfp, "user.=info			%s.4\n", log);

	while (fgets(buf, sizeof(buf), sfp) != NULL)
		fputs(buf, dfp);

	err = 0;
	err |= ferror(sfp);
	err |= ferror(dfp);
	err |= fclose(sfp);
	err |= fclose(dfp);

	if (err)
	{
		in_rpt("Error copying %s to %s", saveconf, conffile);
		DBUG_RETURN(1);
	}

	/* Send SIGHUP to tell syslogd to re-read config file */
	if (kill(pid, SIGHUP) != 0)
	{
		in_rpt("Error sending SIGHUP to syslogd");
		DBUG_RETURN(1);
	}

	DBUG_RETURN(0);
}

/*
 * Cleanup syslogd. Restores original conf file.
 */
static int
cleanup_syslogd(void)
{
	pid_t pid;

	DBUG_ENTER("cleanup_syslogd");

	/* Get PID of syslogd */
	pid = getsyslogdpid();
	if (pid == -1)
	{
		in_rpt("Can't determine PID of syslogd");
		DBUG_RETURN(1);
	}

	/* Replace log file with original */
	if (filecopy(saveconf, conffile, 0) != 0)
	{
		in_rpt("Error restoring original %s from %s", conffile,
			saveconf);
		DBUG_RETURN(1);
	}

	/* Send SIGHUP to tell syslogd to re-read config file */
	if (kill(pid, SIGHUP) != 0)
	{
		in_rpt("Error sending SIGHUP to syslogd");
		DBUG_RETURN(1);
	}

	/* Remove saved conf file */
	unlink(saveconf);

	DBUG_RETURN(0);
}

int
tet_main(int argc, char **argv)
{
	DBUG_ENTER("tet_main");

	/* Verify the calling process is super-user */
	if (geteuid() != 0)
	{
		in_rpt("Calling process is not super-user");
		DBUG_RETURN(1);
	}

	/* Set real user-ID to root so system() doesn't pull any funnies */
	setuid(0);

	/* Set umask */
	umask(002);

	/* Perform startup or cleanup actions as appropriate */
	if (argc == 3 && strcmp(argv[1], "startup") == 0)
		DBUG_EXIT(setup_syslogd(argv[2]));
	else if (argc == 2 && strcmp(argv[1], "cleanup") == 0)
		DBUG_EXIT(cleanup_syslogd());

	in_rpt("%s called with bad arguments", argv[0]);
	DBUG_EXIT(-1);
}

/*
 * filecopy
 *
 * Copy file.
 *
 *	src	Name of source file.
 *	dest	Name of destination file.
 *	force	If true (non-zero) then dest is removed first.
 *
 * Returns 0 for success; -1 for failure.
 */
int
filecopy(char *src, char *dest, int force)
{
	int		errflag;
	int		err;
	int		sfd;
	int		dfd;
	struct stat	statbuf;
	char		buf[CPBUFSIZE];
	ssize_t		nbytes;

	DBUG_ENTER("filecopy");

	errflag = FALSE;
	errno = 0;

	/* Open source file for reading */
	sfd = open(src, O_RDONLY);
	if (sfd < 0)
	{
		err = errno;
		in_rpt("open(%s): %s", src, strerror(err));
		DBUG_RETURN(-1);
	}

	if (fstat(sfd, &statbuf) != 0)
	{
		err = errno;
		in_rpt("fstat: %s", strerror(err));
		close(sfd);
		DBUG_RETURN(-1);
	}

	/* If required, remove destination file */
	if (force)
	{
		unlink(dest);
	}

	/* Open destination file ready to be written to with the same mode
	 * as the original.
	 */
	dfd = open(dest, O_WRONLY|O_CREAT|O_TRUNC, statbuf.st_mode);
	if (dfd < 0)
	{
		err = errno;
		in_rpt("open(%s): %s", dest, strerror(err));
		close(sfd);
		DBUG_RETURN(-1);
	}

	/* Read from source and write to destination */
	errflag = FALSE;

	while ((nbytes = read(sfd, buf, sizeof(buf))) != 0)
	{
		if (nbytes < 0)
		{
			err = errno;
			in_rpt("read: %s", strerror(err));
			errflag = TRUE;
			break;
		}

		if (write(dfd, buf, (size_t)nbytes) < 0)
		{
			err = errno;
			in_rpt("write: %s", strerror(err));
			errflag = TRUE;
			break;
		}
	}

	/* Close open file descriptors */
	if (close(sfd) != 0)
	{
		err = errno;
		in_rpt("close: %s", strerror(err));
		errflag = TRUE;
	}

	if (close(dfd) != 0)
	{
		err = errno;
		in_rpt("close: %s", strerror(err));
		errflag = TRUE;
	}

	DBUG_RETURN(errflag ? -1 : 0);
}
