/*
* Copyright (c) 1997 X/Open Company Ltd., A member of The Open Group
*
* All rights reserved. No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* Motif, OSF/1 and UNIX are registered trademarks and X/Open,
* the "X Device" and The Open Group are trademarks of The Open Group.
*
* Project: LSB-OS
*
* File: tset/LSB.os/procprim/execl/execl_check.c
*
* Modifications:
* $Log: execl_check.c,v $
* Revision 1.2  2001/10/02 11:14:55  ajosey
* Update the test cases not to use functionality outside the LSB spec
*
* Revision 1.1  2001/05/27 06:59:19  cyeoh
* Adds files for lsb-os test suite
*
*
*/

/*
*	This file contains the code to test the exec* functions for
*	inheritance behavior.
*
*	It checks:
*	1) named semaphores are closed
*	2) memory mappings are unmapped
*	3) scheduling policy parameters are unchanged
*	4) timers are deleted
*	5) message queues are closed
*
*	check exits with 0 if the behavior in question is correct,
*	otherwise it exits with an error indication.
*/

#include	<vsrt/vsrt.h>
#ifdef _POSIX_PRIORITY_SCHEDULING
#include	<sched.h>
#endif
#include	<time.h>
#include	<sys/mman.h>
#include	<setjmp.h>
#include	<setprv.h>

struct sigaction Sigaction;

static int got_sigsegv = 0;
static int got_sigbus = 0;
static int setjmp_called = 0;
static sigjmp_buf env;

char	*msg1="get_param failed in execed process";
char	*msg2="get_scheduler failed in execed process";
char	*msg3="setprv failed in execed process";

extern	int	setprv();

static void mmapsig_fn(s)
int	s;
{
	if (s == SIGSEGV)
		got_sigsegv++;
	else if (s == SIGBUS)
		got_sigbus++;
	if (setjmp_called) {
		setjmp_called = 0;
		siglongjmp(env, 1);
	}
}

char ebuf[512];

int main(int, char **);

int main(int argc, char **argv)
{
	int	*adr1;
	int	pid;
	int	status;
	timer_t	timer_id;
	volatile int i;
	int	policy, priority;
	int	npolicy;
#ifdef _POSIX_PRIORITY_SCHEDULING
	struct	sched_param param;
#endif

	argv++; argc--;
	if (strcmp(*argv, "semaphores") == 0) {
		/* inform parent that we are up and running */
		if (write(1, "EXECED_PROG_IS_RUNNING", 22) != 22)
			exit(99);
		
		/* wait a while for parent to complete test */
		sleep(10);

		if (write(1, "PASS", 4) != 4)
			exit(99);
		exit(0);
	}

	if (strcmp(*argv, "mappings") == 0) {

		adr1 = (int *)strtoul(*(++argv), NULL, 16);
		if (adr1 == 0)
			exit(99);

		if (sigemptyset(&(Sigaction.sa_mask)) == -1)
			exit(99);
		Sigaction.sa_handler = mmapsig_fn;
		Sigaction.sa_flags = 0;
		if (sigaction(SIGSEGV,&Sigaction,(struct sigaction *)NULL) == -1)
			exit(99);
		if (sigemptyset(&(Sigaction.sa_mask)) == -1)
			exit(99);
		Sigaction.sa_handler = mmapsig_fn;
		Sigaction.sa_flags = 0;
		if (sigaction(SIGBUS,&Sigaction,(struct sigaction *)NULL) == -1)
			exit(99);

		if ((pid = fork()) == 0) {
			setjmp_called = 1;
			if (sigsetjmp(env,1) == 0)
				i = *adr1;
	
			if ((got_sigbus == 0) && (got_sigsegv == 0)) {
				if(write(1, "FAIL", 4) != 4)
					exit(99);
			}
			if(write(1, "PASS", 4) != 4)
				exit(99);
			exit(0);
		}
		pid = wait(&status);
		if (pid == -1)
			exit(99);

		if (WIFEXITED(status)) {
			if (WEXITSTATUS(status) == 99)
				exit(99);
		} else
			exit(99);
		exit(0);
	}

#ifdef _POSIX_PRIORITY_SCHEDULING
	/*we're passed our expected policy and priority*/
	if (strcmp(*argv, "scheduling") == 0) {
		policy = atoi(*(++argv));
		priority = atoi(*(++argv));
		if (setprv(PRV_GETRTSCHED) != 0) {
			if (write(1, msg3, msg3) != strlen(msg3))
				exit(99);
			exit(0);
       		}
		if (sched_getparam(0, &param) == -1) {
			if (write(1, msg1, msg1) != strlen(msg1))
				exit(99);
			exit(0);
		}
		if ((npolicy = sched_getscheduler(0)) == -1) {
			if (write(1, msg2, msg2) != strlen(msg2))
				exit(99);
			exit(0);
		}
		if ((param.sched_priority != priority) || (npolicy != policy)) {
			sprintf(ebuf, "Expected policy = %d, priority %d, actually %d and %d", policy, priority, npolicy, param.sched_priority);
			if (write(1, ebuf, strlen(ebuf)) != strlen(ebuf))
				exit(99);
			exit(0);
		}
		
		if (write(1, "PASS", 4) != 4)
			exit(99);

		exit(0);
	}
#endif

	if (strcmp(*argv, "timers") == 0) {
		timer_id = (timer_t)strtoul(*(++argv), NULL, 16);
		if ((timer_delete(timer_id) == -1) && (errno == EINVAL)) {
			if (write(1, "PASS", 4) != 4)
				exit(99);
		} else {
			if (write(1, "FAIL", 4) != 4)
				exit(99);
		}

		exit(0);
	}
	if (strcmp(*argv, "mqueues") == 0) {
		/* inform parent that we are up and running */
		if (write(1, "EXECED_PROG_IS_RUNNING", 22) != 22)
			exit(99);
		
		/* wait a while for parent to complete test */
		sleep(10);

		if (write(1, "PASS", 4) != 4)
			exit(99);
		exit(0);
	}
	exit(99);
}
